;;;; above-below.el -- part of John's elisp change-ringing system
;;; Time-stamp: <99/03/19 15:15:05 johnstu>

(eval-when (compile load eval)
  (provide 'above-below)
  (require 'place-notation)
  (require 'bell-numbers)
  (require 'beer)
  (require 'method-library)
  )

;;;; Combining methods above and below the treble

(defun ringing:splitter-output-function (row
					 remaining-changes
					 method
					 rounds
					 below-holder
					 above-holder)
  "Output (display) function for splitting a method around the treble."
  (let* ((treble-position (string-match "1" row))
	 (this-change (car remaining-changes))
	 (changes-length (length this-change))
	 (below-changes (concat (substring this-change 0 (1+ treble-position))
				(make-string (- changes-length (1+ treble-position)) 0)))
	 (above-changes (concat (make-string treble-position 0)
				(substring this-change treble-position))))
    (rplacd below-holder (cons below-changes (cdr below-holder)))
    (rplacd above-holder (cons above-changes (cdr above-holder)))))

(defun ringing:split-around-treble (method)
  "Split METHOD around the treble, returning a cons for below . above treble,
where below and above are methods."
  (let* ((method-name (ringing:method-name method))
	 (method-data (ringing:find-named-method method))
	 (ringing:observation-number (change-ringing-method-nbells method-data))
         (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	 (below (make-change-ringing-method
		 :title (concat method-name " below the treble")
		 :nbells (change-ringing-method-nbells method-data)
		 :plain nil
		 :work nil
		 :plain-lead-end (make-string
				  (length (change-ringing-method-plain-lead-end method-data))
				  0)
		 :bob (make-string
		       (length (change-ringing-method-bob method-data))
		       0)
		 :single (make-string
			  (length (change-ringing-method-single method-data))
			  0)
		 :special-start (change-ringing-method-special-start method-data)
		 :autocalls (change-ringing-method-autocalls method-data)
		 :defining-file 'constructed
		 :original-description-string 'constructed))
	 (above (make-change-ringing-method
		 :title (concat method-name " above the treble")
		 :nbells (change-ringing-method-nbells method-data)
		 :plain nil
		 :work nil
		 :plain-lead-end (change-ringing-method-plain-lead-end method-data)
		 :bob (change-ringing-method-bob method-data)
		 :single (change-ringing-method-single method-data)
		 :special-start (change-ringing-method-special-start method-data)
		 :autocalls (change-ringing-method-autocalls method-data)
		 :defining-file 'constructed
		 :original-description-string 'constructed))
	 (below-change-holder (cons t nil))
	 (above-change-holder (cons t nil))
	 (place-marker (list 'work)))
	
    (ringing:ring-from-rounds-plain method-data
				    'ringing:ring-one-lead-helper-function
				    (list place-marker)
				    'ringing:splitter-output-function
				    (list below-change-holder above-change-holder))

    (let ((collected-below (reverse (cdr below-change-holder)))
	  (collected-above (reverse (cdr above-change-holder))))
   
      (setf (change-ringing-method-plain below) collected-below
	    (change-ringing-method-work below) (nreverse (cdr (reverse collected-below)))
	    (change-ringing-method-plain above) collected-above
	    (change-ringing-method-work above) (nreverse (cdr (reverse collected-above)))))
  
    (push (cons (change-ringing-method-title below) below) ringing:constructed-methods)
    (push (cons (change-ringing-method-title above) above) ringing:constructed-methods)

    (cons below above)))

(defun ringing:combined-change (b a)
  "Combine the changes B and A (typically below and above the treble."
  (if (and (or (stringp a) (arrayp a))
	   (or (stringp b) (arrayp b))
	   (= (length a) (length b)))
      (let ((c (make-string (length a) 0)))
	(dotimes (i (length a))
	  (let ((ac (aref a i))
		(bc (aref b i)))
	    (cond
	     ((= ac bc) (aset c i ac))
	     ((zerop ac) (aset c i bc))
	     ((zerop bc) (aset c i ac))
	     (t (error "Cannot combine these rows: %s %s" b a)))))
	c)
    nil))

(defun ringing:combine (b a)
  "Return the combination of methods B and A (typically below and above treble)."
  (let* ((method-data-b (ringing:find-constructed-method b))
	 (method-data-a (ringing:find-constructed-method a))
	 (changes-b (change-ringing-method-plain method-data-b))
	 (changes-a (change-ringing-method-plain method-data-a))
	 (method-data-c (make-change-ringing-method
			 :title (concat (change-ringing-method-title method-data-b)
					" with "
					(change-ringing-method-title method-data-a))
			 :nbells (change-ringing-method-nbells method-data-b)
			 :plain nil
			 :work nil
			 :plain-lead-end nil
			 :bob (change-ringing-method-bob method-data-b)
			 :single (change-ringing-method-single method-data-b)
			 :special-start (or (change-ringing-method-special-start method-data-a)
					    (change-ringing-method-special-start method-data-b))
			 :autocalls (or (change-ringing-method-autocalls method-data-a)
					(change-ringing-method-autocalls method-data-b))

			 :original-description-string "constructed"))
	 (changes-c nil)
	 (c nil)
	 )
    (unless (= (change-ringing-method-nbells method-data-b)
	       (change-ringing-method-nbells method-data-a))
      (error "Cannot combine methods at different stages"))
    (while (setq c (ringing:combined-change (pop changes-b)
					    (pop changes-a)))
      (push c changes-c))
    
    (setf (change-ringing-method-plain method-data-c) (reverse changes-c)
	  (change-ringing-method-plain-lead-end method-data-c) (list (car changes-c))
	  (change-ringing-method-work method-data-c) (reverse (cdr changes-c))
	  (change-ringing-method-original-description-string method-data-c) (place-notation:recreate-string-from-changes
									     (change-ringing-method-work method-data-c)))

    (push (cons (change-ringing-method-title method-data-c) method-data-c) ringing:constructed-methods)
 
    method-data-c))

(defvar ringing:constructed-methods nil
  "Methods that have been constructed by manipulating other methods.")

(defun ringing:find-constructed-method (description)
  "Find a constructed method as per DESCRIPTION
in which the allowable constructs are
X
X above the treble
X below the treble
X with Y
X/Y (equivalent to X above the treble with Y below the treble)."
  (if (change-ringing-method-p description)
      description
    (let ((pair (assoc description ringing:constructed-methods)))
      (if pair
	  (cdr pair)
	(let ((ordinary (ringing:find-named-method description t)))
	  (if ordinary
	      ordinary
	    (save-match-data
	      (cond
	       ((string-match "^\\(.+\\)/\\(.+\\)$" description)
		(ringing:find-constructed-method
		 (format "%s above the treble with %s below the treble"
			 (substring description (match-beginning 1) (match-end 1))
			 (substring description (match-beginning 2) (match-end 2)))))
	       ((string-match "^\\(.+\\) with \\(.+\\)$" description)
		(ringing:combine (ringing:find-constructed-method
				  (substring description (match-beginning 1) (match-end 1)))
				 (ringing:find-constructed-method
				  (substring description (match-beginning 2) (match-end 2)))))
	       ((string-match "^\\(.+\\) below the treble$" description)
		(car (ringing:split-around-treble
		      (substring description (match-beginning 1) (match-end 1)))))
	       ((string-match "^\\(.+\\) above the treble$" description)
		(cdr (ringing:split-around-treble
		      (substring description (match-beginning 1) (match-end 1)))))))))))))

(defun ringing:identify-method-above-and-below (method &optional just-work)
  "Try to identify the parts of METHOD above or below the treble
against well-known methods. The well-known methods are defined in
a method group whose name is \"naming-%d\" where the %d is the number
of bells.
With optional JUST-WORK, compare only the work, not the lead ends."
  (interactive (list (ringing:read-method-name "Spot elements of: ")
		     (y-or-n-p "Compare just the work? ")))
  (let* ((method-data (ringing:find-named-method method))
	 (stage (change-ringing-method-nbells method-data))
	 (reference-file-name (expand-file-name (format "naming-%d" stage)
						method-groups:groups-directory)))
    (if (file-exists-p reference-file-name)
	(let* ((split-method (ringing:split-around-treble method-data))
	       (below (car split-method)) (above (cdr split-method))
	       (below-found nil) (above-found nil)
	       (reference-methods-group (method-groups:file-group-data reference-file-name))
	       (split-references (method-groups:split-methods reference-methods-group)))

	  (dolist (reference-method-halves split-references)
	    (when (ringing:equal-methods below (car reference-method-halves) just-work)
	      (setq below-found (car reference-method-halves)))
	    (when (ringing:equal-methods above (cdr reference-method-halves) just-work)
	      (setq above-found (cdr reference-method-halves))))

	  (when (interactive-p)
	    (message "%s and %s"
		     (if above-found (change-ringing-method-title above-found) "not recognized above the treble")
		     (if below-found (change-ringing-method-title below-found) "not recognized below the treble")))
	  (cons below-found above-found))
      (progn
	(when (interactive-p)
	  (error "No naming group for %s -- create one in %s if you want one"
		 (ringing:stage-name stage) reference-file-name))
	nil))))

;;; end of above-below.el
