;;;; analyze-group.el -- part of John's elisp change-ringing system
;;; Analyze groups of methods
;;; Time-stamp: <99/10/12 14:38:40 johnstu>

(eval-when (compile load eval)
  (provide 'analyze-group)
  (require 'cl)
  (require 'method-groups)
  (require 'above-below))

(defun ringing:method-short-name (long-name &optional not-this)
  "Return a short form of LONG-NAME.
If it's only above or below the treble, remove mention thereof.
If optional NOT-THIS matches the full name, return nil."
  (let* ((name (ringing:method-name long-name))
	 (anyway (if (string-match " ..... the treble" name)
		     (substring name 0 -16)
		   long-name)))
    (if (and (stringp not-this)
	     (string-match not-this anyway))
	nil
      (capitalize (microSiRiL:name-used-in-file anyway)))))

(defun method-group:assemble-report (group &optional just-work)
  "Assemble the data for a report on GROUP.
With optional JUST-WORK, compare only the work, not the lead ends."
  (let ((methods (method-groups:group-methods group))
	(report nil))
    (dolist (method-pair methods)
      (let ((method-name (car method-pair)))
	(message "Matching %s" method-name)
	(let* ((method-data (cdr method-pair))
	       (above-below (ringing:identify-method-above-and-below method-data just-work))
	       (similar (method-library:methods-matching-work
			 method-data
			 (microSiRiL:parse-file-superficially (change-ringing-method-defining-file method-data)))))
	  (push (vector (car method-pair)
			(change-ringing-method-original-description-string method-data)
			(if (car above-below) (ringing:method-short-name (car above-below) method-name) nil)
			(if (cdr above-below) (ringing:method-short-name (cdr above-below) method-name) nil)
			(if (car similar) (mapconcat 'ringing:method-short-name (mapcar 'car (car similar)) ", ") nil)
			(if (cdr similar) (mapconcat 'ringing:method-short-name (mapcar 'car (cdr similar)) ", ") nil))
		report))))
    report))

(defun method-group:add-to-cluster (method clusters)
  "Place METHOD in CLUSTERS."
  (catch 'found
    (dolist (set (cdr clusters))
      (dolist (some-method (cdr set))
	(when (ringing:equal-methods method some-method)
	    (push method (cdr set))
	    (throw 'found t))))
    (push (list 'set method)
	  (cdr clusters))))

(defun method-group:sort-cluster (cluster)
  "Sort CLUSTER according to number of elements in each set."
  (rplacd cluster
	  (sort (cdr cluster)
		(function
		 (lambda (a b)
		   (> (length a) (length b)))))))
			    

(defun method-group:cluster-methods-by-halves (group)
  "Work out which methods in GROUP have the same half-methods."
  (let* ((group-data (method-groups:find-named-group group))
	 (halves (method-groups:split-methods group-data))
	 (above-clusters (list 'clusters))
	 (below-clusters (list 'clusters)))
    (dolist (half halves)
      (let ((below (car half))
	    (above (cdr half)))
	(method-group:add-to-cluster below below-clusters)
	(method-group:add-to-cluster above above-clusters)))
    (cons (method-group:sort-cluster below-clusters)
	  (method-group:sort-cluster above-clusters))))

(defun method-group:print-cluster (cluster)
  "Print a cluster to stdout."
  (dolist (set (cdr cluster))
    (princ "  ")
    (princ (mapconcat 'identity (cdr set) ", "))
    (when (> (length (cdr set)) 1)
      (princ (format " (%d methods)" (length (cdr set)))))
    (princ "\n")))

(defun method-group:make-basename-clusters (cluster)
  "Make a version of CLUSTER with just the base names in it."
  (cons 'clusters (mapcar (function (lambda (set)
				     (cons 'set
					   (mapcar 'ringing:method-short-name
						   (cdr set)))))
			 (cdr cluster))))

(defun method-group:delete-from-clusters (method cluster)
  "Delete METHOD from CLUSTER, destructively."
  (let ((method-name (ringing:method-short-name method)))
    (dolist (set (cdr cluster))
      (rplacd set
	      (delete-if (function (lambda (a) (string= method-name a)))
			 set))))
  cluster)

(defun method-group:sort-clusters-prioritizing (cluster famous)
  "Sort sets in clusters, bringing any in FAMOUS to the front."
  (dolist (set (cdr cluster))
    (rplacd set (sort (cdr set)
		      (function (lambda (a b)
				  (or
				   (member a famous)
				   (string< a b)))))))
  cluster)  

(defun method-group:report (group &optional just-work)
  "Report on the methods in GROUP.
With optional JUST-WORK, compare only the work, not the lead ends.

The \"Above Treble\" and \"Below Treble\" columns are drawn from a
specified set of methods (defined by the file \"naming-%d\", where the
number of bells is substituted into the %d, in the directory given in
method-groups:groups-directory), as are the \"Clusters\" listings at
the bottom, while \"Same except lead end\" and \"Same except
half-lead\" for each method are drawn from the entire mSiRiL methods file
containing the definition of that method."
  (interactive (list (method-group:read-group-name "Report on group: ")
		     (y-or-n-p "Just use work in comparing half-methods? ")))
  (let* ((group-data (method-groups:find-named-group group))
	 (title (method-groups:group-name group-data))
	 (report-data (method-group:assemble-report group just-work))
	 (report (cons (vector "Name" "Place Notation" "Below Treble" "Above Treble" "Same except lead end" "Same except half-lead")
		       report-data))
	 (n-columns (length (car report)))
	 (columns (make-vector n-columns 0))
	 (formats (make-vector n-columns nil))
	 (blanks (make-vector n-columns nil))
	 (stage (method-groups:method-group-stage group-data))
	 (reference-file-name (expand-file-name (format "naming-%d" stage)
						method-groups:groups-directory))
	 (reference-methods-group (method-groups:file-group-data reference-file-name))
	 (famous-names (mapcar 'ringing:method-short-name (mapcar 'car (method-groups:method-group-methods reference-methods-group))))
	 )
    (dolist (row report)
      (dotimes (col n-columns)
	(let ((cell (aref row col)))
	  (when (and (stringp cell)
		     (> (length cell) (aref columns col)))
	    (aset columns col (length cell))))))
    (dotimes (col n-columns)
      (aset formats col (format "%% -%ds" (aref columns col)))
      (aset blanks col (make-string (aref columns col) ? )))
    (with-output-to-temp-buffer (format "*Report on %s*" title)
      (princ (format "Report on method similarities for %s\n\n" title))
      (dolist (row report)
	(dotimes (col n-columns)
	  (let ((cell (aref row col)))
	    (if (stringp cell)
		(princ (format (aref formats col)
			       cell))
	      (princ (aref blanks col))))
	  (princ " ")
	  )
	(princ "\n"))
      (princ "\nClusters:\n")
      (let* ((clusters (method-group:cluster-methods-by-halves group-data))
	     (above-clusters (method-group:sort-clusters-prioritizing (method-group:make-basename-clusters (cdr clusters)) famous-names))
	     (below-clusters (method-group:sort-clusters-prioritizing (method-group:make-basename-clusters (car clusters)) famous-names)))
	(princ " Same above the treble:\n")
	(method-group:print-cluster above-clusters)
	(princ " Same below the treble:\n")
	(method-group:print-cluster below-clusters)))))

(defun method-group:cluster-by-changes-in-lead (group &optional just-work)
  "List which methods in GROUP have the same changes in each lead,
that is, can be interchanged as far as falseness goes.

With optional argument JUST-WORK, exclude the lead ends from the comparisons."
  (interactive (list (method-group:read-group-name "Report methods with same changes per lead in group: ")
		     (y-or-n-p "Just use work in comparing methods? ")))
  (let* ((group-data (method-groups:find-named-group group))
	 (title (method-groups:group-name group-data))
	 (methods (method-groups:group-methods group-data))
	 (methods-lead-changes (mapcar
				(function
				 (lambda (method)
				   (cons (car method)
					 (ringing:method-lead-changes (cdr method) just-work))))
				methods))
	 (method-clusters nil)
	 )
    (dolist (method-desc methods-lead-changes)
      (let ((pair (assoc (cdr method-desc) method-clusters)))
	(if pair
	    (push (car method-desc) (cdr pair))
	  (push (cons (cdr method-desc) (list (car method-desc))) method-clusters))))
    (dolist (cluster method-clusters)
      (rplacd cluster (mapconcat 'identity
				 (sort (cdr cluster) 'string<)
				 ", ")))
    (let ((cluster-names (mapcar 'cdr method-clusters)))
      (setq cluster-names (sort cluster-names 'string<))
      (with-output-to-temp-buffer (format "*Equivalence report on %s*" title)
	(dolist (name cluster-names)
	  (princ (format "%s\n" name)))))
    method-clusters))

;;; end of analyze-group.el
