;;;; beer.el -- part of John's elisp change-ringing system
;;; This file implements the Basic Emacs Engine for Ringing
;;; Time-stamp: <99/07/05 12:56:12 johnstu>

(eval-when (compile load eval)
  (provide 'beer)
  (require 'bell-colours))

;;;; ringing system driven by place notation parser's data structures

;;;; main entry points
;;; ringing:ring-one-change
;;; ringing:ring-changes
;;; ringing:ring-from-rounds
;;; ringing:ring-from-rounds-plain
;;; ringing:advance-changes
;;; ringing:retreat-changes

(defvar ringing:most-recent-change nil
  "The most recent change to have been rung.
May be used for drawing the crossings between two rows.")

(defvar ringing:most-recent-row nil
  "The most recent row to have been rung.")

(defun ringing:ring-one-change (row change)
  "Starting with ROW, ring CHANGE.
Each argument may be a string or an array -- anything aref and
aset will work on. Any elements beyond those affected by the
changes are copied unaffected; this lets you put a line ending character
at the end of the string for simplicity of outputting changes into
a buffer."
  (setq ringing:most-recent-change change)
  (let* ((nc (length change))
	 (nr (length row))
	 (new-row (if (stringp row)
		      (make-string nr 0)
		    (make-vector nr 0)))
	 (i 1))
    (aset new-row 0 (aref row 0))
    (while (< i nc)
      (aset new-row (aref change i) (aref row i))
      (setq i (1+ i)))
    (while (< i nr)
      (aset new-row i (aref row i))
      (setq i (1+ i)))
    (setq ringing:most-recent-row new-row)
    new-row))
		
(defun ringing:ring-changes (row changes continue-fn continue-args display-fn display-args)
  "From ROW ring CHANGES calling CONTINUE-FN on the resulting row and remaining CHANGES and CONTINUE-ARGS each time, showing it first also to DISPLAY-FN with added DISPLAY-ARGS.
If CONTINUE-FN returns a string or array, that is used as the next row;
if it returns a list, the car of it is used as the next row, and the
cdr of it as a continuation of CHANGES.
Return the resulting row when all of CHANGES (including any continuations
added by FN) have been used."
  (catch 'stand
    (while changes
      (setq row (ringing:ring-one-change row (car changes)))
      (apply display-fn row changes display-args)
      (setq changes (cdr changes)
	    ringing:backstroke (not ringing:backstroke))
      (let ((fnres (apply continue-fn row changes continue-args)))
	(cond
	 ((or (stringp fnres)
	      (arrayp fnres))
	  (setq row fnres))
	 ((consp fnres)
	  (setq row (car fnres)
		changes (cdr fnres))))))
    row))

;;;; main program entry point for ringing from parsed data

(defvar ringing:current-stage nil
  "The current stage.
Used by auto-hunt, at least.")

(defun ringing:ring-from-rounds (method
				 continue-function continue-args
				 display-function display-args)
  "Ring METHOD starting at rounds using CONTINUE-FUNCTION, CONTINUE-ARGS after each change, showing it first also to DISPLAY-FUNCTION with added DISPLAY-ARGS.
CONTINUE-FUNCTION is as described for ringing:ring-changes, with the method
structure (wrapped in a list, so you can rplaca it for splicing), and the
rounds string as its extra arguments, plus any given here as ARGS.
Also does various output-related setup; use ringing:ring-from-rounds-plain
if you don't want this."
  (unless (or ringing:show-numbers ringing:show-lines)
    (error "Neither number plotting or line plotting enabled"))
  (let* ((nbells (change-ringing-method-nbells method))
	 (method-holder (list method))
	 (method-display:lead-end-line
	  (method-display:make-lead-end-line nbells))
	 (method-display:row-worth-of-space (make-string (+ nbells 2) ? ))
	 (ringing:observation-number (if ringing:observation-number
					 ringing:observation-number
				       nbells))
	 (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	 (ringing:other-observation-number (if ringing:green-line-too
					       (if (evenp ringing:observation-number)
						   (1- ringing:observation-number)
						 (1+ ringing:observation-number))))
	 (ringing:other-observation (if ringing:other-observation-number
					(ringing:bell-number-to-char ringing:other-observation-number)
				      nil))
	 (rounds (ringing:make-rounds (change-ringing-method-nbells method)))
	 (ringing:backstroke nil))
    (setq ringing:previous-row rounds
	  ringing:current-stage nbells
	  ringing:window-configuration-before-help nil)
    (ringing:insert-row-lines rounds)
    (ringing:ring-changes rounds
			  (if (change-ringing-method-special-start method)
			      (change-ringing-method-special-start method)
			    (if (change-ringing-method-work method)
				(change-ringing-method-work method)
			      (change-ringing-method-plain method)))
			  continue-function
			  (append (list method-holder rounds) continue-args)
			  display-function
			  (append (list method-holder rounds) display-args))))

(defun ringing:ring-from-rounds-plain (method
				       continue-function continue-args
				       display-function display-args)
  "Ring METHOD starting at rounds using CONTINUE-FUNCTION, CONTINUE-ARGS after each change, showing it first also to DISPLAY-FUNCTION with added DISPLAY-ARGS.
FUNCTION is as described for ringing:ring-changes, with the method
structure (wrapped in a list, so you can rplaca it for splicing), and the
rounds string as its extra arguments, plus any given here as ARGS.
ringing:ring-from-rounds is similar but also binds some variables used
by the usual output routines.
This plain version is meant for composition proving and so on, where the
output is not displayed."
  (let ((rounds (ringing:make-rounds (change-ringing-method-nbells method)))
	(method-holder (list method))
	(ringing:backstroke nil))
    (setq ringing:window-configuration-before-help nil)
    (ringing:ring-changes rounds
			  (if (change-ringing-method-special-start method)
			      (change-ringing-method-special-start method)
			    (if (change-ringing-method-work method)
				(change-ringing-method-work method)
			      (change-ringing-method-plain method)))
			  continue-function
			  (append (list method-holder rounds) continue-args)
			  display-function
			  (append (list method-holder rounds) display-args))))

(defun ringing:advance-changes (row changes)
  "Return the result of starting from ROW and ringing CHANGES.
ROW is not modified. No callbacks are used."
  (while changes
    (setq row (ringing:ring-one-change row (car changes))
	  changes (cdr changes)))
  row)

(defun ringing:retreat-changes (row changes)
  "Return the row from which you would have started to get to ROW by ringing CHANGES.
ROW, CHANGES are not modified. No callbacks are used."
  (let ((back-changes (reverse changes)))
    (while back-changes
      (setq row (ringing:ring-one-change row (car back-changes))
	    back-changes (cdr back-changes)))
    row))

;;;; parameter not used in this file but in some of its callers

(defvar ringing:auto-ring-inter-row-delay .125
  "How long to wait between rows, when showing methods non-interactively.")

;;;; variables used by many callers

(defvar ringing:observation-number nil
  "The observation bell, as a number.
If you bind this, you should also bind ringing:observation.")

(defvar ringing:observation nil
  "The observation bell, as a character naming a bell.
If you bind this, you should bind also ringing:observation-number.")

(defvar ringing:green-line-too nil
  "Whether to show a green line for the other handbell rung by the observation ringer.")

(defvar ringing:other-observation-number nil
  "The other observation bell, as a number,
that is, the other handbell rung by the observation ringer.
If you bind this, you should also bind ringing:observation.")

(defvar ringing:other-observation nil
  "The other observation bell, as a character naming a bell,
that is, the other handbell rung by the observation ringer.
If you bind this, you should bind also ringing:observation-number.")

;;; end of beer.el
