;;;; bell-colours.el -- part of John's elisp change-ringing system
;;; try to set text colours for method lines
;;; Time-stamp: <98/07/30 09:19:02 john>

(provide 'bell-colours)

(defvar ringing:highlight-bad-rows t
  "Whether to highlight rows on which the user has made a mistake.")

(defvar ringing:mark-backstroke nil
  "Whether to mark backstrokes (in italic if possible, else underlined).")

(defconst ringing:end-of-row-char ?\n
  "The character to put at the end of each row,
where it will not be affected by the changes but will be output
into buffers etc.

Normally newline, but for multi-column output may be re-bound
to space or tab.")

(defvar method-display:asterisk-manual-bell nil
  "Whether to mark the observation bell with an asterisk instead of a digit.")
  
(defun ringing:make-rounds (n &optional no-colours)
  "Construct the rounds for N bells, using ringing:end-of-row-char at the end.
We try to mark the treble and observation with the appropriate colours;
ringing:observation gives which bell to mark in blue, otherwise we
take the last one.
With optional argument NO-COLOURS, don't set the colours."
  (let* ((rounds (make-string (+ n 2) 0))
	 (i 1))
    (aset rounds 0 ? )
    (if (<= n 16)
	(while (<= i n)
	  (aset rounds i
		(cond
		 ((= i 10) ?0)
		 ((= i 11) ?E)
		 ((= i 12) ?T)
		 ((>= i 13) (+ (- i 13) ?A))
		 (t (+ ?0 i))))
	  (setq i (1+ i)))
      (while (<= i n)
	(let ((j (1+ (mod (1- i) 54))))
	  (aset rounds i
		(cond
		 ((<= j 26) (+ ?` j))
		 ((<= j 54) (+ ?& j)))))
	(setq i (1+ i))))
    (if (and method-display:asterisk-manual-bell
	     ringing:observation-number)
	(aset rounds ringing:observation-number ?*))
    (aset rounds (1+ n) ringing:end-of-row-char)
    (unless no-colours
      (ringing:set-colours rounds))
    rounds))

(defun ringing:roundsp (row rounds)
  "Return whether ROW is ROUNDS according to that definition of ROUNDS."
  (string= row rounds))

(defconst ringing:bad-background-colour "yellow"
  "The colour for highlighting bad rows.")

(defun ringing:copy-face (old-face new-face &optional frame new-frame)
  "Safe caller for copy-face, which does not exist on TTY-only emacsen."
  (if (fboundp 'copy-face)
      (copy-face old-face new-face frame new-frame)
    new-face))

(defun ringing:set-face-foreground (face color &optional frame)
  "Safe caller for set-face-foreground, which does not exist on TTY-only emacsen."
  (when (fboundp 'copy-face) (set-face-foreground face color frame)))

(defun ringing:set-face-background (face color &optional frame)
  "Safe caller for set-face-background, which does not exist on TTY-only emacsen."
  (when (fboundp 'copy-face) (set-face-background face color frame)))

(defun ringing:set-face-underline-p (face underline-p &optional frame)
  "Safe caller for set-face-underlined-p, which does not exist on TTY-only emacsen."
  (when (fboundp 'copy-face) (set-face-underline-p face underline-p frame)))

(defconst ringing:default-face
  (ringing:copy-face 'default 'ringing:default-face)
  "Face for showing the other lines.")
(defconst ringing:red-line-face
  (ringing:copy-face 'default 'ringing:red-line-face)
  "Face for showing the red line")
(ringing:set-face-foreground ringing:red-line-face "red")
(defconst ringing:blue-line-face
  (ringing:copy-face 'default 'ringing:blue-line-face)
  "Face for showing the blue line")
(ringing:set-face-foreground ringing:blue-line-face "blue")
(defconst ringing:green-line-face
  (ringing:copy-face 'default 'ringing:green-line-face)
  "Face for showing the green line")
(ringing:set-face-foreground ringing:green-line-face "ForestGreen")
(defconst ringing:bad-red-line-face
  (ringing:copy-face 'default 'ringing:bad-red-line-face)
  "Face for showing the red line when the user has made a mistake")
(ringing:set-face-foreground ringing:bad-red-line-face "red")
(ringing:set-face-background ringing:bad-red-line-face ringing:bad-background-colour)
(defconst ringing:bad-blue-line-face
  (ringing:copy-face 'default 'ringing:bad-blue-line-face)
  "Face for showing the blue line when the user has made a mistake")
(ringing:set-face-foreground ringing:bad-blue-line-face "blue")
(ringing:set-face-background ringing:bad-blue-line-face ringing:bad-background-colour)
(defconst ringing:bad-green-line-face
  (ringing:copy-face 'default 'ringing:bad-green-line-face)
  "Face for showing the green line when the user has made a mistake")
(ringing:set-face-foreground ringing:bad-green-line-face "ForestGreen")
(ringing:set-face-background ringing:bad-green-line-face ringing:bad-background-colour)
(defconst ringing:bad-face
  (ringing:copy-face 'default 'ringing:bad-face)
  "Face for showing when the user has made a mistake")
(ringing:set-face-background ringing:bad-face ringing:bad-background-colour)

(defconst ringing:face-colour-names ["black" "red" "green" "blue"]
  "Names for the face colours.")

(defvar ringing:use-bold nil
  "Whether to try using bold.")

(defun ringing:setup-faces ()
  "Set up a vector to go in ringing:faces, which see."
  (when (fboundp 'copy-face)
    (let ((outer (make-vector 2 nil)))
      (dotimes (outeri 2)
	(let ((mid (make-vector 2 nil)))
	  (aset outer outeri mid)
	  (dotimes (midi 2)
	    (let ((inner (make-vector 4 nil)))
	      (aset mid midi inner)
	      (dotimes (inneri 4)
		(let ((this-face (ringing:copy-face 'default
						    (intern
						     (format "ringing:face-%s-%s-%s"
							     (if (= outeri 1) "bad" "ok")
							     (if (= midi 1) "backstroke" "handstroke")
							     (aref ringing:face-colour-names inneri)))))
		      (got-italic nil)
		      (got-bold nil))
		  (when (= outeri 1)
		    (ringing:set-face-background this-face ringing:bad-background-colour))
		  (when (= midi 1)
		    (condition-case no-italic
			(progn
			  (make-face-italic this-face)
			  (setq got-italic t))
		      (error (set-face-underline-p this-face t))))
		  (when (and ringing:use-bold (= midi 0))
		    (condition-case no-bold
			(progn
			  (make-face-bold this-face)
			  (setq got-bold t))
		      (error nil)))
		  (case inneri
		    (0 nil)
		    (1 (ringing:set-face-foreground this-face "red"))
		    (2 (ringing:set-face-foreground this-face "ForestGreen"))
		    (3 (ringing:set-face-foreground this-face "blue")))
		  (aset inner inneri this-face)))))))
      outer)))

(makunbound 'ringing:faces)

(defvar ringing:faces
  (ringing:setup-faces)
  "Storage for the font faces for the ringing system,
as a nest of vectors.
Outermost dimension is [good bad];
next is [handstroke backstroke];
innermost is [plain red green blue].")

(defconst ringing:black 0
  "Black constant for ringing:get-face.")
(defconst ringing:red 1
  "Red constant for ringing:get-face.")
(defconst ringing:green 2
  "Green constant for ringing:get-face.")
(defconst ringing:blue 3
  "Blue constant for ringing:get-face.")

(defun ringing:get-face (bad backstroke colour)
  "Return the face for BAD BACKSTROKE COLOUR."
  (aref
   (aref (aref ringing:faces (if bad 1 0))
	 (if backstroke 1 0))
   colour))

(defun ringing:get-colour-face (colour)
  "Return the face for COLOUR according to settings for bad and backstroke."
  (ringing:get-face (and ringing:highlight-bad-rows ringing:bad-row)
		    (and ringing:mark-backstroke ringing:backstroke)
		    colour))

(defun ringing:set-colours (row)
  "Set the colours for ROW, with ringing:observation-number (tenor if nil)."
  (when (fboundp 'copy-face)
    (let ((ringing:observation-number
	   (if (null ringing:observation-number)
	       (- (length row) 2)
	     ringing:observation-number)))
      (put-text-property 1 2 'face
			 (ringing:get-face nil nil ringing:red)
			 row)
      (put-text-property ringing:observation-number (1+ ringing:observation-number)
			 'face
			 (ringing:get-face nil nil ringing:blue)
			 row)
      (when (numberp ringing:other-observation-number)
	(put-text-property ringing:other-observation-number (1+ ringing:other-observation-number)
			   'face
			   (ringing:get-face nil nil ringing:green)
			   row)))))

(defun ringing:put-text-property (start end property value &optional object)
  "Safe wrapper for put-text-property, which does not exist on TTY-only emacsen."
  (when (fboundp 'copy-face)
    (put-text-property start end property value object)))

(defconst ringing:blue-line-up "\\ "
  "Blue line for a bell going up.")

(defconst ringing:blue-line-places " |"
  "Blue line for a bell making places.")

(defconst ringing:blue-line-down "/ "
  "Blue line for a bell going down.")

(defconst ringing:red-line-up "\\ "
  "Red line for a bell going up.")

(defconst ringing:red-line-places " |"
  "Red line for a bell making places.")

(defconst ringing:red-line-down "/ "
  "Red line for a bell going down.")

(defconst ringing:green-line-up "\\ "
  "Green line for a bell going up.")

(defconst ringing:green-line-places " |"
  "Green line for a bell making places.")

(defconst ringing:green-line-down "/ "
  "Green line for a bell going down.")

(ringing:put-text-property 0 2 'face ringing:blue-line-face ringing:blue-line-up)
(ringing:put-text-property 0 2 'face ringing:blue-line-face ringing:blue-line-places)
(ringing:put-text-property 0 2 'face ringing:blue-line-face ringing:blue-line-down)

(ringing:put-text-property 0 2 'face ringing:red-line-face ringing:red-line-up)
(ringing:put-text-property 0 2 'face ringing:red-line-face ringing:red-line-places)
(ringing:put-text-property 0 2 'face ringing:red-line-face ringing:red-line-down)

(ringing:put-text-property 0 2 'face ringing:green-line-face ringing:green-line-up)
(ringing:put-text-property 0 2 'face ringing:green-line-face ringing:green-line-places)
(ringing:put-text-property 0 2 'face ringing:green-line-face ringing:green-line-down)

(defconst ringing:bad-blue-line-up "\\ "
  "Blue line for a bell going up.")

(defconst ringing:bad-blue-line-places " |"
  "Blue line for a bell making places.")

(defconst ringing:bad-blue-line-down "/ "
  "Blue line for a bell going down.")

(defconst ringing:bad-red-line-up "\\ "
  "Red line for a bell going up.")

(defconst ringing:bad-red-line-places " |"
  "Red line for a bell making places.")

(defconst ringing:bad-red-line-down "/ "
  "Red line for a bell going down.")

(defconst ringing:bad-green-line-up "\\ "
  "Green line for a bell going up.")

(defconst ringing:bad-green-line-places " |"
  "Green line for a bell making places.")

(defconst ringing:bad-green-line-down "/ "
  "Green line for a bell going down.")

(ringing:put-text-property 0 2 'face ringing:bad-blue-line-face ringing:bad-blue-line-up)
(ringing:put-text-property 0 2 'face ringing:bad-blue-line-face ringing:bad-blue-line-places)
(ringing:put-text-property 0 2 'face ringing:bad-blue-line-face ringing:bad-blue-line-down)

(ringing:put-text-property 0 2 'face ringing:bad-red-line-face ringing:bad-red-line-up)
(ringing:put-text-property 0 2 'face ringing:bad-red-line-face ringing:bad-red-line-places)
(ringing:put-text-property 0 2 'face ringing:bad-red-line-face ringing:bad-red-line-down)

(ringing:put-text-property 0 2 'face ringing:bad-green-line-face ringing:bad-green-line-up)
(ringing:put-text-property 0 2 'face ringing:bad-green-line-face ringing:bad-green-line-places)
(ringing:put-text-property 0 2 'face ringing:bad-green-line-face ringing:bad-green-line-down)

(defun ringing:aset-blue (string index character)
  "Do ASET on the arguments, setting the colour to blue."
  (aset string index character)
  (when (fboundp 'copy-face)
    (ringing:put-text-property index (1+ index) 'face
			       (ringing:get-colour-face ringing:blue)
			       string))
  character)

(defun ringing:aset-red (string index character)
  "Do ASET on the arguments, setting the colour to red."
  (aset string index character)
  (when (fboundp 'copy-face)
    (ringing:put-text-property index (1+ index) 'face
			       (ringing:get-colour-face ringing:red)
			       string))
  character)

(defun ringing:aset-green (string index character)
  "Do ASET on the arguments, setting the colour to green."
  (aset string index character)
  (when (fboundp 'copy-face)
    (ringing:put-text-property index (1+ index) 'face
			       (ringing:get-colour-face ringing:green)
			       string))
  character)

(defvar ringing:bad-row nil
  "Whether the present row is bad, that is, to be drawn with a different background.
Nil generally, bound to non-nil as needed.")

(defvar ringing:backstroke nil
  "Whether we are on a backstroke.")

(defun ringing:apply-bell-colours (row)
  "To ROW, apply colouring, according to ringing:observation.
If ambient ringing:bad-row non-nil, use a different background colour."
  (when (fboundp 'copy-face)
    (let ((i 0)
	  (n (length row)))
      (ringing:put-text-property 0 n 'face (ringing:get-colour-face ringing:black) row)
      (while (< i n)
	(cond ((eq (aref row i) ?1)
	       (ringing:put-text-property i (1+ i) 'face
					  (ringing:get-colour-face ringing:red)
					  row))
	      ((or (eq (aref row i) ringing:observation)
		   (eq (aref row i) ?*))
	       (ringing:put-text-property i (1+ i) 'face
					  (ringing:get-colour-face ringing:blue)
					  row))
	      ((and ringing:green-line-too
		    (not (eq ringing:observation-number 2))
		    (or (eq (aref row i) ringing:other-observation)
			(eq (aref row i) ?#)))
	       (ringing:put-text-property i (1+ i) 'face
					  (ringing:get-colour-face ringing:green)
					  row)))
	(incf i)))))

;;; end of bell-colours.el
