;;;; bell-numbers.el -- part of John's elisp change-ringing system
;;; Convert numbers between numbers and extended digits, etc
;;; Time-stamp: <99/07/05 12:57:35 johnstu>

(provide 'bell-numbers)

(defvar ringing:input-help-buffer-name "*Ringing input help*"
  "The name of the buffer for showing the user what they can type.")

(defvar ringing:bell-input-extra-help-function nil
  "Output additional help (to stdout); bound according to context.")

(defvar ringing:window-configuration-before-help nil
  "The window configuration before displaying help.
Saved so you can get rid of the help at a suitable time.")

(defun ringing:bell-input-help (&optional previous)
  "Display help about ringing input.
If ringing:bell-input-extra-help-function is non-nil, it is called to write
further help (to stdout), allowing context-dependent help to be included.
The output from this goes after the standard help."
  (interactive)
  (when (null ringing:window-configuration-before-help)
    (setq ringing:window-configuration-before-help
	  (current-window-configuration)))
  (with-output-to-temp-buffer ringing:input-help-buffer-name
    (when ringing:use-timing
      (princ
       "Either:\n press the space bar with the appropriate timing\nor:\n"))
    (princ
     "Enter the place you wish to ring in next, as a digit, with 0 being
ten, E being eleven, and T being twelve, and A to D for 16-up.\n")
    (when (numberp previous)
      (princ
       (format
	"You can also use left or right arrow keys to move up or down from
your present position of %d, or the down arrow key to make places.\n"
	previous))
      (when ringing:enable-auto-hunt
	(princ "\nC-Left hunts in to the lead, and C-Right hunts to the back\n")
	(princ "M-Right treble bobs out to the back, and C-M-Right also, but starting differently\n")))
    (princ "\nThe up arrow key calls an immediate stand.\n")
    (when (not (null ringing:bell-input-extra-help-function))
      (funcall ringing:bell-input-extra-help-function))))

(defvar ringing:use-timing nil
  "Whether to use real-time reading of bell events.

If true, you can press the space bar at the appropriate time to ring;
if false, you must use arrow or number keys.

See ringing:time-step for further details.")

(defvar ringing:time-step .1
  "If non-nil, use for real-time reading of bell events,
in which case it should be a floating-point number of seconds
for use as the time step in reading the input, and the input is
taken as given unless it takes more than ringing:time-late steps
which is taken as going up a place;
if the timed input is the space bar, if it is sooner than
ringing:time-early it is down a place, if later than ringing:time-late
is is up a place, otherwise it is making places.

Since the normal settings of ringing:time-late and ringing:time-early are
centred around 10, ringing:time-step should be set to one-tenth of the
average change time -- i.e. typically around 0.2.")

(defvar ringing:time-late 13
  "How many ringing:time-step units (which see) to wait for before
counting a stroke as being up.")

(defvar ringing:time-early 7
  "How few ringing:time-step units (which see) to wait for the space
bar to count as ringing early.")

(defvar ringing:time-max 20
  "How long to wait before giving up on the user pulling a stroke.")

(defun ringing:make-time-messages ()
  "Re-fill the time messages array.
Use this after changing ringing:time-early, ringing:time-late,
or ringing:time-max"
  (let ((array (make-vector (1+ ringing:time-max) nil)))
	(dotimes (i (1+ ringing:time-max))
	  (aset array i (make-string i ?.)))
	(let* ((half (/ ringing:time-max 2))
	      (i (1+ half)))
	  (while (<= i ringing:time-max)
	    (aset (aref array i) half ?|)
	    (incf i)))
	(let* ((i (1+ ringing:time-early)))
	  (while (<= i ringing:time-max)
	    (aset (aref array i) ringing:time-early ?:)
	    (incf i)))
	(let* ((i (1+ ringing:time-late)))
	  (while (<= i ringing:time-max)
	    (aset (aref array i) ringing:time-late ?:)
	    (incf i)))
	array))

(makunbound 'ringing:time-step-messages)

(defvar ringing:time-step-messages
  (ringing:make-time-messages)
  "Messages for showing where in the stroke timeing the system is.")

(defvar ringing:use-timing-here nil
  "Whether to use timing to read positions. This is not a user variable --
use ringing:time-step nil or numeric to switch timing on or off -- but for
making timing be used only while ringing, not reading starting places etc.")

(defun ringing:provide-event-from-timing ()
  "Generate a bell number by using the timing sweep."
  (let ((i 0))
    (sit-for 0)
    (catch 'pulled
      (while (and (not (input-pending-p))
		  (< i ringing:time-max))
	(incf i)
	(message (aref ringing:time-step-messages i))
	(sit-for ringing:time-step))
      (let ((the-event (read-event)))
	(if (eq the-event ? )
	    (throw 'pulled
		   (if (>= i ringing:time-late)
		       'right
		     (if (< i ringing:time-early)
			 'left
		       'down)))
	  (throw 'pulled the-event))))))

(defvar ringing:auto-hunt-state nil
  "Whether we are hunting automatically (all the way in or out on one keystroke)
and if so which way.")

(defvar ringing:auto-hunt-position nil
  "Where the automatic hunting is.")

(defvar ringing:enable-auto-hunt nil
  "*Whether to allow automatic hunting.")

(defun ringing:next-auto-hunt-event ()
  "Provide the next automatic hunting event."
  (ringing:bell-number-to-char
   (case ringing:auto-hunt-state

     (hunt-out
      (if (= ringing:auto-hunt-position (1- ringing:current-stage))
	  (progn
	    (setq ringing:auto-hunt-state nil)
	    ringing:current-stage)
	(incf ringing:auto-hunt-position)))

     (hunt-in
      (if (= ringing:auto-hunt-position 2)
	  (progn
	    (setq ringing:auto-hunt-state nil)
	    1)
	(decf ringing:auto-hunt-position)))


     (treble-bob-out-1
      (setq ringing:auto-hunt-state 'treble-bob-out-2)
      (incf ringing:auto-hunt-position))

     (treble-bob-out-2
      (setq ringing:auto-hunt-state 'treble-bob-out-3)
      (incf ringing:auto-hunt-position))

     (treble-bob-out-3
      (setq ringing:auto-hunt-state 'treble-bob-out-4)
      (incf ringing:auto-hunt-position))

     (treble-bob-out-4
      (setq ringing:auto-hunt-state 
	    (if (= ringing:auto-hunt-position (1- ringing:current-stage))
		nil
	      'treble-bob-out-1))
      (decf ringing:auto-hunt-position))

     (t
      nil))))

(defvar ringing:timeout nil
  "The timeout for manual input, or nil if no timeout.
After this time (in seconds) without any input, the appropriate arrow key will
be \"typed\" for you. This behaviour is suppressed on the first row of a touch,
to give you time to read any header in the buffer, and so on.")

(defvar mini-ropes:use-for-input nil
  "Whether to use the mini-ropes for input instead of just display.")

(defun ringing:read-event-with-optional-timeout ()
  "Read an event, applying timeout if enabled."
  (if (and ringing:timeout ringing:use-timing-here)
      (progn
	(sit-for ringing:timeout)
	(if (input-pending-p)
	    (read-event)
	  'timeout))
    (read-event)))

(defun ringing:read-bell-event ()
  "Read an event referring to ringing a bell.
This will normally just do read-event, but may try to be real-time according
to the variable ringing:time-step."
  (if mini-ropes:use-for-input
      (let ((n (length ringing:most-recent-row))
	    (i 1)
	    (pulled nil))
	(require 'mini-ropes)
	(while (< i n)
	  (sit-for mini-ropes:input-one-bell-delay)
	  (when (input-pending-p) (setq pulled i))
	  (mini-ropes:ring (aref ringing:most-recent-row i))
	  (incf i))
	(if (null pulled)
	    'right
	  (+ ?0 pulled)))
    (if (or (not ringing:use-timing)
	    (null ringing:time-step)
	    (not ringing:use-timing-here))
	(if (and ringing:enable-auto-hunt
		 ringing:auto-hunt-state)
	    (ringing:next-auto-hunt-event)
	  (ringing:read-event-with-optional-timeout))
      (ringing:provide-event-from-timing))))

(defvar ringing:prompt-relative  "%s (1 .. <-- %d --> .. %s; ? for help): "
  "Prompt string for when the previous place is known.")

(defvar ringing:prompt-absolute  "%s (1..%s; ? for help): "
  "Prompt string for when the previous place is not known.")

(defun ringing:number-range-as-string (n)
  "Return the string describing 1..N as a string, in ringing terms."
  (if (stringp n)
      n
    (substring (format "1..%d,0,e,t,a,b,c,d" (min n 9))
	       0
	       (max 4 (+ 4 (* (- n 9) 2))))))

(defvar ringing:conductor-key-fn nil
  "A function to call on unrecognized keys when ringing interactively.
This allows the user to conduct, for example, it could implement keys
for bob, single, and change of method.
It is called with the input character, the result of calling ringing:bell-char-to-number
on that character, and the previous place in which the manual bell rang.")

(defvar ringing:error-off-end-of-row t
  "*Whether to give an error on going off the end of a row.
If this is nil, ignore movements that would take you off the end.")

(defun ringing:read-number-up-to (prompt limit &optional previous)
  "Using PROMPT, read a number in 1..LIMIT, using 0 E T A B C D above 9.
If PREVIOUS is non-nil, you can go up, down or make places from there
using the cursor keys or the rows of letters ZXC or QWD or QWE.
The prompt is used as the first format argument to ringing:prompt-relative
if PREVIOUS is given, otherwise for ringing:prompt-absolute.
Up arrow throws to 'stand.
Only those keys return a value from this function, any others are handled
within a loop that keeps on going until one of these is entered.
? displays help, and other keys are passed to ringing:conductor-key-fn
if it is non-nil."
  (catch 'got-number
    (while t
      (if (numberp previous)
	  (message ringing:prompt-relative prompt previous (ringing:number-range-as-string limit))
	(message ringing:prompt-absolute prompt (ringing:number-range-as-string limit)))
      (let* ((char (ringing:read-bell-event))
	     (charnumber (ringing:bell-char-to-number char previous)))
	(cond
	 ((and (numberp charnumber) (> charnumber 0) (<= charnumber limit))
	  (throw 'got-number charnumber))
	 ((and ringing:error-off-end-of-row (numberp charnumber))
	  (throw 'got-number 0))
	 ((eq charnumber 'help)
	  (ringing:bell-input-help previous))
	 ((eq charnumber 'stand)
	  (message "Stand!")
	  ;; main catcher for this is in ringing:ring-changes in ./beer.el
	  (throw 'stand 'user-stand))
	 ((eq charnumber 'timeout)
	  (throw 'got-number 'timeout))
	 ((not (null ringing:conductor-key-fn))
	  (funcall ringing:conductor-key-fn char charnumber previous)))))))

(defun ringing:bell-number-to-char (bellnumber)
  "Return BELLNUMBER's identification as a character -- digit or E or T."
  (cond ((<= bellnumber 9) (+ bellnumber ?0))
	((= bellnumber 10) ?0)
	((= bellnumber 11) ?E)
	((= bellnumber 12) ?T)
	((>= bellnumber 13) (+ (- bellnumber 13) ?A))))

(defun ringing:bell-char-to-number (bellchar &optional previous)
  "Return BELLCHAR as a number, allowing cursor arrow symbols if PREVIOUS is given.
The rows of letters \"zxc\" or \"qwd\" or \"qwe\" may be used instead of the cursor keys.
If the character can't be interpreted as a number, return nil."
  (case bellchar

    ((?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9)
     (- bellchar ?0))
    (?0
     10)
    (?E
     11)
    (?T
     12)

    ((?A ?B ?C ?D)
     (+ 13 (- bellchar ?A)))

    ((down ?x ?w)
     previous)
    ((left ?z ?q)
     (if (numberp previous)
	 (1- previous)
       nil))
    ((right ?c ?e ?d)
     (if  (numberp previous)
	 (1+ previous)
       nil))

    (C-left
     (setq ringing:auto-hunt-state 'hunt-in
	   ringing:auto-hunt-position
	   (if (numberp previous)
	       (1- previous)
	     nil)))
    (C-right
     (setq ringing:auto-hunt-state 'hunt-out
	   ringing:auto-hunt-position
	   (if  (numberp previous)
	       (1+ previous)
	     nil)))

    (M-right
     (setq ringing:auto-hunt-state 'treble-bob-out-3
	   ringing:auto-hunt-position
	   (if  (numberp previous)
	       (1+ previous)
	     nil)))

    (C-M-right
     (setq ringing:auto-hunt-state 'treble-bob-out-1
	   ringing:auto-hunt-position
	   (if  (numberp previous)
	       (1+ previous)
	     nil)))

    (timeout
     'timeout)
    (??
     'help)
    (up 'stand)
    (t (ringing:user-set-key bellchar previous))))

;;;; mostly contributed code:

(defun ringing:user-set-key (bellchar previous)
  "Try BELLCHAR for being a user-defined key, relative to PREVIOUS."
  (cond
   ((eq bellchar (aref ringing:keys 0))
    (if (numberp previous)
	 (1- previous)
       nil))
   ((eq bellchar (aref ringing:keys 1))
    previous)
   ((eq bellchar (aref ringing:keys 2))
    (if  (numberp previous)
	 (1+ previous)
       nil))
   ((eq bellchar (aref ringing:keys 3))
    'stand)))

(defvar ringing:keys '[[left] [down] [right] [up]]
  "User-defined keys used whilst ringing a method.")

(defun ringing:set-keys ()
  "Set up the Emacs ringing keys."
  (interactive)
  (let ((keys nil))
    (push (ringing:check-mouse (read-key-sequence "Press Key for moving DOWN a place")) keys)
    (push (ringing:check-mouse (read-key-sequence "Press Key for making a PLACE")) keys)
    (push (ringing:check-mouse (read-key-sequence "Press Key for moving UP a place")) keys)
    (push (ringing:check-mouse (read-key-sequence "Press Key for STAND")) keys)
    (setq ringing:keys (apply 'vector (nreverse keys)))
    ringing:keys))

(defun ringing:check-mouse (event)
  "If EVENT is a mouse event, return a vector of just the
key and swallow the release event."
  (if (vectorp event) ;; mouse event
      (let ((key (car (append event nil))))
        (if (equal (event-modifiers key) '(down))
            (read-event)) ;; throw away mouse key release event
        (vector (event-basic-type key)))
    event))

;; end of bell-numbers.el
