;;;; composition-checker.el -- part of John's elisp change-ringing system
;;; check compositions
;;; Time-stamp: <98/06/08 07:36:02 john>

(eval-when (compile load eval)
  (provide 'composition-checker)
  (require 'touches))

;;;; checking compositions

(defun ringing:touch-checker-display-function (row
					       remaining-changes
					       method-holder
					       rounds
					       so-far-holder
					       truth-holder)
  "Display function for checking touches"
  (let ((so-far (car so-far-holder)))
    (if (bufferp so-far)
	(progn
	  (set-buffer so-far)
	  (when (and (not (equal row rounds))
		     (save-excursion
		       (goto-char (point-min))
		       (search-forward row (point-max) t)))
	    (rplaca truth-holder nil))
	  (insert row ?\n))
      (progn
	(when (and (not (equal row rounds))
		   (member row so-far))
	  (rplaca truth-holder nil))
	(push row (car so-far-holder))))))

(defun ringing:check-touch (touch-calls &optional method so-far helper-function)
  "Check a touch of TOUCH-CALLS with METHOD giving the actions at each lead end.
If METHOD is not given, looks for a \"go\" call in the touch.
SO-FAR, if given, contains rows already rung (not including the initial rounds).
This may be a buffer, or a list of strings.
An optional HELPER-FUNCTION overrides the normal one.
For interactive calling, only TOUCH-CALLS and METHOD are used.
Return is a list of two items: whether the touch so far is true (as t
or nil), and the new value of SO-FAR, as a list of strings or as a
buffer, according to what was supplied. If a buffer is given, it is
extended to contain the extra rows. If a list is given, a list is
returned whose tail is the original one. Changes are added at the
front of this list.
SO-FAR may also be a vector of two elements, the first of which is
a list of changes so far, and the second of which is the method in
effect at that point. If it is provided in this form, the result is
in this form, but in a new vector (the original one is not modified)."
  (interactive
   (let* ((method (ringing:read-method-name "Check touch of: ")))
     (list
      method
      (read-from-minibuffer "Touch: "))))
  (when (null method)
    (setq method (ringing:starting-method-for-touch touch-calls)))
  (when (and (null method) (vectorp so-far))
    (setq method (aref so-far 1)))
  (let* ((method-data (ringing:find-named-method method))
	 (nbells (change-ringing-method-nbells method-data))
	 (method-name (ringing:method-name method))
	 (ringing:observation nbells)
	 (ringing:observation-number (ringing:bell-char-to-number ringing:observation))
	 (ringing:end-of-row-char ? )	; before making the rounds
	 (method-display:block-top-marker (make-marker))
	 (ringing:call-touch-at-changes-remaining 1)
	 (calls (ringing:convert-to-touch touch-calls method-data))
	 (place-marker (list 'work))
	 (count-holder (list 0))
	 (truth-holder (list t))
         (so-far-holder (list (if (arrayp so-far) (aref so-far 0) so-far)))
	 (method-holder (list method-data))
	 (ringing:backstroke nil)
	 (rounds (ringing:make-rounds nbells))
	 (starting-row
	  (cond
	   ((bufferp so-far)
	    (save-window-excursion
	      (set-buffer so-far)
	      (goto-char (point-max))
	      (if (re-search-backward " [0-9ETA-Deta-d]+ " (point-min) t)
		  (buffer-substring (match-beginning 0) (match-end 0))
		rounds)))
	   ((vectorp so-far)
	    (if (null (aref so-far 0))
		rounds
	      (car (aref so-far 0))))
	   (t
	    (if (null so-far)
		rounds
	      (car so-far))))))
    (save-window-excursion
      (if (bufferp so-far)
	  (switch-to-buffer so-far))
      (ringing:ring-changes starting-row
			    (if (change-ringing-method-special-start method-data)
				(change-ringing-method-special-start method-data)
			      (if (change-ringing-method-work method-data)
				  (change-ringing-method-work method-data)
				(change-ringing-method-plain method-data)))
			    (if helper-function
				helper-function
			      'ringing:show-touch-helper-function)
			    (list method-holder rounds place-marker calls count-holder truth-holder)
			    'ringing:touch-checker-display-function
			    (list method-holder rounds so-far-holder truth-holder)))
    (list (car truth-holder)
	  (if (vectorp so-far)
	      (vector (car so-far-holder)
		      (car method-holder))
	    (car so-far-holder)))))

(defun ringing:default-aesthetics (n)
  "Return the default aesthetics structure for N bells."
  (let* ((result nil)
	(max-pattern-base-raw (ringing:make-rounds n t))
	(max-pattern-base (substring max-pattern-base-raw 1 (- (length max-pattern-base-raw) 1)))
	(i 2)
	(k (1- n)))
    (while (< i k)
      (push (list (substring max-pattern-base (- i)) i) result)
      (push (list (concat (substring max-pattern-base (- i)) "\\>") (* 2 i)) result)
      (incf i))
    result))

(defun ringing:musicality (changes &optional aesthetics)
  "Check CHANGES against AESTHETICS (which defaults suitably).
An aesthetics specification is a list each of whose members is a list,
the first element of which is a regular expression and the second the
number representing how highly you value changes matching that regular
expression. The patterns are applied in order to each row, until one
matches. The overall result of this function is the sum of these values, 
divided by the number of changes."
  (unless aesthetics (setq aesthetics
			   (ringing:default-aesthetics
			    (ringing:n-bells-from-row-string (car changes)))))
  (/ (float (apply '+
		   (mapcar (lambda (row)
			     (catch 'matched
			       (dolist (a aesthetics)
				 (when (string-match (first a) row)
				   (throw 'matched (second a))))
			       0))
			   changes)))
     (length changes)))

;; end of composition-checker.el
