;;;; full-grid-mode.el -- part of John's elisp change-ringing system
;;; edit place notation, displaying the method dynamically
;;; Time-stamp: <99/05/17 17:36:25 johnstu>

(eval-when (compile load eval)
  (provide 'full-grid-mode)
  (require 'beer)
  (require 'method-lines)
  (require 'place-notation)
  (require 'ringing-methods)
  (require 'bell-colours)		; use yellow for giant cursor
  (require 'cl))

(defvar full-grid-mode:nbells nil
  "How many bells in this method.")

(defvar full-grid-mode:changes nil
  "The changes for this method.")

(defvar full-grid-mode:symmetric nil
  "Whether this method is obviously symmetric.
See also full-grid-mode:symmetric-string.")

(defvar full-grid-mode:symmetric-string ""
  "String saying Whether this method is obviously symmetric.
See also full-grid-mode:symmetric.")

(defvar full-grid-mode:row-string nil
  "String saying where in the line we are.")

(defvar full-grid-mode:place-string nil
  "String saying where in the row method we are.")

(defvar full-grid-mode:rounds nil
  "The rounds for this method.")

(defvar full-grid:whole-changes-string nil
  "The place notation string representing full-grid-mode:changes.")

(defvar full-grid-mode:place-notation-string nil
  "The latest result of converting the method back to place notation")

(defvar full-grid-mode:always-identify t
  "*Whether to try to keep track of what method this is.")

(defvar full-grid-mode:method nil
  "The method being edited -- a scratch structure for those that need it.")

(defvar place-notation-mode:method-holder nil
  "A holder list for the method being edited, as required for callbacks.")

(defvar full-grid-mode:keymap (make-keymap)
  "Keymap for full-grid-mode.")

(defvar full-grid-mode:read-only-keymap (make-keymap)
  "Keymap for full-grid-mode without the editing facilities.")

(suppress-keymap full-grid-mode:keymap)

(suppress-keymap full-grid-mode:read-only-keymap)

(define-key full-grid-mode:keymap "x" 'full-grid-mode:cross)
(define-key full-grid-mode:keymap "|" 'full-grid-mode:places)
;; (define-key full-grid-mode:keymap " " 'full-grid-mode:swap-cross-and-place)
(define-key full-grid-mode:keymap "*" 'full-grid-mode:symbol-here)
(define-key full-grid-mode:keymap [down-mouse-3] 'full-grid-mode:mouse-symbol-here)
(define-key full-grid-mode:keymap "s" 'full-grid-mode:save-to-file)
(define-key full-grid-mode:keymap "\M-p" 'full-grid-mode:print)
(define-key full-grid-mode:keymap "v" 'full-grid-mode:preview)

(defun full-grid-mode:define-key-both (keystring command)
  "Define KEYSTRING to do COMMAND in exploring and editing place notation."
  (if (listp keystring)
      (dolist (key keystring)
	(full-grid-mode:define-key-both key command))
    (progn
      (define-key full-grid-mode:keymap keystring command)
      (define-key full-grid-mode:read-only-keymap keystring command))))

(full-grid-mode:define-key-both "?" 'full-grid-mode:internal-info)
(full-grid-mode:define-key-both '("n" "\C-n" [down]) 'full-grid-mode:next-line)
(full-grid-mode:define-key-both '("p" "\C-p" [up]) 'full-grid-mode:previous-line)
(full-grid-mode:define-key-both '("f" "\C-f" [right]) 'full-grid-mode:forward-char)
(full-grid-mode:define-key-both '("b" "\C-b" [left]) 'full-grid-mode:backward-char)
(full-grid-mode:define-key-both '("a" "\C-a") 'full-grid-mode:beginning-of-line)
(full-grid-mode:define-key-both '("e" "\C-e") 'full-grid-mode:end-of-line)
(full-grid-mode:define-key-both '("<" "\M-<") 'full-grid-mode:beginning-of-buffer)
(full-grid-mode:define-key-both '(">" "\M->") 'full-grid-mode:end-of-buffer)
(full-grid-mode:define-key-both [down-mouse-1] 'full-grid-mode:mouse-point-here)
(full-grid-mode:define-key-both "%" 'full-grid-mode:mark-bell)

(mapcar 'make-variable-buffer-local
	'(full-grid-mode:nbells
	  full-grid-mode:changes
	  full-grid-mode:rounds
	  full-grid-mode:method
	  place-notation-mode:method-holder))

;;;; top-level functions

(defvar full-grid-mode:open-method-buffer-hooks nil
  "Hooks to run on starting a place notation edit buffer.")

(defun full-grid-mode (&optional read-only)
  "Major mode for editing place notation.
With optional arg, read-only, as a method explanation.

If editing the method, the following commands are available:
\\{full-grid-mode:keymap}

If simply exploring the method, the following commands are available:
\\{full-grid-mode:read-only-keymap}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map (if read-only
		     full-grid-mode:read-only-keymap
		   full-grid-mode:keymap))
  (buffer-disable-undo)
  (toggle-read-only 1)
  (setq mode-line-format (full-grid:hack-mode-line default-mode-line-format))
  ;; (unless read-only (add-hook 'post-command-hook 'full-grid-mode:move-underlining nil 'local))
  (setq major-mode 'full-grid-mode
	mode-name "Full-Grid"
	full-grid-mode:underlined-start nil
	)
  (message "post-command-hook is %s in buffer %s" post-command-hook (buffer-name))
  ;; (debug nil post-command-hook buffer-name)
  (run-hooks 'full-grid-mode:open-method-buffer-hooks)
  )

(defvar full-grid:start-edit-hooks nil
  "Hooks to run on starting to edit a method.")

(defun full-grid-edit-or-explore (n original read-only)
  "Edit or explore place notation on N bells starting with ORIGINAL, perhaps READ-ONLY."
  (let ((ringing:end-of-row-char ? )	; before making the rounds
	
	)
    (switch-to-buffer
     (ringing:get-buffer-create
      (if read-only
	  "*place notation grid*"
	"*place notation edit*")))
    (toggle-read-only -1)
    (erase-buffer)
    (full-grid-mode read-only)
    (setq full-grid-mode:nbells n
	  full-grid-mode:changes (place-notation:parse-string original n)
	  full-grid:whole-changes-string (place-notation:create-fully-expanded-from-changes
					      full-grid-mode:changes)
	  full-grid-mode:rounds (ringing:make-rounds n)
	  method-display:block-top-marker (make-marker)
	  full-grid-mode:method (make-change-ringing-method
				     :title "Unnamed"
				     :nbells full-grid-mode:nbells
				     :plain full-grid-mode:changes
				     :bob nil
				     :single nil
				     :original-description-string "Scratch method")
	  place-notation-mode:method-holder (list full-grid-mode:method)
	  ringing:most-recent-change nil)
    (let* ((ringing:observation-number full-grid-mode:nbells)
	   (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	   (ringing:draw-crossings t)
	   (ringing:show-lines nil)
	   (ringing:show-numbers t)
	   (ringing:show-lines-with-lines nil)
	   (ringing:auto-ring-inter-row-delay nil)
	   (method-display:main-display-function 'method-display:single-column-display-function))
      (toggle-read-only -1)
      (insert (format "%s of %s (%s)"
		      (if read-only "Exploration" "Edit")
		      original
		      full-grid:whole-changes-string))

 ;; (insert "\nhere is some more string\n")

      (set-marker method-display:block-top-marker (point))
      (require 'ermm)
      (ermm:init nil)
      (if full-grid-mode:always-identify
	  (insert (format "\nMethod: %s\n" (method-library:find-name-for-method full-grid-mode:method)))
	(insert "\n\n"))
      (ringing:insert-row-lines full-grid-mode:rounds)
      (full-grid-mode:edit-ring-from-here-onwards full-grid-mode:rounds
						      full-grid-mode:changes)
      (unless read-only (run-hooks 'full-grid:start-edit-hooks))
      (goto-char (point-min))
      (beginning-of-line 3)
      (forward-char 2)
 ;; (recursive-edit)
      (full-grid-mode:move-underlining)
      (toggle-read-only 1)
      (full-grid-mode:update-status))))

(defun full-grid-edit (n &optional original)
  "Edit place notation on N bells starting with ORIGINAL."
  (interactive "nNumber of bells: 
sInitial place notation: ")
  (if (null original) (setq original "-."))
  (full-grid-edit-or-explore n original nil))

(defun full-grid-explore (n place-notation)
  "Display N-bell method with PLACE-NOTATION."
  (interactive "nNumber of bells: 
sInitial place notation: ")
  (full-grid-edit-or-explore n place-notation t))

(defun full-grid:edit-based-on (method-name)
  "Edit a method based on METHOD-NAME (typically from the libraries)."
  (interactive (list (ringing:read-method-name "Edit a method based on: ")))
  (let* ((method (ringing:find-named-method method-name))
	 (nbells (change-ringing-method-nbells method)))
    (full-grid-edit nbells
			(place-notation:recreate-string-from-changes (change-ringing-method-plain method)))))

;;;; status of the method

(defun full-grid:hack-mode-line (template)
  "Hack TEMPLATE to be as needed for place-notation editing."
  (cond
   ((null template)
    nil)
   ((eq (car template) 'mode-line-process)
    (append '(full-grid-mode:symmetric-string) (full-grid:hack-mode-line (cdr template))))
   ((and (consp (car template))
	 nil				; I can't see a chance for us to update it each time, so don't bother
	 (eq (car (car template)) 'line-number-mode))
    (cons 'full-grid-mode:row-string (full-grid:hack-mode-line (cdr template))))
   ((and (consp (car template))
	 nil				; I can't see a chance for us to update it each time, so don't bother
	 (eq (car (car template)) 'column-number-mode))
    (cons 'full-grid-mode:place-string (full-grid:hack-mode-line (cdr template))))
   (t (cons (car template)
	 (full-grid:hack-mode-line (cdr template))))))

(defun full-grid-mode:update-status ()
  "Update the status of the method edit."
  (setq full-grid-mode:symmetric (ringing:symmetric-p full-grid-mode:changes)
	full-grid-mode:symmetric-string (if full-grid-mode:symmetric
					    ":symmetric" ":asymmetric")
	;; full-grid-mode:row-string (format "R%d" pnm:edit-row-number)
	;; full-grid-mode:place-string (format "P%d" pnm:left-place)
	)
  (force-mode-line-update))

;;;; track the cursor with an underlined pair of bells

(defconst full-grid-mode:underlined (ringing:copy-face 'default 'full-grid-mode:underlined))
(ringing:set-face-underline-p full-grid-mode:underlined t)
(ringing:set-face-background full-grid-mode:underlined ringing:bad-background-colour)

(defconst full-grid-mode:marked (ringing:copy-face 'default 'full-grid-mode:marked))
(ringing:set-face-foreground full-grid-mode:marked "blue")

(defvar full-grid-mode:underlined-start nil
  "The place which is underlined.")

(defun full-grid-mode:move-underlining ()
  "Move the underlining to where it is needed."
  ;; (message "Moving underlining... ") (sit-for 1) (message nil)
  (full-grid-mode:clear-bell-colour)
  (save-excursion
    (let* ((was-read-only buffer-read-only)
	   (buffer-read-only nil))
      (when full-grid-mode:underlined-start
	(put-text-property full-grid-mode:underlined-start
			   (+ full-grid-mode:underlined-start 3)
			   'face 'default))
      (full-grid:set-line-variables)
      (full-grid-mode:mark-bell (ringing:bell-char-to-number (full-grid-mode:bell-at-point)))
      (let ((bol (save-excursion (beginning-of-line 1) (point))))
	(setq full-grid-mode:underlined-start
              (+ bol (1- (* 2 pnm:left-place))))
	(put-text-property full-grid-mode:underlined-start
			   (+ full-grid-mode:underlined-start 3)
			   'face full-grid-mode:underlined)))))

(defun full-grid-mode:next-line (arg)
  "Like next-line but updates full-grid cursor."
  (interactive "p")
  (next-line arg)
  (full-grid-mode:move-underlining))

(defun full-grid-mode:previous-line (arg)
  "Like previous-line but updates full-grid cursor."
  (interactive "p")
  (previous-line arg)
  (full-grid-mode:move-underlining))

(defun full-grid-mode:forward-char (arg)
  "Like forward-char but updates full-grid cursor."
  (interactive "p")
  (when (< (current-column) (1- (* full-grid-mode:nbells 2)))
    (forward-char arg))
  (full-grid-mode:move-underlining))

(defun full-grid-mode:backward-char (arg)
  "Like backward-char but adjusts full-grid cursor."
  (interactive "p")
  (when (> (current-column) 1)
    (backward-char arg))
  (full-grid-mode:move-underlining))

(defun full-grid-mode:beginning-of-line (arg)
  "Like beginning-of-line but adjusts full-grid cursor."
  (interactive "p")
  (beginning-of-line arg)
  (move-to-column 1)
  (full-grid-mode:move-underlining))

(defun full-grid-mode:end-of-line (arg)
  "Like end-of-line but adjusts full-grid cursor."
  (interactive "p")
  (end-of-line arg)
  (move-to-column (1- (* full-grid-mode:nbells 2)))
  (full-grid-mode:move-underlining))

(defun full-grid-mode:beginning-of-buffer ()
  "Go to the beginning of the buffer."
  (interactive)
  (goto-char (point-min))
  (beginning-of-line 2)
  (forward-char 3)
  (full-grid-mode:move-underlining))
  
(defun full-grid-mode:end-of-buffer ()
  "Go to the end of the buffer."
  (interactive)
  (goto-line (+ 3 (* 2 (length full-grid-mode:changes))))
  (move-to-column (1- (* full-grid-mode:nbells 2)))
  (full-grid-mode:move-underlining))

;;;; Revising the line from the edit downwards

(defun full-grid-mode:edit-helper-function (row
						remaining-changes
						method
						rounds
						place-marker)
  "Helper function for editing lines."
  (if (string= row rounds)
      nil
    (if remaining-changes
	t
      nil)))

(defun full-grid-mode:edit-ring-from-here-onwards (row remaining-changes)
  "Starting from ROW, ring REMAINING-CHANGES with output as used in the place notation editor."
  (when remaining-changes
    (message "Redrawing lines...")
    (save-excursion
      (let* ((ringing:observation-number full-grid-mode:nbells)
	     (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	     (nbells full-grid-mode:nbells)
	     (ringing:draw-crossings t)
	     (ringing:show-lines nil)
	     (ringing:show-numbers t)
	     (ringing:show-lines-with-lines nil)
	     (ringing:auto-ring-inter-row-delay nil)
	     (place-marker (list 'work))
	     (callback-arglist (list place-notation-mode:method-holder full-grid-mode:rounds place-marker))
	     (ringing:write-place-notation-beside t))
	(ringing:ring-changes row
			      remaining-changes
			      'full-grid-mode:edit-helper-function
			      callback-arglist
			      'method-display:single-column-display-function
			      callback-arglist)
	(set-marker method-display:block-top-marker nil)))
    (full-grid-mode:update-status)
    (message "Redrawing lines... done")))

;;;; framework for performing each edit and the associated update

(defvar pnm:edit-cursor-line nil
  "The line the cursor is in at the start (and, usually, end) of the edit.")

(defvar pnm:edit-cursor-column nil
  "The column the cursor is in at the start (and, usually, end) of the edit.")

(defvar pnm:edit-row-number nil
  "The row number for editing.")

(defvar pnm:left-place nil
  "The place number for editing.")

(defvar pnm:right-place nil
  "The adjacent place for editing.")

(defvar pnm:edit-change-data nil
  "The change data array.")

(defvar pnm:edit-places-string nil
  "The places string.")

(defvar pnm:row-for-restart nil
  "The current row for restarting ringing.")

(defvar pnm:edit-on-row nil
  "Whether we are currently on a row (else on the change between them).")

(defvar pnm:edit-on-place nil
  "Whether we are currently on a place (else in the space between them).")

(defvar pnm:edit-on-place-to-left-of-cross nil
  "Whether we are currently on a left-hand place (only relevant when pnm:edit-on-place).")

(defvar pnm:edit-on-place-to-right-of-cross nil
  "Whether we are currently on a right-hand place (only relevant when pnm:edit-on-place).")

(defvar pnm:debug nil
  "Whether to do various debugging things")

(defun full-grid-mode:enable-debug (off)
  "Set whether to output debug messages for full-grid editing.
With prefix arg, disable debug, else enable it."
  (interactive "P")
  (if off
      (setq pnm:debug nil)
    (setq pnm:debug t))
  (message "full-grid-mode debugging %s" (if pnm:debug "enabled" "disabled")))

(defmacro pnm:bell-at (place)
  "Return the bell at PLACE."
  `(aref pnm:row-for-restart ,place))

(defmacro pnm:next-at (position)
  "Return the next position for the bell at POSITION."
  `(aref pnm:edit-change-data ,position))

(defmacro pnm:making-places (bell)
  "Return whether BELL is making places for this change."
  `(= ,bell (pnm:next-at ,bell)))

(defmacro pnm:going-up (bell)
  "Return whether BELL is making places for this change."
  `(= (1+ ,bell) (pnm:next-at ,bell)))

(defmacro pnm:going-down (bell)
  "Return whether BELL is making places for this change."
  `(= (1- ,bell) (pnm:next-at ,bell)))

(defmacro pnm:set-next (where next)
  "Say that bell in place WHERE goes to NEXT."
  `(aset pnm:edit-change-data ,where ,next))

(defun full-grid-mode:bell-at-point ()
  "Return the bell character at or near point."
  (full-grid:set-line-variables)
  (let* ((restart (full-grid-mode:row-for-restart))
	 (row (ringing:ring-one-change restart pnm:edit-change-data))
	 (pos (min (1+ (/ (current-column) 2)) full-grid-mode:nbells)))
    (message "Restart \"%s\", row=\"%s\", pos %d" restart row pos)
    (aref row pos)))

(defun full-grid:set-line-variables ()
  "Set some internal variables according to point.
Don't do anything that requires closing afterwards."
  (setq pnm:edit-cursor-line (count-lines 
			      ;;; method-display:block-top-marker
			      (point-min)
			      (point)
			      )
	pnm:edit-cursor-column (current-column)
	pnm:edit-row-number (/ (- pnm:edit-cursor-line 4) 2)
	pnm:edit-on-row (oddp pnm:edit-cursor-line)
	pnm:left-place (/ pnm:edit-cursor-column 2)
	pnm:right-place (1+ pnm:left-place)
	pnm:edit-on-place (oddp pnm:edit-cursor-column)
	pnm:edit-change-data (nth pnm:edit-row-number full-grid-mode:changes)
	pnm:edit-places-string (ringing:change-to-place-notation pnm:edit-change-data)
	pnm:edit-on-place-to-left-of-cross (and pnm:edit-on-place
						(pnm:going-up pnm:right-place))
	pnm:edit-on-place-to-right-of-cross (and pnm:edit-on-place
						 (pnm:going-down pnm:right-place))
	pnm:row-for-restart (full-grid-mode:row-for-restart))
  (unless (< pnm:left-place full-grid-mode:nbells)
    (decf pnm:left-place)
    (decf pnm:right-place))
  (when (zerop pnm:left-place)
    (incf pnm:left-place)
    (incf pnm:right-place)))

(defun full-grid-mode:open-line-edit ()
  "Set up some internal variables according to point.
This should be followed with full-grid-mode:close-line-edit.
This is useful at the start of most edit commands."
  (full-grid-mode:move-underlining)
  (full-grid:set-line-variables)
  (when pnm:debug
    (with-output-to-temp-buffer "*place notation edit debug*"
      (princ (place-notation:debug-readifier pnm:edit-change-data)) (princ "\n")
      (princ (make-string  pnm:left-place ?-)) (princ "^\n")
      (princ (format "Cursor is on places %d (bell %c, moving to %c) and %d (bell %c, moving to %c)\n"
		     pnm:left-place (pnm:bell-at pnm:left-place)
		     (place-notation:debug-readify-char (aref pnm:edit-change-data pnm:left-place))
		     pnm:right-place (pnm:bell-at pnm:right-place)
		     (place-notation:debug-readify-char (aref pnm:edit-change-data  pnm:right-place))))
      (princ (full-grid-mode:internal-info-string)) (princ "\n")
      (princ (if pnm:edit-on-place "on place: " "on crossing: "))
      (when pnm:edit-on-place-to-left-of-cross (princ "(place to left of cross): "))
      (when pnm:edit-on-place-to-right-of-cross (princ "(place to right of cross): "))
      (princ (if (pnm:making-places pnm:left-place)
		 "left making places; " "left crossing; "))
      (princ (if (pnm:making-places pnm:right-place)
		 "right making places\n" "right crossing\n")))))

(defun full-grid-mode:row-for-restart ()
  "Return the row for restarting the method drawing."
  (save-excursion
    (if (= pnm:edit-row-number 0)
	full-grid-mode:rounds
      (beginning-of-line (if pnm:edit-on-row -1 0))
      (let ((result '(? )))
	(while (not (looking-at "$"))
	  (if (looking-at "[0-9eta-d]")
	      (push (char-after (point)) result))
	  (goto-char (1+ (point))))
	(push ?  result)
	(coerce (reverse result) 'string)))))

(defun full-grid-mode:change-method-from-here ()
  "Alter the method to be consistent from the current editing point."
  (save-excursion
    (rplaca (nthcdr pnm:edit-row-number full-grid-mode:changes)
	    pnm:edit-change-data)
    (setq pnm:edit-change-data (nth pnm:edit-row-number full-grid-mode:changes))
    (let ((restart-changes (nthcdr (- pnm:edit-row-number 0) full-grid-mode:changes)))
      (end-of-line (if pnm:edit-on-row -1 0))
      (delete-region (point) (point-max))
      (full-grid-mode:edit-ring-from-here-onwards pnm:row-for-restart restart-changes)
      ;; (message "restart-row=%s restart-changes=%s" pnm:row-for-restart restart-changes)
      )))

(defun full-grid-mode:close-line-edit ()
  "Various actions for the end of an edit."
  (let ((was-read-only buffer-read-only)
	(buffer-read-only nil)) 
    (full-grid-mode:change-method-from-here)
    (save-excursion
      (goto-char (point-min))
      (when (re-search-forward "Edit of \\(.+\\)" (point-max) t)
	(delete-region (match-beginning 1) (match-end 1))
	(setq full-grid-mode:place-notation-string (place-notation:recreate-string-from-changes full-grid-mode:changes)
	      full-grid:whole-changes-string (place-notation:create-fully-expanded-from-changes
					      full-grid-mode:changes))
	(setf (change-ringing-method-plain full-grid-mode:method) full-grid-mode:changes)
	(insert full-grid-mode:place-notation-string " (" full-grid:whole-changes-string")"))))



    (when full-grid-mode:always-identify
      (goto-char (point-min))
      (if (re-search-forward "Method: \\(.+\\)$" (point-min) t)
	(replace-match (method-library:find-name-for-method full-grid-mode:method) t t nil 1)
	(message "Could not find method name to replace")
	))

  (goto-line pnm:edit-cursor-line)
  (move-to-column pnm:edit-cursor-column)
  (full-grid-mode:move-underlining))

(defmacro full-grid-mode:with-open-line-edit (&rest edit-command-body)
  "Perform EDIT-COMMAND-BODY with the full-grid editing variables set up
and the result being taken into account."
  `(progn
     (full-grid-mode:open-line-edit)
     (prog1
	 (progn
	   ,@edit-command-body)
       (full-grid-mode:close-line-edit))))

;;;; editing (etc) commands

(defun full-grid-mode:internal-info-string ()
  "Return a string giving a line of useful information about the current point."
  (format "Line %d %s %s column %d ==> change %d place %d ==> places %s; restart \"%s\""
	  (count-lines (point-min) (point))
	  (if pnm:edit-on-row "(on row)" "(on change between rows)")
	  (if pnm:edit-on-place "(on place)" "(on gap/crossing)")
	  (current-column)
	  pnm:edit-row-number pnm:left-place pnm:edit-places-string
	  (full-grid-mode:row-for-restart)))

(defun full-grid-mode:internal-info ()
  "Display internal information, mostly for development and debugging purposes."
  (interactive)
  (full-grid-mode:open-line-edit)
  (message (full-grid-mode:internal-info-string))
  (full-grid-mode:close-line-edit))

(defun full-grid-mode:trace-bell (ibell form)
  "Trace the path of IBELL by evalling FORM with point set to each of its positions."
  (save-excursion
    (goto-char (point-min))
    (let ((buffer-read-only nil))
      (beginning-of-line 3)
      (let ((place ibell))
	(dolist (change full-grid-mode:changes)
	  (move-to-column (1- (* place 2)))
	  (setq place (aref change place))
	  (eval form)
	  (forward-line 2))))))

(defmacro full-grid-mode:map-bell-positions (ibell &rest form)
  "With point set to each position of IBELL in turn, eval FORM."
  `(full-grid-mode:trace-bell ,ibell '(progn ,@form)))
			     
(defvar full-grid-mode:marked-bell nil
  "Which bell has most recently been coloured in.")

(defun full-grid-mode:clear-bell-colour ()
  "Clear any bell colouring."
  (when full-grid-mode:marked-bell
    (full-grid-mode:map-bell-positions
     full-grid-mode:marked-bell
     (put-text-property (point) (1+ (point)) 'face 'default (current-buffer)))
    (setq full-grid-mode:marked-bell nil)))

(defun full-grid-mode:mark-bell (bell)
  "Colour in the path of BELL."
  (interactive
   (if (eq major-mode 'full-grid-mode)
       (progn
	 (full-grid:set-line-variables)
	 (list (ringing:bell-char-to-number (full-grid-mode:bell-at-point))))
     (list (string-to-int (read-from-minibuffer "Mark bell: " 1)))))
  (full-grid-mode:clear-bell-colour)
  (full-grid-mode:map-bell-positions
   bell
   (put-text-property (point) (1+ (point)) 'face full-grid-mode:marked (current-buffer)))
  (setq full-grid-mode:marked-bell bell))

(defun full-grid-mode:cross ()
  "Make a crossing."
  (interactive)
  (full-grid-mode:with-open-line-edit
   (message "X")
   (if (pnm:making-places pnm:left-place)
       (progn
	 (ringing:message "Lower bell already making places")
	 (if (pnm:making-places pnm:right-place)
	     (progn
	       (ringing:message "Lower bell already making places and upper bell already making places")
	       (pnm:set-next pnm:left-place pnm:right-place)
	       (pnm:set-next pnm:right-place pnm:left-place)
	       )
	   (progn
	     (ringing:message "Lower bell already making places but upper bell not already making places")
	     ))
	 )
     (progn
       (ringing:message "Lower bell not already making places")
       (if (pnm:making-places pnm:right-place)
	   (progn
	     (ringing:message "Lower bell not already making places but upper bell already making places")
	     )
	 (progn
	   (ringing:message "Lower bell not already making places and upper bell not already making places")
	   ))))))

(defun full-grid-mode:places ()
  "Make a bell (or two) make places."
  (interactive)
  (full-grid-mode:with-open-line-edit
   (ringing:message "| |")
   (if (pnm:making-places pnm:left-place)
       (progn
	 (ringing:message "Lower bell already making places")
	 (if (pnm:making-places pnm:right-place)
	     (progn
	       (ringing:message "Lower bell already making places and upper bell already making places")
	       )
	   (progn
	     (ringing:message "Lower bell already making places but upper bell not already making places")
	     ))
	 )
     (progn
       (ringing:message "Lower bell not already making places")
       (if (pnm:making-places pnm:right-place)
	   (progn
	     (ringing:message "Lower bell not already making places but upper bell already making places")
	     )
	 (progn
	   (ringing:message "Lower bell not already making places and upper bell not already making places")
	   (pnm:set-next pnm:left-place pnm:left-place)
	   (pnm:set-next pnm:right-place pnm:right-place)
	   ))))))

(defun full-grid-mode:swap-cross-and-place ()
  "Change cross-place to place-cross."
  (interactive)
  (full-grid-mode:with-open-line-edit
   (ringing:message "|  X <--> X  |")

   )
  (error "Sorry, not implemented yet")
  )

(defun full-grid-mode:ensure-on-change ()
  "Give an error if not on a change."
  (full-grid:set-line-variables)  
  (when pnm:edit-on-row
    (error "This command may only be used on a change, not a row.")))

(defun pnm:is-treble (position)
 "Returns whether the bell at POSITION is the treble."
 (= (pnm:bell-at position) ?1))

(defvar full-grid:ok-to-disturb-treble nil
  "*Whether the user to change the treble path without being warned.")

(defun pnm:check-edit (&rest positions)
  "Check that it's OK to change any of POSITIONS."
  (if (or full-grid:ok-to-disturb-treble
	  (not (eval (cons 'or (mapcar 'pnm:is-treble positions))))
	  (y-or-n-p "OK to disturb treble path? "))
      t
    nil))

(defun full-grid-mode:symbol-here ()
  "Re-arrange the place notation such that there is a symbol at point."
  (interactive)
  (full-grid-mode:ensure-on-change)
  (full-grid-mode:with-open-line-edit
   (let* ((left-making-places (pnm:making-places pnm:left-place))
	  (left-crossing (not left-making-places))
	  (right-making-places (pnm:making-places pnm:right-place))
	  (right-crossing (not right-making-places))
	  (on-place pnm:edit-on-place)
	  (on-crossing (not on-place))
	  )
     (when pnm:debug
       (ringing:message "%s; %s; %s"
			(if on-place "on place" "on crossing")
			(if left-making-places "left making places" "left crossing")
			(if right-making-places "right making places" "right crossing")))
     (cond
      ((and on-place
	    left-making-places
	    right-making-places)
       (when pnm:debug (ringing:message "Nothing to do (on pair of places)")))
      ((and on-place
	    left-making-places
	    right-crossing)
       (let* ((right pnm:right-place)
	      (far-right (1+ right)))
	 (when pnm:debug (ringing:message "\"| *X \" --> \"| | |\""))
         (when (pnm:check-edit right far-right)
	   (pnm:set-next right right)
	   (pnm:set-next far-right far-right))))
      ((and on-place
	    left-crossing
	    right-making-places)
       (when pnm:debug (ringing:message "Nothing to do (on place)")))
      ((and on-place
	    left-crossing
	    right-crossing)
       (if pnm:edit-on-place-to-right-of-cross
	   (let ((left pnm:left-place)
		 (right pnm:right-place))
	     (when pnm:debug (ringing:message "\" X*  X \" --> \"| |  X \""))
	     (when (pnm:check-edit left right)
	       (pnm:set-next left left)
	       (pnm:set-next right right)))
	 (let* ((right (if (= pnm:left-place 1)
			   pnm:left-place
			 pnm:right-place))
		(far-right (1+ right)))

	   ;; This one gives a problem when done in the leftmost column
           ;; I think this is now fixed by the (if (= left 1) left right) above
 
	   (when pnm:debug
	     (ringing:message "(left=%d) right=%d far-right=%d" pnm:left-place right far-right)
	     (ringing:message "Old change data was %s" pnm:edit-change-data)
	     (ringing:message "\" X  *X \" --> \" X  | |\"")
	     )
	   (when (pnm:check-edit right far-right)
	     (pnm:set-next right right)
	     (pnm:set-next far-right far-right))
	   (when pnm:debug (ringing:message "New change data is %s" pnm:edit-change-data))
	   )))
      ((and on-crossing
	    left-making-places
	    right-making-places)
       (let ((left pnm:left-place)
	     (right pnm:right-place))
	 (when pnm:debug (ringing:message "\"|*|\" --> \" X \""))
	 (when (pnm:check-edit left right)
	   (pnm:set-next left right)
	   (pnm:set-next right left))))
      ((and on-crossing
	    left-making-places
	    right-crossing)
       (let* ((left pnm:left-place)
	      (right pnm:right-place)
	      (far-right (1+ right)))
	 (when pnm:debug (ringing:message "\"|* X \" --> \" X  |\""))
	 (when (pnm:check-edit left right far-right)
	   (pnm:set-next left right)
	   (pnm:set-next right left)
	   (pnm:set-next far-right far-right))))
      ((and on-crossing
	    left-crossing
	    right-making-places)
       (let* ((left pnm:left-place)
	      (far-left (1- left))
	      (right pnm:right-place))
	 (when pnm:debug (ringing:message "\" X *|\" --> \"|  X \""))
	 (when (pnm:check-edit far-left left right)
	   (pnm:set-next far-left far-left)
	   (pnm:set-next left right)
	   (pnm:set-next right left))))
      ((and on-crossing
	    left-crossing
	    right-crossing)
       (if (and (pnm:going-up pnm:left-place)
		(pnm:going-down pnm:right-place))
	   (when pnm:debug (ringing:message "Nothing to do -- already on crossing"))
	 (let* ((left pnm:left-place)
		(far-left (1- left))
		(right pnm:right-place)
		(far-right (1+ right)))
	   (when pnm:debug (ringing:message "\" X * X \" --> \"|  X  |\""))
	   (when (pnm:check-edit far-left far-right left right)
	     (pnm:set-next far-left far-left)
	     (pnm:set-next left right)
	     (pnm:set-next right left)
	     (pnm:set-next far-right far-right)))))))))

(defun full-grid-mode:mouse-symbol-here (click)
  "Put a symbol where the mouse was clicked, on a change.
This makes the minimum alterations to the change to get a symbol into that
position (i.e. to make a crossing or place there)."
  (interactive "@e")
  (goto-char (posn-point (event-start click)))
  (full-grid:set-line-variables)
  (full-grid-mode:symbol-here))

(defun full-grid-mode:mouse-point-here (click)
  "Put the point where the mouse was clicked, and update cursor."
  (interactive "@e")
  (goto-char (posn-point (event-start click)))
  (full-grid-mode:move-underlining))

;;;; Writing edited methods out to the rest of the system

(defun full:grid-mode:current-method (&optional name)
  "Return the current method, as a change-ringing-method structure. Optionally give it NAME."
  ;; In the current implementation, this is updated after each change anyway
  (let ((method (copy-change-ringing-method full-grid-mode:method)))
    (setf (change-ringing-method-title method)
	  (if name
	      name
	    (gensym "anon-method-")))
    method))

(defun full-grid-mode:save-to-file (file method-name)
  "Save the current method to FILE, giving it METHOD-NAME."
  (interactive "*FSave method to file:
sName to save method under: ")
  (save-window-excursion
    (find-file file)
    (save-excursion
      (goto-char (point-min))
      (if (re-search-forward (format "^%s " method-name) (point-max) t)
	  (if (y-or-n-p (format "Replace existing definition of %s in %s? " method-name file))
	      (let ((start (progn (beginning-of-line 1) (point))))
		(beginning-of-line 2)
		(delete-region start (point))
		(mSiRiL-mode:insert-description (full:grid-mode:current-method)))
	    (message "File not changed"))
	(progn
	  (while (and (re-search-forward "^\\([a-z]+\\) " (point-max) t)
		      (string< (buffer-matched 1) method-name)))
	  (mSiRiL-mode:insert-description (full:grid-mode:current-method)))))))

(defun full-grid-mode:get-postscript-parameters-interactively ()
  "Get the parameters for printing a method from the full grid editor."
  (let* ((title (read-from-minibuffer "Title for method: "))
	 (params (ringing:get-postscript-parameters-interactively)))
    (cons title params)))

(defun full-grid-mode:print-or-preview (method output-function
					       style style-parameters
					       header-style header-style-parameters
					       grid-style grid-style-parameters)
  "Print or preview METHOD, via OUTPUT-FUNCTION  and with parameters as specified."
  (save-window-excursion
    (let* ((method-name (ringing:method-name method))
	   (ps-file-name (expand-file-name (format "%s.ps"
						   (ringing:name-as-file-name method-name))
					   mSiRiL-mode:ps-temp-files-directory)))
      (find-file ps-file-name)
      (ringing:postscript-for-method method (current-buffer)
				     style style-parameters
				     header-style header-style-parameters
				     grid-style grid-style-parameters)
      (basic-save-buffer)
      (funcall output-function ps-file-name)
      (unless ringing:keep-postscript-buffers (kill-buffer nil))
      (unless ringing:keep-postscript-files (delete-file ps-file-name)))))
  
(defun full-grid-mode:print (name
			     style style-parameters
			     header-style header-style-parameters
			     grid-style grid-style-parameters)
  "Print current method, titling it with NAME and with parameters as specified."
  (interactive (full-grid-mode:get-postscript-parameters-interactively))
  (full-grid-mode:print-or-preview (full:grid-mode:current-method name)
				   'ringing:send-postscript-to-printer
				   style style-parameters
				   header-style header-style-parameters
				   grid-style grid-style-parameters))

(defun full-grid-mode:preview (name
			       style style-parameters
			       header-style header-style-parameters
			       grid-style grid-style-parameters)
  "Preview current method, titling it with NAME and with parameters as specified."
  (interactive (full-grid-mode:get-postscript-parameters-interactively))
  (full-grid-mode:print-or-preview (full:grid-mode:current-method name)
				   'ringing:send-postscript-to-previewer
				   style style-parameters
				   header-style header-style-parameters
				   grid-style grid-style-parameters))
  
;;; end of full-grid-mode.el
