;;;; handbells.el -- part of John's elisp change-ringing system
;;; Basic ringing on handbells
;;; Time-stamp: <99/09/16 08:11:21 johnstu>

(eval-when (compile load eval)
  (provide 'handbells)
  (require 'plain-courses)		; requires our other requisites for us
  )

(defun handbells:right-hand-bell (bell)
  "Return the right-hand bell of the pair containing BELL."
  (cond
   ((oddp bell) bell)
   ((zerop bell) 1)
   (t (1- bell))))

(defvar handbells:previous-rows nil
  "The most recently rung rows, for handbells input help.")

(defvar handbells:help-level 0
  "The level of help displayed so far, for this row.
Successive requests for help (without doing anything else inbetween)
produce some more help.")

(defvar handbells:show-positions-in-prompt t
  "*Whether to show the current positions in the prompt for handbell ringing.")

(defvar handbells:show-fully nil
  "*Whether to show the lines fully instead of just the current row.")

(defun handbells:display-function (row
				   remaining-changes
				   method-holder
				   rounds
				   place-marker
				   right-hand-bell-char
				   left-hand-bell-char
				   stage)
  "Display (and input) function for ringing handbells."
  (let* ((prev-row (if (car handbells:previous-rows)
		       (car handbells:previous-rows)
		     rounds))
	 (pos1 (position right-hand-bell-char prev-row))
	 (pos2 (position left-hand-bell-char prev-row))
	 (where1 (ringing:bell-number-to-char pos1))
	 (where2 (ringing:bell-number-to-char pos2))
	 (direction (if (< where1 where2) "--->" "<---"))
	 (prompt1 (if handbells:show-positions-in-prompt (format "%s %s {%c,%c} first: " prev-row direction where1 where2) "first: "))
	 (prompt2 (if handbells:show-positions-in-prompt (format "%s %s {%c,%c} second: " prev-row direction where1 where2) "second: "))
	 (first-position nil) (first-bell nil)
	 (second-position nil) (second-bell nil)
	 (faulted nil))
    (unless (null ringing:window-configuration-before-help)
      (set-window-configuration ringing:window-configuration-before-help))
    (setq handbells:help-level 0)
    (while (not (progn
		  (setq first-position (ringing:read-number-up-to prompt1 stage)
			first-bell (aref row first-position)
			second-position (ringing:read-number-up-to prompt2 stage)
			second-bell (aref row second-position))
		  (or (and (= first-bell right-hand-bell-char)
			   (= second-bell left-hand-bell-char))
		      (and (= first-bell left-hand-bell-char)
			   (= second-bell right-hand-bell-char)))))
      (ringing:message "Your bells are %c and %c rang in %d (bell %c) and %d (bell %c)"
		       right-hand-bell-char left-hand-bell-char
		       first-position first-bell second-position second-bell)
      (unless faulted
	(setq prompt1 (concat "Wrong: " prompt1)
	      prompt2 (concat "Wrong: " prompt2)
	      faulted t)))
    ;; (message "handbells: %s" row)
    (push row handbells:previous-rows)
    (let ((ringing:ambient-bad-row faulted))
      (when handbells:show-fully
	(funcall
	 method-display:main-display-function
	 row remaining-changes method-holder rounds place-marker)))))

(defun handbells:help-n-rows (n)
  "Display the last N rows, for handbell help."
  (let ((n-rows (min (length handbells:previous-rows) n)))
    (princ (format "\nLast %d rows:\n" n-rows))
    (dolist (row (reverse (subseq handbells:previous-rows 0 n-rows)))
      (princ row)
      (princ (ringing:row-lines row))
      (princ "\n"))))

(defun handbells:extra-help-function ()
  "Extra help function for use when ringing handbells."
  (cond
   ((= handbells:help-level 0)
    (handbells:help-n-rows 4))
   ((= handbells:help-level 1)
    (handbells:help-n-rows 10))
   ((= handbells:help-level 2)
    (handbells:help-n-rows 16))
   (t
    (princ "\nSo far:\n")
    (dolist (row (reverse handbells:previous-rows))
      (princ row)
      (princ (ringing:row-lines row))
      (princ "\n")
      )))
  (incf handbells:help-level))

(defun handbells:ring-method (method right-hand-bell)
  "Ring a plain course of METHOD interactively, using RIGHT-HAND-BELL
and the corresponding left-hand bell."
  (interactive
   (let* ((method-name (ringing:read-method-name "Ring on handbells plain course of: "))
	  (manual (ringing:read-number-up-to
		   "Right-hand bell"
		   (change-ringing-method-nbells
		    (ringing:find-named-method method-name)))))
     (list method-name manual)))
  (let* ((method-data (ringing:find-named-method method))
	 (place-marker (list 'work))
	 (ringing:green-line-too t)
	 (ringing:end-of-row-char ? )
	 (ringing:observation (ringing:bell-number-to-char right-hand-bell))
	 (ringing:observation-number right-hand-bell)
         (ringing:other-observation-number (1+ right-hand-bell))
	 (ringing:other-observation (ringing:bell-number-to-char ringing:other-observation-number))
	 (ringing:bell-input-extra-help-function 'handbells:extra-help-function)
	 (method-display:block-top-marker (if handbells:show-fully (make-marker) nil))
	 (ringing:test-user nil)	; we do this ourselves
	 )
    (when handbells:show-fully
      (let ((ringing-buffer (ringing:get-buffer-create
			     (format ringing:method-buffer-name-format (ringing:method-name method)))))
	(switch-to-buffer ringing-buffer)
	(erase-buffer)
	(delete-other-windows)
	(insert (format "Plain course of %s %s on handbells\n"
			(capitalize (ringing:method-name method))
			(change-ringing-method-original-description-string method-data)))
	(set-marker method-display:block-top-marker (point))
	(if (eq method-display:main-display-function 'method-display:multi-column-display-function)
	    (progn
	      (insert ?\n)
	      (when ringing:label-columns
		(insert (centre-string-in-width (format "%d,%d" ringing:observation-number ringing:other-observation-number)
						(length (method-display:make-lead-end-line (change-ringing-method-nbells method-data))))
			?\n)
		(end-of-line 1)))
	  (progn
	    ))	))
    (setq handbells:previous-rows nil)
    (funcall (if handbells:show-fully
		 'ringing:ring-from-rounds
	       'ringing:ring-from-rounds-plain)
	     method-data
	     'ringing:plain-course-helper-function
	     (list place-marker)
	     'handbells:display-function
	     (list place-marker
		   (ringing:bell-number-to-char right-hand-bell)
		   (ringing:bell-number-to-char (1+ right-hand-bell))
		   (change-ringing-method-nbells method-data)))
    (when handbells:show-fully
      (set-marker method-display:block-top-marker nil))
    (unless (null ringing:window-configuration-before-help)
      (set-window-configuration ringing:window-configuration-before-help))))

;;; end of handbells.el
