;;;; leads.el -- part of John's elisp change-ringing system
;;; Time-stamp: <99/10/29 17:54:32 johnstu>

(provide 'leads)
(require 'place-notation)
(require 'ringing-methods)

(defvar ringing:next-lead-end-action '(plain)
  "The next lead end action to take: (plain), (bob), (single) or (change-of-method method)")

(defvar ringing:call-at-remaining-rows 3
  "How many rows to call before reaching the lead end.")

(defun ringing:ring-one-lead-helper-function (row
					      remaining-changes
					      method-holder
					      rounds
					      place-marker)
  "Helper function for ringing one lead of a method."
  (if (and nil (string= row rounds))
      nil
    (if remaining-changes
	t
      (let ((method (car method-holder)))
	(cons row
	      (case (car place-marker)
		((lead-end)
		 nil)
		((work)
		 (rplaca place-marker 'lead-end)
		 (case (car ringing:next-lead-end-action)
		   (plain (change-ringing-method-plain-lead-end method))
		   (bob (change-ringing-method-bob method))
		   (single (change-ringing-method-single method))))))))))

(defvar ringing:lead-ends 'shadow-me
  "For counting leads. This global copy should not be touched, so is not a number.")

(defun ringing:record-lead-heads-output-function (row
						  remaining-changes
						  method
						  rounds
						  place-marker)
  "Output callback for counting leads."
  (when (eq (car place-marker) 'lead-end)
    (push (substring row 1 -1) ringing:lead-ends)))

(defun ringing:plain-course-leads (method)
  "Count the leads in a plain course of METHOD."
  (interactive
   (let* ((method-name (ringing:read-method-name "Count leads in plain course of: ")))
     (list method-name)))
  (let* ((method-name (ringing:method-name method))
	 (method-data (ringing:find-named-method method))
	 (ringing:observation-number (change-ringing-method-nbells method-data))
         (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	 (ringing:lead-ends nil)
	 (place-marker (list 'work))
	 (lead-ends (nreverse
		     (progn
		       (ringing:ring-from-rounds-plain method-data
						       'ringing:plain-course-helper-function
						       (list place-marker)
						       'ringing:record-lead-heads-output-function
						       (list place-marker))
		       ringing:lead-ends))))
    (when (interactive-p) (message "%s has leads %s in the plain course" method-name lead-ends))
    lead-ends))

(defun ringing:count-plain-course-leads (method)
  "Count the leads in a plain course of METHOD."
  (interactive
   (let* ((method-name (ringing:read-method-name "Count leads in plain course of: ")))
     (list method-name)))
  (let* ((method-data (ringing:find-named-method method))
	 (lead-count (length (ringing:plain-course-leads method-data))))
    (when (interactive-p) (message "%s has %d leads in the plain course"
				   (change-ringing-method-title method-data)
				   lead-count))
    lead-count))

(defun ringing:count-lead-rows (method)
  "Count the rows in a lead of METHOD."
  (interactive
   (let* ((method-name (ringing:read-method-name "Count leads in plain course of: ")))
     (list method-name)))
  (let* ((method-name (ringing:method-name method))
	 (method-data (ringing:find-named-method method))
	 (lead-length (+ (length (change-ringing-method-work method-data))
			 (length (change-ringing-method-plain-lead-end method-data)))))
    (when (interactive-p) (message "%s has %d rows per lead" method-name lead-length))
    lead-length))

;;;; single leads

(defun ringing:record-lead-head-output-function (row
						 remaining-changes
						 method
						 rounds)
  "Output callback for recording one lead."
  (when (eq (car place-marker) 'lead-end)
    nil
    ))

(defun ringing:plain-lead-matrices (method)
  "Return a cons of arrays saying where bells go to and come from in a plain lead of METHOD."
  (let* ((method-data (ringing:find-named-method method))
	 (nbells (change-ringing-method-nbells method-data))
	 (method-holder (list method-data))
	 (place-marker (list 'work))
	 (rounds (let ((r (make-vector (1+ nbells) nil)))
		   (dotimes (i (1+ nbells))
		     (aset r i i))
		   r))
	 (lead-end (ringing:ring-changes rounds
					 (if (change-ringing-method-special-start method-data)
					     (change-ringing-method-special-start method-data)
					   (if (change-ringing-method-work method-data)
					       (change-ringing-method-work method-data)
					     (change-ringing-method-plain method-data)))
					 'ringing:ring-one-lead-helper-function
					 (list method-holder rounds place-marker)
					 'ringing:record-lead-head-output-function
					 (list method-holder rounds)))
	 (forward-matrix (make-vector (1+ nbells) nil)))
    ;; produce the inverted matrix:
    (dotimes (i (1+ nbells))
      (aset forward-matrix (aref lead-end i) i))
    (cons forward-matrix lead-end)))

(defvar ringing:lead-changes 'shadow-me
  "For recording changes in a lead. This global copy should not be touched/")

(defun ringing:record-lead-changes-output-function (row
						  remaining-changes
						  method
						  rounds
						  place-marker)
  "Output callback for recording changes in a lead."
    (push (substring row 1 -1) ringing:lead-changes))

(defun ringing:method-lead-changes (method &optional just-work)
  "Return the lead changes of METHOD. With optional JUST-WORK, ignore the lead ends.
This is a string containing all the changes that occur in the first lead
starting from rounds. The changes are sorted, and separated by spaces."
  (interactive
   (let* ((method-name (ringing:read-method-name "Show the changes in a lead of: ")))
     (list method-name)))
  (let* ((method-name (ringing:method-name method))
	 (method-data (ringing:find-named-method method))
	 (ringing:observation-number (change-ringing-method-nbells method-data))
         (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	 (ringing:lead-changes nil)
	 (place-marker (list 'work))
	 (lead-changes
	  (progn
	    (ringing:ring-from-rounds-plain method-data
					    'ringing:ring-one-lead-helper-function
					    (list place-marker)
					    'ringing:record-lead-changes-output-function
					    (list place-marker))
	    ringing:lead-changes)))
    (setq lead-changes (sort lead-changes 'string<))
    (when (interactive-p)
      (with-output-to-temp-buffer (format "*Changes in lead of %s*" method-name)
	(dolist (c lead-changes)
	  (print c))))
    lead-changes
    (mapconcat 'identity lead-changes " ")))

;;; end of leads.el
