;;;; manual-ringing.el -- part of John's elisp change-ringing system
;;; handle input from the user, for the ringing simulator
;;; Time-stamp: <99/03/05 14:55:16 johnstu>

(eval-when (compile load eval)
  (provide 'manual-ringing))

(defvar ringing:prev-place nil
  "The previous position the user rang at (skipping clashes).")

(defvar ringing:show-clashes-warning-big t
  "Whether to pop up a buffer for clashes.
If this is nil, the message telling you that you've clashed appears in the
minibuffer for a moment.")

(defvar ringing:clashes 0
  "How many mistakes the user has made in this touch.")

(defvar ringing:clash-message-buffer-name "*Clash!*"
  "The name of the buffer for indicating that there's been a clash.")

(defvar ringing:prompt-user-extra-string nil
  "Extra string to insert at start of user ringing prompt.")

;; This can probably be replaced by ringing:observation-place-number from method-display.el:

(defun ringing:what-user-should-have-typed (row)
  "Return the place in which the user should have rung in ROW."
  (let ((here 0)
	(len (length row)))
    (while (and (/= (aref row here) ringing:observation)
		(/= (aref row here) ?*)
		(<= here len))
      (incf here))
    (setq ringing:bad-row t)
    here))

(defvar ringing:allow-timeout-this-row t
  "Whether to allow timeout this row. Not a user parameter.
This is for disabling timeout on the first row of a touch, to give the user
time to see which method it is and so on.")

(defun ringing:get-user-stroke (row
				remaining-changes
				method
				rounds
				place-marker)
  "This is for calling from callback functions for the ringing system."
  (when (null ringing:prev-place) (setq ringing:prev-place
					ringing:observation-number;; (ringing:bell-char-to-number ringing:observation)
					))
    (let* ((nbells (change-ringing-method-nbells method))
	   (ringing:use-timing-here ringing:allow-timeout-this-row)
	   (prompt (if ringing:prompt-user-extra-string
		       (format "%sRing at: " ringing:prompt-user-extra-string)
		     "Ring at: "))
	   (place (ringing:read-number-up-to prompt
					     nbells ringing:prev-place))
	   (clashed nil))
      (if (eq place 'timeout)
	  (setq ringing:prev-place (ringing:what-user-should-have-typed row))
	(while (and (/= (aref row place) ringing:observation)
		    (/= (aref row place) ?*))
	  (setq clashed t
		ringing:bad-row t
		ringing:auto-hunt-state nil)
	  (incf ringing:clashes)
	  (let ((error-message
		 (format "Wrong: you rang bell %c in %d, where %c should be!"
			 ringing:observation
			 place (aref row place))))
	    (if ringing:show-clashes-warning-big
		(save-window-excursion
		  (with-output-to-temp-buffer ringing:clash-message-buffer-name
		    (princ (format "%d clashes so far\n" ringing:clashes))
		    (princ error-message))
		  (setq place
			(ringing:read-number-up-to
			 prompt nbells ringing:prev-place)))
	      (progn
		(ringing:message error-message)
		(setq place
		      (ringing:read-number-up-to
		       prompt nbells ringing:prev-place)))))
	  (when (eq place 'timeout)
	    (setq place (ringing:what-user-should-have-typed row))))
	(setq ringing:prev-place place))
      ))

(defun ringing:get-user-stroke-OLD (row
				remaining-changes
				method
				rounds
				place-marker)
  "This is for calling from callback functions for the ringing system."
  (if (null ringing:prev-place) (setq ringing:prev-place (ringing:bell-char-to-number ringing:observation)))
  (let* ((nbells (change-ringing-method-nbells method))
	 (ringing:use-timing-here t)
	 (prompt (if ringing:prompt-user-extra-string
		     (format "%sRing at: " ringing:prompt-user-extra-string)
		   "Ring at: "))
	 (place (ringing:read-number-up-to prompt
					   nbells ringing:prev-place))
	 (clashed nil))
    (if (eq place 'timeout)
	(ringing:what-user-should-have-typed row)
      (while (and (/= (aref row place) ringing:observation)
		  (/= (aref row place) ?*))
	(setq clashed t
	      ringing:bad-row t)
	(incf ringing:clashes)
	(let ((error-message 
	       (format "Wrong: you rang bell %c in %d, where %c should be!"
		       ringing:observation
		       place (aref row place))))
	  (if ringing:show-clashes-warning-big
	      (save-window-excursion
		(with-output-to-temp-buffer ringing:clash-message-buffer-name
		  (princ (format "%d clashes so far\n" ringing:clashes))
		  (princ error-message))
		(setq place
		      (ringing:read-number-up-to
		       prompt nbells ringing:prev-place)))
	    (progn
	      (ringing:message error-message)
	      (setq place
		    (ringing:read-number-up-to
		     prompt nbells ringing:prev-place)))))
	(when (eq place 'timeout)
	  (setq place (ringing:what-user-should-have-typed row)))))
    (setq ringing:prev-place place)))

;;;; time the user doing things, out of interest

(defun seconds-between (start end)
  "Return the number of seconds between START and END,
which are given in the format described by (current-time)."
  (let* ((start-large (first start))
	 (start-small (second start))
	 (end-large (first end))
	 (end-small (second end))
	 (difference-large (- end-large start-large))
	 (difference-small (- end-small start-small)))
    (+ (* 65536 difference-large) difference-small)))

(defun plural-ending (n)
  "Return the plural ending for N."
  (if (= n 1) "" "s"))

(defun time-string (time-in-sec &optional with-seconds)
  "Return TIME-IN-SEC as a %d hours %d minutes %d seconds sort of string.
If time is greater than 10 minutes, only give times with non-nil &optional with-seconds."
  (let* ((total (floor time-in-sec))
	 (seconds (mod total 60))
	 (minutes (/ (mod total 3600) 60))
	 (hours (/ total 3600)))
    (if (>= hours 1)
	(if with-seconds
	    (if (zerop minutes)
		(format "%d hour%s %d second%s" hours (plural-ending hours) seconds (plural-ending seconds))
	      (format "%d hour%s %d minute%s %d second%s" hours (plural-ending hours) minutes (plural-ending minutes) seconds (plural-ending seconds)))
	  (if (zerop minutes)
	      (format "%d hour%s" hours (plural-ending hours))
	    (format "%d hour%s %d minute%s" hours (plural-ending hours) minutes (plural-ending minutes))))
      (if (zerop minutes)
	  (format "%d second%s" seconds (plural-ending seconds))
	(if (and (not (zerop seconds)) (or with-seconds (<= minutes 10)))
	    (format "%d minute%s %d second%s" minutes (plural-ending minutes) seconds (plural-ending seconds))
	  (format "%d minute%s" minutes (plural-ending minutes)))))))

;;; end of manual-ringing.el
