;;;; method-display.el -- part of John's elisp change-ringing system
;;; Time-stamp: <99/09/13 09:45:34 johnstu>
;;; method-display functions: output lines of bells, numbers etc

(eval-when (compile load eval)
  (provide 'method-display)
  (require 'method-lines)
  (require 'manual-ringing)
  (require 'ermm))

(defvar ringing:test-user nil
  "Whether to ask the user for each stroke, and verify what they enter.")

(defvar method-display:show-final-rounds nil
  "*Whether to show the final rounds again in displaying a method.
You may find the output with this off neater, as the rounds go in
column of their own.")

(defvar method-display:draw-lead-twice t
  "*Whether to draw the lead end at the bottom of each column
as well as at the top.")

(defvar method-display:lead-end-line "\n"
  "Thing to draw at the end of each lead.
Normally bound by whatever passes the callback that uses it.")

(defun method-display:column-width (n)
  "Return the width of a column with N bells."
  (+ (if ringing:show-lines-with-lines (* 2 2) 2)
     (if ringing:show-numbers n 0)
     (if ringing:show-lines
	 (if ringing:show-lines-with-lines (* 2 n) n)
       0)
     (if (and ringing:show-numbers
	      ringing:show-lines)
	 2
       0)))

(defun method-display:make-lead-end-line (n)
  "Make a lead end marker line for N bells."
  (let ((line (make-string (method-display:column-width n) ? ))
	(i 1))
    (aset line (1- (length line)) ringing:end-of-row-char)
    (while (<= i n)
      (aset line (if ringing:show-lines-with-lines (* 2 i) i) ?-)
      (if ringing:show-lines-with-lines (aset line (1- (* 2 i)) ?-))
      (if (and ringing:show-numbers
	       ringing:show-lines)
	  (aset line (+ i
			(if ringing:show-lines-with-lines (* 2 n) n)
			(if ringing:show-lines-with-lines (* 2 2) 2)) ?-))
      (incf i))
    line))

(defun ringing:observation-place-number (row)
  "Return the place the observation bell is in in ROW."
  (catch 'found-observation
    (let ((i 0)
	  (n (length row)))
      (while (< i n)
	(let ((bell (aref row i)))
	  (when (or (= bell ringing:observation)
		    (= bell ?*))
	    (throw 'found-observation i)))
	(incf i)))))

(defun ringing:other-observation-place-number (row)
  "Return the place the observation bell is in in ROW."
  (if ringing:other-observation-number
      (catch 'found-other-observation
	(let ((i 0)
	      (n (length row)))
	  (while (< i n)
	    (let ((bell (aref row i)))
	      (when (or (= bell ringing:other-observation)
			(= bell ?#))
		(throw 'found-other-observation i)))
	    (incf i))))
    nil))

(defun centre-string-in-width (string width)
  "Return STRING centred in WIDTH with spaces around it."
  (let* ((remaining (- width (length string)))
	 (left (/ remaining 2)))
    (concat (make-string left ? )
	    string
	    (make-string (- remaining left) ? ))))

(defvar ringing:label-columns t
  "*Whether to label the columns in multi-column displays.")

(defun method-display:label-column (placebell width &optional other-place-bell)
  "Label the current column for PLACEBELL where columns are WIDTH wide."
  (when ringing:label-columns
    (end-of-line 2)
    (insert (centre-string-in-width
	     (if other-place-bell
		 (format "%d,%d" placebell other-place-bell) 
	       (format "%d" placebell))
	     width))
    (end-of-line 1)
    ;; I'm moving this into mSiRiL-mode:lines-for-method as it's only wanted for the first one, I think
    (when nil (eobp) (insert ?\n))
    ))

(defvar method-display:form-feed-between-blocks nil
  "*Whether to insert a form feed between the blocks of columns,
in multi-column output.")

(defvar method-display:block-top-marker nil
  "Where to go back up to when starting the next column.")

(defvar method-display:page-width nil
  "*Width of page in characters, for choosing when to start new block of columns.
Non-number or zero means use the frame width.")

(defvar ringing:ambient-bad-row nil
  "The value of ringing:bad-row to be seen at the start of the display function.
This should normally be nil, unless you have your own way of quizzing the user,
as is done in the handbells code.")

(defun method-display:multi-column-display-function (row
						     remaining-changes
						     method-holder
						     rounds
						     place-marker)
  "Display callback for multi-column output.
This outputs ROW, with all the trimmings, and handles moving from one column
to the next, and, when the columns have gone all the way across the page, down
to the next row of columns."
  (let ((ringing:bad-row ringing:ambient-bad-row)
	(method (car method-holder)))
    (when ringing:test-user
      (ringing:get-user-stroke row
			       remaining-changes
			       method
			       rounds
			       place-marker)
      (setq ringing:allow-timeout-this-row t))
    (ringing:apply-bell-colours row)
    (ermm:output-row row)
    (let ((down-a-line (if ringing:show-lines-with-lines 2 2)))

      (when (and (null (cdr remaining-changes))
		 (eq (car place-marker) 'lead-end))


	(when method-display:draw-lead-twice

	  ;; draw dashed line across the bottom, followed
	  ;; by the lead end row (which will be repeated
	  ;; at the top of the next column)

	  (when method-display:lead-end-line
	    (end-of-line down-a-line)
	    (if (eobp) (insert ?\n))
	    (insert method-display:lead-end-line))

	  (end-of-line down-a-line)
	  (if (eobp) (insert ?\n))
	  (ringing:insert-row-lines row))



	;; end of the column -- go to top of the next one
	(end-of-line down-a-line)
	(if (eobp) (insert "\n"))
	(goto-char (if (and (boundp 'method-display:block-top-marker)
			    method-display:block-top-marker)    
		       method-display:block-top-marker
		     (point-min)))



	;; test whether we have reached the right-hand margin
	(when (>= (save-excursion (end-of-line 2) (current-column))
		  (if (and (numberp method-display:page-width)
			   (not (zerop method-display:page-width)))
		      method-display:page-width
		    (- (frame-width) 2 (length method-display:lead-end-line))))
	  (goto-char (point-max))
	  (when method-display:form-feed-between-blocks
	    (insert "\n"))
	  ;; (push (marker-position method-display:block-top-marker) old-markers)
          ;; (yes-or-no-p "Here is where the marker will be")
	  (set-marker method-display:block-top-marker (point))
	  (insert "\n"))

	;; label the column with which place bell this is
	(when (or method-display:show-final-rounds
		  (not (string= row rounds)))
	  (method-display:label-column (ringing:observation-place-number row)
				       (length method-display:lead-end-line)
				       (ringing:other-observation-place-number row)
				       )))

      ;; having dealt with the end-of-column stuff if necessary,
      ;; now do the ordinary outputting of a row; but usually we
      ;; will not want to do this if it is the final rounds, as
      ;; that would then stand in a little column all by itself

      (when (or method-display:show-final-rounds
		(not (string= row rounds)))
	(end-of-line down-a-line)
	(if (eobp) (insert "\n"))
	(ringing:insert-row-lines row)
	(when ringing:auto-ring-inter-row-delay
	  (sit-for ringing:auto-ring-inter-row-delay))))))

(defun method-display:single-column-display-function (row
						      remaining-changes
						      method-holder
						      rounds
						      place-marker)
  "Helper function for ringing plain courses interactively."
  (let ((ringing:bad-row ringing:ambient-bad-row)
	(method (car method-holder)))
    (when ringing:test-user
      (ringing:get-user-stroke row
			       remaining-changes
			       method
			       rounds
			       place-marker)
      (setq ringing:allow-timeout-this-row t))
    (insert "\n")
    (ringing:apply-bell-colours row)
    (ermm:output-row row)
    (ringing:insert-row-lines row)
    (when ringing:auto-ring-inter-row-delay
      (sit-for ringing:auto-ring-inter-row-delay))
    (if (string= row rounds)
	(insert "\nBack to rounds\n"))))

(defvar method-display:main-display-function
  'method-display:multi-column-display-function
  "The display function to use in most places.")

;;; end of method-display.el
