;;;; method-groups.el -- part of John's elisp change-ringing system
;;; Handle groups of methods -- "standard eight" and so on
;;; Time-stamp: <99/11/03 09:54:25 johnstu>

(eval-when (compile load eval)
  (provide 'method-groups)
  (require 'stages)
  (require 'method-library)
  (require 'ringing-methods))

(defvar method-groups:groups-directory (expand-file-name "~/Ringing/method-groups/")
  "Where to look for the groups of methods, such as standard-eight.
Each file simply contains method names, one per line. The place
notation is not given, so the grouping may be applied to different
stages without trying to extend automatically.")

(defconst method-groups:standard-network-groups-URL
  "http://www.cb1.com/~john/ringing/method-groups/"
  "Where on the network some group files may be found.")

(defun method-groups:ftp-groups ()
  "Fetch a new version of the method groups."
  (interactive)
  (ringing:ftp-directory-files method-groups:standard-network-groups-URL
			       method-groups:groups-directory))

(defstruct method-groups:method-group
  "Description of a group of methods."
  title stage
  ;; There are two forms of this: if stage=0, methods is an alist of method
  ;; names (in their short form, e.g. "Cambridge") to nil; if stage>0, it
  ;; is an alist of method names (in their full form, e.g. "Cambridge Surprise Major")
  ;; to method structures.
  methods
  split-methods				; conses of below . above partial methods
  code-letters
  title-style-default
  style-default
  layout-default)

(defvar method-groups:all-groups
  nil
  "The currently known method groups.
You should read this list as a whole via the function method-groups:all-groups,
and should get individual groups via the function method-groups:group-methods.

Some special groups are not stored on the list in this variable, but are
handled by those functions.

In fact, it is stored as an alist of group name to methods;
if the group has not yet been used, the methods list is nil.
The method list is a alist of method kind string (such as \"s\" for surprise)
dotted with the name of the method.")

(defconst method-groups:special-names
  '(("*Touch book*")
    ("*Methods rung this session*"))
  "Names of special groups of methods used and maintained by the internals
of the ringing software.")

(defun method-groups:list-ephemeral-groups ()
  "List the ephemeral groups."
  (let ((ephimerides nil))
    (dolist (buffer (buffer-list))
      (let ((ephemeris (mSiRiL-mode:buffer-ephemeral-group buffer)))
	(when ephemeris (push
			 ephemeris
			 ephimerides))))
    ephimerides))

(defun touch-book:visited-buffer-groups ()
  "List the groups for currently loaded method library buffers."
  (let ((buffer-groups nil))
    (dolist (buffer (buffer-list))
      (let ((buffer-group (mSiRiL-mode:buffer-group buffer)))
	(when buffer-group (push
			 buffer-group
			 buffer-groups))))
    buffer-groups))
  
(defvar method-groups:include-touch-book-stage-groups nil
  "*Whether to include the touch book stage groups.
This may cause more methods to be read than you'd really like.")

(defvar method-groups:include-visited-buffer-groups nil
  "*Whether to include the groups for all currently loaded method library buffers.
This may take a little while, as they say. It's only likely to be useful if you're
collecting statistics on methods, or something like that.")

(defun method-groups:all-groups (&optional force-new)
  "Return the list of method groups, each dotted with nil or a group structure.
Don't use the result for the cdr of each group description, but call
method-groups:group-methods, which ensures that it is filled in properly, even
in the case of groups which are special (i.e internal \"magic\" of the system)
and will always have nil as the members list from this function."
  (if (or force-new (null method-groups:all-groups))
      (setq method-groups:all-groups
	    (mapcar (function
		     (lambda (file)
		       (cons (file-name-nondirectory file) nil)))
		    (directory-files method-groups:groups-directory
				     nil
				     "[^.]$" t))))
  (append method-groups:special-names
	  (method-groups:list-ephemeral-groups)
	  (if method-groups:include-touch-book-stage-groups
	      (touch-book:touch-book-stage-groups)
	    nil)
	  (if method-groups:include-visited-buffer-groups
	      (touch-book:visited-buffer-groups)
	    nil)
	  method-groups:all-groups))

(defun method-groups:get-misc-data (data-name)
  "Get DATA-NAME from the group description in the current buffer; or nil."
  (save-excursion
    (goto-char (point-min))
    (if (re-search-forward (format "^%s=\\(.+\\)$" data-name)
			   (point-max) t)
	(buffer-matched 1)
      nil)))

(makunbound 'method-group-mode-map)
(defvar method-group-mode-map (make-keymap "Method Group")
  "Keymap for method group mode")

(define-key method-group-mode-map "?" 'method-group-mode:display-method)
(define-key method-group-mode-map "\M-." 'method-group-mode:find-method)
(define-key method-group-mode-map "r" 'method-group-mode:re-teach-random-leads)
(define-key method-group-mode-map "q" 'method-group-mode:quiz-random-leads)
(define-key method-group-mode-map "a" 'method-group-mode:test-all-leads)
(define-key method-group-mode-map "m" 'method-group-mode:ring-all-methods)
(define-key method-group-mode-map "p" 'method-group:print-group)
(define-key method-group-mode-map "v" 'method-group:preview-group)
(define-key method-group-mode-map "z" 'method-group:report)

(defvar method-group-mode:raw-method-data nil
  "List of methods for a group file. Local per file.")

(defvar method-group-mode:group-data nil
  "Group of methods for a group file. Local per file.")

(defvar method-group-mode:all-stages-known nil
  "Whether the stages for all methods in this file are known.")

(defvar method-group-mode:title nil
  "The apparent title for this group.")

(defvar method-group-mode:method-count-string ""
  "Description of how many methods there are in this group, for display with the mode name.")

(mapcar 'make-variable-buffer-local
	'(method-group-mode:raw-method-data
	  method-group-mode:group-data
	  method-group-mode:all-stages-known
	  method-group-mode:title
	  method-group-mode:method-count-string))

;; suggested auto-mode-alist entry:
;; ("method-groups" . method-group-mode)

(defun method-group-mode ()
  "Major mode for viewing (and possibly editing) group files.
Special commands are:
\\{method-group-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (mapcar 'make-local-variable
	  '(method-group-mode:raw-method-data
	    method-group-mode:group-data
	    method-group-mode:all-stages-known
	    method-group-mode:title
	    method-group-mode:method-count-string))
  (setq major-mode 'method-group-mode
	mode-name "Method Group"
	mode-line-format (copy-sequence mode-line-format)
	buffer-read-only t
	method-group-mode:group-data nil)
  (let ((mode-tail (memq 'mode-name mode-line-format)))
    (when mode-tail
      (rplacd mode-tail
	      (append '(method-group-mode:method-count-string)
		      (cdr mode-tail)))))
  (method-group-mode:update-buffer-variables)
  (use-local-map method-group-mode-map)
  (add-hook 'local-write-file-hooks
	    'method-group-mode:buffer-update-hook
	    nil
	    t))

(defun method-group-mode:update-buffer-variables ()
  "Update the buffer variables for this group buffer."
  (setq method-group-mode:raw-method-data (method-group-mode:parse-buffer)
	method-group-mode:all-stages-known (method-group-mode:all-stages-known method-group-mode:raw-method-data)
	method-group-mode:title
	(let ((explicit-title (method-groups:get-misc-data "group-title")))
	  (if explicit-title
	      explicit-title
	    (save-excursion
	      (goto-char (point-min))
	      (if (looking-at "^# \\(.+\\)$")
		  (buffer-matched 1)
		(file-name-nondirectory (buffer-file-name))))))
	method-group-mode:method-count-string
	(format ": %d methods" (length method-group-mode:raw-method-data))
	method-group-mode:group-data (if method-group-mode:group-data
					  (method-group-mode:group-data)
					method-group-mode:group-data)))

(defvar method-group-mode:line-regexp "^\\([a-z]+[0-9]*\\) \\(.+\\)$"
  "Pattern matching a line of a method group buffer.")

(defun method-group-mode:parse-buffer ()
  "Return the raw parsed method group data for the current buffer.
This is an list of (method-name type size location)."
  (let ((result nil))
    (save-excursion
      (goto-char (point-min))
      (while (re-search-forward method-group-mode:line-regexp (point-max) t)
	(let* ((location (buffer-matched 1))
	       (name-and-code (buffer-matched 2))
	       (code (if (string-match "(\\([a-z]\\))$" name-and-code)
			 (substring name-and-code (match-beginning 1) (match-end 1))
		       nil))
	       (name (if (string-match "([a-z])$" name-and-code)
			 (substring name-and-code 0 (match-beginning 0))
		       name-and-code))
	       (munged-name (microSiRiL:munge-name name))
	       (stage-number (if (string-match "[0-9]+$" location)
				 (string-to-int (substring location (match-beginning 0)))
			       nil))
	       (stage-name (if stage-number
			       (ringing:stage-name stage-number)
			     nil))
	       (method-type (ringing:figure-out-kind location))
	       (method-type-string (ringing:kind-by-name method-type)))
	  (push (list munged-name method-type method-type-string stage-number stage-name location code)
		result))))
    (nreverse result)))

(defun method-group-mode:all-stages-known (parsed-data)
  "Return whether all pairs in PARSED-DATA indicate their stage."
  (catch 'found-unstaged
    (dolist (entry parsed-data)
      (unless (null (nth 3 entry))
	(throw 'found-unstaged nil)))
    t))

(defun method-group-mode:buffer-update-hook ()
  "Things to do when a method group buffer has been edited."
  (method-group-mode:update-buffer-variables)
  ;; this does not count as having written the file:
  nil)

(defun method-group-mode:raw-method-data-at-place (where)
  "Return the method at WHERE."
  (save-excursion
    (let ((eol (progn (end-of-line 1) (point))))
      (beginning-of-line 1)
      (if (re-search-forward method-group-mode:line-regexp eol t)
	  (let* ((method-name-and-code (buffer-matched 2))
		 (method-name (if (string-match "([a-z])$" method-name-and-code)
				  (substring method-name-and-code 0 (match-beginning 0))
				method-name-and-code))
		 (method-data (assoc method-name method-group-mode:raw-method-data)))
	    method-data)
	(error "Could not identify method on this line")))))  

(defun method-group-mode:method-data-at-place (where)
  "Return the method at WHERE."
  (save-excursion
    (let ((eol (progn (end-of-line 1) (point))))
      (beginning-of-line 1)
      (if (re-search-forward method-group-mode:line-regexp eol t)
	  (let* ((group (method-group-mode:group-data))
		 (method-name-and-code (buffer-matched 2))
		 (method-name (if (string-match "([a-z])$" method-name-and-code)
				  (substring method-name-and-code 0 (match-beginning 0))
				method-name-and-code))
		 (method-raw-data (assoc method-name method-group-mode:raw-method-data))
		 (method-full-name (microSiRiL:constructed-name
				    (nth 0 method-raw-data)
				    (nth 2 method-raw-data)
				    (if (nth 4 method-raw-data)
					(nth 4 method-raw-data)
				      (ringing:stage-name (method-groups:method-group-stage group)))))
		 (method-data (assoc method-full-name
				     (method-groups:method-group-methods group))))
 ;; (message "method-name-and-code %s; method-name %s; method-full-name %s" method-name-and-code method-name method-full-name)
	    method-data)
	(error "Could not identify method on this line")))))  

(defun method-group-mode:find-method (where)
  "Locate the definition of the method around WHERE (point, interactively)."
  (interactive "d")
  (method-library:locate-method (car (method-group-mode:method-data-at-place where))))
  
(defun method-group-mode:display-method (where)
  "Show the method around WHERE (point, interactively)."
  (interactive "d")
  (mSiRiL-mode:lines-for-method (cdr (method-group-mode:method-data-at-place where)))
  (other-window -1))

(defun method-group-mode:ring-all-methods ()
  "Get the user to ring all methods in this group."
  (interactive)
  (let* ((group-data (method-group-mode:group-data))
	 (methods (method-groups:method-group-methods group-data))
	 (stage (method-groups:method-group-stage group-data))
	 (ringing:method-buffer-name-format (format "*%s practice*" (method-groups:method-group-title group-data))))
    (catch 'stood
      (dolist (method methods)
	(let* ((method-data (cdr method))
	       (lowest-places (ringing:lowest-working-bell-making-places
			       (car (change-ringing-method-plain-lead-end method-data)))))
	  (if (numberp lowest-places)
	      lowest-places
	    (change-ringing-method-nbells method-data))
	  (when (eq (ringing:ring-plain-course (car method) lowest-places) 'user-stand)
	    (unless (y-or-n-p "Continue at next method? ")
	      (throw 'stood t))))))))

(defun ringing:quiz-whole-methods-at-stage (stage)
  "Get the user to ring all the methods in the quiz group for STAGE.
See ringing:quiz-at-stage for explanation of quiz groups."
  (interactive (list (ringing:read-stage-from-user "Quiz whole methods at stage: ")))
  (let ((quiz-file (expand-file-name (format "quiz-%d" (ringing:figure-out-number-of-bells stage))
				     method-groups:groups-directory)))   
    (if (file-exists-p quiz-file)
	(save-window-excursion
	  (find-file quiz-file)
	  (method-group-mode:ring-all-methods))
      (error "No quiz file defined for %s (would be called %s)"
	     stage quiz-file))))

(defun method-group-mode:re-teach-random-leads ()
  "Ring, with preview of the line, random leads from methods in this group file."
  (interactive)
  (save-window-excursion
    (ringing:teach-random-methods-from-group
     (method-group-mode:group-data))))

(defun method-group-mode:quiz-random-leads (&optional n)
  "Ring random leads from methods in this group file.
With optional / prefix argument other than 1, stop after that number of leads."
  (interactive "p")
  (save-window-excursion
    (ringing:quiz-random-methods-from-group
     (method-group-mode:group-data)
     n)))

(defun ringing:quiz-at-stage (stage)
  "Quiz the user on random leads of methods at STAGE.
The methods are defined in the method group file called quiz-n where n
is the stage as a number (in the method groups directory, as defined by
method-groups:groups-directory."
  (interactive (list (ringing:read-stage-from-user "Quiz on leads of methods at stage: ")))
  (let ((ringing:quiz-between-leads (if (and (null ringing:quiz-between-leads)
					     (eq major-mode 'shell-mode))
					(current-buffer)
				      ringing:quiz-between-leads))
	(quiz-file (expand-file-name (format "quiz-%d" (ringing:figure-out-number-of-bells stage))
				     method-groups:groups-directory)))
    (if (file-exists-p quiz-file)
	(save-window-excursion
	  (find-file quiz-file)
	  (method-group-mode:quiz-random-leads))
      (error "No quiz file defined for %s (would be called %s)"
	     stage quiz-file))))

(defun randomize-list-order (orig)
  "Return a list containing the same elements as ORIG but in a random order."
  (let* ((vec (apply 'vector orig))
	 (n (length vec)))
    (dotimes (i n)
      (let* ((j (random n))
	     (a (aref vec i)))
	(aset vec i (aref vec j))
	(aset vec j a)))
    (coerce vec 'list)))

(defun method-group-mode:test-all-leads ()
  "Test the user on all leads in this group file.
The testing is done in random order, but each lead is removed from the pool
of leads waiting to be tested once you have got it right."
  (interactive)
  (save-window-excursion
    (let* ((group-data (method-group-mode:group-data))
	   (methods (method-groups:method-group-methods group-data))
	   (stage (method-groups:method-group-stage group-data))
	   (leads nil)
	   )
      (dolist (method methods)
	(dotimes (i stage)
	  (unless (zerop i) (push (cons (car method) (1+ i)) leads))))
      (setq leads (randomize-list-order leads))
      (while (not (null leads))
	(let* ((lead (car leads))
	       (result (ringing:try-user-on-lead (car lead) (cdr lead))))
	  (if (eq result 'user-stand)
	      (setq leads nil)
	    (progn
	      (setq leads (cdr leads))
	      (if (>= ringing:clashes ringing:quiz-acceptable-faults)
		  (let ((retry (nthcdr (random (length (cdr leads))) leads)))
		    (ringing:message "Not good enough -- this lead will re-appear later")
		    (rplacd retry (cons lead (cdr retry))))
		(progn
		  (ringing:message "Good enough")
		  )
		)
	      (ringing:message "%d leads left to try" (length leads))
	      ))))
      (ringing:message "End of testing"))))

(defun method-group-mode:make-title-adding-stage-if-necessary (title stage-name)
  "Onto TITLE put STAGE-NAME if TITLE does not already indicate the stage."
  (if (string-match stage-name title)
      title
    (concat title " at " stage-name)))

(defun method-group-mode:group-data (&optional stage-getter force)
  "Return the full group data for the current file.
If the stage cannot be worked out, and STAGE-GETTER is given,
call stage-getter with a prompt as argument to get the stage."
  (if (eq major-mode 'method-group-mode)
      (if (and method-group-mode:group-data (not force))
	  method-group-mode:group-data
	(setq method-group-mode:group-data
	      (let ((stages nil))
		(dolist (method method-group-mode:raw-method-data)
		  (pushnew (nth 3 method) stages))
		(when (null stage-getter) (setq stage-getter 'ringing:read-stage-from-user))
		(let* ((default-stage
			 (if (= (length stages) 1)
			     (if (null (car stages))
				 (funcall stage-getter
					  (format "Number of bells for methods in group %s: "
						  method-group-mode:title))
			       (car stages))
			   (if (memq nil stages)
			       (if (= (length stages) 2)
				   (if (null (car stages))
				       (cadr stages)
				     (car stages))
				 (funcall stage-getter
					  (format "Default number of bells for methods in group %s: "
						  method-group-mode:title)))
			     (car stages))))
		       (default-stage-name (ringing:stage-name default-stage))
		       (staged-methods nil)
		       (code-table nil))
		  (dolist (method method-group-mode:raw-method-data)
		    (let* ((method-full-name (microSiRiL:constructed-name
					      (nth 0 method)
					      (nth 2 method)
					      (if (nth 4 method)
						  (nth 4 method)
						default-stage-name)))
			   (method-data (ringing:find-named-method method-full-name)))
		      ;; (message "group-data parser has name %s and method data %s" method-full-name method-data)
		      (push (cons method-full-name method-data) staged-methods)
		      (when (nth 6 method)
			(push (cons (nth 6 method) method-full-name) code-table))))
		  (let* ((group-data (make-method-groups:method-group
				      :title
				      (method-group-mode:make-title-adding-stage-if-necessary
				       method-group-mode:title
				       default-stage-name)
				      :methods staged-methods
				      :stage default-stage
				      :code-letters code-table
				      :title-style-default (method-groups:get-misc-data "method-title-style")
				      :style-default (method-groups:get-misc-data "line-style")
				      :layout-default (method-groups:get-misc-data "layout-style")))
			 (pair-by-filename (assoc (file-name-nondirectory (buffer-file-name)) method-groups:all-groups))
			 (pair-by-title (assoc (method-groups:method-group-title group-data) method-groups:all-groups))
			 )
		    (if pair-by-filename
			(rplacd pair-by-filename group-data)
		      (push (cons (file-name-nondirectory (buffer-file-name)) group-data) method-groups:all-groups))
		    (if pair-by-title
			(rplacd pair-by-title group-data)
		      (push (cons (method-groups:method-group-title group-data) group-data) method-groups:all-groups))
		    group-data)))))
    (error "Not a method group file")))

(defun method-groups:file-group-data (file &optional stage-getter)
  "Return the full group data for FILE.
If the stage cannot be worked out, and STAGE-GETTER is given,
call stage-getter with a prompt as argument to get the stage."
  (interactive (list (read-file-name "Test file group data for: ") 'ringing:read-stage-from-user))
  (save-window-excursion
    (let ((was-displaying (ringing:get-file-window file)))
      (find-file file)
      (let ((result (method-group-mode:group-data stage-getter)))
	(unless was-displaying (bury-buffer))
	result))))

(defun method-groups:find-named-group (group)
  "Return the data structure for GROUP."
  (cond
   ((method-groups:method-group-p group)
    group)
   ((string= group "*Touch book*")
    (require 'touch-book)
    (touch-book:touch-book-as-group))
   ((string= group "*Methods rung this session*")
    (touch-book:methods-rung-this-session-as-group))
   ((string-match "\\*Select.+ methods\\*" group)
    (cdr (assoc group (method-groups:list-ephemeral-groups))))
   ((assoc group (method-groups:all-groups))
    (cdr (assoc group (method-groups:all-groups))))
    ((string-match "\\*My \\([a-z]+\\) methods\\*" group)
     (let ((stage-name (substring group (match-beginning 1) (match-end 1))))
       (touch-book:touch-book-as-group stage-name)))
   (t
    (message "Looking for group file %s" group)
    (let ((group-data (method-groups:file-group-data
		       (expand-file-name group method-groups:groups-directory))))
      ;; (message "Got data %s" group-data)
      group-data))))

(defun method-groups:singleton-group (method-name)
  "Return a method group containing only METHOD-NAME."
  (let* ((method (ringing:find-named-method method-name))
	 (really-name (format "*singleton group of %s*" (ringing:method-name method-name)))
	 (found-already-pair (assoc really-name method-groups:all-groups)))
    (if found-already-pair
	(cdr found-already-pair)
      (let ((group (cons really-name
			 (make-method-groups:method-group
			  :title really-name
			  :stage (change-ringing-method-nbells method)
			  :methods (list (cons really-name method))
			  :code-letters nil
			  :title-style-default nil
			  :style-default nil
			  :layout-default nil))))
	(push group
	      method-groups:all-groups)
	group))))

(defun method-groups:group-methods (group)
  "Return the method names for GROUP.
Each name is dotted with its method data structure."
  (method-groups:method-group-methods
   (method-groups:find-named-group group)))

(defun method-groups:group-name (group-or-name)
  "Return the name of GROUP-OR-NAME, or return unchanged if a string."
  (cond
   ((stringp group-or-name) group-or-name)
   ((method-groups:method-group-p group-or-name)
    (method-groups:method-group-title group-or-name))
   (t group-or-name)))

(defun method-group:group-stage (group-name)
  "Return the stage of GROUP-NAME if known, else nil."
  (let ((group-data (method-groups:find-named-group group-name)))
  (catch 'sussed
    (dolist (method-pair (method-groups:group-methods group-name))
      (when (change-ringing-method-p (cdr method-pair))
	(throw 'sussed
	       (change-ringing-method-nbells (cdr method-pair)))))
    nil)))

(defun method-group:random-method-from-group (group &optional not-this-one)
  "From GROUP return any method, but optionally NOT-THIS-ONE."
  (let* ((the-methods (method-groups:group-methods group))
	 (n-methods (length the-methods)))
    (if (= n-methods 1)
	(car (first the-methods))
      (let ((chosen (car (nth (random n-methods) the-methods))))
	(if (null not-this-one)
	    chosen
	  (let ((not-this-name (ringing:method-name not-this-one)))
	    ;; think of something known to terminate for this: !!!!!!!!!!!!!!!!
	    (while (string= not-this-name chosen)
	      (setq chosen (car (nth (random n-methods) the-methods))))
	    chosen))))))

(defun method-group:make-group-buffer (group name)
  "Make a group buffer for GROUP.
This is meant for creating a group from the ephemeral group in an mSiRiL
method library buffer."
  (interactive (list (completing-read "Make group buffer for group: "
				      (method-groups:all-groups)
				      nil
				      t
				      (car (method-group:buffer-default-group))
				      )
		     (read-from-minibuffer "Group name: ")))
  (let* ((the-methods (method-groups:group-methods group))
	 (n-methods (length the-methods)))
    (when (get-buffer name)
      (setq name (generate-new-buffer-name name)))
    (switch-to-buffer (get-buffer-create name))
    (erase-buffer)
    (insert (format "# %s methods\n" name))
    (dolist (method (reverse the-methods))
      (let* ((full-name (car method))
	     (short-name (microSiRiL:name-used-in-file full-name))
	     (file-format (method-library:file-format-for-method full-name)))
	(insert (format "%s %s\n"
			(format file-format (change-ringing-method-nbells (cdr method)))
			short-name))))
    (method-group-mode)
    (setq default-directory method-groups:groups-directory)))

;;;; split methods

(defun method-groups:split-methods (group)
  "Return the split methods for GROUP."
  (let ((split-methods (method-groups:method-group-split-methods group)))
    (when (null split-methods)
      (setq split-methods 
	    (mapcar (function
		     (lambda (pair)
		       (ringing:split-around-treble (cdr pair))))
		    (method-groups:method-group-methods group)))
      (setf (method-groups:method-group-split-methods group) split-methods))
    split-methods))

;;;; Lead matrices

(defun method-group:lead-matrix (group)
  "Return the lead matrix for GROUP."
  (let* ((group-data (method-groups:find-named-group group))
	 (methods (method-groups:method-group-methods group-data))
	 (stage (method-groups:method-group-stage group-data))
	 (from-place-bells (make-vector (1+ stage) nil))
	 )
    (dotimes (i (1+ stage))
      (unless (zerop i)
	(aset from-place-bells i (make-vector (1+ stage) nil))))
    (dolist (method methods)
      (let* ((matrices (ringing:plain-lead-matrices (car method)))
	     (forward (car matrices))
	     (backward (cdr matrices)))
	(dotimes (i (1+ stage))
	  (when (> i 1)
	    (let ((from-this-place (aref from-place-bells i)))
	      (push (cdr method) (aref from-this-place 0))
	      (push (cdr method) (aref from-this-place (aref forward i))))))))
    from-place-bells))

(defun method-group:display-lead-matrix (group-or-matrix)
  (interactive (list (completing-read "Display lead matrix for group: "
				      (method-groups:all-groups)
				      nil
				      t
				      (method-group:buffer-default-group))))
  "Display the lead matrix for GROUP-OR-MATRIX."
  (let* ((matrix (if (vectorp group-or-matrix)
		     group-or-matrix
		   (method-group:lead-matrix group-or-matrix)))
	 (stage (1- (length matrix))))
    (with-output-to-temp-buffer "*Lead matrix*"
      (dotimes (from (1+ stage))
	(when (> from 1)
	  (let ((from-here (aref matrix from)))
	    (princ (format "Methods to take you from place bell %d...\n" from))
	    (dotimes (to (1+ stage))
	      (when (and (/= from to) (> to 1))
		(princ (format "   ... to place bell %d:\n" to))
		(dolist (method (aref from-here to))
		  (princ (format "       %s\n" (change-ringing-method-title method))))))))))))

;;;; Printing and previewing groups

(defun method-group:print-group (group
				 &optional style style-parameters
				 title-style title-style-parameters
				 grid-style grid-style-parameters
				 group-style group-style-parameters
				 printed-title)
  "Print GROUP via PostScript, with optional style args as for ringing:postscript-for-group."
  (interactive (method-group:print-or-preview-group-parameters (method-group:read-group-name "Print group: ")))
  (method-group:print-or-preview-group group
				       'ringing:send-postscript-to-printer
				       style style-parameters
				       title-style title-style-parameters
                                       grid-style grid-style-parameters
				       group-style group-style-parameters
				       printed-title))

(defun method-group:preview-group (group
				   &optional style style-parameters
				   title-style title-style-parameters
				   grid-style grid-style-parameters
				   group-style group-style-parameters
				   printed-title)
  "Preview STAGE, GROUP via PostScript, with optional style args as for ringing:postscript-for-group."
  (interactive (method-group:print-or-preview-group-parameters (method-group:read-group-name "Preview group: ")))
  (method-group:print-or-preview-group group
				       'ringing:send-postscript-to-previewer
				       style style-parameters
				       title-style title-style-parameters
                                       grid-style grid-style-parameters
				       group-style group-style-parameters
				       printed-title))

(defun method-group:buffer-default-group ()
  "Return a default group, sensibly from the current buffer."
  (case major-mode
    (mSiRiL-mode (mSiRiL-mode:buffer-ephemeral-group (current-buffer)))
    (method-group-mode (cons (method-groups:method-group-title (method-group-mode:group-data)) (method-group-mode:group-data)))
    (t nil)))

(defun method-group:read-group-name (prompt)
  "Read a group name, with PROMPT and sensible defaults."
  (completing-read prompt
		   (method-groups:all-groups)
		   nil
		   t
		   (car (method-group:buffer-default-group))))

(defun method-group:print-or-preview-group-parameters (group-name)
  "Get the parameter list to go with GROUP-NAME, for callers of method-group:print-or-preview-group."
  (require 'ps-line)
  (let* ((group-data (method-groups:find-named-group group-name))
	 (printed-title (read-from-minibuffer "Title to put on page: "
					      (method-groups:method-group-title group-data)))
	 (params (ringing:get-postscript-parameters-interactively t)))
    (append (list group-name)
	    params
	    (list printed-title))))

(defun method-group:print-or-preview-group (group
					    output-function
					    &optional style style-parameters
					    title-style title-style-parameters
					    grid-style grid-style-parameters
					    group-style group-style-parameters
					    printed-title)
  "Print or preview GROUP via PostScript, with optional style args as for ringing:postscript-for-group."
  (require 'ps-line)
  (let ((ps-file-name (expand-file-name (format "%s.ps"
						(ringing:name-as-file-name group))
					mSiRiL-mode:ps-temp-files-directory)))
    (save-window-excursion
      (find-file ps-file-name)
      (erase-buffer)
      (ringing:postscript-for-group group
				    (current-buffer)
				    style style-parameters
				    title-style title-style-parameters
				    group-style group-style-parameters
				    grid-style grid-style-parameters
				    printed-title)
      (require 'mSiRiL-mode)		; previewer variable is herein
      (basic-save-buffer)
      (apply output-function (list ps-file-name))
      (unless ringing:keep-postscript-buffers
	(kill-buffer nil))
      (unless ringing:keep-postscript-files
	(delete-file ps-file-name)))))

;;; end of method-groups.el
