;;;; method-library.el -- part of John's elisp change-ringing system
;;; Read microSiRiL method library files
;;; Time-stamp: <99/11/03 09:50:10 johnstu>

(eval-when (compile load eval)
  (provide 'method-library)
  (require 'place-notation)
  (require 'mSiRiL-mode))

(defconst ringing:method-kinds-by-file
  '((alliance "alliance" "^a[0-9]*$" "a%d")
    (delight "delight" "^d[0-9]*$" "d%d")
    (hybrid "hybrid" "^h[0-9]*$" "h%d")
    (surprise "surprise" "^s[0-9]*$" "s%d")
    (treble-bob "treble bob" "^t[0-9]*$" "t%d")
    (treble-place "treble place" "^tp[0-9]*$" "tp%d")
    (plain "" "^p[0-9]*$" "plain" "p%d")
    (principle "" "^o[0-9]*$" "principle" "o%d")
    (dixonoid "dixonoid" "^x[0-9]*$" "x%d"))
  "What kinds of method should be in each kind of file.
The first element of each list is a symbol naming that kind of method.
The second element is the string to insert in the name of such methods.
The third element is a regular expression matching the last element of
filenames for files that should contain that kind of method in the
method library. The fourth element is a format string taking the number
of bells, to recreate the file name from the kind and the stage.")

(defconst ringing:method-kind-names
  (delete-if (lambda (str) (string= str ""))
	     (mapcar 'second ringing:method-kinds-by-file))
  "The kinds of methods.")

(defconst ringing:method-kind-names-regexp
  (concat "\\(" (mapconcat 'identity ringing:method-kind-names "\\)\\|\\(") "\\)")
  "Regular expression matching method kind names")

(defun ringing:figure-out-kind (file)
  "Figure out which kinds of methods should be in FILE."
  (let ((kinds ringing:method-kinds-by-file)
	(last-part (file-name-nondirectory file)))
    (catch 'got-kind
      (while kinds
	(let ((kind (car kinds)))
	  (if (or (string-match (second kind) file)
		  (string-match (third kind) last-part))
	      (throw 'got-kind (first kind))))
	(setq kinds (cdr kinds)))
      nil)))

(defun ringing:kind-by-name (kind &optional always)
  "Return the string describing methods of KIND."
  (let* ((descr (assoc kind ringing:method-kinds-by-file))
	(kind (second descr)))
    (if (and (zerop (length kind))
	     always
	     (>= (length descr) 4))
	(nth 3 descr)
      kind)))

(defvar microSiRiL:methods-in-files nil
  "Alist of filenames to methods in those files.")

(defun microSiRiL:constructed-name (main-name kind stage)
  "Construct a full method name from MAIN-NAME KIND STAGE."
  (microSiRiL:munge-name
   (if (and (stringp kind) (not (zerop (length kind))))
       (concat main-name " " kind " " stage)
     (concat main-name " " stage))))

(defun strip-extra-spaces (string)
  "Remove leading, trailing, and internal repeated spaces from STRING"
  (if (string-match "^ *\\(.+[^ ]+\\) *$" string)
      (progn
	(setq string (substring string (match-beginning 1) (match-end 1)))
	(while (string-match "  +" string)
	  (setq string (concat (substring string 0 (1+ (match-beginning 0)))
			       (substring string (match-end 0)))))
	string)
    string))

(defvar microSiRiL:separable-name-words
  '("pleasure"
    "court"
    "place"
    "bob"
    "college"
    "little"
    )
  "List of strings to make into separate words in method names,
which is done by microSiRiL:munge-name")

(defconst microSiRiL:separable-name-words-2
  '("pleasure" "court" "place" "bob" "college" "delight"
    ;; "slow"
    "course"
    "common" "single" 
    ;; "double"
    "reverse" "fen" "springs" "way" "little" "abbey"
    "scholars" "victory"
    ;; "new"
    "bishops" "county" "supreme" "festival"
    "upper" "lower" "exercise" "norton"
    )
  "List of strings to make into separate words in method names,
which is done by microSiRiL:munge-name")

(defun microSiRiL:munge-name (whole-name)
  "Try to make a better real-world name from WHOLE-NAME."
  (remove-text-properties 0 (1- (length whole-name)) '(face nil) whole-name)
  (capitalize
   (let ((case-fold-search t))
     (when (string-match "^st\\([bcdfghjklmnpqstvwxz].+\\)$" whole-name)
       (setq whole-name (concat "St "
				(capitalize (substring whole-name
						       (match-beginning 1)
						       (match-end 1))))))
     (dolist (word microSiRiL:separable-name-words)
       (when (string-match (format "^\\(.*\\)\\(%s\\)\\(.*\\)$" word) whole-name)
	 (setq whole-name (concat (substring whole-name
					     (match-beginning 1)
					     (match-end 1))
				  " "
				  (substring whole-name
					     (match-beginning 2)
					     (match-end 2))
				  " "
				  (substring whole-name
					     (match-beginning 3)
					     (match-end 3))))))
     (strip-extra-spaces whole-name))))

(defconst microSiRiL:line-parts-regexp
  "^\\([a-z0-9]+\\) +\\([0-9]*[a-z]+[0-9]?\\) +\\(.+\\)$"
  "How to pull apart a microSiRiL library line.")

(defvar microSiRiL:load-start-time 0
  "When the latest load started")

(defvar microSiRiL:load-end-time 0
  "When the latest load started")

(defun microSiRiL:apologize-for-delay-on (file &optional n-loaded)
  "Put up a message about the delay in reading FILE; or say that N-LOADED were loaded."
  (let ((old-buffer (current-buffer))
	(new-buffer (ringing:get-buffer-create "*Apologies for the delay*")))
    (switch-to-buffer-other-window new-buffer)
    (set-buffer new-buffer)
    (goto-char (point-max))
    (if (null n-loaded)
	(progn
	  (insert "Reading " file
		  " (" (ringing:kind-by-name (ringing:figure-out-kind file) t)
		  " " (ringing:stage-name (ringing:figure-out-number-of-bells file)) ")")
	  (setq microSiRiL:load-start-time (current-time)))
      (progn
	(setq microSiRiL:load-end-time (current-time))
	(insert (format " -- %d methods loaded in %s\n"
			n-loaded
			(time-string
			 (seconds-between
			  microSiRiL:load-start-time
			  microSiRiL:load-end-time))))))
    (sit-for 0)
    (other-window -1)))

(defvar microSiRiL:find-method-library-file-hook nil
  "Hook to run on reading a microSiRiL method library file.")

(defvar method-library:full-progress-messages t
  "Whether to display each method name while an mSiRiL file is scanned.
A nice way of showing progress, but also slows the progress down a bit (on
the author's system, this is quite a small effect; with a slow network
connection it might make more difference)..") 

(defun microSiRiL:parse-file-superficially (file &optional force)
  "Read the method names and descriptions from FILE,
adding them to the list of known methods, but don't actually parse the place
notation into method structures -- just leave it as strings for parsing as
needed. If called on a directory, will recurse through it.
Uses a cache in microSiRiL:methods-in-files unless optional &force is non-nil."
  (interactive "fScan methods in file: ")
  (setq file (file-truename file))
  (let ((from-cache (or force (assoc file microSiRiL:methods-in-files))))
    (if (or force (null from-cache))
	(if (file-directory-p file)
	    (let ((the-files (directory-files file nil nil t)))
	      (while the-files
		(let ((this-file (car the-files)))
		  (if (not (string-match "^\\.\\.?$" this-file))
		      (microSiRiL:parse-file-superficially
		       (expand-file-name this-file file))))
		(setq the-files (cdr the-files))))
	  (save-window-excursion
	    (let ((file-old-buffer (find-buffer-visiting file))
		  (this-file-methods nil)
		  (nbells (ringing:stage-name
			   (ringing:figure-out-number-of-bells file)))
		  (kind-string (ringing:kind-by-name (ringing:figure-out-kind file))))
	      (microSiRiL:apologize-for-delay-on file nil)
	      (find-file file)
	      (run-hooks 'mSiRiL-mode-hook 'microSiRiL:find-method-library-file-hook)
	      (message "Scanning %s" file)
	      (let ((case-fold-search t))
		(save-excursion
		  (save-restriction
		    (widen)
		    (goto-char (point-min))
		    (while (re-search-forward microSiRiL:line-parts-regexp
					      (point-max) t)
		      (when method-library:full-progress-messages (message (buffer-matched 1)))
		      (let* (;; save all the bits from the buffer, as microSiRiL:munge-name destroys match-data
			     (base-name  (buffer-matched 1))
			     (group-name (intern (buffer-matched 2)))
			     (place-string (buffer-matched 3))
			     (this-method (list (microSiRiL:constructed-name
						 base-name
						 kind-string
						 nbells)
						group-name
						place-string)))
			(push this-method
			      ringing:known-methods)
			(push this-method
			      this-file-methods))))))
	      (if (null file-old-buffer) (kill-buffer nil))
	      (if (assoc file microSiRiL:methods-in-files)
		  (rplacd (assoc file microSiRiL:methods-in-files)
			  this-file-methods)
		(push (cons file this-file-methods)
		      microSiRiL:methods-in-files))
	      (microSiRiL:apologize-for-delay-on file (length this-file-methods))
	      this-file-methods)))
      (cdr from-cache))))

(defvar microSiRiL:parse-whole-files t
  "*Whether to load the whole file into our data structures when you look at a method in the file.")

(defun microSiRiL:name-used-in-file (name-as-given)
  "Like the reverse of microSiRiL:constructed-name just for the main part of the name."
  (let ((stage-start (string-match (concat " \\(" ringing:stage-names-regexp "\\)$") name-as-given)))
    (when stage-start
      (setq name-as-given (substring name-as-given 0 stage-start))))
  ;; (message "Removed stage, now \"%s\"" name-as-given)
  (let ((kind-start (string-match (concat " " ringing:method-kind-names-regexp " *$") name-as-given)))
    (when kind-start
      (setq name-as-given (substring name-as-given 0 kind-start))))
  ;; (message "Removed kind, now  \"%s\"" name-as-given)
  (while (string-match " +" name-as-given)
    (setq name-as-given (concat (substring name-as-given 0 (match-beginning 0))
				(substring name-as-given (match-end 0)))))
  (downcase name-as-given))

(defun method-library:locate-method (method-name)
  "Find the definition of METHOD-NAME."
  (let* ((method (ringing:find-named-method method-name))
	 (name-in-file (microSiRiL:name-used-in-file method-name)))
    (find-file-other-window (change-ringing-method-defining-file method))
    (goto-char (point-min))
    (re-search-forward (format "^%s " name-in-file))
    (beginning-of-line 1)))

(defun microSiRiL:method-from-file (file method-name &optional no-error force)
  "Look in FILE for METHOD-NAME."
  (if microSiRiL:parse-whole-files
      (progn
	(microSiRiL:parse-file-superficially file force)  
	(let ((pair (assoc method-name ringing:known-methods)))
	  (if pair
	      (place-notation:as-method pair file)
	    (if no-error
		nil
	      (error "Could not find definition of %s" method-name)))))
    (save-match-data
      (save-window-excursion
	(find-file file)
	(save-excursion
	  (goto-char (point-min))
	  (if (re-search-forward (format "^%s " (microSiRiL:name-used-in-file method-name))
				 (point-max)
				 t)
	      (progn
		(goto-char (match-beginning 0))
		(if (re-search-forward microSiRiL:line-parts-regexp
				       (save-excursion (end-of-line 1) (point))
				       t)
		    (microSiRiL:convert-from-string
		     method-name ;; (concat (buffer-matched 1) " " "size")
						    (intern (buffer-matched 2))
						    (buffer-matched 3))
		  nil))
	    nil))))))

(defun microSiRiL:ends (code nbells)
  "Return list of place notation (plain bob single) for method CODE and NBELLS."
  (let* ((code-string (symbol-name code))
	 (first-letter (aref code-string 0))
	 (is-x (string-match "x$" code-string))
	 (is-z (string-match "\\([0-9et]+\\)z" code-string))
	 (high-bob-and-single
	  (list (format "1%d" (- nbells 2))
		(format "1%d%d%d" (- nbells 2) (1- nbells) nbells)))
	 (low-bob-and-single (list "14" "1234")))
    (if is-z
	(let ((plain-string (substring code-string (match-beginning 1) (match-end 1))))
	  (cons plain-string high-bob-and-single)
	  )
      (case first-letter
	((?a ?b ?c ?d ?e ?f)
	 (cons "12" (if is-x high-bob-and-single low-bob-and-single)))
	((?g ?h ?j ?k ?l ?m)
	 (cons (format "1%c" (ringing:bell-number-to-char nbells))
	       (if is-x low-bob-and-single high-bob-and-single)))
	(t nil)))))

(defun microSiRiL:just-autocalls-string (string)
  "Return just the autocalls part of STRING."
  (let ((semicolonlocation (string-match ";" string)))
    (if semicolonlocation
	(substring string semicolonlocation)
      nil)))

(defun microSiRiL:without-autocalls-string (string)
  "Return STRING stripped of any autocall place notation."
  (let ((semicolonlocation (string-match ";" string)))
    (if semicolonlocation
	(substring string 0 semicolonlocation)
      string)))

(defun microSiRiL:convert-from-string (method-name method-kind method-string &optional file-of-origin)
  "Make a method from METHOD-NAME METHOD-KIND and METHOD-STRING.
If FILE-OF-ORIGIN is given, record it as coming from that file."
  (place-notation:debug "Converting %s %s %s\n" method-name method-kind method-string)
  (let* ((nbells (ringing:figure-out-number-of-bells method-name))
         (normal-method-string (microSiRiL:without-autocalls-string method-string))
	 (base-changes-list (place-notation:parse-string
			     normal-method-string
			     nbells))
	 (given-entire-plain (string-match " *\\+" normal-method-string))
	 (reversed-base-if-needed (if given-entire-plain
				      (reverse base-changes-list)
				    nil))
	 (ends (microSiRiL:ends method-kind nbells))
	 (plain-end (if given-entire-plain
			(list (car reversed-base-if-needed))
		      (place-notation:parse-string (first ends) nbells)))
	 (method-data (make-change-ringing-method
		       :title method-name
		       :nbells nbells
		       :plain (if given-entire-plain
				  base-changes-list
				(append base-changes-list plain-end))
		       :work (if given-entire-plain
				 (reverse (cdr reversed-base-if-needed))
			       base-changes-list)
		       :plain-lead-end plain-end
		       :bob (place-notation:parse-string (second ends) nbells)
		       :single (place-notation:parse-string (third ends) nbells)
		       :special-start (if (and nil (string-match "^Stedman [a-z]+$" method-name))
					  (nthcdr (/ (length base-changes-list) 2) base-changes-list)
					nil)
		       :autocalls (place-notation:parse-autocalls
				   (microSiRiL:just-autocalls-string method-string) nbells)
		       :original-description-string (format "%s %s" method-kind method-string)
		       :defining-file file-of-origin)))
    (place-notation:debug "Converted %s %s %s --> %s (%s)\n"
			  method-name
			  method-kind
			  method-string
			  method-data
			  (if given-entire-plain "no implied lead" "implied lead"))
    (place-notation:debug-dump-method method-data)
    method-data))

(defun ringing:equal-methods (a b &optional just-work)
  "Return whether methods A and B are the same.
Their works and lead ends are compared, but not their names.
With optional JUST-WORK, compare only the work, not the lead ends."
  (and (equal (change-ringing-method-work a)
	      (change-ringing-method-work b))
       (or just-work
	   (equal (change-ringing-method-plain-lead-end a)
		  (change-ringing-method-plain-lead-end b)))))

(defun method-library:find-name-for-method (method)
  "Try to identify METHOD by name.
This is done by searching all method files for that stage, looking for that place notation."
  "Place notation search not implemented"
  (let ((plain (change-ringing-method-plain method))
	(work (change-ringing-method-work method)))
    (if (or plain work)
	(let ((work-string 
	       (if plain
		   (method-library:just-work-string (place-notation:recreate-string-from-changes plain t))
		 (method-library:just-work-string (place-notation:recreate-string-from-changes work t))))
	      (files (method-library:files-for-n-bells
		      (change-ringing-method-nbells method))))
	  ;; (message "Looking for method with work-string \"%s\"" work-string)
	  (catch 'found
	    (dolist (method ringing:known-methods)
	      ;; (message "Comparing %s against %s (%s)" work-string (car method) (method-library:method-work-string (cdr method)))
	      (when (string= (method-library:method-work-string (cdr method)) work-string)
		(throw 'found (car method))))
	    "Anon"))
      "no work for this method")))

(defun place-notation:as-method (pair &optional file-of-origin)
  "Return the method described by PAIR,
converting from place notation and stashing back in PAIR if appropriate.
PAIR is a list of which the car is the method name, and the cdr is the
method kind code and the place notation.
If FILE-OF-ORIGIN is given, mark that as the file it came from, if making the method structure."
  (let ((meth (cdr pair)))
    (if (change-ringing-method-p meth)
	meth
      (rplacd pair (microSiRiL:convert-from-string
		    (car pair)
		    (first meth)
		    (second meth)
		    file-of-origin)))))

(defvar method-library:directory
  (expand-file-name "~/Ringing/method-libraries/")
  "*Where to look for the mSiRiL method libraries.
The filenames in this directory should be as in the Internet central repository
of methods, pointed to by method-library:standard-network-library-URL.
See also place-notation:default-methods-directory, for where to put added
methods of your own, in a format similar to that down the sides of methods
in the Ringing World diary.")

(defconst method-library:standard-network-library-URL
  "http://sunsite.doc.ic.ac.uk/recreation/change-ringers/ringing-programs/microsiril/method-libraries/"
  "The usual change-ringing method repository on the network.")

(defun method-library:ftp-libraries ()
  "Fetch a new version of the method libraries."
  (interactive)
  (ringing:ftp-directory-files method-library:standard-network-library-URL
			       method-library:directory))

(defun ringing:ftp-directory-files (source-URL
				    destination-directory)
  "Copy the files from source-URL to DESTINATION-DIRECTORY."
  (interactive "sURL:
DDirectory: ")
  (unless (file-directory-p destination-directory)
    (make-directory destination-directory t))
  (if (string-match "^http://\\([.a-z0-9]+\\)/\\(.+\\)$" source-URL)
      (let* ((remote-host (substring source-URL
				     (match-beginning 1) (match-end 1)))
	     (remote-local-directory-name (substring source-URL
						     (match-beginning 2) (match-end 2)))
	     (local-remote-directory-name (format "/anonymous@%s:/%s"
						  remote-host remote-local-directory-name))
	     (remote-files (directory-files local-remote-directory-name
					    nil "[a-z]+[0-9]+$" t)))
	(dolist (method-library-name remote-files)
	  (let ((remote-file-name (expand-file-name method-library-name
						    local-remote-directory-name))
		(local-file-name (expand-file-name method-library-name
						   destination-directory)))
	    (copy-file remote-file-name local-file-name t)))))
  (place-notation:forget-methods))

(defun method-library:files-for-n-bells (nbells)
  "Return an alist of method library files whose names seem to match NBELLS."
  (let ((all-method-files (directory-files method-library:directory
					   t (format "[a-z]+%d$" nbells))))
    (if (null all-method-files)
	(with-output-to-temp-buffer "*Whence the methods come*"
	  (princ "The methods library directory is empty.\n")
	  (princ "The standard library may be fetched from ")
	  (princ method-library:standard-network-library-URL)
	  (princ " and should be placed in ")
	  (princ method-library:directory)
	  (princ "\nThe command method-library:ftp-libraries may be used to do this.\n")
	  (error "Method files not found"))
      (mapcar (lambda (fn) (cons (file-name-nondirectory fn) fn))
	      all-method-files))))

(defun method-library:choose-method-file (prompt-base nbells &optional auto)
  "Prompt based on PROMPT-BASE to get a list containing a file chosen by the user for NBELLS methods.
With optional third argument non-nil, don't prompt the user, but return
all method files for NBELLS."
  (let ((possibles (method-library:files-for-n-bells nbells))
	(completion-ignore-case t))
    (if auto
	(mapcar 'cdr possibles)
      (list (cdr (assoc (completing-read (concat prompt-base "Method file: ")
					 possibles nil t)
			possibles))))))

(defvar method-library:files-by-description nil
  "Cache of alist of types and stages of method to library filename.
Example entry: (\"Surprise Minor\" . \"~/methods/s6\").
Normally, use the function of same name to access the list, in case
the directory has changed.")

(defun method-library:files-by-description (&optional force)
  "Return cache of alist of types and stages of method to library filename.
Example entry: (\"Surprise Minor\" . \"~/methods/s6\")."
  (if (and (not force) method-library:files-by-description)
      method-library:files-by-description
    (progn
      (setq method-library:files-by-description nil)
      (dolist (file (directory-files method-library:directory t "[a-z]+[0-9]$"))
	(let* ((stage (ringing:stage-name (ringing:figure-out-number-of-bells file)))
	       (kind (ringing:figure-out-kind file))
	       (description
		(case kind
		  (plain
		   (concat "Plain " stage))
		  (principle
		   (concat stage " Principles"))
		  (otherwise
		   (concat (ringing:kind-by-name kind)
			   " "
			   stage)))))
	  (push (cons (capitalize description)
		      file)
		method-library:files-by-description)))
      method-library:files-by-description)))

(defvar method-library:prompt-stage-and-type-separately nil
  "*Whether to prompt for stage and type separately.")

(defun method-library:find-file-by-description (description)
  "Find a method library file by its description."
  (interactive
   (let ((completion-ignore-case t))
     (list (completing-read "Find file containing: "
			    (method-library:files-by-description)
			    nil t))))
  (find-file (cdr (assoc description method-library:files-by-description))))

(defun method-library:choose-method-interactively (prompt-base
						   &optional user-choose-file)
  "Prompt the user for a method using PROMPT-BASE for the prompts.
With optional second argument non-nil, don't prompt the user for which method
collection file to use, but look in all of them for the chosen number of bells."
  (let* ((nbells (ringing:read-stage-from-user
		  (concat prompt-base "Number of bells: ")))
	 (files (method-library:choose-method-file
		 prompt-base nbells
		 (not user-choose-file)))
	 (method-lists (mapcar 'microSiRiL:parse-file-superficially files))
	 (methods (apply 'concatenate 'list method-lists))
	 (completion-ignore-case t)
	 (method-base-name (completing-read prompt-base methods nil t)))
    method-base-name
    ))

(defun method-library:report-on-loaded-methods (&optional load-all)
  "List the currently loaded methods.
With optional / prefix argument, load all the method files first,
which may take several minutes."
  (interactive "P")
  (with-output-to-temp-buffer "*Summary of methods currently loaded*"
    (let ((total 0)
	  (stage 4))
      (princ "Number of methods found in each file opened, by stage:\n")
      (while (<= stage 16)
	(let ((stage-total 0)
	      (this-file-methods (method-library:files-for-n-bells stage))
	      )
	  (princ (format "\n%s methods:\n" (ringing:stage-name stage)))
	  (mapcar (function
		   (lambda (file-short-and-full)
		     (let* ((full-name (cdr file-short-and-full))
			    (short-name (car file-short-and-full))
			    (pair (assoc full-name microSiRiL:methods-in-files))
			    (rest (cdr pair)))
		       (when rest
			 (incf total (length rest))
			 (incf stage-total (length rest)))
		       (princ
			(format "  %s (%s %s): "
				short-name
				(ringing:kind-by-name (ringing:figure-out-kind full-name) t)
				(ringing:stage-name (ringing:figure-out-number-of-bells full-name))))
		       (princ
			(if rest
			    (int-to-string (length rest))
			  "none"))
		       (princ "\n"))))
		  ;; microSiRiL:methods-in-files
		  this-file-methods)
	  (princ (format " %d %s methods loaded\n" stage-total (ringing:stage-name stage))))
	(incf stage))
      (princ (format "Total: %d methods loaded\n" total)))))

(defun method-library:file-format-for-method (method-name)
  "Return the format string for the filename for methods like METHOD-NAME."
  (let ((kinds ringing:method-kinds-by-file))
    (catch 'matched
      (while kinds
	(let* ((kind (car kinds))
	       (kindstring (second kind)))
	  (when (and (stringp kindstring)
		     (not (string= kindstring ""))
		     (string-match kindstring method-name))
	    (throw 'matched (fourth kind)))
	  (setq kinds (cdr kinds))))
      "p%d")))

(defun method-library:proper-file-for-method (method-name)
  "Return the name of the file that should define METHOD-NAME,
according to the mSiRiL file naming conventions."
  (let* ((kind-format (method-library:file-format-for-method method-name))
	 (stage-number (ringing:figure-out-number-of-bells method-name)))
    (if (and (stringp kind-format) (numberp stage-number))
	(expand-file-name (format kind-format stage-number)
			  method-library:directory)
      nil)))

(defun method-library:display-method (method &optional display-buffer-function)
  "Display METHOD, optionally using DISPLAY-BUFFER-FUNCTION."
  (let* ((method-name (ringing:method-name method))
	 (name-in-file (microSiRiL:name-used-in-file method-name))
	 (buffer (find-file-noselect method-name)))
    (funcall (if display-buffer-function
		 display-buffer-function
	       'switch-to-buffer)
	     buffer)
    (goto-char (point-min))
    (unless (search-forward name-in-file (point-max) t)
      (error "Could not find %s in expected file" method-name))))

(defun method-library:load-all-files (&optional force)
  "Load all the files in method-library:directory.
With optional / prefix arg, re-load any that are already loaded."
  (interactive "P")
  (when force
    ;; force unload, so when we observe garbage collection results,
    ;; we get an idea of the data size
    (place-notation:forget-methods))
  (let* ((start-garbage-collect (garbage-collect))
	 (start-time (current-time))
	 (load-results
	  (mapcar
	   (function
	    (lambda (filename)
	      (unless (file-directory-p filename)
		(microSiRiL:parse-file-superficially filename force))))
	   (directory-files method-library:directory t nil t)))
	 (end-time (current-time))
	 (end-garbage-collect (garbage-collect)))
    (require 'manual-ringing)		; for time-string and seconds-between
    (message "Loading all methods took %s%s"
	     (time-string (seconds-between start-time end-time))
	     (if force
		 (format "; %d more conses, %d bytes more strings"
			 (- (caar end-garbage-collect) (caar start-garbage-collect))
			 (- (fourth end-garbage-collect) (fourth start-garbage-collect)))
	       ""))))

(defun method-library:random-method-from-file (file)
  "Return a method chosen at random from those in FILE."
  (interactive "fPick a random method from file: ")
  (let* ((methods (microSiRiL:parse-file-superficially file)))
    (nth (random (length methods)) methods)))

(defun ringing:lowest-working-bell-making-places (changes)
  "Return the number of the lowest bell apart from treble in CHANGES that makes places, else NIL."
  (catch 'found
    (let ((i 2)
	  (n (length changes)))
      (while (< i n)
	(when (= i (aref changes i))
	  (throw 'found i))
	(incf i)))
    nil))

(defun method-library:just-work-string (allstring)
  "Return just the work part of ALLSTRING,
that is, without lead end code."
  (let* ((has-preamble (string-match "[^,] " allstring))
	 (workstart (if has-preamble (match-end 0) 0))
	 (workend (string-match "," allstring)))
    (substring allstring workstart workend)))

(defun method-library:work-minus-half-lead (work)
  "Return WORK minus the half-lead, or nil if half-lead cannot be distinguished."
  (if (string-match "^&\\([-.0-9eta-d]+[-.]\\)\\([0-9eta-d]+\\)$" work)
      (substring work (match-beginning 1) (match-end 1))
    nil))
 
(defun method-library:method-work-string (method)
  "Return the work string for METHOD."
  (cond
   ((consp method) (second method))
   ((change-ringing-method-p method)
    (method-library:just-work-string (change-ringing-method-original-description-string method)))
   (t nil)))

(defun method-library:methods-matching-work (method methods &optional not-this)
  "Return a pair of lists of methods whose work is similar to the work of METHOD, from the list METHODS;
METHODS and the result are in the same form as the list of methods loaded from a file.
The car of the pair is those differing only at the lead end, and
the cdr is those differing at the half-lead.
METHOD may be a method, or the work string of a method.
If NOT-THIS is given, the method of that name is not included."
  (let* ((work (method-library:method-work-string method))
	 (minus-half-lead (method-library:work-minus-half-lead work))
	 (differing-at-lead-end-only nil)
	 (differing-at-half-lead nil))
    (when (and (null not-this)
	       (change-ringing-method-p method))
      (setq not-this (change-ringing-method-title method)))
    (dolist (method methods)
      (when (or (null not-this)
		(not (string= not-this (car method))))
	(let* ((this-work (method-library:method-work-string (cdr method)))
	       (this-minus-half-lead (method-library:work-minus-half-lead this-work)))
	  ;; (when this-minus-half-lead (message "%s = %s, %s = %s? (%s)" work this-work minus-half-lead this-minus-half-lead (car method)))
	  (if (string= work this-work)
	      (progn
		;; (message "lead end match")
		(push method differing-at-lead-end-only))
	    (when (and this-minus-half-lead (string= minus-half-lead this-minus-half-lead))
	      ;; (message "Half-lead match")
	      (push method differing-at-half-lead))))))
    (cons differing-at-lead-end-only differing-at-half-lead)))

(defun method-library:insert-method-list (methods)
  "Insert a list of methods."
  (let ((the-format (format "  %% %ds: %%s\n" (apply 'max (mapcar 'length (mapcar 'car methods))))))
    (dolist (method methods)
      (let ((method-data (ringing:find-named-method (car method))))
	(insert (format the-format
			(capitalize (change-ringing-method-title method-data))
			(place-notation:recreate-string-from-changes (change-ringing-method-plain method-data) t)))))))  

(defun method-library:insert-similar-methods (method methods &optional not-this)
  "Insert methods whose work is that of METHOD, chosen from METHODS, except NOT-THIS if given."
  (let* ((similar-methods (method-library:methods-matching-work
			   method
			   methods
			   not-this))
	 (at-end (car similar-methods))
	 (at-half (cdr similar-methods)))
    (when at-end
      (goto-char (point-max))
      (insert "Methods differing at lead end only:\n")
      (method-library:insert-method-list at-end)
      (insert "\n"))
    (when at-half
      (goto-char (point-max))
      (insert "Methods differing at half-lead:\n")
      (method-library:insert-method-list at-half)
      (insert "\n"))))

(defun method-library:map-methods-in-file (file fn &rest fnargs)
  "To each method in FILE, apply FN with remaining FNARGS, collecting the results."
  (let* ((raw-file-data (microSiRiL:parse-file-superficially file))
	 (result nil))
    (dolist (method-entry raw-file-data)
      (let ((method-data (ringing:find-named-method (car method-entry))))
	(message (change-ringing-method-title method-data))
	(push (apply fn method-data fnargs)
	      result)))
    result))

;;; end of method-library.el
