;;;; method-lines.el -- part of John's elisp change-ringing system
;;; make ASCII graphic method lines from rows of numbers
;;; Time-stamp: <98/05/23 14:32:53 john>

(eval-when (compile load eval)
  (provide 'method-lines)
  (require 'bell-colours))

(defvar ringing:underline-faults t
  "Whether to underline faults.
Like highlighting them, but shows up when you print a buffer in a plain manner.")

(defvar ringing:row-lines-treble ?.
  "The character with which to draw the treble line.
Normally . unless that is being used to provide a grid by putting it in
ringing:row-lines-other-bells.")

(defvar ringing:row-lines-marked-bell ?*
  "The character with which to draw the marked bell.")

(defvar ringing:row-lines-other-marked-bell ?#
  "The character with which to draw the second marked bell, for handbells.")

(defvar ringing:row-lines-other-bells ? 
  "The character with which to draw other bells.
Normally a space, unless you want dots to draw a grid, in which case you
should probably set ringing:row-lines-treble to something other than dot.")

(defvar ringing:row-lines-start-char ? 
  "The character to put at the start of each line of \"line\" characters.")

(defvar ringing:row-lines-end-char ? 
  "The character to put at the end of each line of \"line\" characters.")

(defun ringing:enable-grid (off)
  "Set whether a grid is drawn.
With prefix arg, disable grid, otherwise enable it."
  (interactive "P")
  (if off
      (setq ringing:row-lines-treble ?.
	    ringing:row-lines-other-bells ? )
    (setq ringing:row-lines-treble ?+
	    ringing:row-lines-other-bells ?.)))

(defun ringing:row-lines (row &optional widely)
  "Convert ROW to \"blue\" and \"red\" lines, with ringing:observation as blue.
If optional third arg WIDELY non-nil, add spaces between characters.
Blue is done with a \"*\", red with a \".\", and all others as spaces.
The characters drawn may be changed by setting ringing:row-lines-treble,
ringing:row-lines-marked-bell, and ringing:row-lines-other-bells."
  (let* ((i 0)
	 (space-char (if (and ringing:underline-faults
			      ringing:bad-row) ?_ ringing:row-lines-other-bells))
	 (n (length row))
	 (w (if widely (* 2 n) n))
	 (lines (make-string w ? )))
    (when (and ringing:highlight-bad-rows ringing:bad-row)
      (put-text-property 0 w 'face ringing:bad-face lines))
    (while (< i n)
      (let ((bell (aref row i))
	    (str-index (if widely (* 2 i) i)))
	(cond ((= bell ?1)
	       (ringing:aset-red lines str-index ringing:row-lines-treble))
	      ((or (= bell ringing:observation) (= bell ringing:row-lines-marked-bell))
	       (ringing:aset-blue lines str-index ringing:row-lines-marked-bell))
	      ((and ringing:green-line-too
		    (or (= bell ringing:other-observation) (= bell ringing:row-lines-other-marked-bell)))
	       (ringing:aset-green lines str-index ringing:row-lines-other-marked-bell))
	      (t
	       (aset lines str-index space-char))))
      (if widely (aset lines (1+ (* i 2)) ? ))
      (incf i))
    (aset lines 0 ringing:row-lines-start-char)
    (aset lines (1- w) ringing:row-lines-end-char)
    lines))

(defvar ringing:show-lines-with-lines nil
  "*Whether to draw \\, |, / between rows.

If you do this, the resulting buffer has twice as many lines of text as
it would otherwise, since it still has the * for the blue line and . for
the red line, as before.")

(defvar ringing:previous-row nil
  "The previous row given to ringing:insert-row-lines,
for drawing lines between rows.")

(defvar method-display:row-worth-of-space ""
  "A string of spaces, as long as one row.")

(defun ringing:draw-lines-between-rows (first second &optional comment)
  "Draw the line between rows FIRST and SECOND for ringing:observation (a character).
Do this for the digits and the lines of * and . , according to what output
is enabled."
  (when ringing:show-lines
    (insert ? )
    (ringing:draw-coloured-lines-between-rows first second))
  (when ringing:show-numbers
    (if ringing:show-lines
	(insert method-display:row-worth-of-space)
      (ringing:draw-coloured-lines-between-rows first second)))
  ;; (insert "!")
  (end-of-line 2)
  (when comment (insert comment))
  (if (eobp) (insert ?\n)))

(defun ringing:draw-coloured-lines-between-rows (first second)
  "Draw all the lines between rows FIRST and SECOND."
  (if ringing:green-line-too
      (ringing:draw-lines-between-rows-1
       first second
       ringing:observation ?*
       '(ringing:draw-one-line-segment
	 i first second ringing:other-observation ?#
	 ringing:green-line-down
	 ringing:green-line-places
	 ringing:green-line-up
	 '(ringing:draw-one-line-segment
	   i first second ?1 ?.
	   ringing:red-line-down
	   ringing:red-line-places
	   ringing:red-line-up)))
    (ringing:draw-lines-between-rows-1
     first second ringing:observation ?*
     '(ringing:draw-one-line-segment
       i first second ?1 ?.
       ringing:red-line-down
       ringing:red-line-places
       ringing:red-line-up))))

(defun ringing:draw-one-line-segment (i
				      first second
					mark-me1 mark-me2
					line-down line-places line-up
					&optional further-fun)
  "Draw one piece of line segment.
By the time you get to looking at this function, you should have figured out
what the args mean; else see ringing:draw-lines-between-rows-1."
  (if (or (eq (aref first i) mark-me1)
	  (eq (aref first i) mark-me2))
      (if (eq (aref second i) (aref first i))
	  (insert line-places)
	(if (eq (aref second (1- i)) (aref first i))
	    (insert line-down)
	  (if further-fun
	      (eval further-fun)
	    (insert "  "))))
    (if (and (or (eq (aref second i) mark-me1)
		 (eq (aref second i) mark-me2))
	     (eq (aref first (1- i))
		 (aref second i)))
	(insert line-up)
      (if further-fun
	  (eval further-fun)
	(insert "  ")))))

(defun ringing:draw-lines-between-rows-1 (first second mark-me1 mark-me2 &optional further-fun)
  "Draw the line between rows FIRST and SECOND for MARK-ME1 / MARK-ME2 (characters).
If FURTHER-FUN is given, eval this to fill in the ones we haven't drawn in; otherwise
put spaces therein.
Do this only for one of: the digits; the lines of * and ."
  (let ((len (length first)))
    (if (eq len (length second))
	(let ((i 1))
	  (while (< i len)
	    (ringing:draw-one-line-segment
	     i
	     first second
	     mark-me1 mark-me2
	     ringing:blue-line-down ringing:blue-line-places ringing:blue-line-up
	     further-fun)
	    (incf i))
	  (insert ? )))))

(defun ringing:change-to-place-notation (&optional change)
  "Return a string representing CHANGE (or ringing:most-recent-change if not given) as place notation."
  (let* ((the-change (if change change ringing:most-recent-change))
	(placechars nil)
	(i 1)
	(n (length the-change))
	(any nil))
    (while (< i n)
      (let ((c (aref the-change i)))
	(when (eq c i)
	  (setq any t)
	  (push (ringing:bell-number-to-char c) placechars))
	(incf i)))
    (unless any (push ?x placechars))
    (coerce (nreverse placechars) 'string)))

(defvar ringing:write-place-notation-beside nil
  "Whether to write the place notation beside the crossings.
Used in the place notation / line editor.")

(defun ringing:draw-crossings ()
  "Draw the crossings represented by the most recent change."
  (when ringing:most-recent-change
    (insert ? )
    (let ((i 1)
	  (n (length ringing:most-recent-change)))
      (while (< i n)
	(let ((c (aref ringing:most-recent-change i)))
	  (cond
	   ((eq c i)
	    (insert ?| ? ))
	   ((eq c (1+ i))
	    (insert ?  ?X))
	   (t
	    (insert ?  ?  ))))
	(incf i)))
    (when ringing:write-place-notation-beside
      (insert ?  (ringing:change-to-place-notation))
      (end-of-line 2)
      (if (eobp) (insert ?\n)))))

(defvar ringing:show-lines t
  "*Whether to show the lines as well as the numbers --
or possibly instead, according to ringing:show-numbers.")

(defvar ringing:show-numbers t
  "*Whether to show the numbers as well as the lines --
or possibly instead, according to ringing:show-lines.")

(defvar ringing:draw-crossings nil
  "Whether to draw the crossings.
This is really meant for full-grid-mode only, at least for now.")

(defvar ringing:display-course-bells-only nil
  "Whether to display just the course bells in the numbers column,
or, alternatively, all the bell numbers.")

(defun ringing:insert-only-observation-and-course (row)
  "Of ROW, insert just the observation and course bells."
  (let ((n (length row))
	(i 0)
	(ni 1))
    (while (< i n)
      (setq ni (1+ i))
      (if (= ni n) (setq ni (1- n)))
      (if (or (eq (aref row i) ringing:observation)
	      (eq (aref row i) ?*)
	      (eq (aref row i) ?\n)
	      (eq (aref row  ni) ringing:observation)
	      (eq (aref row ni) ?*))
	  (insert (aref row i))
	(insert ? ))
      (incf i))))

(defun ringing:insert-row-lines (row &optional comment)
  "Insert ROW as \"blue\" and \"red\" lines, with ringing:observation (a character) as blue.
Blue is done with a \"*\", red with a \".\", and all others as spaces.
Optional argument gives comment string to insert on right, after lines-between-lines if present
else after numbers.
The characters drawn may be changed by setting ringing:row-lines-treble,
ringing:row-lines-marked-bell, and ringing:row-lines-other-bells."
  (if ringing:draw-crossings
      (let ((n (length row))
	    (i 1))
	(ringing:draw-crossings)
	(insert ? )
	(while (< i n)
	  (insert (aref row i) ? )
	  (incf i)))
    (progn    
      (when ringing:show-lines
	(when (and ringing:show-lines-with-lines (not (null ringing:previous-row)))
	  (ringing:draw-lines-between-rows ringing:previous-row row comment))
	(insert (ringing:row-lines row ringing:show-lines-with-lines)))
      (when ringing:show-numbers
	(if ringing:display-course-bells-only
	    (ringing:insert-only-observation-and-course row)
	  (insert row)))
      (when (and (not ringing:show-lines-with-lines) comment) (insert comment))))
  (setq ringing:previous-row row))

;; end of method-lines.el
