;;;; place-bell-lines.el -- part of John's elisp change-ringing system
;;; Time-stamp: <99/11/11 15:57:17 johnstu>

(eval-when (compile load eval)
  (provide 'place-bell-lines)
  (require 'beer)
  (require 'leads)
)

(defun ringing:record-place-bell-lines-row (starting-row change bell-lines-array row-index)
  "Record for each bell in STARTING-ROW what CHANGE makes it do, in BELL-LINES-ARRAY at ROW-INDEX."
  (let* ((position 1)
	 (nbells (1- (length change)))
	 )
    (while (<= position nbells)
      (let* ((bell (ringing:bell-char-to-number (aref starting-row position)))
	     (next (aref change position))
	     (direction (if (= next position) ?| (if (> next position) ?> ?<)))
	     )
	(aset (aref bell-lines-array (1- bell)) row-index direction)
	(incf position)))))

(defun ringing:record-place-bell-lines-output-function (row
							remaining-changes
							method-holder
							rounds
							bell-lines-array
							row-index-holder
							previous-row-holder)
  "Output function for recording lines by place bell."
  (let* ((method (car method-holder))
	 (nbells (change-ringing-method-nbells method))
	 (change (car remaining-changes))
	 (row-index (car row-index-holder))
	 (previous-row (car previous-row-holder))
	 )
    (if (null previous-row) (setq previous-row rounds))
    (ringing:record-place-bell-lines-row previous-row change bell-lines-array row-index)
    (rplaca row-index-holder (1+ row-index))
    (rplaca previous-row-holder row)))

(defun ringing:place-bell-lines (method &optional force)
  "Return the place bell lines array for METHOD.
This is cached in the method, and re-used unless option FORCE given."
  (let ((method-data (ringing:find-named-method method)))
    (when (or force (null (change-ringing-method-place-bell-lines method-data)))
      (let* ((nbells (change-ringing-method-nbells method-data))
	     (place-bell-lines (make-vector nbells nil))
	     (method-holder (list method-data))
	     (place-marker (list 'work))
	     (row-index-holder (list 0))
	     (nrows (ringing:count-lead-rows method-data))
	     (ringing:next-lead-end-action (list 'plain))
	     (rounds (let ((r (make-vector (1+ nbells) nil)))
		       (dotimes (i (1+ nbells))
			 (aset r i i))
		       r)))
	(dotimes (i nbells)
	  (aset place-bell-lines i (make-string (1+ nrows) ?-)))
	(setf (change-ringing-method-place-bell-lines method-data) place-bell-lines)
	(ringing:ring-from-rounds-plain method-data
					'ringing:ring-one-lead-helper-function
					(list place-marker)
					'ringing:record-place-bell-lines-output-function
					(list place-bell-lines row-index-holder (list nil)))
	(let ((lead-end-row (copy-sequence ringing:most-recent-row))
	      (first-lead-in-change (car (change-ringing-method-work method-data))))
	  (ringing:record-place-bell-lines-row lead-end-row first-lead-in-change place-bell-lines (car row-index-holder))))
      (change-ringing-method-place-bell-lines method-data))))

(defun ringing:lead-end-order-starting-at (method starting-bell)
  "Return the lead end order of METHOD, starting at STARTING-BELL."
  (interactive
   (let* ((method-name (ringing:read-method-name "Show lead end order for: "))
	  (starting (ringing:read-number-up-to
		     "Starting bell"
		     (change-ringing-method-nbells
		      (ringing:find-named-method method-name)))))
     (list method-name starting)))
  (let* ((method-data (ringing:find-named-method method))
	 (forward-matrix (car (ringing:plain-lead-matrices method-data)))
	 (current-bell (aref forward-matrix starting-bell))
	 (leads (list starting-bell)))
    (while (/= current-bell starting-bell)
      (push current-bell leads)
      (setq current-bell (aref forward-matrix current-bell)))
    (setq leads (nreverse leads))
    (if (interactive-p)
	(message "%s" leads))
    leads))

(defun ringing:work-line (method starting-bell &optional force)
  ;; this is wrong, I think it may be repeating things at the lead ends
  "Return the directions string for METHOD starting from STARTING-BELL;
this may use cached information, unless optional argument FORCE is non-nil.

This currently assumes that METHOD is not a principle -- or something like that."
  (interactive
   (let* ((method-name (ringing:read-method-name "Recite the work for: "))
	  (manual (ringing:read-number-up-to
		   "Manual bell"
		   (change-ringing-method-nbells
		    (ringing:find-named-method method-name))))
	  (force (y-or-n-p "Force recalculation? ")))
     (list method-name manual force)))
  (let* ((method-data (ringing:find-named-method method))
	 (place-bell-lines (ringing:place-bell-lines method-data force))
	 (order (ringing:lead-end-order-starting-at method-data starting-bell))
	 (combined-line
	  (apply 'concat (mapcar (function
				  (lambda (place-bell)
				    (substring
				     (aref place-bell-lines (1- place-bell))
				     0 -1
				     )
				    ))
				 order))))
    (message "Work is %s" combined-line)
    combined-line))

;;;; for testing it

(defun ringing:display-line (start line)
  "With the bell initially at START display LINE."
  (dotimes (i (length line))
    (let* ((direction-char (aref line i))
	   (increment (if (= direction-char ?|) 0 (if (= direction-char ?>) 1 -1))))
      (princ (make-string (1- start) ?.))
      (princ "*\n")
      (incf start increment))))

(defun foolines ()
  "Test the place-bell-lines code"
  (interactive)
  (let* ((method "Cambridge Surprise Minor")
	 (c6lines (ringing:place-bell-lines method t)))
    (with-output-to-temp-buffer "*Lines*"
      (dotimes (placebell 6)
	(let ((line (aref c6lines placebell)))
	  (princ (format "%d: %s\n" (1+ placebell) line))
	  (ringing:display-line (1+ placebell) line)
	  ))
      (let ((whole-work-line (ringing:work-line method 2 t)))
	(princ (format "\nWhole method starting from 2:\n%s\n" whole-work-line))
	(ringing:display-line 2 whole-work-line)
	))))

;;; end of place-bell-lines.el
