;;;; place-notation.el -- part of John's elisp change-ringing system
;;; parse mSiRiL and the diary notation
;;; Time-stamp: <99/08/25 19:45:11 johnstu>

(eval-when (compile load eval)
  (provide 'place-notation)
  (require 'stages)
  (require 'cl)			; otherwise compiler complains about this not defining map, when it auto-loads
  )

;;;; debugging, tracing etc

(defvar place-notation:debugging nil
  "Whether to log parsing of place notation.")

(defvar old-message-log-max 200
  "message-log-max before place-notation:debugging got at it.")

(defun place-notation:debugging (debug)
  "Enable or disable DEBUG of place notation reading."
  (interactive (list (yes-or-no-p "Debug reading of place notation? ")))
  (when (numberp message-log-max) (setq old-message-log-max message-log-max))
  (setq place-notation:debugging debug
	message-log-max (if debug t old-message-log-max)))

(defun place-notation:debug-readify-char (char)
  "Return a more legible version of CHAR."
  (if (<= char 12)
      (+ char ?0)
    char))

(defun place-notation:debug-readifier (listorstring)
  "Return a more legible version of LISTORSTRING."
  (cond ((integerp listorstring)
	 (place-notation:debug-readify-char listorstring))
	((listp listorstring)
	 (mapcar 'place-notation:debug-readifier listorstring))
	((stringp listorstring)
	 (concat (map 'vector
		      'place-notation:debug-readify-char
		      listorstring)))
	(t (format "%s" listorstring))))

(defun place-notation:debug (&rest args)
  "Output debugging into the parse log buffer"
  (when place-notation:debugging
    (princ (apply 'format
		  ;; (mapcar 'place-notation:debug-readifier
			  args
			  ;;  )
		  ))))

(defun place-notation:debug-dump-method (method)
  "Output METHOD into the parse log buffer."
  (when place-notation:debugging
    (princ (format "title: %s
nbells: %s
plain: %s
work: %s
plain-lead-end: %s
bob: %s
single: %s
original-description-string: %s
"
		   (change-ringing-method-title method)
		   (change-ringing-method-nbells method)
		   (change-ringing-method-plain method)
		   (change-ringing-method-work method)
		   (change-ringing-method-plain-lead-end method)
		   (change-ringing-method-bob method)
		   (change-ringing-method-single method)
		   (change-ringing-method-original-description-string method)))))

(defun place-notation:forget-methods ()
  "Forget all the methods, to force a re-read -- really for debugging."
  (interactive)
  (setq ringing:known-methods nil
	microSiRiL:methods-in-files nil
	method-groups:all-groups nil
	ringing:constructed-methods nil))

;;;; helper for low-level parser: handle half-leads

(defun place-notation:add-reverse-half-lead (so-far)
  "Add the cdr of SO-FAR to the front of SO-FAR and return that."
  (let ((rev
	 ;; note that at this stage, "so-far" is already in
	 ;; the reverse order -- the last thing our caller
	 ;; does it to reverse it into the correct order and
	 ;; return that
	 (cdr so-far)
	 ))
    (while rev
      (push (pop rev) so-far)))
  so-far)

;;;; helper routine for low-level parser

(defun place-notation:construct-change-from-places (places cover)
  "From PLACES (list of bells making places, highest number first) and COVER (last bell plus one)
construct a change (string of where each bell goes next)."
  ;; to do: handle implicit places
  (place-notation:debug
   "Constructing change from %s with cover at %d\n" places cover)
  (let ((bell 1)
	nextbell next-place
	(change (make-string cover 0)))
    ;; handle implied places at back
    (when (and (not (null places)) (evenp (- cover (car places))))
      (push (1- cover) places))
    (push cover places)
    (setq places (nreverse places))
    ;; handle implied places at lead
    (when (and (not (null places)) (evenp (car places)))
      (push 1 places))
    (while places
      (setq next-place (car places)
	    places (cdr places))
      (place-notation:debug "Next place: %d\n" next-place)
      (while (< bell next-place)
	(setq nextbell (1+ bell))
	(place-notation:debug "Exchanging %d and %d\n" bell nextbell)
	(aset change bell nextbell)
	(aset change nextbell bell)
	(setq bell (+ 2 bell)))
      (when (< bell cover)
	(place-notation:debug "%d making places\n" bell)
	(aset change bell bell)
	(setq bell (1+ next-place))
	))
    (place-notation:debug
     "Constructed change %s\n"
     (prin1-to-string (place-notation:debug-readifier change)))
    change))

;;;; the low-level parser (sorry, it's written in lowish-level elisp)

(defun place-notation:parse-region (a b n)
  "Parse region from A to B as place notation for N bells."
  (place-notation:parse-string (buffer-substring a b) n))

(defun place-notation:parse-string (place-string n)
  "Parse PLACE-STRING as place notation for N bells.
This parses place notation as in the diary, or mSiRiL method library format.
There should be either a newline or a semi-colon between changes;
an X implies this, too..
A ^ after a change means that a half-lead has been given, and so what
has just been parsed is inserted again but in reverse at that point.

This function simply returns a list of change strings:
We store place notation as lists of strings (nasty binary in strings),
where each character is the position that the bell in that position
next moves to.

** We number the bells from 1 up, so the first character in each string
is available for labelling things. **

All the strings in a block have a character for each bell, so we fill
in trailing places as we read the place notation."
  (let ((result nil)
	(places nil)
	(a 0)
	(b (length place-string))
	c				; character being worked on
	(cover (1+ n))
	last-was-change
	(just-done-change nil)
	;; The following two flags indicate whether we are doing
	;; a half or full lead in msiril notation; if neither is
	;; true, we do not appear to be using msiril notation.
	(siril-half-lead-only nil)
	(siril-full-lead nil)
	)
    (place-notation:debug "Parsing %s (%d bells)\n" place-string n)
    (while (<= a b)
      (setq c (if (= a b)
		  ?\n			; hack on an end marker to flush the last change
		(aref place-string a))
	    last-was-change just-done-change
	    just-done-change nil)
      (place-notation:debug "%s[%d]=%c\n%s^\n" place-string a c (make-string a ? ))
      (case c

	;; Anything that marks or implies the end of a row will cause
	;; the completion of the row data for that row (i.e. any remaining
	;; bells not marked as making places will swap) and the pushing
	;; of that row onto the list of changes.
	((?\n ?x ?X 59 ?^ ?- ?, ?.)
	 ;; 59 in the cases above is semi-colon; if I put an
	 ;; actual semi-colon there, it confuses the syntax colouring!
 	 (place-notation:debug "\nAt change end: \"%c\" last-was-change=%s\n" c last-was-change)
	 (when (or;; (eq c ?-)		; experimental
		;; ah, I think it's more than that -- doesn't "-" mean there is a change before the "-" AND one after it?
		(not last-was-change))
	   (place-notation:debug "\nStarting change at %c\n" c)
           (setq just-done-change t)
	   (push (place-notation:construct-change-from-places
		  places cover)
		 result)
	   ;; (setq places nil)
	   ;; for SiRiL notation, "-" means complete the preceding change
	   ;; and add a cross, except if it occurs at the start of the
	   ;; notation, in which case it only adds a cross
	   (place-notation:debug "Considering implied cross: c=%c places=%s so-far=%s\n" c places result)
	   (when (and (eq c ?-)
		      ;; there is still something wrong in this decision, I think!
		      ;; (not (null (cdr result)))
		      ;; (not (null result))
		      (not (null places))
		      )
	     (place-notation:debug "Doing implied cross\n")
	     (push (place-notation:construct-change-from-places
		    nil cover)
		   result))
	   (setq places nil)
	   )
	 (when (or (= c ?^)
		   (and siril-half-lead-only
			(or (= c ?,) (= c ?\n))))
	   (place-notation:debug
	    "Adding reverse of reverse of %s\n"
	    (prin1-to-string (place-notation:debug-readifier (cdr result))))
	   (setq result (place-notation:add-reverse-half-lead result)
		 siril-half-lead-only nil ; we've used up the symmetry marker
		 just-done-change t)))

	;; Bell numbers for changes; we use 1234567890ETABCD
	((?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9)
	 (push (- c ?0) places))
	((?0)
	 (push 10 places))
	((?E ?e)
	 (push 11 places))
	((?T ?t)
	 (push 12 places))
	((?A ?a)
	 (push 13 places))
	((?B ?b)
	 (push 14 places))
	((?C ?c)
	 (push 15 places))
	((?D ?d)
	 (push 16 places))

	;; SIRIL compatibility
	((?&)
	 (place-notation:debug "Noticed that this is a SiRiL half lead\n")
	 (setq siril-half-lead-only t))
	((?+)
	 (place-notation:debug "Noticed that this is a SiRiL full lead\n")
	 (setq siril-full-lead t)))

      ;; advance to next character in region
      (setq a (1+ a)))

    ;; We have been pushing successive changes onto the head
    ;; of the list; now give them back in the expected order,
    ;; that is, first change first.
    (let ((forward-result (nreverse result)))
      (place-notation:debug "Parsed %s (%d bells) --> %s\n" place-string n forward-result)
      forward-result)))

;;;; parse autocalls for such things as Dixons

(defun place-notation:parse-autocalls (autocalls-string nbells)
  "Parse a description of the rules for automatic calls such as in Dixons.
AUTOCALLS-STRING should be a series of segments for which the syntax of each is
;Pattern:Notation
Where Pattern describes when to apply Notation. Pattern is made into a regular
expression for pattern matching, by suffixing it with the pattern matching the
start of a change (for example, \"4\" matches bell 4 leading, whereas \"..2\"
matches bell 2 being in 3rd place). The Notation part of each entry is normal
microSiRiL place notation.
The result is an alist of Pattern . Notation in the order given, where Notation
is a changes list."
  (if autocalls-string
      (let ((start 0)
	    (result nil))
	(while (string-match ";\\([^:]+\\):\\([^;]+\\)" autocalls-string start)
	  (push (cons (concat " " (substring autocalls-string (match-beginning 1) (match-end 1)))
		      (place-notation:parse-string
		       (substring autocalls-string (match-beginning 2) (match-end 2))
		       nbells))
		result)
	  (setq start (match-end 0)))
	(nreverse result))
    nil))

;;;; Combining the lists into a complete description of a method

(defstruct change-ringing-method
  title					; the name of the method
  nbells				; number of bells including hunt but excluding cover, e.g. 6 = minor
  plain					; work plus plain lead end
  work					; just the work, no lead end
  plain-lead-end			; just the plain lead end
  bob					; just the bob lead end
  single				; just the single lead end
  special-start				; ringing will start here if non-nil,
					; but continue at work subsequently --
					; this is how we start Stedman
  autocalls				; for things such as Dixons
  original-description-string		; as in the mSiRiL method library
  defining-file				; the file from which we read this method
  place-bell-lines			; optional array saying what each place bell does
					; use ringing:place-bell-lines to access creating if necessary
  )

(defun ringing:method-name (method)
  "Return the name of METHOD (or METHOD if it is already a string).
Just a bundling of the accessor change-ringing-method-title, such that
it's OK if you give it just a name already (as some functions in this
system take either a name or a method)."
  (if (change-ringing-method-p method)
      (change-ringing-method-title method)
    (let ((capped (capitalize method)))
      (remove-text-properties 0 (length capped) '(face nil) capped)
      capped)))

;;;; parsing complete methods

(defvar ringing:known-methods nil
  "The methods we've read so far.")

(defun place-notation:parse-method-in-region (a b)
  "Return the change-ringing-method described by the text from A to B,
which is expected to contain one to three blocks of place notation, one
labelled \"Method: <name>\", and optionally ones labelled \"Bob:\" and
\"Single:\".
See documentation for place-notation:parse-region for the syntax
accepted within each block."
  (with-output-to-temp-buffer
      (format "*place notation parser log for %s*" (buffer-name))
    (let ((title "Unknown")
	  (nbells 0)
	  (i 3)
	  start plain-end end
	  (plain nil)
	  (bob nil)
	  (single nil))
      (goto-char a)
      (if (re-search-forward "Method: \\(.+\\)$" b t)
	  (setq title (buffer-substring (match-beginning 1)
					(match-end 1))))
      (setq nbells
	    (catch 'matched
	      (while (< i (length ringing:number-of-bells-names))
		(if (string-match (aref ringing:number-of-bells-names i)
				  title)
		    (throw 'matched i))
		(incf i))
	      (error "Could not work out number of bells for %s" title)
	      0))

      (goto-char a)
      (re-search-forward "^[-xX1-9eEtT;^.,&+]+$")
      (setq start (match-beginning 0))
      (re-search-forward "^$")
      (setq plain-end (match-beginning 0))
      (place-notation:debug "Parsing main line of method:\n")
      (setq plain (place-notation:parse-region start plain-end nbells))

      (when (search-forward "Bob:" b t)
	(when (looking-at "$") (beginning-of-line 2))
	(setq start (point))
	(re-search-forward "^$")
	(setq end (match-beginning 0))
	(place-notation:debug "Parsing Bob:\n")
	(setq bob (place-notation:parse-region start end nbells)))

      (goto-char plain-end)

      (when (search-forward "Single:" b t)
	(when (looking-at "$") (beginning-of-line 2))
	(setq start (point))
	(re-search-forward "^$")
	(setq end (match-beginning 0))
	(place-notation:debug "Parsing Single:\n")
	(setq single (place-notation:parse-region start end nbells)))

      (let ((method (make-change-ringing-method
		     :title title
		     :nbells nbells
		     :plain plain
		     :bob bob
		     :single single)))
	(push (cons title method) ringing:known-methods)
	method))))

(defun place-notation:parse-method-in-file (filename)
  "Parse the method in FILENAME; see place-notation:parse-method-in-region
for the structure of the file, and place-notation:parse-region for the
syntax."
  (interactive "fParse place notation in file: ")
  (save-window-excursion
    (find-file filename)
    (place-notation:parse-method-in-region (point-min) (point-max))))

(defun place-notation:parse-methods-in-directory (dir)
  "Parse all methods in DIR/*.pn, using place-notation:parse-method-in-file."
  (interactive "DParse place notation in directory: ")
  (mapcar 'place-notation:parse-method-in-file
	  (directory-files dir t "\\.pn$")))

(defvar place-notation:default-methods-directory
  (expand-file-name "~/ringing/methods/")
  "The default place in which to look for methods.
This is not the mSiRiL method library -- that is pointed to by
method-library:directory -- but a collection of files in a format
similar to that down the side of the methods in the Ringing World diary.")

;;;; reverse of parsing

(defun ringing:symmetric-p (method)
  "Return whether METHOD is obviously symmetric."
  (let* ((plain (if (consp method) method (change-ringing-method-work (ringing:find-named-method method))))
	 (nailp (reverse plain)))
    (or (equal plain nailp)
	(equal (cdr nailp) (reverse (cdr nailp))))))

(defun place-notation:recreate-string-from-changes-old (changes &optional marked suppress-lead-end)
;;; superceded by one from Dave
  "Turn CHANGES back into place notation.
With optional argument MARKED, put & or + at the start as appropriate,
showing only first half if &
Unless optional SUPPRESS-LEAD-END given, append the lead end if &."
  (let* ((placechars nil)
	 (is-symmetric (ringing:symmetric-p changes))
	 (displayed-length (+ 0 (/ (length changes) 2)))
	 (displayed-changes (if (and marked is-symmetric)
				(subseq changes 0 displayed-length)
			      changes))
	 (remainder (if (and marked is-symmetric)
			(subseq changes (1- (* 2 (+ 0 displayed-length))))
		      nil))
	 (first-row t)
	 (prev-was-cross nil))
;;    (message (format "is-symmetric=%S\n changes=%S\n (length changes)=%d\n displayed-length=%d\n (length displayed-changes)=%d\n (length remainder)=%d\n "
;;		     is-symmetric changes (length changes) displayed-length (length displayed-changes) (length remainder)))
    (while displayed-changes
      (let* ((the-change (car displayed-changes))
	     (i 1)
	     (n (1- (length the-change)))
	     (any nil)
	     (any-internal nil)
	     (any-non-internal nil))
;;	(message "  Working on row %S, string so far is %s" the-change (coerce (reverse placechars) 'string))
	(while (< i n)
	  (let ((c (aref the-change i)))
	    (when (eq c i)
	      ;; making places
	      (if (and (not any) (not first-row))
		  ;; at the first one, decide whether to start this row with . or with -
		  (if prev-was-cross
		      (push ?- placechars)
		    (push ?. placechars)))
	      (setq any t)
	      (if (and (not (eq i 1)) (not (eq i n)))
		  (progn
		    (setq any-internal t)
		    (push (ringing:bell-number-to-char c) placechars))
		(setq any-non-internal t)))
	    (incf i)))
	(when (and any-non-internal
		   (not any-internal))
	  (push ?1 placechars))
	(setq prev-was-cross (not any)
	      first-row nil))
      (setq displayed-changes (cdr displayed-changes)))
    (let ((place-string (coerce (nreverse placechars) 'string)))
;;      (message "That makes %s" place-string)
       ;;;;;;;;;;;;;;;; problem may well be in this area -- final "-" missing where relevant
      (if marked
	  (if is-symmetric
	      (let ((remainder-string (place-notation:recreate-string-from-changes remainder nil)))
		(if (or suppress-lead-end (string= remainder-string ""))
		    (format "&%s" place-string)
		  (format "&%s, %s" place-string remainder-string)))
	    (format "+%s" place-string))
	place-string))))

(defun place-notation:recreate-string-from-changes (changes &optional marked suppress-lead-end)
;;; based on the one superceded by one from Dave, for tinkering with
  "Turn CHANGES back into place notation.
With optional argument MARKED, put & or + at the start as appropriate,
showing only first half if &
Unless optional SUPPRESS-LEAD-END given, append the lead end if &."
  (let* ((placechars nil)
	 (is-symmetric (ringing:symmetric-p changes))
	 (displayed-length (+ 0 (/ (length changes) 2)))
	 (displayed-changes (if (and marked is-symmetric)
				(subseq changes 0 displayed-length)
			      changes))
	 (remainder (if (and marked is-symmetric)
			(subseq changes (1- (* 2 (+ 0 displayed-length))))
		      nil))
	 (first-row t)
	 (prev-was-cross nil))
    (message (format "is-symmetric=%S\n changes=%S\n (length changes)=%d\n displayed-length=%d\n (length displayed-changes)=%d\n (length remainder)=%d\n "
		     is-symmetric changes (length changes) displayed-length (length displayed-changes) (length remainder)))
    (while displayed-changes
      (let* ((the-change (car displayed-changes))
	     (i 1)
	     (n (1- (length the-change)))
	     (any nil)
	     (any-internal nil)
	     (any-non-internal nil))
	(message "  Working on row %S, string so far is %s" the-change (coerce (reverse placechars) 'string))
	(while (< i n)
	  (let ((c (aref the-change i)))
	    (when (eq c i)
	      ;; making places
	      (if (and (not any) (not first-row))
		  ;; at the first one, decide whether to start this row with . or with -
		  (if prev-was-cross
		      (push ?- placechars)
		    (push ?. placechars)))
	      (setq any t)
	      (if (and (not (eq i 1)) (not (eq i n)))
		  (progn
		    (setq any-internal t)
		    (push (ringing:bell-number-to-char c) placechars))
		(setq any-non-internal t)))
	    (incf i)))
	(when (and any-non-internal
		   (not any-internal))
	  (push ?1 placechars))
	(setq prev-was-cross (not any)
	      first-row nil))
      (setq displayed-changes (cdr displayed-changes)))
    (let ((place-string (coerce (nreverse placechars) 'string)))
;;      (message "That makes %s" place-string)
      (if marked
	  (if is-symmetric
	      (let ((remainder-string (place-notation:recreate-string-from-changes remainder nil)))
		(if (or suppress-lead-end (string= remainder-string ""))
		    (format "&%s" place-string)
		  (format "&%s, %s" place-string remainder-string)))
	    (format "+%s" place-string))
	place-string))))

(defun place-notation:recreate-string-from-changes-from-Dave (changes &optional marked suppress-lead-end)
;;; this one is from Dave
  "Turn CHANGES back into place notation.
With optional argument MARKED, put & or + at the start as appropriate,
showing only first half if &
Unless optional SUPPRESS-LEAD-END given, append the lead end if &."
  (let* ((placechars nil)
	 (is-symmetric (ringing:symmetric-p changes))
	 (displayed-length (+ 1 (/ (length changes) 2)))
	 (displayed-changes (if (and marked is-symmetric)
				(subseq changes 0 displayed-length)
			      changes))
	 (remainder (if (and marked is-symmetric)
			(subseq changes (1- (* 2 (- displayed-length 1))))
		      nil))
	 (first-row t)
	 (prev-was-cross nil))
;;    (message (format "is-symmetric=%S\n changes=%S\n (length changes)=%d\n displayed-length=%d\n (length displayed-changes)=%d\n (length remainder)=%d\n "
;;		     is-symmetric changes (length changes) displayed-length (length displayed-changes) (length remainder)))
    (while displayed-changes
      (let* ((the-change (car displayed-changes))
	     (i 1)
	     (n (1- (length the-change)))
	     (any nil)
	     (any-internal nil)
	     (any-non-internal nil))
;;        (message "  Working on row %S, string so far is %s" the-change (coerce (reverse placechars) 'string))
	(while (< i n)
	  (let ((c (aref the-change i)))
	    (when (eq c i)
	      ;; making places
	      (if (and (not any) (not first-row))
		  ;; at the first one, decide whether to start this row with . or with -
		  (if prev-was-cross
		      (push ?- placechars)
		    (push ?. placechars)))
	      (setq any t)
	      (if (and (not (eq i 1)) (not (eq i n)))
		  (progn
		    (setq any-internal t)
		    (push (ringing:bell-number-to-char c) placechars))
		(setq any-non-internal t)))
	    (incf i)))
	(when (and any-non-internal
		   (not any-internal))
	  (push ?1 placechars))
	(setq prev-was-cross (not any)
	      first-row nil))
      (setq displayed-changes (cdr displayed-changes)))
    (let ((place-string (coerce (nreverse placechars) 'string)))
;;      (message "That makes %s" place-string)
      (if marked
	  (if is-symmetric
	      (let ((remainder-string (place-notation:recreate-string-from-changes remainder nil)))
		(if (or suppress-lead-end (string= remainder-string ""))
		    (format "&%s" place-string)
		  (format "&%s, %s" place-string remainder-string)))
	    (format "+%s" place-string))
	place-string))))

(defun place-notation:create-fully-expanded-from-changes (changes)
  "Turn CHANGES back into place notation, fully expanded as for Method Maker etc."
  (let* ((placechars nil)
	 (displayed-changes changes)
	 (first-row t)
	 (prev-was-cross nil))
    (while displayed-changes
      (let* ((the-change (car displayed-changes))
	     (i 1)
	     (n (1- (length the-change)))
	     (any nil)
	     (any-internal nil)
	     (any-non-internal nil))
	(while (<= i n)
	  (let ((c (aref the-change i)))
	    (when (eq c i)
	      (when (and (not any) (not first-row))
		;; at the first one, decide whether to start this row with . or with .X.
		(if prev-was-cross
		    (progn
		      (push ?. placechars)
		      (push ?X placechars)))
		(push ?. placechars))
	      (setq any t)
	      (push (ringing:bell-number-to-char c) placechars))
	    (incf i)))
	(setq prev-was-cross (not any)
	      first-row nil))
      (setq displayed-changes (cdr displayed-changes)))
    (let ((place-string (coerce (nreverse placechars) 'string)))
      place-string)))

(defun place-notation:insert-method-maker-URL-for-method (method)
  "Insert the method maker URL for METHOD.
This includes the title and the fully expanded place notation."
  (interactive (list (method-library:choose-method-interactively "Insert Method Maker URL for method: ")))
  (let* ((URLstring (concat
		     "http://www.yacc.co.uk/cgi-bin/method?T=" method
		     "&P=" (place-notation:create-fully-expanded-from-changes
			    (change-ringing-method-plain
			     (ringing:find-named-method method)))))
	 (len (length URLstring)))
    (dotimes (i len)
      (when (= (aref URLstring i) ? ) (aset URLstring i ?+)))
    (insert URLstring)))

;;;; display what we have parsed, for debugging mostly

(defun place-notation:show-one-line-place-notation (row-of-changes)
  "Display ROW-OF-CHANGES as place notation, i.e. list those which do not change."
  (let ((i 1)
	(changed nil)
	(l (length row-of-changes)))
    (while (< i l)
      (when (eq i (aref row-of-changes i))
	(setq changed t)
	(princ (char-to-string (+ i ?0 ))))
      (setq i (1+ i)))
    (unless changed (princ "X"))))

(defun place-notation:show-place-notation (title pn)
  "Using TITLE, display list of changes PN into stdout."
  (when title
    (princ (format "\n%s:\n" title)))
  (let ((p pn))
    (while p
      (place-notation:show-one-line-place-notation (car p))
      (princ "\n" )
      (setq p (cdr p)))))

(defun place-notation:show-method-place-notation (method-name)
  "Display METHOD-NAME in fully expanded place notation.
Written originally to check the parser, but might also be useful with
things that modify methods."
  (interactive
   (progn
     (require 'ringing-methods)
     (list (ringing:read-method-name "Show place notation for: "))))
  (with-output-to-temp-buffer
      (format "*Place notation for %s*" method-name)
    (let ((method
	   (ringing:find-named-method method-name)
	   ;; (cdr (assoc method-name ringing:known-methods))
	   ))
      (princ (format "Method: %s\n"
		     (ringing:method-name method)))
      (place-notation:show-place-notation
       nil
       (change-ringing-method-plain method))
      (place-notation:show-place-notation
       "Bob"
       (change-ringing-method-bob method))
      (place-notation:show-place-notation
       "Single"
       (change-ringing-method-single method)))))

(defun place-notation:show-change-notation (title pn &optional places-too)
  "Using TITLE, display list of changes PN into stdout.
With optional third arg PLACES-TOO, show also the place notation"
  (when title
    (princ (format "\n%s:\n" title)))
  (let ((p pn))
    (while p
      (let* ((change (car p))
	     (i 1)
	     (l (length change)))
	(while (< i l)
	  (princ (char-to-string (+ (aref change i) ?0 )))
	  (setq i (1+ i)))
	(when places-too
	    (princ " ")
	    (place-notation:show-one-line-place-notation change))
	(princ "\n" )
	(setq p (cdr p))))))

(defun place-notation:show-method-change-notation (method-name &optional places-too)
  "Display METHOD-NAME in the change notation used internally.
Written originally to check the parser, but might also be useful with
things that modify methods.
Note that this is the internal notation of the program, not place notation.

With optional second argument PLACES-TOO (always enabled interactively), show the
place notation alongside the change notation."
  (interactive
   (progn
     (require 'ringing-methods)
     (list (ringing:read-method-name "Show change notation for: ") t)))
  (with-output-to-temp-buffer
      (format "*Change notation for %s*" method-name)
    (let ((method
	   (ringing:find-named-method method-name)
	   ;; (cdr (assoc method-name ringing:known-methods))
	   ))
      (if places-too
	  (progn
	    (princ "*** The place notation is on the right ***\n")
	    (princ "*** On the left is the internal notation used by this program ***\n\n")
	    )
	(progn
	  (princ "*** This is not place notation ***\n")
	  (princ "*** It is the internal notation used by this program ***\n\n")))
      (princ (format "Change notation for method: %s\n"
		     (ringing:method-name method)))
      (place-notation:show-change-notation
       nil
       (change-ringing-method-plain method)
       places-too)
      (place-notation:show-change-notation
       "Bob"
       (change-ringing-method-bob method)
       places-too)
      (place-notation:show-change-notation
       "Single"
       (change-ringing-method-single method)
       places-too))))

;;;; end of place notation parser
