;;;; placenotation-mode.el -- part of John's elisp change-ringing system
;;; edit place notation, displaying the method dynamically
;;; Time-stamp: <98/05/11 15:16:22 john>

(eval-when (compile load eval)
  (provide 'placenotation-mode)
  (require 'beer)
  (require 'method-lines)
  (require 'place-notation)
  (require 'ringing-methods)
  (require 'bell-colours)		; use yellow for giant cursor
  (require 'cl))

(defvar placenotation-mode:nbells nil
  "How many bells in this method.")

(defvar placenotation-mode:changes nil
  "The changes for this method.")

(defvar placenotation-mode:symmetric nil
  "Whether this method is obviously symmetric.
See also placenotation-mode:symmetric-string.")

(defvar placenotation-mode:symmetric-string ""
  "String saying Whether this method is obviously symmetric.
See also placenotation-mode:symmetric.")

(defvar placenotation-mode:row-string nil
  "String saying where in the line we are.")

(defvar placenotation-mode:place-string nil
  "String saying where in the row method we are.")

(defvar placenotation-mode:rounds nil
  "The rounds for this method.")

(defvar placenotation:whole-changes-string nil
  "The place notation string representing placenotation-mode:changes.")

(defvar placenotation-mode:method nil
  "The method being edited -- a scratch structure for those that need it.")

(defvar place-notation-mode:method-holder nil
  "A holder list for the method being edited, as required for callbacks.")

(defvar placenotation-mode:keymap (make-keymap)
  "Keymap for placenotation-mode.")

(defvar placenotation-mode:read-only-keymap (make-keymap)
  "Keymap for placenotation-mode without the editing facilities.")

(suppress-keymap placenotation-mode:keymap)

(suppress-keymap placenotation-mode:read-only-keymap)

(define-key placenotation-mode:keymap "x" 'placenotation-mode:cross)
(define-key placenotation-mode:keymap "|" 'placenotation-mode:places)
;; (define-key placenotation-mode:keymap " " 'placenotation-mode:swap-cross-and-place)
(define-key placenotation-mode:keymap "*" 'placenotation-mode:symbol-here)
(define-key placenotation-mode:keymap [down-mouse-3] 'placenotation-mode:mouse-symbol-here)

(defun placenotation-mode:define-key-both (keystring command)
  "Define KEYSTRING to do COMMAND in exploring and editing place notation."
  (if (listp keystring)
      (dolist (key keystring)
	(placenotation-mode:define-key-both key command))
    (progn
      (define-key placenotation-mode:keymap keystring command)
      (define-key placenotation-mode:read-only-keymap keystring command))))

(placenotation-mode:define-key-both "?" 'placenotation-mode:internal-info)
(placenotation-mode:define-key-both '("n" "\C-n" [down]) 'placenotation-mode:next-line)
(placenotation-mode:define-key-both '("p" "\C-p" [up]) 'placenotation-mode:previous-line)
(placenotation-mode:define-key-both '("f" "\C-f" [right]) 'placenotation-mode:forward-char)
(placenotation-mode:define-key-both '("b" "\C-b" [left]) 'placenotation-mode:backward-char)
(placenotation-mode:define-key-both '("a" "\C-a") 'placenotation-mode:beginning-of-line)
(placenotation-mode:define-key-both '("e" "\C-e") 'placenotation-mode:end-of-line)
(placenotation-mode:define-key-both '("<" "\M-<") 'placenotation-mode:beginning-of-buffer)
(placenotation-mode:define-key-both '(">" "\M->") 'placenotation-mode:end-of-buffer)
(placenotation-mode:define-key-both [down-mouse-1] 'placenotation-mode:mouse-point-here)

(mapcar 'make-variable-buffer-local
	'(placenotation-mode:nbells
	  placenotation-mode:changes
	  placenotation-mode:rounds
	  placenotation-mode:method
	  place-notation-mode:method-holder))

;;;; top-level functions

(defvar placenotation-mode:open-method-buffer-hooks nil
  "Hooks to run on starting a place notation edit buffer.")

(defun placenotation-mode (&optional read-only)
  "Major mode for editing place notation.
With optional arg, read-only, as a method explanation.

If editing the method, the following commands are available:
\\{placenotation-mode:keymap}

If simply exploring the method, the following commands are available:
\\{placenotation-mode:read-only-keymap}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map (if read-only
		     placenotation-mode:read-only-keymap
		   placenotation-mode:keymap))
  (buffer-disable-undo)
  (toggle-read-only 1)
  (setq mode-line-format (placenotation:hack-mode-line default-mode-line-format))
  ;; (unless read-only (add-hook 'post-command-hook 'placenotation-mode:move-underlining nil 'local))
  (setq major-mode 'placenotation-mode
	mode-name "PlaceNotation"
	)
  (message "post-command-hook is %s in buffer %s" post-command-hook (buffer-name))
  ;; (debug nil post-command-hook buffer-name)
  (run-hooks 'placenotation-mode:open-method-buffer-hooks)
  )

(defvar placenotation:start-edit-hooks nil
  "Hooks to run on starting to edit a method.")

(defun placenotation-edit-or-explore (n original read-only)
  "Edit or explore place notation on N bells starting with ORIGINAL, perhaps READ-ONLY."
  (let ((ringing:end-of-row-char ? )	; before making the rounds
	
	)
    (switch-to-buffer
     (ringing:get-buffer-create
      (if read-only
	  "*place notation grid*"
	"*place notation edit*")))
    (toggle-read-only -1)
    (erase-buffer)
    (placenotation-mode read-only)
    (setq placenotation-mode:nbells n
	  placenotation-mode:changes (place-notation:parse-string original n)
	  placenotation:whole-changes-string (place-notation:create-fully-expanded-from-changes
					      placenotation-mode:changes)
	  placenotation-mode:rounds (ringing:make-rounds n)
	  method-display:block-top-marker (make-marker)
	  placenotation-mode:method (make-change-ringing-method
				     :title "Unnamed"
				     :nbells placenotation-mode:nbells
				     :plain placenotation-mode:changes
				     :bob nil
				     :single nil
				     :original-description-string "Scratch method")
	  place-notation-mode:method-holder (list placenotation-mode:method)
	  ringing:most-recent-change nil)
    (let* ((ringing:observation-number placenotation-mode:nbells)
	   (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	   (ringing:draw-crossings t)
	   (ringing:show-lines nil)
	   (ringing:show-numbers t)
	   (ringing:show-lines-with-lines nil)
	   (ringing:auto-ring-inter-row-delay nil)
	   (method-display:main-display-function 'method-display:single-column-display-function))
      (toggle-read-only -1)
      (insert (format "%s of %s (%s)"
		      (if read-only "Exploration" "Edit")
		      original
		      placenotation:whole-changes-string))
      (set-marker method-display:block-top-marker (point))
      (require 'ermm)
      (ermm:init nil)
      (insert "\n\n")
      (ringing:insert-row-lines placenotation-mode:rounds)
      (placenotation-mode:edit-ring-from-here-onwards placenotation-mode:rounds
						      placenotation-mode:changes)
      (unless read-only (run-hooks 'placenotation:start-edit-hooks))
      (goto-char (point-min))
      (beginning-of-line 2)
      (forward-char 3)
      (placenotation-mode:move-underlining)
      (toggle-read-only 1)
      (placenotation-mode:update-status))))

(defun placenotation-edit (n &optional original)
  "Edit place notation on N bells starting with ORIGINAL."
  (interactive "nNumber of bells: 
sInitial place notation: ")
  (if (null original) (setq original "-."))
  (placenotation-edit-or-explore n original nil))

(defun placenotation-explore (n place-notation)
  "Display N-bell method with PLACE-NOTATION."
  (interactive "nNumber of bells: 
sInitial place notation: ")
  (placenotation-edit-or-explore n place-notation t))

(defun placenotation:edit-based-on (method-name)
  "Edit a method based on METHOD-NAME (typically from the libraries)."
  (interactive (list (ringing:read-method-name "Edit a method based on: ")))
  (let* ((method (ringing:find-named-method method-name))
	 (nbells (change-ringing-method-nbells method)))
    (placenotation-edit nbells
			(place-notation:recreate-string-from-changes (change-ringing-method-plain method)))))

;;;; status of the method

(defun placenotation:hack-mode-line (template)
  "Hack TEMPLATE to be as needed for place-notation editing."
  (cond
   ((null template)
    nil)
   ((eq (car template) 'mode-line-process)
    (append '(placenotation-mode:symmetric-string) (placenotation:hack-mode-line (cdr template))))
   ((and (consp (car template))
	 nil				; I can't see a chance for us to update it each time, so don't bother
	 (eq (car (car template)) 'line-number-mode))
    (cons 'placenotation-mode:row-string (placenotation:hack-mode-line (cdr template))))
   ((and (consp (car template))
	 nil				; I can't see a chance for us to update it each time, so don't bother
	 (eq (car (car template)) 'column-number-mode))
    (cons 'placenotation-mode:place-string (placenotation:hack-mode-line (cdr template))))
   (t (cons (car template)
	 (placenotation:hack-mode-line (cdr template))))))

(defun placenotation-mode:update-status ()
  "Update the status of the method edit."
  (setq placenotation-mode:symmetric (ringing:symmetric-p placenotation-mode:changes)
	placenotation-mode:symmetric-string (if placenotation-mode:symmetric
						":symmetric" ":asymmetric")
	;; placenotation-mode:row-string (format "R%d" pnm:edit-row-number)
	;; placenotation-mode:place-string (format "P%d" pnm:left-place)
	)
  (force-mode-line-update)
)

;;;; track the cursor with an underlined pair of bells

(defconst placenotation-mode:underlined (copy-face 'default 'placenotation-mode:underlined))
(set-face-underline-p placenotation-mode:underlined t)
(set-face-background placenotation-mode:underlined ringing:bad-background-colour)

(defvar placenotation-mode:underlined-start nil
  "The place which is underlined.")

(defun placenotation-mode:move-underlining ()
  "Move the underlining to where it is needed."
  ;; (message "Moving underlining... ") (sit-for 1) (message nil)
  (save-excursion
    (let* ((was-read-only buffer-read-only)
	   (buffer-read-only nil))
      (when placenotation-mode:underlined-start
	(put-text-property placenotation-mode:underlined-start
			   (+ placenotation-mode:underlined-start 3)
			   'face 'default))
      (placenotation:set-line-variables)
      (let ((bol (save-excursion (beginning-of-line 1) (point))))
	(setq placenotation-mode:underlined-start
              (+ bol (1- (* 2 pnm:left-place))))
	(put-text-property placenotation-mode:underlined-start
			   (+ placenotation-mode:underlined-start 3)
			   'face placenotation-mode:underlined)))))

(defun placenotation-mode:next-line (arg)
  "Like next-line but updates placenotation cursor."
  (interactive "p")
  (next-line arg)
  (placenotation-mode:move-underlining))

(defun placenotation-mode:previous-line (arg)
  "Like previous-line but updates placenotation cursor."
  (interactive "p")
  (previous-line arg)
  (placenotation-mode:move-underlining))

(defun placenotation-mode:forward-char (arg)
  "Like forward-char but updates placenotation cursor."
  (interactive "p")
  (when (< (current-column) (1- (* placenotation-mode:nbells 2)))
    (forward-char arg))
  (placenotation-mode:move-underlining))

(defun placenotation-mode:backward-char (arg)
  "Like backward-char but adjusts placenotation cursor."
  (interactive "p")
  (when (> (current-column) 1)
    (backward-char arg))
  (placenotation-mode:move-underlining))

(defun placenotation-mode:beginning-of-line (arg)
  "Like beginning-of-line but adjusts placenotation cursor."
  (interactive "p")
  (beginning-of-line arg)
  (move-to-column 1)
  (placenotation-mode:move-underlining))

(defun placenotation-mode:end-of-line (arg)
  "Like end-of-line but adjusts placenotation cursor."
  (interactive "p")
  (end-of-line arg)
  (move-to-column (1- (* placenotation-mode:nbells 2)))
  (placenotation-mode:move-underlining))

(defun placenotation-mode:end-of-buffer ()
  "Go to the end of the buffer."
  (interactive)
  (goto-line (+ 3 (* 2 (length placenotation-mode:changes))))
  (move-to-column (1- (* placenotation-mode:nbells 2)))
  (placenotation-mode:move-underlining))
  
(defun placenotation-mode:beginning-of-buffer ()
  "Go to the beginning of the buffer."
  (interactive)
  (goto-char (point-min))
  (beginning-of-line 2)
  (forward-char 3)
  (placenotation-mode:move-underlining))

;;;; Revising the line from the edit downwards

(defun placenotation-mode:edit-helper-function (row
						remaining-changes
						method
						rounds
						place-marker)
  "Helper function for editing lines."
  (if (string= row rounds)
      nil
    (if remaining-changes
	t
      nil)))

(defun placenotation-mode:edit-ring-from-here-onwards (row remaining-changes)
  "Starting from ROW, ring REMAINING-CHANGES with output as used in the place notation editor."
  (when remaining-changes
    (message "Redrawing lines...")
    (save-excursion
      (let* ((ringing:observation-number placenotation-mode:nbells)
	     (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	     (nbells placenotation-mode:nbells)
	     (ringing:draw-crossings t)
	     (ringing:show-lines nil)
	     (ringing:show-numbers t)
	     (ringing:show-lines-with-lines nil)
	     (ringing:auto-ring-inter-row-delay nil)
	     (place-marker (list 'work))
	     (callback-arglist (list place-notation-mode:method-holder placenotation-mode:rounds place-marker))
	     (ringing:write-place-notation-beside t))
	(ringing:ring-changes row
			      remaining-changes
			      'placenotation-mode:edit-helper-function
			      callback-arglist
			      'method-display:single-column-display-function
			      callback-arglist)
	(set-marker method-display:block-top-marker nil)))
    (placenotation-mode:update-status)
    (message "Redrawing lines... done")))

;;;; framework for performing each edit and the associated update

(defvar pnm:edit-cursor-line nil
  "The line the cursor is in at the start (and, usually, end) of the edit.")

(defvar pnm:edit-cursor-column nil
  "The column the cursor is in at the start (and, usually, end) of the edit.")

(defvar pnm:edit-row-number nil
  "The row number for editing.")

(defvar pnm:left-place nil
  "The place number for editing.")

(defvar pnm:right-place nil
  "The adjacent place for editing.")

(defvar pnm:edit-change-data nil
  "The change data array.")

(defvar pnm:edit-places-string nil
  "The places string.")

(defvar pnm:edit-on-row nil
  "Whether we are currently on a row (else on the change between them).")

(defvar pnm:edit-on-place nil
  "Whether we are currently on a place (else in the space between them).")

(defvar pnm:edit-on-place-to-left-of-cross nil
  "Whether we are currently on a left-hand place (only relevant when pnm:edit-on-place).")

(defvar pnm:edit-on-place-to-right-of-cross nil
  "Whether we are currently on a right-hand place (only relevant when pnm:edit-on-place).")

(defvar pnm:debug nil
  "Whether to do various debugging things")

(defun placenotation-mode:enable-debug (off)
  "Set whether to output debug messages for placenotation editing.
With prefix arg, disable debug, else enable it."
  (interactive "P")
  (if off
      (setq pnm:debug nil)
    (setq pnm:debug t))
  (message "placenotation-mode debugging %s" (if pnm:debug "enabled" "disabled")))

(defmacro pnm:bell-at (place)
  "Return the bell at PLACE."
  `(aref (placenotation-mode:row-for-restart) ,place))

(defmacro pnm:next-at (position)
  "Return the next position for the bell at POSITION."
  `(aref pnm:edit-change-data ,position))

(defmacro pnm:making-places (bell)
  "Return whether BELL is making places for this change."
  `(= ,bell (pnm:next-at ,bell)))

(defmacro pnm:going-up (bell)
  "Return whether BELL is making places for this change."
  `(= (1+ ,bell) (pnm:next-at ,bell)))

(defmacro pnm:going-down (bell)
  "Return whether BELL is making places for this change."
  `(= (1- ,bell) (pnm:next-at ,bell)))

(defmacro pnm:set-next (where next)
  "Say that bell in place WHERE goes to NEXT."
  `(aset pnm:edit-change-data ,where ,next))

(defun placenotation:set-line-variables ()
  "Set some internal variables according to point.
Don't do anything that requires closing afterwards."
  (setq pnm:edit-cursor-line (count-lines (point-min) (point))
	pnm:edit-cursor-column (current-column)
	pnm:edit-row-number (/ (- pnm:edit-cursor-line 4) 2)
	pnm:edit-on-row (oddp pnm:edit-cursor-line)
	pnm:left-place (/ pnm:edit-cursor-column 2)
	pnm:right-place (1+ pnm:left-place)
	pnm:edit-on-place (oddp pnm:edit-cursor-column)
	pnm:edit-change-data (nth pnm:edit-row-number placenotation-mode:changes)
	pnm:edit-places-string (ringing:change-to-place-notation pnm:edit-change-data)
	pnm:edit-on-place-to-left-of-cross (and pnm:edit-on-place
						(pnm:going-up pnm:right-place))
	pnm:edit-on-place-to-right-of-cross (and pnm:edit-on-place
						(pnm:going-down pnm:right-place)))
  (unless (< pnm:left-place placenotation-mode:nbells)
    (decf pnm:left-place)
    (decf pnm:right-place))
  (when (zerop pnm:left-place)
    (incf pnm:left-place)
    (incf pnm:right-place)))

(defun placenotation-mode:open-line-edit ()
  "Set up some internal variables according to point.
This should be followed with placenotation-mode:close-line-edit.
This is useful at the start of most edit commands."
  (placenotation-mode:move-underlining)
  (placenotation:set-line-variables)
  (when pnm:debug
    (with-output-to-temp-buffer "*place notation edit debug*"
      (princ (place-notation:debug-readifier pnm:edit-change-data)) (princ "\n")
      (princ (make-string  pnm:left-place ?-)) (princ "^\n")
      (princ (format "Cursor is on places %d (bell %c, moving to %c) and %d (bell %c, moving to %c)\n"
		     pnm:left-place (pnm:bell-at pnm:left-place)
		     (place-notation:debug-readify-char (aref pnm:edit-change-data pnm:left-place))
		     pnm:right-place (pnm:bell-at pnm:right-place)
		     (place-notation:debug-readify-char (aref pnm:edit-change-data  pnm:right-place))))
      (princ (placenotation-mode:internal-info-string)) (princ "\n")
      (princ (if pnm:edit-on-place "on place: " "on crossing: "))
      (when pnm:edit-on-place-to-left-of-cross (princ "(place to left of cross): "))
      (when pnm:edit-on-place-to-right-of-cross (princ "(place to right of cross): "))
      (princ (if (pnm:making-places pnm:left-place)
		 "left making places; " "left crossing; "))
      (princ (if (pnm:making-places pnm:right-place)
		 "right making places\n" "right crossing\n")))))

(defun placenotation-mode:row-for-restart ()
  "Return the row for restarting the method drawing."
  (save-excursion
    (beginning-of-line (if pnm:edit-on-row -1 0))
    (let ((result '(? )))
      (while (not (looking-at "$"))
	(if (looking-at "[0-9et]")
	    (push (char-after (point)) result))
	(goto-char (1+ (point))))
      (push ?  result)
      (coerce (reverse result) 'string))))

(defun placenotation-mode:change-method-from-here ()
  "Alter the method to be consistent from the current editing point."
  (save-excursion
    (rplaca (nthcdr pnm:edit-row-number placenotation-mode:changes)
	    pnm:edit-change-data)
    (setq pnm:edit-change-data (nth pnm:edit-row-number placenotation-mode:changes))
    (let ((restart-changes (nthcdr (- pnm:edit-row-number 0) placenotation-mode:changes))
	  (restart-row (placenotation-mode:row-for-restart)))
      (end-of-line (if pnm:edit-on-row -1 0))
      (delete-region (point) (point-max))
      (placenotation-mode:edit-ring-from-here-onwards restart-row restart-changes)
      ;; (message "restart-row=%s restart-changes=%s" restart-row restart-changes)
      )))

(defun placenotation-mode:close-line-edit ()
  "Various actions for the end of an edit."
  (let ((was-read-only buffer-read-only)
	(buffer-read-only nil)) 
    (placenotation-mode:change-method-from-here)
    (save-excursion
      (goto-char (point-min))
      (when (re-search-forward "Edit of \\(.+\\)" (point-max) t)
	(delete-region (match-beginning 1) (match-end 1))
	(setq placenotation:whole-changes-string (place-notation:recreate-string-from-changes placenotation-mode:changes))
	(insert placenotation:whole-changes-string))))
  (goto-line pnm:edit-cursor-line)
  (move-to-column pnm:edit-cursor-column)
  (placenotation-mode:move-underlining))

(defmacro placenotation-mode:with-open-line-edit (&rest edit-command-body)
  "Perform EDIT-COMMAND-BODY with the placenotation editing variables set up
and the result being taken into account."
  `(progn
     (placenotation-mode:open-line-edit)
     (prog1
	 (progn
	   ,@edit-command-body)
       (placenotation-mode:close-line-edit))))

;;;; editing (etc) commands

(defun placenotation-mode:internal-info-string ()
  "Return a string giving a line of useful information about the current point."
  (format "Line %d %s %s column %d ==> change %d place %d ==> places %s; restart \"%s\""
	  (count-lines (point-min) (point))
	  (if pnm:edit-on-row "(on row)" "(on change between rows)")
	  (if pnm:edit-on-place "(on place)" "(on gap/crossing)")
	  (current-column)
	  pnm:edit-row-number pnm:left-place pnm:edit-places-string
	  (placenotation-mode:row-for-restart)))

(defun placenotation-mode:internal-info ()
  "Display internal information, mostly for development and debugging purposes."
  (interactive)
  (placenotation-mode:open-line-edit)
  (message (placenotation-mode:internal-info-string))
  (placenotation-mode:close-line-edit))

(defun placenotation-mode:cross ()
  "Make a crossing."
  (interactive)
  (placenotation-mode:with-open-line-edit
   (message "X")
   (if (pnm:making-places pnm:left-place)
       (progn
	 (ringing:message "Lower bell already making places")
	 (if (pnm:making-places pnm:right-place)
	     (progn
	       (ringing:message "Lower bell already making places and upper bell already making places")
	       (pnm:set-next pnm:left-place pnm:right-place)
	       (pnm:set-next pnm:right-place pnm:left-place)
	       )
	   (progn
	     (ringing:message "Lower bell already making places but upper bell not already making places")
	     ))
	 )
     (progn
       (ringing:message "Lower bell not already making places")
       (if (pnm:making-places pnm:right-place)
	   (progn
	     (ringing:message "Lower bell not already making places but upper bell already making places")
	     )
	 (progn
	   (ringing:message "Lower bell not already making places and upper bell not already making places")
	   ))))))

(defun placenotation-mode:places ()
  "Make a bell (or two) make places."
  (interactive)
  (placenotation-mode:with-open-line-edit
   (ringing:message "| |")
   (if (pnm:making-places pnm:left-place)
       (progn
	 (ringing:message "Lower bell already making places")
	 (if (pnm:making-places pnm:right-place)
	     (progn
	       (ringing:message "Lower bell already making places and upper bell already making places")
	       )
	   (progn
	     (ringing:message "Lower bell already making places but upper bell not already making places")
	     ))
	 )
     (progn
       (ringing:message "Lower bell not already making places")
       (if (pnm:making-places pnm:right-place)
	   (progn
	     (ringing:message "Lower bell not already making places but upper bell already making places")
	     )
	 (progn
	   (ringing:message "Lower bell not already making places and upper bell not already making places")
	   (pnm:set-next pnm:left-place pnm:left-place)
	   (pnm:set-next pnm:right-place pnm:right-place)
	   ))))))

(defun placenotation-mode:swap-cross-and-place ()
  "Change cross-place to place-cross."
  (interactive)
  (placenotation-mode:with-open-line-edit
   (ringing:message "|  X <--> X  |")

   )
  (error "Sorry, not implemented yet")
  )

(defun placenotation-mode:ensure-on-change ()
  "Give an error if not on a change."
  (placenotation:set-line-variables)  
  (when pnm:edit-on-row
    (error "This command may only be used on a change, not a row.")))

; This one gives a problem when in the left-most column:
; on place; left crossing; right crossing
; " X  *X " --> " X  | |"

(defun placenotation-mode:symbol-here ()
  "Re-arrange the place notation such that there is a symbol at point."
  (interactive)
  (placenotation-mode:ensure-on-change)
  (placenotation-mode:with-open-line-edit
   (let* ((left-making-places (pnm:making-places pnm:left-place))
	  (left-crossing (not left-making-places))
	  (right-making-places (pnm:making-places pnm:right-place))
	  (right-crossing (not right-making-places))
	  (on-place pnm:edit-on-place)
	  (on-crossing (not on-place))
	  )
     (when pnm:debug
       (ringing:message "%s; %s; %s"
			(if on-place "on place" "on crossing")
			(if left-making-places "left making places" "left crossing")
			(if right-making-places "right making places" "right crossing")))
     (cond
      ((and on-place
	    left-making-places
	    right-making-places)
       (when pnm:debug (ringing:message "Nothing to do (on pair of places)")))
      ((and on-place
	    left-making-places
	    right-crossing)
       (let* ((right pnm:right-place)
	      (far-right (1+ right)))
	 (when pnm:debug (ringing:message "\"| *X \" --> \"| | |\""))
	 (pnm:set-next right right)
	 (pnm:set-next far-right far-right)))
      ((and on-place
	    left-crossing
	    right-making-places)
       (when pnm:debug (ringing:message "Nothing to do (on place)")))
      ((and on-place
	    left-crossing
	    right-crossing)
       (if pnm:edit-on-place-to-right-of-cross
	   (let ((left pnm:left-place)
		 (right pnm:right-place))
	     (when pnm:debug (ringing:message "\" X*  X \" --> \"| |  X \""))
	     (pnm:set-next left left)
	     (pnm:set-next right right))
	 (let* ((right pnm:right-place)
		(far-right (1+ right)))
	   (when pnm:debug (ringing:message "\" X  *X \" --> \" X  | |\""))
	   (pnm:set-next right right)
	   (pnm:set-next far-right far-right))))
      ((and on-crossing
	    left-making-places
	    right-making-places)
       (let ((left pnm:left-place)
	     (right pnm:right-place))
	 (when pnm:debug (ringing:message "\"|*|\" --> \" X \""))
	 (pnm:set-next left right)
	 (pnm:set-next right left)))
      ((and on-crossing
	    left-making-places
	    right-crossing)
       (let* ((left pnm:left-place)
	      (right pnm:right-place)
	      (far-right (1+ right)))
	 (when pnm:debug (ringing:message "\"|* X \" --> \" X  |\""))
	 (pnm:set-next left right)
	 (pnm:set-next right left)
	 (pnm:set-next far-right far-right)))
      ((and on-crossing
	    left-crossing
	    right-making-places)
       (let* ((left pnm:left-place)
	      (far-left (1- left))
	      (right pnm:right-place))
	 (when pnm:debug (ringing:message "\" X *|\" --> \"|  X \""))
	 (pnm:set-next far-left far-left)
	 (pnm:set-next left right)
	 (pnm:set-next right left)))
      ((and on-crossing
	    left-crossing
	    right-crossing)
       (if (and (pnm:going-up pnm:left-place)
		(pnm:going-down pnm:right-place))
	   (when pnm:debug (ringing:message "Nothing to do -- already on crossing"))
	 (let* ((left pnm:left-place)
		(far-left (1- left))
		(right pnm:right-place)
		(far-right (1+ right)))
	   (when pnm:debug (ringing:message "\" X * X \" --> \"|  X  |\""))
	   (pnm:set-next far-left far-left)
	   (pnm:set-next left right)
	   (pnm:set-next right left)
	   (pnm:set-next far-right far-right))))))))

(defun placenotation-mode:mouse-symbol-here (click)
  "Put a symbol where the mouse was clicked, on a change.
This makes the minimum alterations to the change to get a symbol into that
position (i.e. to make a crossing or place there)."
  (interactive "@e")
  (goto-char (posn-point (event-start click)))
  (placenotation-mode:symbol-here))

(defun placenotation-mode:mouse-point-here (click)
  "Put the point where the mouse was clicked, and update cursor."
  (interactive "@e")
  (goto-char (posn-point (event-start click)))
  (placenotation-mode:move-underlining))
  
;;; end of placenotation-mode.el
