;;;; plain-courses.el -- part of John's elisp change-ringing system
;;; display and ring plain courses
;;; Time-stamp: <99/06/24 09:09:26 johnstu>

(eval-when (compile load eval)
  (provide 'plain-courses)
  (require 'ringing-methods)
  (require 'beer)			; but you should probably load the lot anyway
  (require 'method-display)
  (require 'method-lines)
  (require 'bell-numbers)
  (require 'manual-ringing)
  (require 'touch-book))

(defvar ringing:plain-course-row-count 0
  "Number of rows rung this time.")

(defvar ringing:lead-end-callback nil
  "A function to call at each lead end.
Called with row, remaining-changes, method, and ringing:lead-end-callback-args.")

(defvar ringing:lead-end-callback-args nil
  "Extra arguments for ringing:lead-end-callback.")

(defun ringing:plain-course-helper-function (row
					     remaining-changes
					     method-holder
					     rounds
					     place-marker)
  "Helper function for ringing or showing plain courses."
  (incf ringing:plain-course-row-count)
  (if (string= row rounds)
      nil
    (if remaining-changes
	t
      (let ((method (car method-holder)))
	(cons row
	      (case (car place-marker)
		((lead-end)

		 (when ringing:lead-end-callback
		   (apply ringing:lead-end-callback
			  row
			  remaining-changes
			  method
			  ringing:lead-end-callback-args))

		 (rplaca place-marker 'work)
		 (change-ringing-method-work method))
		((work)
		 (rplaca place-marker 'lead-end)
		 (change-ringing-method-plain-lead-end method))))))))

;;; interactive entry points

(defvar ringing:method-buffer-name-format "*%s*"
  "Format for buffer names based on a method name string")

(defun ringing:show-plain-course (method-name manual-bell)
  "Show a plain course of METHOD-NAME marking out MANUAL-BELL."
  (interactive
   (let* ((method-name (ringing:read-method-name "Show plain course of: "))
	  (manual (ringing:read-number-up-to
		   "Observation bell"
		   (change-ringing-method-nbells
		    (ringing:find-named-method method-name)))))
     (list method-name manual)))
  (let* ((method (ringing:find-named-method method-name))
	 (ringing:observation-number manual-bell)
	 (ringing:observation (ringing:bell-number-to-char manual-bell)))
    (require 'mSiRiL-mode)
    (require 'ermm)
    (ermm:init (change-ringing-method-nbells method))
    (mSiRiL-mode:lines-for-method method nil
				  'switch-to-buffer)))

(defun ringing:ring-plain-course (method-name manual-bell)
  "Ring a plain course of METHOD-NAME, with the user ringing MANUAL-BELL."
  (interactive
   (let* ((method-name (ringing:read-method-name "Ring plain course of: "))
	  (manual (ringing:read-number-up-to
		   "Manual bell"
		   (change-ringing-method-nbells
		    (ringing:find-named-method method-name)))))
     (list method-name manual)))
  (setq ringing:clashes 0
	ringing:plain-course-row-count 0)
  (setq ringing:prev-place (ringing:bell-char-to-number manual-bell))
  (let* ((ringing-buffer (ringing:get-buffer-create
			  (format ringing:method-buffer-name-format (ringing:method-name method-name))))
	 (method (ringing:find-named-method method-name))
	 (ringing:observation-number manual-bell)
	 (ringing:observation (ringing:bell-number-to-char manual-bell))
	 (ringing:allow-timeout-this-row nil) ; will be set to t after one row
	 (nbells (change-ringing-method-nbells method))
	 (ringing:test-user t)
	 (ringing:end-of-row-char ? )	; before making the rounds
	 (method-display:block-top-marker (make-marker))
	 ;; (the-rounds (ringing:make-rounds nbells))
	 (place-marker (list 'work)))
    (switch-to-buffer ringing-buffer)
    (erase-buffer)
    (delete-other-windows)
    (insert (format "Plain course of %s %s\n"
		    (capitalize (ringing:method-name method-name))
		    (change-ringing-method-original-description-string method)))
    (set-marker method-display:block-top-marker (point))
    (require 'ermm)
    (ermm:init (change-ringing-method-nbells method))
    (if (eq method-display:main-display-function 'method-display:multi-column-display-function)
	(progn
	  (insert ?\n)
	  (when ringing:label-columns
	    (insert (centre-string-in-width (format "%d" ringing:observation-number)
					    (length (method-display:make-lead-end-line nbells)))
		    ?\n)
	    (end-of-line 1)))
      (progn
	))
    ;;    (insert "\n")
    (let* ((ringing:lead-end-callback 'ringing:lead-end-record-clashes)
	   (ringing:lead-end-callback-args (list (cons ringing:clashes ringing:observation-number)))
	   (start-time (current-time))
	   (ringing-result
	    (ringing:ring-from-rounds method
				      'ringing:plain-course-helper-function
				      (list place-marker)
				      method-display:main-display-function
				      (list place-marker)))
	   (end-time (current-time))
	   (time-taken (seconds-between start-time end-time)))
      (set-marker method-display:block-top-marker nil)
      (if (eq ringing-result 'user-stand)
	  (progn
	    (kill-buffer ringing-buffer)
	    'user-stand)
	(progn
	  (touch-book:record-clash-level method-name
					 (cdar ringing:lead-end-callback-args)
					 (- ringing:clashes (caar ringing:lead-end-callback-args)))
	  (with-output-to-temp-buffer "*Ringing report*"
	    (princ (format "Touch of %d changes finished with %d clashes, in %d seconds\nPeal time would be %s (%d changes per minute)"
			   ringing:plain-course-row-count
			   ringing:clashes
			   time-taken
			   (time-string (* (float time-taken)
					   (/ 5000.0
					      (float ringing:plain-course-row-count))))
			   (floor (/ (float ringing:plain-course-row-count) (/ (float time-taken) 60.0)))
			   )))
	  t)))))

(defun ringing:lead-end-record-clashes (row remaining-changes method clash-per-place-counter)
  "Record clashes per lead."
  (let* ((clashes-before-this-lead (car clash-per-place-counter))
	 (clashes-this-lead (- ringing:clashes clashes-before-this-lead))
	 (which-lead-that-was (cdr clash-per-place-counter))
	 (which-lead-is-next (string-match (format "%c\\|\\*" ringing:observation) row))
	 )
    (touch-book:record-clash-level method which-lead-that-was clashes-this-lead)
    (rplaca clash-per-place-counter ringing:clashes)
    (rplacd clash-per-place-counter which-lead-is-next)))

;; end of plain-courses.el
