;;;; ps-line.el -- part of John's elisp change-ringing system
;;; Time-stamp: <99/11/03 10:17:56 johnstu>

(provide 'ps-line)
(require 'ps-config)
(require 'beer)
(require 'leads)
(require 'place-notation)
(require 'bell-numbers)
(require 'method-library)
(require 'method-groups)
(require 'ringing)			; need to know ringing-el-dir

(defvar ringing:keep-postscript-files t
  "Whether to leave the PostScript files around after printing.")

(defvar ringing:keep-postscript-buffers nil
  "Whether to keep the PostScript buffers around after printing.")

(defvar ringing:keep-postscript-comments nil
  "Whether to leave full comments in the PostScript files produced.")

(defun ringing:changes-to-moves (changes n)
  "Convert CHANGES to line movements (only for first N bells)."
  (let ((i 1)
	(moves (make-string n ? )))
    (while (<= i n)
      (aset moves (1- i)
	    (let ((next (aref changes i)))
	      (cond
	       ((= next i) ?|)
	       ((< next i) ?<)
	       (t ?>))))
      (incf i))
    moves))

(defun ringing:postscript-output-function (row
					   remaining-changes
					   method-holder
					   rounds
					   place-marker)
  "Display helper function for outputting PostScript from methods."
  ;; (insert (format "%% %s\n" remaining-changes))
  (let ((method (car method-holder))
	(this-change (car remaining-changes)))
    (when (eq (car place-marker) 'lead-end)
      (insert "="))
    (when this-change
      (insert (ringing:changes-to-moves this-change
					(change-ringing-method-nbells method))
	      ?,)
      (when (> (current-column) 72)
	(insert "\n            ")))))

(defvar ringing:postscript-styles-dir (expand-file-name "ps-styles" ringing-el-dir)
  "Where to look for the PostScript style definitions.")

(defun ringing:ps-styles ()
  "The possible styles for methods."
  (mapcar 'list
	  (directory-files ringing:postscript-styles-dir
			   nil
			   "[a-z].*[^~]$")))

(defvar ringing:latest-ps-style "default"
  "The most recently used PS printing style.")

(defvar ringing:postscript-title-styles-dir (expand-file-name "ps-title-styles" ringing-el-dir)
  "Where to look for the PostScript title style definitions.")

(defun ringing:ps-title-styles ()
  "The possible styles for titles."
  (mapcar 'list
	  (directory-files ringing:postscript-title-styles-dir
			   nil
			   "[a-z].*[^~]$")))

(defvar ringing:latest-ps-title-style "default"
  "The most recently used PS title printing style.")

(defvar ringing:postscript-group-styles-dir (expand-file-name "ps-group-styles" ringing-el-dir)
  "Where to look for the PostScript group style definitions.")

(defun ringing:ps-group-styles ()
  "The possible styles for groups."
  (mapcar 'list
	  (directory-files ringing:postscript-group-styles-dir
			   nil
			   "[a-z].*[^~]$")))

(defvar ringing:latest-ps-group-style "default"
  "The most recently used PS group printing style.")

(defvar ringing:postscript-grid-styles-dir (expand-file-name "ps-grid-styles" ringing-el-dir)
  "Where to look for the PostScript grid style definitions.")

(defun ringing:ps-grid-styles ()
  "The possible styles for grids."
  (mapcar 'list
	  (directory-files ringing:postscript-grid-styles-dir
			   nil
			   "[a-z].*[^~]$")))

(defvar ringing:latest-ps-grid-style "none"
  "The most recently used PS grid printing style.")

(defvar ringing-style-prompter nil
  "The style prompter.
Nil globally, may be set in file variables of style files.")

(defvar ringing-style-default-parameters nil
  "The non-interactive equivalent of the style prompter.
Nil globally, may be set in file variables of style files.")

(defvar ringing-suggested-layout nil
  "The suggested layout.
Nil globally, may be set in file variables of method style files.")

(make-variable-buffer-local 'ringing-style-prompter)
(make-variable-buffer-local 'ringing-suggested-layout)
(make-variable-buffer-local 'ringing-style-default-parameters)

(defun ringing:prompt-for-style-parameters (file)
  "Prompt for any parameters wanted by FILE."
  (let ((file-was-in-buffer (find-buffer-visiting file)))
    (save-window-excursion
      (find-file file)
      ;; have had problems, I think may have been if buffer loaded by state saver
      ;; before the ringing system loaded?
      (hack-local-variables)
      (prog1
	  (if ringing-style-prompter
	      (eval ringing-style-prompter)
	    nil)
	(if file-was-in-buffer
	    (bury-buffer file-was-in-buffer)
	  (kill-buffer (current-buffer)))))))

(defun ringing:default-style-parameters (file)
  "Get defaults for any parameters wanted by FILE.
Like ringing:prompt-for-style-parameters, but doesn't prompt the user for anything."
  (let ((file-was-in-buffer (find-buffer-visiting file)))
    (save-window-excursion
      (find-file file)
      ;; have had problems, I think may have been if buffer loaded by state saver
      ;; before the ringing system loaded?
      (hack-local-variables)
      (prog1
	  (if ringing-style-default-parameters
	      (eval ringing-style-default-parameters)
	    nil)
	(if file-was-in-buffer
	    (bury-buffer file-was-in-buffer)
	  (kill-buffer (current-buffer)))))))

(defun ringing:get-suggested-layout (style)
  "Get the suggested layout for STYLE."
  (let* ((file (expand-file-name style ringing:postscript-styles-dir))
	 (file-was-in-buffer (find-buffer-visiting file)))
    (save-window-excursion
      (find-file file)
      ;; have had problems, I think may have been if buffer loaded by state saver
      ;; before the ringing system loaded?
      (hack-local-variables)
      (prog1
	  ringing-suggested-layout
	(if file-was-in-buffer
	    (bury-buffer file-was-in-buffer)
	  (kill-buffer (current-buffer)))))))

(defun ringing:postscript-header-stuff (title &optional
					      style style-parameters
					      title-style title-style-parameters
					      grid-style grid-style-parameters
					      group-style group-style-parameters)
  "Output all the header stuff for TITLE."
  (when (null style) (setq style "default"))
  (when (null title-style) (setq title-style "default"))
  (insert "%!PS-Adobe-2.0 EPSF-1.2\n")
  (insert (format "%%Creator: emacs-ringing %s\n" ringing-version))
  (insert (format "%%%%For: %s\n" (user-login-name)))
  (insert (format "%%%%Title: %s\n" title))
  (insert "%%Pages: 1\n")
  (insert (format "%%%%CreationDate: %s\n" (current-time-string)))
  (insert "%%BoundingBox: 0 0 0 0\n")	; don't really know until PS runs!
  (insert "%%DocumentFonts: Helvetica Courier\n") ; don't really know until PS runs!
  (insert "%%TitleStyle: " title-style "\n")
  (when group-style (insert "%%GroupStyle: " group-style "\n"))
  (insert "%%MethodStyle: " style "\n")
  (when grid-style (insert "%%GridStyle: " grid-style "\n"))
  (insert "%%EndComments\n")
  (insert "save /ringingsave exch def\n")
  (ringing:postscript-write-configuration-dictionary)
  (insert "ringingparameters begin\n")
  (insert "% header files follow:\n")
  (insert-file-contents (expand-file-name "general-header.ps" ringing-el-dir) nil) (goto-char (point-max)) (insert ?\n)
  (when (stringp title-style-parameters)
    (insert title-style-parameters ?\n))
  (insert-file-contents (expand-file-name title-style ringing:postscript-title-styles-dir) nil)
  (goto-char (point-max))
  (insert ?\n)
  (when group-style
    (when (stringp group-style-parameters)
      (insert group-style-parameters ?\n))
    (insert-file-contents (expand-file-name group-style ringing:postscript-group-styles-dir) nil)
    (goto-char (point-max))
    (insert ?\n))
  (when (stringp style-parameters)
    (insert style-parameters ?\n))
  (insert-file-contents (expand-file-name style ringing:postscript-styles-dir) nil)
  (goto-char (point-max))
  (insert ?\n)
  (if grid-style
      (progn
	(when (stringp grid-style-parameters)
	  (insert grid-style-parameters ?\n))
	(insert-file-contents (expand-file-name grid-style ringing:postscript-grid-styles-dir) nil)
	(goto-char (point-max))
	(insert ?\n))
    (progn
      (insert "/gridrow { pop pop pop } def\n")))
  (unless ringing:keep-postscript-comments
    (goto-char (point-min))
    (when (search-forward "header files follow" (point-max) t)
      (beginning-of-line 1)
      (save-excursion
	(while (re-search-forward "%.*$" (point-max) t)
	  (replace-match "")))
      (save-excursion
	(while (re-search-forward "^[ 	]+" (point-max) t)
	  (replace-match "")))
      (delete-matching-lines "^$")))
  (goto-char (point-max))
  (insert "%%EndProlog\n"))

(defun ringing:postscript-trailer-stuff ()
  "Output all the trailer stuff."
  (insert "%%Trailer\n")
  (insert "end % ringingparameters\n")
  (insert "ringingsave restore\n"))

(defun ringing:open-formats (style title-style &optional group-style)
  "Open the format resources as required."
  (insert (format "\n/%s /ChangeRingingMethodFormat findresource begin\n"
		  (if style style "Default")))
  (insert (format "  /%s /ChangeRingingTitleFormat findresource begin\n"
		  (if title-style title-style "Default")))
  (when group-style
    (insert (format "    /%s /ChangeRingingMethodGroupLayout findresource begin\n" group-style))))

(defun ringing:close-formats (&optional including-group)
  "Close the format resources."
  (when including-group (insert " end\n"))
  (insert "\n  end\nend\n"))

(defun ringing:method-name-for-PostScript-use (name)
  "Return NAME in a form suitable for use as a PostScript name."
  (let ((psname (copy-sequence name)))
    (dotimes (i (length psname))
      (when (memq (aref psname i) '(? ?*))
	(aset psname i ?_)))
    psname))

(defun ringing:definition-postscript-for-one-method (method)
  "Output the PostScript for ringing METHOD into the current buffer.
Called from ringing:postscript-for-method which also sets up the
stuff around the edge; could also be used repeatedly to output several
methods into one PostScript buffer."
  (let* ((nbells (change-ringing-method-nbells method))
	 (method-name (ringing:method-name method))
	 (nleads (ringing:count-plain-course-leads method))
	 (ringing:end-of-row-char ? )
	 (ringing:observation-number
	  (if (null ringing:observation-number)
	      (let ((lowest-places (ringing:lowest-working-bell-making-places
				    (car (change-ringing-method-plain-lead-end method)))))
		;; (insert (format "%% lowest bell making places is %d\n" lowest-places))
		(if (numberp lowest-places)
		    lowest-places
		  1))
	    ringing:observation-number))
	 (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	 (ringing:observation-number (ringing:bell-char-to-number ringing:observation))
	 (method-display:asterisk-manual-bell nil)
	 (place-marker (list 'work)))

    (insert (format "\n/%s <<\n" (ringing:method-name-for-PostScript-use method-name)))

    (insert (format "  /methodtitlestring (%s)\n"
		    (capitalize method-name)))
    (insert (format "  /placenotationstring (%s)\n"
		    (concat (place-notation:recreate-string-from-changes (change-ringing-method-work method) t t)
			    ", "
			    (place-notation:recreate-string-from-changes (change-ringing-method-plain-lead-end method)))))
    (insert (format "  /expandedplacenotationstring (%s)\n"
		    (place-notation:create-fully-expanded-from-changes (change-ringing-method-plain method))))
    (insert (format "  /nleads %d\n" nleads))
    (insert (format "  /nrows %d\n" (ringing:count-lead-rows method)))
    (insert (format "  /nbells %d\n" nbells))
    (insert (format "  /observation %d\n" ringing:observation-number))
    (insert (format "  /latestchange %d string\n" nbells))
    (insert (format "  /rounds (%s)\n"
		    (substring (ringing:make-rounds nbells t) 1 (1+ nbells))))
    (insert "  /changes (")
    (ringing:ring-from-rounds-plain method
				    'ringing:ring-one-lead-helper-function
				    (list place-marker)
				    'ringing:postscript-output-function
				    (list place-marker))
    (insert "#)\n")

    (insert ">> definemethod\n")))

(defun ringing:printing-postscript-for-one-method (method)
  "Output the PostScript for ringing METHOD into the current buffer.
Called from ringing:postscript-for-method which also sets up the
stuff around the edge; could also be used repeatedly to output several
methods into one PostScript buffer."
  (let* ((nbells (change-ringing-method-nbells method))
	 (method-name (ringing:method-name method))
	 (nleads (ringing:count-plain-course-leads method))
	 (ringing:end-of-row-char ? )
	 (ringing:observation (if (null ringing:observation)
				  ;; (ringing:bell-number-to-char nbells)
				  (let ((lowest-places (ringing:lowest-working-bell-making-places
							(car (change-ringing-method-plain-lead-end method)))))
				    (if (numberp lowest-places)
					(ringing:bell-number-to-char lowest-places)
				      ringing:observation))
				ringing:observation))
	 (ringing:observation-number (ringing:bell-char-to-number ringing:observation))
	 (place-marker (list 'work)))
    (insert (format "      /%s printmethod\n" (ringing:method-name-for-PostScript-use method-name)))))

(defun ringing:definition-postscript-for-one-group (group-name methods)
  "Output the PostScript definition for group named GROUP-NAME containing METHODS."
  (insert (format "\n/%s <<\n"
		  (ringing:method-name-for-PostScript-use group-name)))
  (insert (format "  /grouptitlestring (%s)\n" group-name))
  (insert "  /methods [\n")
  (mapcar (function
	   (lambda (method)
	     (insert (format "    /%s\n" (ringing:method-name-for-PostScript-use (car method))))))
	  methods)
  (insert "  ]\n")
  (insert (format "  /nmethods %d\n" (length methods)))
  (insert ">> definegroup\n"))

(defun ringing:printing-postscript-for-one-group (group-name)
  "Output the PostScript to print defined group named GROUP-NAME."
  (insert (format "        /%s printgroup\n"
		  (ringing:method-name-for-PostScript-use group-name))))

(defun ringing:get-postscript-parameters-interactively (&optional for-group)
  "Return a list of style style-parameters title-style title-style-parameters
and optionally group-style and group-style-parameters."
  (let* ((style (setq ringing:latest-ps-style
		      (completing-read "Style: "
				       (ringing:ps-styles)
				       nil t
				       ringing:latest-ps-style)))
	 (style-file (expand-file-name style
				       ringing:postscript-styles-dir))
	 (style-parameters (ringing:prompt-for-style-parameters
			    style-file))
	 (suggested-layout (ringing:get-suggested-layout style-file))
	 (title-style (setq ringing:latest-ps-title-style
			    (completing-read "Title style: "
					     (ringing:ps-title-styles)
					     nil t
					     (if suggested-layout
						 suggested-layout
					       ringing:latest-ps-title-style))))
	 (title-style-parameters (ringing:prompt-for-style-parameters
				  (expand-file-name title-style
						    ringing:postscript-title-styles-dir)))
	 (grid-style (setq ringing:latest-ps-grid-style
			   (completing-read "Grid style: "
					    (ringing:ps-grid-styles)
					    nil t
					    ringing:latest-ps-grid-style)))
	 (grid-style-parameters (ringing:prompt-for-style-parameters
				 (expand-file-name grid-style
						   ringing:postscript-grid-styles-dir))))
    (if for-group
	(let* ((group-style (setq ringing:latest-ps-group-style
				  (completing-read "Layout style: "
						   (ringing:ps-group-styles)
						   nil t
						   ringing:latest-ps-group-style)))
	       (group-style-parameters (ringing:prompt-for-style-parameters
					(expand-file-name group-style
							  ringing:postscript-group-styles-dir))))
	  (list style style-parameters
		title-style title-style-parameters
		grid-style grid-style-parameters
		group-style group-style-parameters))
      (list style style-parameters
	    title-style title-style-parameters
	    grid-style grid-style-parameters))))

(defun ringing:postscript-for-method (method ps-buffer
					     &optional
					     style style-parameters
					     title-style title-style-parameters
					     grid-style grid-style-parameters)
  "Output METHOD as PostScript into PS-BUFFER,
with optional STYLE and TITLE-STYLE naming postscript drawing routine
preamble files."
  (interactive
   (let* ((method-name (ringing:read-method-name "Print method: "))
	  (ps-buffer-name (read-from-minibuffer "Buffer for PostScript: "
						(concat method-name ".ps")))
	  (params (ringing:get-postscript-parameters-interactively)))
     (append (list method-name ps-buffer-name) params)))
  (let ((method-data (ringing:find-named-method method)))
    (set-buffer (ringing:get-buffer-create ps-buffer))
    (erase-buffer)
    (ringing:postscript-header-stuff (ringing:method-name method)
				     style style-parameters
				     title-style title-style-parameters
				     grid-style grid-style-parameters)
    (ringing:definition-postscript-for-one-method method-data)
    (insert (format "%%%%Page: %s-1 1\n" (buffer-name)))
    (ringing:open-formats style title-style)
    (insert "\ncommonbeginpage\n")
    (ringing:printing-postscript-for-one-method method-data)
    (insert "\ncommonendpage\n")
    (ringing:close-formats)
    (ringing:postscript-trailer-stuff)
    (switch-to-buffer-other-window ps-buffer)
    ps-buffer))

(defvar ringing:sort-group-for-printing nil
  "*Whether to print methods in a group in alphabetical order.
Otherwise, they are printed in the order given.")

(defun ringing:postscript-for-group (group ps-buffer &optional
					   style style-parameters
					   title-style title-style-parameters
					   group-style group-style-parameters
					   grid-style grid-style-parameters
					   printed-title)
  "Output GROUP as PostScript into PS-BUFFER,
with optional STYLE and TITLE-STYLE naming PostScript drawing routine
preamble files, and GROUP-STYLE for the imposition PostScript.
With PRINTED-TITLE, use that on the printout instead of the natural
name for the group -- useful for strangely named groups such as ephemeral
ones."
  (interactive
   (progn
     (require 'method-groups)
     (let* ((group-name (completing-read "Print group: "
					 (method-groups:all-groups)
					 nil
					 t))
	    (printed-title (read-from-minibuffer "Title to put on page: " group-name))
	    (ps-buffer-name (read-from-minibuffer "Buffer for PostScript: "
						  (concat group-name ".ps")))
	    (params (ringing:get-postscript-parameters-interactively t)))
       (append (list group-name ps-buffer-name)
	       params
	       (list printed-title)))))
  (let* ((group-data (method-groups:find-named-group group))
	 (group-title (if (stringp printed-title)
			  printed-title
			(method-groups:method-group-title group-data)))
	 (group-methods (method-groups:method-group-methods group-data)))
    
      (setq group-methods (if ringing:sort-group-for-printing
			      (sort (copy-sequence group-methods)
				(function
				 (lambda (a b)
				   (string< (car a) (car b)))))
			    (reverse group-methods)))
    (set-buffer (ringing:get-buffer-create ps-buffer))
    (erase-buffer)
    (ringing:postscript-header-stuff (if printed-title printed-title group-title)
				     style style-parameters
				     title-style title-style-parameters
                                     grid-style grid-style-parameters
				     group-style group-style-parameters)
    (insert (format "%%%%Page: %s-1 1\n" (buffer-name)))
    (insert (format "%% start of group %s\n" group-title))
    (dolist (method group-methods)
      (message "Preparing definition postscript for %s" (car method))
      (ringing:definition-postscript-for-one-method (cdr method)))
    (ringing:definition-postscript-for-one-group group-title group-methods)
    (ringing:open-formats style title-style group-style)
    (insert "\n      commonbeginpage\n")
    (ringing:printing-postscript-for-one-group group-title)
    (insert "      commonendpage\n")
    (ringing:close-formats t)
    (insert (format "%% end of group %s\n" group))
    (ringing:postscript-trailer-stuff)
    (switch-to-buffer-other-window ps-buffer)
    ps-buffer))

(defun ringing:postscript-styles-sampler (sample-method)
  "Produce a page demonstrating all the print styles."
  (interactive
   (let* ((method-name (ringing:read-method-name "Method for sample prints: ")))
     (list method-name)))
  (find-file (expand-file-name (format "%s-sampler.ps"
				       (ringing:name-as-file-name sample-method))
			       mSiRiL-mode:ps-temp-files-directory))
  (erase-buffer)
  (insert "%!PS-Adobe-2.0 EPSF-1.2\n")
  (insert (format "%%Creator: emacs-ringing %s\n" ringing-version))
  (insert (format "%%%%For: %s\n" (user-login-name)))
  (insert (format "%%%%Title: %s\n" "sampler"))
  (insert "%%Pages: 1\n")
  (insert (format "%%%%CreationDate: %s\n" (current-time-string)))
  (insert "%%BoundingBox: 0 0 0 0\n")	; don't really know until PS runs!
  (insert "%%DocumentFonts: Helvetica Courier\n") ; don't really know until PS runs!
  (insert "%%EndComments\n")
  (insert "save /ringingsave exch def\n")
  (insert "/RingingPSdebug false def\n")
  (insert "/gridrow { pop pop pop } def\n")
  (ringing:postscript-write-configuration-dictionary)
  (insert "ringingparameters begin\n")
  (insert "% header files follow:\n")
  (insert-file-contents (expand-file-name "general-header.ps" ringing-el-dir) nil) (goto-char (point-max)) (insert ?\n)
  (goto-char (point-max))
  (insert ?\n)
  (let* ((styles (ringing:ps-styles))
	 (nstyles (length styles))
	 (styles-across (ceiling (sqrt nstyles)))
	(styles-and-layouts nil)
	(title-styles-wanted '("default"))
	(method (ringing:find-named-method sample-method)))
    (insert (format "/methodsacrosspage %d def\n/methodsdownpage %d def\n/numberpages false def\n"
		    styles-across (1+ (/ nstyles styles-across))))
    (insert-file-contents (expand-file-name "landscape-n-by-m" ringing:postscript-group-styles-dir) nil)
    (goto-char (point-max))
    (insert ?\n)
    (insert (format "/nmethods %d def\n" nstyles))
    (insert "/styletitlebuff 256 string def\n")
    (dolist (style styles)
      (let* ((filename (expand-file-name (car style) ringing:postscript-styles-dir))
	     (layout-style-wanted (ringing:get-suggested-layout filename)))
	(insert-file-contents filename nil)
	(when layout-style-wanted
	  (pushnew layout-style-wanted title-styles-wanted :test 'equal))
	(push (cons style (if layout-style-wanted layout-style-wanted "default")) styles-and-layouts))
      (goto-char (point-max))
      (insert ?\n))
    (dolist (title-style title-styles-wanted)
      (let* ((title-file (expand-file-name title-style ringing:postscript-title-styles-dir))
	     (title-params (ringing:default-style-parameters title-file)))
	(when title-params (insert "\n" title-params "\n"))
	(insert-file-contents title-file nil)
	(goto-char (point-max))))
    (ringing:definition-postscript-for-one-method method)
    (insert (format "%%%%Page: %s-1 1\n" (buffer-name)))
    (insert (format "    /landscape-n-by-m /ChangeRingingMethodGroupLayout findresource begin\n"))
    (insert "\ncommonbeginpage\n")
    (insert "firstmethod\n")
    (let* ((z (car styles-and-layouts))
	   (s (reverse (cdr styles-and-layouts))))
      (insert "[ ")
      (while s
	(let ((style-pair (car s)))
	  (insert (format "\n  [ /%s /%s ]"
			  (car (car style-pair))
			  (cdr style-pair))))
	(setq s (cdr s)))
      (insert "\n] {\n")
      (insert "  dup ==\n")
      (insert "  dup 1 get /ChangeRingingTitleFormat findresource begin\n")
      (insert "  0 get\n")
      (insert "  dup styletitlebuff cvs /Courier 8 selectfont xmid ymax 72 sub moveto methodtitlecentreshow\n")
      (insert "  /ChangeRingingMethodFormat findresource begin\n")
      (ringing:printing-postscript-for-one-method method)
      (insert "    end\n  end\n  nextmethod\n")
      (insert "} forall\n")
      (insert "[ /" (car (car z)) " /" (cdr z) " ]")
      (insert "  dup ==\n")
      (insert "  dup 1 get /ChangeRingingTitleFormat findresource begin\n")
      (insert "  0 get\n")
      (insert "  dup styletitlebuff cvs /Courier 8 selectfont xmid ymax 72 sub moveto methodtitlecentreshow\n")
      (insert "  /ChangeRingingMethodFormat findresource begin ")
      (insert ?\n)
      (ringing:printing-postscript-for-one-method method)
      (insert "    end  end\n"))
    (insert "lastmethod\n")
    (insert "commonendpage\n")
    (insert "end\n")
    (ringing:postscript-trailer-stuff)
    (basic-save-buffer))
  (current-buffer))

(defvar ringing:ps-print-command "lpr"
  "Command to send a named PostScript file to a suitable printer.")

(defun ringing:send-postscript-to-printer (ps-file-name)
  "Send PS-FILE-NAME to the printer."
  (message "Sending %s to printer" ps-file-name)
  (shell-command (format "%s %s" ringing:ps-print-command
			 (ringing:localize-file-name ps-file-name)))
  (message "%s queued for printing" ps-file-name))

(defun ringing:send-postscript-to-previewer (ps-file-name)
  "Send PS-FILE-NAME to the previewer."
  (message "Sending %s to previewer" ps-file-name)
  (shell-command (format ringing:simple-preview-method-printing-command-format
			 ps-file-name)))

(defvar ringing:simple-preview-method-printing-command-format
  "gs %s &"
  "Format for previewer command for ringing:send-postscript-to-previewer")

;;; end of ps-line.el
