;;;; recite-work.el -- part of John's elisp change-ringing system
;;; Time-stamp: <99/11/12 15:00:57 johnstu>

(eval-when (compile load eval)
  (provide 'recite-work)
  (require 'place-bell-lines))

(defun ringing:longest-work-at (work-string pieces)
  "Look in WORK-STRING for the longest of PIECES that matches it from the start.
PIECES is an alist of names consed with regexps.
The result is the PIECE consed with the end position reached."
  (let ((best-piece nil)
	(best-end-position 0))
    (dolist (piece pieces)
      (let ((this-end-position (and
				(string-match (first piece) work-string)
				(match-end 0))))
	(when (and this-end-position
		   (> this-end-position best-end-position))
	  (setq best-end-position this-end-position
		best-piece piece))))
    (cons best-piece best-end-position)))

(makunbound 'ringing:pieces-of-work)

(defvar ringing:pieces-of-work
  '(("^><>" up-dodge "up dodge" 1)
    ("^<><" down-dodge "down dodge" -1)
    ("^><><" double-dodge-up "double dodge up" 0)
    ("^<><>" double-dodge-down "double dodge down" 0)
    ("^><><" fishtail "fishtail" 0)
    ("^<<|><|>>" whole-turn "whole turn" 0)
    ("^>><<" half-turn "half turn" 0)
    ("^><>|<|><>|<|><>" cambridge-places-up "cambridge places up" 1)
    ("^<><|>|<><|>|<><" cambridge-places-down "cambridge places down" -1)
    ("^<><|>|<>" london-frontwork-up "london frontwork up" 0)
    ("^<>|<|><>" london-frontwork-down "london frontwork down" 0)
    ("^>" up "up" 1)
    ("^<" down "down" -1)
    ("^|" make-places "make places" 0))
  "Named pieces of work, as lists of (name regexp delta)")

(defun ringing:recitation-from-work-string (work-string place nbells)
  "Recite WORK-STRING, with the bell starting in PLACE of NBELLS.
The algorithm used is a greedy one, that is, it always takes the longest
description that matches the next section of the work string.
For some particular combinations of descriptions, it thus might not always
get the best one."
  (let ((recitation nil))
    (while (not (zerop (length work-string)))
      (let* ((next (ringing:longest-work-at work-string ringing:pieces-of-work))
	     (piece (car next))
	     (name (second piece))
	     (pretty-name (third piece))
	     (delta (fourth piece))
	     (whereunto (cdr next)))
        (case name
	  (make-places
	   (cond
	    ((= place 1) (setq name 'lead pretty-name "lead"))
	    ((= place nbells) (setq name 'lie pretty-name "lie"))
	    ((and (>= place 2) (<= place 11))
	     (setq pretty-name (concat "make " (aref ["seconds" "thirds" "fourths" "fifths" "sixths"
				    "sevenths" "eighths" "ninths" "tenths" "elevenths"]
				   (- place 2))))))))
	(push pretty-name recitation)
	(setq work-string (substring work-string whereunto))
	(incf place delta)))
    (nreverse recitation)))


(defun ringing:recite-method (method starting-bell)
  "Recite the work for METHOD, beginning at STARTING-BELL."
  (interactive
   (let* ((method-name (ringing:read-method-name "Recite work for: "))
	  (starting (ringing:read-number-up-to
		     "Starting bell"
		     (change-ringing-method-nbells
		      (ringing:find-named-method method-name)))))
     (list method-name starting)))
  (let* ((method-data (ringing:find-named-method method))
	 (raw-work-string (ringing:work-line method-data starting-bell t))
	 (recitation (ringing:recitation-from-work-string
		      raw-work-string starting-bell
		      (change-ringing-method-nbells method-data))))
    (with-output-to-temp-buffer (format "*Recitation of work for %s*" (change-ringing-method-title method-data))
      (dolist (piece recitation)
	(princ (format "%s\n" piece))))))

(defun foorecite ()
  "Test the recitation."
  (interactive)
  (let* ((method "Cambridge Surprise Minor")
	 (raw-work-string (ringing:work-line method 2 t))
	 (recitation (ringing:recitation-from-work-string raw-work-string 2 6)))
    (with-output-to-temp-buffer "*Recitation*"
      (dolist (piece recitation)
	(princ (format "%s\n" piece))))))

;;; end of recite-work.el
