;;;; ringing-methods.el -- part of John's elisp change-ringing system
;;; This file manages the list of methods
;;; Time-stamp: <99/03/11 09:24:36 johnstu>

(eval-when (compile load eval)
  (provide 'ringing-methods)
  (require 'place-notation)
  (require 'method-library))

;;;; finding methods and managing the list of them

(defvar ringing:always-use-method-library t
  "Whether to go straight to the method library,
skipping the predefined methods directory.

The predefined methods directory is really a leftover from an earlier
stage of this package, and on the whole the method libraries are more useful,
so it's probably best to keep this one switched on.

The main exception to this will be if you're composing new methods.")

(defvar ringing:always-scan-all-libraries t
  "Whether to look in all the libraries, rather than asking the user which one.

This takes a few seconds on a typical system, but saves you having to remember
the naming convention for the library files. This delay will happen only the first
time in the session that you ring on a particular number of bells.")

(defun ringing:ensure-methods-loaded (&optional force)
  "Load methods if they're not yet loaded."
  (when (and (or force (null ringing:known-methods))
	     (not ringing:always-use-method-library))
    (place-notation:parse-methods-in-directory
     (setq place-notation:default-methods-directory
	   (read-file-name
	    "No methods defined. Scan directory: "
	    place-notation:default-methods-directory
	    place-notation:default-methods-directory
	    t)))))

(defun ringing:find-named-method (method-name &optional no-error)
  "Find the method named METHOD-NAME.
If a method structure is given, it is returned."
  (if (change-ringing-method-p method-name)
      method-name
    (save-match-data
      (ringing:ensure-methods-loaded)
      (let ((pair (assoc method-name ringing:known-methods))
	    (likely-file (method-library:proper-file-for-method method-name)))
	(if pair
	    (place-notation:as-method pair likely-file)
	  ;; look for more methods, and try again
	  (let* ((from-file (if likely-file
				(microSiRiL:method-from-file likely-file method-name no-error (not no-error))
			      nil)))
	    (if from-file
		from-file
	      (if no-error
		  nil
		(error "Could not find definition of %s" method-name)))))))))

(defun ringing:name-data-pair (method)
  "From method (a string naming it, or the data structure) return a cons
of name and data. Will return such a pair as is, too."
  (cond
   ((and (consp method) (stringp (car method)))
    (if (change-ringing-method-p (cdr method))
	method
      (cons (car method) (ringing:find-named-method (car method) t))))
   ((stringp method)
    (cons method (ringing:find-named-method method t)))
   ((change-ringing-method-p method)
    (cons (change-ringing-method-title method) method))
   (t (error "Could not make a name-method pair from %s" method))))

(defun ringing:read-method-name (prompt)
  "Prompt the user for a method name, with PROMPT.
A more sophisticated routine, which loads parts of a larger method library
selectively, is method-library:choose-method-interactively, defined in 
method-library.el in this directory."
  (if ringing:always-use-method-library
      (method-library:choose-method-interactively
       prompt
       (not ringing:always-scan-all-libraries))
    (let ((completion-ignore-case t))
      (ringing:ensure-methods-loaded)
      (completing-read prompt
		       ringing:known-methods
		       nil t))))

(defun ringing:store-method (name data file)
  "Store method NAME described by DATA, in appropriate places including the methods from FILE."
  (let* ((file-methods (cdr (assoc file microSiRiL:methods-in-files)))
	 (method-in-file-pair (assoc name file-methods))
	 (main-list-pair (assoc name ringing:known-methods)))
    (if method-in-file-pair
	(rplacd method-in-file-pair data)
      (push (cons name data) file-methods))
    (if main-list-pair
	(rplacd main-list-pair data)
      (push (cons name data) ringing:known-methods))
    (setf (change-ringing-method-defining-file data) file)))

(defun ringing:define-anonymous-method (place-notation nbells)
  "Define a method, from a given string."
  (let ((method 
	 (make-change-ringing-method
	  :title place-notation
	  :nbells nbells
	  :plain (place-notation:parse-string place-notation nbells))))
    (ringing:store-method
     place-notation			; re-use as name
     (list nil method)
     "minibuffer")
    method))

(defvar ringing:anonymous-method-history nil
  "History for typing place notation into the minibuffer.")

(defun ringing:anonymous-method-args-reader (prompt)
  "Read the arguments describing anonymous methods."
  (require 'stages)
  (let ((places (read-from-minibuffer prompt
				      nil
				      nil
				      nil
				      'ringing:anonymous-method-history))
	(n (ringing:read-stage-from-user "Number of bells: ")))
    (list places n)))

(defun ringing:display-anonymous-method (place-notation nbells)
  "Display the lines for PLACE-NOTATION on NBELLS."
  (interactive (ringing:anonymous-method-args-reader "Display lines for place notation: "))
  (let ((method (ringing:define-anonymous-method place-notation nbells)))
    (require 'mSiRiL-mode)
    (mSiRiL-mode:lines-for-method method)))

;;; end of ringing-methods.el
