;;;; ringing.el -- part of John's elisp change-ringing system
;;;; This file is the top-level, that loads the other files
;;; Time-stamp: <99/07/02 10:20:29 johnstu>

(provide 'ringing)
(require 'cl)
(load-library "cl-extra")		; something broken in autoload of this

(defconst ringing-version "1.26"
  "The currently loaded version of the emacs ringing system.")

(defvar ringing-el-dir nil
  "The directory containing the emacs-ringing package.
If your version of emacs doesn't define the variable load-file-name,
you should edit the definition of this variable to be a string giving
the directory in which you have installed this package.")

(when (and (null ringing-el-dir)
	   (boundp 'load-file-name))
  (setq ringing-el-dir (file-name-directory load-file-name)))

(when (and (null ringing-el-dir)
	   (not (boundp 'load-file-name)))
  (with-output-to-temp-buffer "*Installation instructions*"
    (princ  "Your emacs has not defined load-file-name.
Please edit the definition of ringing-el-dir near the top of ringing.el"))
  (error "Some further manual installation is required."))

(when (file-directory-p ringing-el-dir)
  (pushnew ringing-el-dir load-path))

(defvar ringing:options-file (expand-file-name "~/.ringing.el")
  "*The filename for storing configuration for this package.")

;;;; an auxiliary routine I use a lot

(if (not (fboundp 'buffer-matched))
    (defmacro buffer-matched (n)
      "Return the Nth pattern-matched string from the current buffer"
      (list 'buffer-substring (list 'match-beginning n) (list 'match-end n))))

;;;; compile all this stuff

(defun ringing:compile-system ()
  "(Re)compile the ringing system."
  (interactive)
  (let ((el-files (directory-files ringing-el-dir t "\.el$")))
  ;; first, load all files, to stop the compiler warning
  ;; about unknown variables
    (mapcar 'load-file el-files)
    ;; compile them all -- byte-recompile-directory skips
    ;; any which previously had no compiled version
    (mapcar 'byte-compile-file el-files)
    ;; now load them all, from the compiled version,
    ;; so we know just where we are
    (mapcar '(lambda (filename)
	       (load-file (concat filename "c")))
	    el-files)))

;;;; autoloads for plain courses

(autoload 'ringing:show-plain-course "plain-courses"
  "Show a plain course of METHOD." t)

(autoload 'ringing:ring-plain-course "plain-courses"
  "Ring a plain course of METHOD-NAME, with the user ringing MANUAL-BELL." t)

(autoload 'ringing:dixonoid-plain-course-helper-function "dixonoids"
  "Helper function for ringing or showing plain courses of things like Dixons." t)

;;;; autoloads for touches

(autoload 'ringing:show-touch "touches"
  "Show a touch of METHOD with TOUCH giving the actions at each lead end." t)

(autoload 'ringing:ring-touch "touches"
  "Ring a touch of METHOD-NAME, with the user ringing MANUAL-BELL, with TOUCH-CALLS." t)

(autoload 'ringing:show-region-as-touch "touches"
  "Take text between FROM and TO as callings for a touch of METHOD." t)

(autoload 'ringing:show-buffer-as-touch "touches"
  "Take text of buffer as callings for touch of METHOD." t)

(autoload 'ringing:define-touch "touches"
  "Define a touch called NAME using CALLS." t)

(autoload 'deftouch "touches"
  "Define TOUCHNAME to call TOUCHCALLS by passing them to ringing:define-touch." nil 'macro)

(autoload 'touch "touches"
  "Return a normalized touch containing TOUCHCALLS.
This is so you can write (touch (go \"Cambridge Surprise Minor\")) like a quoted list constant." nil 'macro)

(autoload 'ringing:find-named-touch "touches"
  "Return the touch called NAME." t)

(autoload 'ringing:show-named-touch "touches"
  "Show the touch named TOUCHNAME." t)

;;;; autoloads for single leads

(autoload 'teach-ringing:do-one-lead "teacher"
  "Ring a lead of METHOD-NAME, with the user ringing MANUAL-BELL if WITH-USER." t)

(autoload 'ringing:ring-one-lead-helper-function "leads"
  "Helper function for ringing one lead of a method." t)

;;;; autoloads for teaching

(autoload 'ringing:teach-random-lead-of-method "teacher"
  "Show the user a random lead of METHOD-NAME, then let them try it themselves." t)

(autoload 'ringing:teach-random-leads-of-method "teacher"
  "Show the user random leads of METHOD-NAME, then let them try them themselves." t)

(autoload 'ringing:teach-random-methods-from-file "teacher"
  "Show the user random leads from random methods in FILE, and get them to try them." t)

(autoload 'ringing:teach-random-methods-from-group "teacher"
  "Show the user random leads from random methods in GROUP, and get them to try them." t)

(autoload 'ringing:teach-lead-until-perfect "teacher"
  "Get the user to try METHOD, PLACE-BELL until they can get it right." t)

(autoload 'ringing:teach-each-lead-of-method-until-perfect "teacher"
  "Get the user to try each lead in turn of METHOD-NAME, repeating each until reliable." t)

;;;; autoloads for method groups

(autoload 'method-groups:all-groups "method-groups"
  "Return the list of method groups, each dotted with nil or known members." t)

(autoload 'method-groups:group-methods "method-groups"
  "Return the method names in GROUP. Use cache unless FORCE-NEW given." t)

(autoload 'method-group:print-group "method-groups"
  "Print STAGE, GROUP via PostScript, with optional style args as for ringing:postscript-for-group." t)

(autoload 'ringing:quiz-at-stage "method-groups"
  "Quiz the user on random leads of methods at STAGE.
The methods are defined in the method group file called quiz-n where n
is the stage as a number (in the method groups directory, as defined by
method-groups:groups-directory." t)

(autoload 'method-group:report "analyze-group"
  "Report on the methods in GROUP." t)

;;;; autoloads for method editor

(autoload 'full-grid-edit "full-grid-mode"
  "Edit place notation on N bells starting with ORIGINAL." t)

(autoload 'full-grid-explore "full-grid-mode"
  "Display N-bell method with PLACE-NOTATION." t)

;;;; some low-level stuff

;;; might want to do this before everything else
(autoload 'place-notation:debugging "place-notation"
  "Enable or disable DEBUG of place notation reading." t)

;;;; mSiRiL viewing / editing mode

(defvar ringing:default-method-library-file-regexp "method-libraries"
  "*How to spot things that should be in mSiRiL mode.")

(autoload 'mSiRiL-mode "mSiRiL-mode"
  "Major mode for microSiRiL files." t)

(if (not (assoc ringing:default-method-library-file-regexp auto-mode-alist))
    (push (cons ringing:default-method-library-file-regexp 'mSiRiL-mode)
	  auto-mode-alist))

;;;; method groups mode

(defvar ringing:default-method-group-file-alist "method-groups[^.]"
  "How to spot things that should be in method-group mode")

(autoload 'method-group-mode "method-groups"
  "Major mode for method group files." t)

(if (not (assoc ringing:default-method-group-file-alist auto-mode-alist))
    (push (cons ringing:default-method-group-file-alist 'method-group-mode)
	  auto-mode-alist))

;;;; autoloads for printing

(autoload 'ringing:postscript-for-method "ps-line"
  "Output METHOD as PostScript." t)

(autoload 'ringing:postscript-debug "ps-config"
  "Enable (or with prefix arg, disable) debug in the PostScript output." t)

;;;; autoloads for composition checking

(autoload 'ringing:check-touch "composition-checker"
  "Check a touch of TOUCH-CALLS with METHOD giving the actions at each lead end.
If METHOD is not given, looks for a \"go\" call in the touch.
SO-FAR, if given, contains rows already rung (not including the initial rounds).
This may be a buffer, or a list of strings.
An optional HELPER-FUNCTION overrides the normal one.
For interactive calling, only TOUCH-CALLS and METHOD are used.
Return is a list of two items: whether the touch so far is true (as t
or nil), and the new value of SO-FAR, as a list of strings or as a
buffer, according to what was supplied. If a buffer is given, it is
extended to contain the extra rows. If a list is given, a list is
returned whose tail is the original one. Changes are added at the
front of this list." t)

;;;; autoloads for composition search

(autoload 'ringing:true-completions-for-touch "composition-search"
  "Return a list of true completions for TOUCH-SO-FAR using METHODS.
An optional HELPER-FUNCTION may be given (to handle anything funny about the
methods.
TOUCH-SO-FAR should be a vector, as described for ringing:check-touch,
which see. METHODS should be a method group, as returned, for example,
by method-groups:find-named-group.")

;;;; autoloads for touch book

(autoload 'touch-book:load "touch-book"
  "Load the touch book from file, if needed; return the touch list." t)

(autoload 'touch-book:list-my-methods "touch-book"
  "List the methods the user has rung." t)

(autoload 'touch-book:worst-leads-first "touch-book"
  "Return a list of leads, as from touch-book:list-all-leads, worst first,
that is, sorted according to the percentage success rate.
Interactively, display the list too." t)

(autoload 'touch-book:teach-worst-lead "touch-book"
  "Teach the ringer their current worst lead." t)

;;;; autoloads for handbells

(autoload 'handbells:ring-method "handbells"
  "Ring a plain course of METHOD interactively, using RIGHT-HAND-BELL
and the corresponding left-hand bell." t)

;;;; customise interface

(when (string-match "^\\([0-9]+\\)\\.\\([0-9]+\\)" emacs-version)
  (let ((major-emacs-version (string-to-int (substring emacs-version (match-beginning 1) (match-end 1))))
	(minor-emacs-version (string-to-int (substring emacs-version (match-beginning 2) (match-end 2)))))
    (message "%d %d" major-emacs-version minor-emacs-version)
    (if (or (> major-emacs-version 19)
	    (and (= major-emacs-version 19)
		 (>= minor-emacs-version 30)))
	(autoload 'ringing:customise "ringing-cus.el"
	  "Customise John's Emacs Ringing System." t)
      (message "Version of emacs is too early to have customize system."))))

;;;; set variables

(defun read-float-from-user (prompt &optional default allow-non-number)
  "Read a number from the user, with PROMPT. Optional DEFAULT, ALLOW-NON-NUMBER"
  (let* ((string (read-from-minibuffer prompt
				       (if default (format "%f" default) nil)))
	 (number (read string)))
    (while (and (not allow-non-number) (not (numberp number)))
      (setq string (read-from-minibuffer (concat "Enter a number! " prompt))
	    number (read string)))
    number))

(defun read-int-from-user (prompt &optional default allow-non-number)
  "Read an integer from the user, with PROMPT. Optional DEFAULT, ALLOW-NON-NUMBER"
  (let* ((string (read-from-minibuffer prompt
				       (if default (format "%d" default) nil)))
	 (number (read string)))
    (while (and (not allow-non-number) (not (integerp number)))
      (setq string (read-from-minibuffer (concat "Enter a number! " prompt))
	    number (read string)))
    number))

(defun ringing:read-char-from-user (prompt default)
  "Read a character from the user, with PROMPT and DEFAULT.
The default is taken if the user types Return."
  (message "%s (default %c)" prompt default)
  (let ((char (read-char)))
    (message "")
    (if (= char 13)
	default
      char)))

(defun read-directory-name (prompt &optional default)
  "Read a directory name from the user, with PROMPT and optional DEFAULT."
  (let ((dir (read-file-name prompt default default t))) ;; t = DIR must exist
    (when (string= dir "") (setq dir default))
    (setq dir (expand-file-name dir))
    (if (not (string-match "[/\\]$" dir)) (setq dir (concat dir "/")))
    dir))

(defun old-read-directory-name (prompt &optional default)
  "Read a directory name from the user, with PROMPT and optional DEFAULT."
  (let ((dir (read-file-name prompt default ".")))
    (while (not (file-directory-p dir))
      (setq dir (read-file-name (concat "Must enter directory: " prompt) default ".")))
    (when (string= dir "") (setq dir default))
    (setq dir (expand-file-name dir))
    (if (not (string-match "[/\\]$" dir)) (setq dir (concat dir "/")))
    dir))

(defun mSiRiL:mark-mode-regexp-from (method-library-directory)
  "Try to make mSiRiL files be recognized as such."
  (pushnew
   (cons (file-name-nondirectory
	  (if (string-match "[/\\]$" method-library-directory)
	      (substring method-library-directory 0 -1)
	    method-library-directory))
	 'mSiRiL-mode)
   auto-mode-alist
   :test 'equal))

(makunbound 'ringing:set-variables)

(defvar ringing:set-variables
  '(("Pop up buffer for clashes" ringing:show-clashes-warning-big
     (y-or-n-p "Pop up buffer for clashes? ")
     plain-courses)
    ("Highlight user mistakes" ringing:highlight-bad-rows
     (y-or-n-p "Highlight user mistakes? ")
     bell-colours)
    ("Underline user mistakes" ringing:underline-faults
     (y-or-n-p "Underline user mistakes? ")
     method-lines)
    ("Allow going off the end of the row, as an error" ringing:error-off-end-of-row
     (y-or-n-p "Allow (and mark as error) going off the end of the row? ")
     bell-numbers)
    ("Mark backstrokes distinctly from handstrokes" ringing:mark-backstroke
     (y-or-n-p "Mark backstrokes distinctly from handstrokes? ")
     bell-colours)
    ("Always use method library" ringing:always-use-method-library
     (y-or-n-p "Go straight to method library directory for more methods? ")
     ringing-methods)
    ("Always scan all libraries" ringing:always-scan-all-libraries
     (y-or-n-p "Always scan all libraries? ")
     ringing-methods)
    ("Method library directory" method-library:directory
     (read-directory-name "Method library directory: " method-library:directory)
     method-library)
    ("Show method names as library is scanned" method-library:full-progress-messages
     (y-or-n-p "Show method names as library is scanned? ")
     method-library)
    ("User's collection of extra methods" place-notation:default-methods-directory
     (if ringing:always-use-method-library
	 place-notation:default-methods-directory
       (read-directory-name "Personal default methods directory: " place-notation:default-methods-directory))
     place-notation)
    ("Method groups directory" method-groups:groups-directory
     (read-directory-name "Method groups directory: " method-groups:groups-directory)
     method-groups)
    ("Include touch book stage groups in group list" method-groups:include-touch-book-stage-groups
     (y-or-n-p "Include touch book stage groups in group list? ")
     method-groups)
    ("Show lines" ringing:show-lines
     (y-or-n-p "Draw \"lines\" (*, .) as well as numbers? ")
     method-lines)
    ("Character (in \"lines\") for treble" ringing:row-lines-treble 
     (ringing:read-char-from-user "Character (in \"lines\") for treble? " ringing:row-lines-treble)
     method-lines)
    ("Character (in \"lines\") for marked bell" ringing:row-lines-marked-bell 
     (ringing:read-char-from-user "Character (in \"lines\") for marked bell? " ringing:row-lines-marked-bell)
     method-lines)
    ("Character (in \"lines\") for other marked bell" ringing:row-lines-other-marked-bell 
     (ringing:read-char-from-user "Character (in \"lines\") for other marked bell? " ringing:row-lines-other-marked-bell)
     method-lines)
    ("Character (in \"lines\") for other bells" ringing:row-lines-other-bells 
     (ringing:read-char-from-user "Character (in \"lines\") for other bells? " ringing:row-lines-other-bells)
     method-lines)
    ("Character (in \"lines\") for start of row" ringing:row-lines-start-char 
     (ringing:read-char-from-user "Character (in \"lines\") for start of row? " ringing:row-lines-start-char)
     method-lines)
    ("Character (in \"lines\") for end of row" ringing:row-lines-end-char 
     (ringing:read-char-from-user "Character (in \"lines\") for end of row? " ringing:row-lines-end-char)
     method-lines)
    ("Multi-column display" method-display:main-display-function
     (if (y-or-n-p "Multi-column display? ")
	 'method-display:multi-column-display-function
       'method-display:single-column-display-function)
     method-display)
    ("Label columns and teaching leads" ringing:label-columns
     (y-or-n-p "Label columns and teaching leads? ")
     method-display)
    ("Page width" method-display:page-width
     (read-int-from-user "Page width: " method-display:page-width t)
     method-display)
    ("Form-feed between blocks" method-display:form-feed-between-blocks
     (if (eq method-display:main-display-function
	     'method-display:multi-column-display-function)
	 (y-or-n-p "Form feed between blocks of columns? ")
       nil)
     method-display)
    ("Draw lines between rows in general" ringing:show-lines-with-lines
     (and ringing:show-lines
	  (y-or-n-p "Draw lines between rows in general? "))
     method-lines)
    ("Draw lines between rows when teaching" ringing:lines-between-rows-when-teaching
     (or ringing:show-lines-with-lines
	 (y-or-n-p "Draw lines between rows just when teaching? "))
     teacher)
    ("Show numbers" ringing:show-numbers
     (or (not ringing:show-lines)
	 (y-or-n-p "Show numbers as well as lines? "))
     method-lines)
    ("Asterisk manual bell" method-display:asterisk-manual-bell
     (and ringing:show-numbers
	  (y-or-n-p "Show observation bell as asterisk instead of number? "))
     bell-colours)
    ("Show only course bells in numbers display" ringing:display-course-bells-only
     (y-or-n-p "Show only course bells in mumbers display? ")
     method-lines)
    ("Green line for other bell of pair generally" ringing:green-line-too
     (y-or-n-p "Show other bell in observation pair in green? ")
     beer)
    ("Green line for other bell of pair while teaching" ringing:green-line-too-while-teaching
     (and ringing:green-line-too
	  (y-or-n-p "While teaching, show other bell in observation pair in green? "))
     beer)
    ("Auto-ring inter-row delay" ringing:auto-ring-inter-row-delay
     (read-float-from-user "Delay between lines (or nil for fast): " ringing:auto-ring-inter-row-delay t)
     beer)
    ("Draw lead twice (at top and bottom)" method-display:draw-lead-twice
     (y-or-n-p "Draw lead end twice in mSiRiL display? ")
     method-display)
    ("Show final rounds again at end" method-display:show-final-rounds
     (y-or-n-p "Show final rounds again in mSiRiL display? ")
     method-display)
    ("List similar methods" mSiRiL-mode:list-similar-methods
     (y-or-n-p "List similar methods? ") mSiRiL-mode)
    ("Identify methods above and beloew the treble" mSiRiL-mode:identify-above-below
     (y-or-n-p "Identify methods above and below the treble? ")
     mSiRiL-mode)
    ("mSiRiL-mode:ring-method's starting point defaults to tenor"
     mSiRiL-mode:default-manual-start-at-tenor
     (y-or-n-p "mSiRiL-mode:ring-method's starting point defaults to tenor? ")
     mSiRiL-mode)
    ("mSiRiL-mode displays methods takes the whole frame"
     mSiRiL-mode:single-window-display
     (y-or-n-p "mSiRiL-mode to display methods using the whole frame? ")
     mSiRiL-mode)
    ("mSiRiL-mode ringing takes the whole frame"
     mSiRiL-mode:single-window-ringing
     (y-or-n-p "mSiRiL-mode methods ringing to use the whole frame? ")
     mSiRiL-mode)
    ("Use timing" ringing:use-timing
     (y-or-n-p "Use timing? ")
     bell-numbers)
    ("Timing speed" ringing:time-step
     (if ringing:use-timing
	 (/
	  (read-float-from-user "Time step (enter time ring a change, in seconds): "
				(* (/ ringing:time-max 2) ringing:time-step))
	  (/ ringing:time-max 2))
       ringing:time-step)
     bell-numbers)
    ("Time steps for boundary between down and places" ringing:time-early
     (if ringing:use-timing
	 (read-int-from-user "End of early time: " ringing:time-early t)
       ringing:time-early)
     bell-numbers)
    ("Time steps for boundary between places and up" ringing:time-late
     (if ringing:use-timing
	 (read-int-from-user "Start of late time: " ringing:time-late t)
       ringing:time-late)
     bell-numbers)
    ("Timeout on not pressing anything" ringing:timeout
     (if ringing:use-timing
	 ringing:timeout
       (read-float-from-user "Timeout: " ringing:timeout t))
     bell-numbers)
    ("Mini ropes" ermm:mini-ropes
     (y-or-n-p "Draw mini-ropes? ")
     ermm)
    ("Font for handstrokes in PostScript printing" ringing:postscript-handstroke-font
     (read-from-minibuffer "Font for handstrokes in PostScript printing: " ringing:postscript-handstroke-font)
     ps-config)
    ("Font for backstrokes in PostScript printing" ringing:postscript-backstroke-font
     (read-from-minibuffer "Font for backstrokes in PostScript printing: " ringing:postscript-backstroke-font)
     ps-config)
    ("Font size for PostScript printing" ringing:postscript-font-size
     (read-float-from-user "Font size for PostScript printing: " ringing:postscript-font-size)
     ps-config)
    ("Leading factor for PostScript printing" ringing:postscript-leading-factor
     (read-float-from-user "Leading factor for PostScript printing: " ringing:postscript-leading-factor)
     ps-config)
    ("Thickness of red lines in PostScript printing" ringing:postscript-red-line-width
     (read-float-from-user "Thickness of red lines in PostScript printing: " ringing:postscript-red-line-width)
     ps-config)
    ("Thickness of blue lines in PostScript printing" ringing:postscript-blue-line-width
     (read-float-from-user "Thickness of blue lines in PostScript printing: " ringing:postscript-blue-line-width)
     ps-config)
    ("Thickness of green lines in PostScript printing" ringing:postscript-green-line-width
     (read-float-from-user "Thickness of green lines in PostScript printing: " ringing:postscript-green-line-width)
     ps-config)
    ("Directory for PostScript temporary files for printing" mSiRiL-mode:ps-temp-files-directory
     (read-directory-name "PostScript temporary files directory: " mSiRiL-mode:ps-temp-files-directory)
     mSiRiL-mode)
    ("Command to send PostScript to printer" ringing:ps-print-command
     (read-from-minibuffer "Command to send PostScript to printer: " ringing:ps-print-command)
     mSiRiL-mode)
    ("Show the user each lead when teaching" teach-ringing:preview-leads
     (y-or-n-p "Show the user each lead when teaching? ")
     teacher)
    ("Never the same random bell consecutively" ringing:never-same-random-consecutive
     (y-or-n-p "Never present the same random bell twice running: ")
     teacher)
    ("Name each teaching lead buffer differently" ringing:lead-buffer-name-format
     (if (y-or-n-p "Name each teaching lead buffer differently: ")
	 "*Place %d of %s*"
       "*One lead*")
     teacher)
    ("Times user must get lead right in intensive learning (teaching phase)"
     teach-ringing:required-perfects-seen
     (read-int-from-user
      "Times user must get lead right in intensive learning (teaching phase): "
      teach-ringing:required-perfects-seen)
     teacher)
    ("Times user must get lead right in intensive learning (testing phase)"
     teach-ringing:required-perfects-unseen
     (read-int-from-user
      "Times user must get lead right in intensive learning (testing phase): "
      teach-ringing:required-perfects-unseen)
     teacher)
    ("Acceptable faults in intensive learning" teach-ringing:tolerable-faults
     (read-int-from-user
      "How many faults are acceptable in intensive learning: "
      teach-ringing:tolerable-faults)
     teacher)
    ("Test user on previous leads in intensive learning"
     teach-ringing:test-previous-leads
     (y-or-n-p "Test user on previous leads in intensive learning? ")
     teacher)
    ("Time to show shell buffer between quiz leads"
     ringing:quiz-between-leads-delay-time
     (read-float-from-user "Timeout to flash up shell buffer during quiz: "
			   ringing:quiz-between-leads-delay-time t)
     teacher)
    ("Syntax coloring in mSiRiL buffers" mSiRiL:do-own-coloring
     (y-or-n-p "Use syntax coloring in mSiRiL buffers? ")
     mSiRiL-mode)
    ("Do syntax coloring lazily in mSiRiL buffers" mSiRiL:color-lazily
     (y-or-n-p "Do syntax coloring lazily in mSiRiL buffers? ")
     mSiRiL-mode))
  "Control parameters to set, with the questions to ask
and the modules required to get the descriptions of the variables.")

(defun ringing:set-one-variable (varspec &optional offer-save)
  "Set a parameter for the emacs change-ringing system.
The result is stored in a file if so desired."
  (interactive
   (list (assoc (let ((completion-ignore-case t))
		  (completing-read "Set ringing parameter: "
				   ringing:set-variables
				   nil 'require-non-null-match))
		ringing:set-variables)
	 t))
  (let* ((varname (second varspec))
	 (question-expr (third varspec))
	 (var-module (fourth varspec)))
    (save-window-excursion
      (if var-module (require var-module))
      (describe-variable varname)
      (set varname (eval question-expr))))
  (when offer-save (ringing:save-variables)))

(defun ringing:set-variables ()
  "Set control parameters for the emacs change-ringing system.
The result is stored in a file if so desired."
  (interactive)

  (mapcar 'ringing:set-one-variable ringing:set-variables)
  
  ;; put ringing:always-use-method-library into effect
  (ringing:ensure-methods-loaded t)

 (ringing:save-variables)

  ;;

  (mSiRiL:mark-mode-regexp-from method-library:directory)

  (message "emacs-ringing configuration complete"))

(defun ringing:save-variables ()
  "Save the ringing system parameters."
  (interactive)
  (when (yes-or-no-p "Save ringing configuration to file? ")
    (save-window-excursion
      (find-file ringing:options-file)
      (erase-buffer)

      (insert (format ";;;; emacs-ringing configuration made at %s\n"
		      (current-time-string)))
      (insert "\n(setq\n")
      (insert (format "  ringing-version-last-saved-config \"%s\"\n" ringing-version))
      (mapcar (lambda (entry)
		(insert (format "  %s %S\n" (second entry)
				(let ((val (eval (second entry))))
				  (if (and (symbolp val) (not (memq val '(t nil))))
				      (list 'quote val)
				    val)
				  ))))
	      ringing:set-variables)
      (insert ")\n;; end of emacs-ringing configuration\n")
      (basic-save-buffer)
      (bury-buffer))))

;;;; bug reporting

(defconst ringing-maintainer-address "john@cb1.com")

(defun ringing:submit-bug-report ()
  "Submit via mail a bug report on the emacs-ringing system."
  (interactive)
  (require 'reporter)
  (let ((reporter-prompt-for-summary-p t))
    (reporter-submit-bug-report
     ringing-maintainer-address
     (concat "ringing.el " ringing-version)
     (mapcar 'second ringing:set-variables))))

;;;; tidying up, etc

(defvar ringing:buffers nil
  "All the buffers made by the ringing system, for ringing:tidy-up to use.")
  
(defun ringing:get-buffer-create (name)
  "Like get-buffer-create, but remembers it so it can be tidied up."
  (let ((buffer (get-buffer-create name)))
    (pushnew buffer ringing:buffers)
    buffer))

(defun ringing:hide-buffers ()
  "Put all the ringing buffers at the bottom of the buffer list,
and try to make sure that none of them are showing on screen."
  (interactive)
  (mapcar 'bury-buffer ringing:buffers)
  (let* ((non-ringing-buffers (remove-if '(lambda (b) (memq b ringing:buffers)) (buffer-list)))
	 (non-ringing-buffer (car non-ringing-buffers)))
    (walk-windows
     (function
      (lambda (window)
	(when (memq (window-buffer window) ringing:buffers)
	  (set-window-buffer window (if non-ringing-buffers
					(pop non-ringing-buffers)
				      non-ringing-buffer)))))
     nil t)))

(defun ringing:tidy-up ()
  "Remove ringing buffers."
  (interactive)
  (mapcar 'kill-buffer ringing:buffers)
  (when (and (boundp 'ringing:clash-message-buffer-name)
	     (or ;; (stringp ringing:clash-message-buffer-name)
		 (bufferp ringing:clash-message-buffer-name)))
    (kill-buffer ringing:clash-message-buffer-name))
  (when (and (boundp 'ringing:input-help-buffer-name)
	     (or ;; (stringp ringing:input-help-buffer-name)
		 (bufferp ringing:input-help-buffer-name)))
    (kill-buffer 'ringing:input-help-buffer-name))
  (when (get-buffer "*Ringing report*")
    (kill-buffer "*Ringing report*"))
  (setq ringing:buffers nil)
  (when (y-or-n-p "Forget method data? ")
      (place-notation:forget-methods))
  (message ""))

(defun ringing:get-file-window (file)
  "Return whether FILE is displayed in a window.
We use this to bury buffers that are not displayed, if visiting them
behind the scenes."
  (let ((buffer (find-buffer-visiting file)))
    (if (null buffer)
	nil
      (get-buffer-window buffer))))

;;;; display messages consistently

(defun ringing:message (message-format &rest args)
  "Usually like (message MESSAGE-FORMAT &rest ARGS) but may delay afterwards
or display elsewere etc."
  (apply 'message message-format args)
  (sit-for 2)
  )

;;;; now actually do the setup

(unless (file-exists-p ringing:options-file)
  (if (yes-or-no-p "Configure ringing system fully now? ")
      (ringing:set-variables)
    (ringing:message "Leaving default settings of everything")))

;; (unless (file-newer-than-file-p (expand-file-name "ringing.elc"
;; 						  ringing-el-dir)
;; 				(expand-file-name "ringing.el"
;; 						  ringing-el-dir))
;;   (when (yes-or-no-p "Compile ringing system? ")
;;     (ringing:compile-system)))
  
(if (file-exists-p ringing:options-file)
    (load-file ringing:options-file)
  (ringing:message "No options file -- taking default settings of everything"))

(mSiRiL:mark-mode-regexp-from method-library:directory)

(defvar ringing:load-hooks nil
  "Hooks to run when the ringing package is loaded.")

(add-hook 'mSiRiL-mode-load-hook
	  '(lambda ()
	     (require 'ringing-menu)))

(run-hooks 'ringing:load-hooks)

;; end of ringing.el
