;;;; rope-graphics.el -- part of John's elisp change-ringing system
;;; draw bell-ropes and sallies
;;; Time-stamp: <97/05/16 12:05:50 john>

(provide 'rope-graphics)

(defstruct rope
  position				; 0 is stood
  direction				; 1 is going up
  handstroke
  tail-end-length
  ringer-reach				; it goes this much further at backstroke
  nadir					; the bottom of the sally can go
					; down this far
  sally-length				; just for drawing it
  x)

(defvar rope-graphics:rope-array nil
  "The current array of ropes, or nil.")

(defvar rope-graphics:deepest-nadir nil
  "The deepest point to which sally bottoms dip.")

(defvar rope-graphics:height-above-standing nil
  "How much below the top of the screen the bottom of the sally at rest is.")

(defvar rope-graphics:leftmost nil
  "Where the leftmost rope is.")

(defvar rope-graphics:spacing nil
  "How far apart the ropes are.")

(defun rope-graphics:init (rope-list)
  "Initialize rope graphics with ROPE-LIST."
  (setq rope-graphics:deepest-nadir (apply 'min (mapcar 'second rope-list)))
  (let* ((n (length rope-list))
	 (i 0))
    (setq rope-graphics:spacing (/ (frame-width) n)
	  rope-graphics:leftmost (/ rope-graphics:spacing 2)
	  rope-graphics:rope-array (make-vector n nil))
    (let ((x rope-graphics:leftmost))
      (while rope-list
	(let ((rope-desc (car rope-list)))
	  (aset rope-graphics:rope-array
		i
		(make-rope :position 0
			   :direction 0
			   :handstroke t
			   :tail-end-length (first rope-desc)
			   :nadir (second rope-desc)
			   :sally-length (third rope-desc)
			   :x x)))
	(setq i (1+ i)
	      x (+ x rope-graphics:spacing)
	      rope-list (cdr rope-list))))
    (rope-graphics:draw-from-scratch)
    ))

(defun rope-graphics:draw-from-scratch ()
  "Re-draw the ropes, from scratch."
  (interactive)
  (switch-to-buffer (ringing:get-buffer-create "*Bell-ropes*"))
  (erase-buffer)
  (let* ((i 0)
	 (n (length rope-graphics:rope-array))
	 (spacer (make-string rope-graphics:leftmost ? ))
	 (h (- (window-height) 4)))
    (setq rope-graphics:height-above-standing (- h rope-graphics:deepest-nadir))
    (goto-char (point-min))
    (while (<= i h)
      (insert "\n")
      (setq i (1+ i)))
    (setq i 0)
    (while (< i n)
      (let ((rope (aref rope-graphics:rope-array i))
	    (j rope-graphics:height-above-standing)
	    )
	(goto-char (point-min))
	(while (> j 0)
	  (end-of-line 2)
	  (insert spacer)
	  (cond
	   ((and (<= j (+ (rope-position rope) (rope-sally-length rope)))
		 (> j (rope-position rope)))
	    (insert "#"))
	   (t
	    (insert "|")))
	  (setq j (1- j))))
      (setq spacer (make-string (1- rope-graphics:spacing) ? )
	    i (1+ i)))))

(defun rope-graphics:test ()
  "Test the rope graphics."
  (interactive)
  (let* ((rope-data '((7 2 5)
		      (8 2 5)
		      (9 2 5)
		      (11 3 5)
		      (13 3 5)
		      (15 3 5)))
	 (n (length rope-data))
	 (i 0))
    (rope-graphics:init rope-data)
    (while (< i n)
      (rope-graphics:pull-off i)
      (incf i))
    (setq i 0)
    (while (< i 100)
      (sit-for 1)
      (rope-graphics:step)
      (incf i))))

(defun rope-graphics:set-char(x y c)
  "Set the character of rope X, Y above the base, to C."
  (goto-line (- rope-graphics:height-above-standing y -1))
  (move-to-column (+ rope-graphics:leftmost (* x rope-graphics:spacing)))
  (delete-char 1)
  (insert c))

(defun rope-graphics:step ()
  "Step rope graphics once for each rope. Return whether any are still moving."
  (interactive)
  (let ((i 0)
	(n (length rope-graphics:rope-array))
	(any-moving nil))
    (while (< i n)
      (let* ((rope (aref rope-graphics:rope-array i))
	     (position (rope-position rope)))

	;; changes in rope position
	(cond

	 ;; going up
	 ((> (rope-direction rope) 0)
	  (rope-graphics:set-char i position ?|)
	  (rope-graphics:set-char i (+ position
				       (rope-sally-length rope)) ?#))

	 ;; going down
	 ((< (rope-direction rope) 0)
	  (rope-graphics:set-char i position ?#)
	  (rope-graphics:set-char i (+ position
				       (rope-sally-length rope)) ?|)))


	
	(setq position (incf (rope-position rope) (rope-direction rope)))

	;; changes in direction of rope movements
	(cond
	 ;; rope is going up, and has reached the top
	 ((and (> (rope-direction rope) 0)
	       (>= position (rope-tail-end-length rope)))
	  (setf (rope-direction rope) 0
		(rope-handstroke rope) nil))

	 ;; rope is going up, and has come to resting position
         ((and (not (rope-handstroke rope))
	       (> (rope-direction rope) 0)
	       (= position 0))
	  (setf (rope-direction rope) 0
		(rope-handstroke rope) t))

	 ;; rope is going down, and has reached the bottom
	 ((and (< (rope-direction rope) 0) ; going down
	       (<= position (rope-nadir rope)))
	  (setf (rope-direction rope) 1)))

	(if (not (zerop (rope-direction rope)))
	    (setq any-moving t)))
      (incf i))
    any-moving))

(defun rope-graphics:pull-off (n)
  "Set rope N moving (downwards, since you can't push ropes)."
  (setf (rope-direction (aref rope-graphics:rope-array n)) -1))

;;; end of rope-graphics.el
