;;;; stages.el -- part of John's elisp change-ringing system
;;; lists the possible numbers of bells
;;; Time-stamp: <98/08/03 16:48:03 john>

(eval-when (compile load eval)
  (provide 'stages)
  (require 'cl))

(defconst ringing:number-of-bells-names
  (vector
   "zerobellscantevennotoccur"
   "justonebellneveroccurs"
   "justtwobellsneveroccur"
   "Singles"
   "Minimus"
   "Doubles"
   "Minor"
   "Triples"
   "Major"
   "Caters"
   "Royal"
   "Cinques"
   "Maximus"
   "Sextuples"
   "On 14"
   "Septuples"
   "On 16"
   )
  "The names for numbers of bells: minimus is 4 and so on.")

(defvar ringing:stages-alist nil
  "Possible numbers of bells, by name.")

(setq ringing:stages-alist nil)
(let ((i 4))
  (while (<= i 16)
    (push (cons (aref ringing:number-of-bells-names i) i)
	  ringing:stages-alist)
    (push (cons (format "on %d" i) i) ringing:stages-alist)
    (push (cons (int-to-string i) i) ringing:stages-alist)
    (incf i)))
(when nil
  (push (cons "maximus" 12) ringing:stages-alist)
  (let ((i 12))
    (while (<= i 16)
      (push (cons (format "on %d" i) i) ringing:stages-alist)
      (incf i))))

(defun ringing:read-stage-from-user (prompt)
  "Get a stage (that is, a number of bells) from the user."
  (let* ((completion-ignore-case t)
	 (stage-name
	 (completing-read prompt
	  ringing:stages-alist
	  nil
	  t)))
    (cdr (assoc stage-name ringing:stages-alist))))

(defun ringing:stage-name (n)
  "Return the string naming things on N bells"
  (when (stringp n) (setq n (ringing:figure-out-number-of-bells n)))
  (if (and (> n 2) (<= n 12))
      (aref ringing:number-of-bells-names n)
    (format "on %d" n)))

(defun ringing:figure-out-number-of-bells (name)
  "Return number of bells for methods in NAME, according to convention.
If the name has a number above 3 in it, we use that, otherwise look for
stage names. This lets the F1 hybrid method through, and London1 to London3,
while also allowing minimus upwards."
  (cond
   ((integerp name) name)
   ((null name) 0)
   (t
    (let ((namepart (file-name-nondirectory name)))
      (cond
       ((string-match " ?\\([a-z]+\\)$" namepart)
	(cdr (assoc (substring namepart (match-beginning 1)
			       (match-end 1))
		    ringing:stages-alist)))
       ((or
	 (string-match "\\(1[0-9]+\\)" namepart)
	 (string-match "\\([04-9]+\\)" namepart))
	(string-to-int (substring namepart
				  (match-beginning 1)
				  (match-end 1))))
       (t nil))))))

(defun ringing:method-at-stage-p (method stage)
  "Return whether METHOD is at STAGE.
METHOD may be a method structure, or a name, or a cons of them."
  (when (consp method) (setq method (car method)))
  (if (stringp method)
      (string-match (concat (ringing:stage-name stage) "$") method)
    (= (change-ringing-method-nbells method)
       (ringing:figure-out-number-of-bells stage))))

(defconst ringing:stage-names-regexp
  (concat "\\("
	  (mapconcat 'identity
		     (subseq ringing:number-of-bells-names 3)
		     "\\)\\|\\(")
	  "\\)")
  "Regular expression matching a stage name.")

(defun ringing:n-bells-from-row-string (string)
  "Return how many bells are being used in STRING."
  (string-match "[0-9a-zA-Z*.]+" string)
  (- (match-end 0) (match-beginning 0)))

;;; end of stages.el
