;;;; test.el -- part of John's elisp change-ringing system
;;; test routines for emacs ringing package
;;; Time-stamp: <99/05/11 17:17:40 johnstu>

(eval-when (compile load eval)
  (require 'method-library)
  (require 'ringing-methods)
  (require 'place-notation))

(defun foobells (method-name)
  "Test bell stuff"
  (interactive
   (progn
     (if (y-or-n-p "Flush loaded methods? ") (place-notation:forget-methods))
     (list (method-library:choose-method-interactively
	    "Foo method: " t))))
  (with-output-to-temp-buffer
      (format "*method trace and data for %s*" method-name)
    (let* ((place-notation:debugging t)
	   (method (ringing:find-named-method method-name)))
      (princ method-name)
      (princ ":\n")
      (princ method)
      (place-notation:show-place-notation
       (format "Test of reading %s" method-name)
       (change-ringing-method-plain method)))))

(defun ringing:test-faces ()
  "Display ringing:faces, which see."
  (interactive)
  (switch-to-buffer-other-window (ringing:get-buffer-create "*ringing faces test*"))
  (erase-buffer)
  (let ((outer ringing:faces))
    (dotimes (outeri 2)
      (let ((mid (aref outer outeri)))
	(dotimes (midi 2)
	  (let ((inner (aref mid midi)))
	    (dotimes (inneri 4)
	      (let* ((this-face (aref inner inneri))
		     (test-string (format "ringing:face-%s-%s-%s\n"
					  (if (= outeri 1) "bad" "ok")
						   (if (= midi 1) "backstroke" "handstroke")
						   (aref ringing:face-colour-names inneri))))
		(put-text-property 0 (length test-string) 'face this-face test-string)
		(insert test-string)))))))))

(defun ringing:produce-configuration-documentation ()
  "Product documentation for all the variables pointed to by ringing:set-variables, which see."
  (interactive)
  (switch-to-buffer (ringing:get-buffer-create "*documentation scratch*"))
  (insert "<dl>\n")
  (mapcar (function
	   (lambda (var)
	     (insert (format "  <dt> %s (<code>%s</code>)\n  <dd>\n" (first var) (second var)))
	     (insert (describe-variable (second var)) "\n")))
	  ringing:set-variables)
  (insert "</dl>\n")

  (goto-char (point-min))
  (delete-matching-lines "^Documentation:$")
  (goto-char (point-min))
  (delete-matching-lines "'s value is ")
  (goto-char (point-min))
  (while (re-search-forward "^\\*" (point-max) t)
    (replace-match "")))

(defun ringing:test-touch-reader ()
  "Test the touch reader."
  (interactive)
  (with-output-to-temp-buffer "*touch reader result*"
    (save-window-excursion
      (find-file "touch-test")
      (let ((result (ringing:read-touch-from-region (point-min) (point-max))))
	;; (prin1 result)
	(prin1 (coerce result 'string))
	(prin1 "\n")
	))
    (save-window-excursion
      (find-file "touches/98diaryP25-1st-onepart")
      (let* ((result (ringing:read-touch-from-region (point-min) (point-max)))
	     (ringing:auto-ring-inter-row-delay nil)
	     (call-string (coerce result 'string))
	     (ringing:touch-buffer-header-format
	      (concat ringing:touch-buffer-header-format
		      "\nComposition: "
		      call-string
		      "\n"))
	    (pbm (ringing:find-named-method "Plain Bob Minor")))
	(prin1 call-string)
	(prin1 "\n")
	(prin1 (ringing:convert-to-touch call-string pbm))
	(prin1 "\n")
	(ringing:show-touch pbm result
			    (function (lambda (buff)
					(message "annotating %s" (buffer-name buff))
					(switch-to-buffer buff)
					;; (goto-char (point-min))
					;; (insert "Composition: ") (insert call-string) (insert "\n")
					))
			    )))))

(defun ringing:define-test-touches ()
  "Define some test touches."
  (interactive)
  (deftouch norwich-and-plain-bob
    (go "Norwich Surprise Minor")
    (method "Plain Bob Minor"))

  (deftouch cambridge-and-plain-bob
    (go "Cambridge Surprise Minor")
    (method "Plain Bob Minor"))

  (deftouch cambridge-and-surfleet-and-beverley
    (go "Cambridge Surprise Minor")
    "Surfleet Surprise Minor"
    "Beverley Surprise Minor")

  (deftouch cambridge-and-before-surfleet-and-beverley
    (go "Cambridge Surprise Minor")
    (wrong)
    (method before "Surfleet Surprise Minor")
    (wrong)
    "Beverley Surprise Minor"
    (wrong))

  (deftouch cambridge-with-bobs
    (go "Cambridge Surprise Minor")
    (wrong)
    (wrong))
)

(defun ringing:test-raw-touch ()
  (ringing:show-touch (ringing:find-named-method "Cambridge Surprise Minor") 
		    `(touch
                      (plain "")
		      (method "" ,(ringing:find-named-method "Beverley Surprise Minor"))
                      (plain "")
		      (method "" ,(ringing:find-named-method "Surfleet Surprise Minor"))
		      )
		    ))

(defun ringing:test-composer ()
  "Test the routines in composition-search.el"
  (interactive)
  (let* ((one-method t)
	 (method "Plain Bob Minor")
	 (completions (ringing:true-completions-for-touch
		       (vector nil (ringing:find-named-method method))
		       (if one-method (method-groups:singleton-group method)
		       (method-groups:find-named-group "my-surprise-really-rung-so-far"))
		       10)))
    (with-output-to-temp-buffer "*compositions*"
      (dolist (completion completions)
	(princ (format "%s\n" completion))))))

(defun ringing:test-name-handling-cycle ()
  "Check that we can go from expanded names back to names used in files."
  (interactive)
  (let* ((files (mapcar 'cdr (method-library:files-for-n-bells 6)))
	 (method-lists (mapcar 'microSiRiL:parse-file-superficially files))
	 (methods (apply 'concatenate 'list method-lists)))
    (with-output-to-temp-buffer "*Name conversion cycle checking*"
      (dolist (method methods)
	(let* ((method-full-name (car method))
	       (method-apparent-file (method-library:proper-file-for-method method-full-name))
	       (method-short-name (microSiRiL:name-used-in-file method-full-name)))
	  (when nil
	    (princ (format "%s (%s:%s)\n" method-full-name method-apparent-file method-short-name)))
	  (princ (format "%s %s\n" method-full-name
			 (save-window-excursion
			   (find-file method-apparent-file)
			   (goto-char (point-min))
			   (if (re-search-forward (concat "^" method-short-name " ") (point-max) t)
			       "ok"
			     (format "failed: %s-->%s:%s" method-full-name method-apparent-file method-short-name))))))))))

(defun ringing:load-source-versions ()
  "Load all the .el files (overloading any .elc files loaded)."
  (interactive)
  (mapcar 'load-file (directory-files ringing-el-dir t "\\.el$")))

(defun ringing:test-finding-methods-individually (file)
  "Try finding each method straight from FILE."
  (interactive "fCheck finding methods in file: ")
  (let* ((methods (microSiRiL:parse-file-superficially file))
	 (microSiRiL:parse-whole-files nil))
    (dolist (method methods)
      (unless (string-match "zzz" (car method))
	(let ((refound (microSiRiL:method-from-file file (car method))))
	  (if refound
	      (message "%s --> %s" method refound)
	    (message "%s not refound" method)
	    ))))))

(defun ringing:test-splitter ()
  "Temporary function."
  (interactive)
  (with-output-to-temp-buffer "*Split test*"
    (prin1
     (ringing:split-around-treble "Cambridge Surprise Minor"))))

(defun ringing:test-lead-count-for-file (file)
  "Check the lead count for all methods in FILE."
  (interactive "fCheck methods in file: ")
  (let ((methods (microSiRiL:parse-file-superficially file)))
    (with-output-to-temp-buffer (format "*Lead counts for %s*" file)
      (dolist (method methods)
	(let* ((method-name (first method))
	       (method-data (ringing:find-named-method method-name))
	       (method-string (change-ringing-method-original-description-string method-data))
	       (nwbells (1- (change-ringing-method-nbells method-data)))
	       (leads (ringing:count-plain-course-leads method-name)))
	  (message method-name)
	  (when (or nil (/= nwbells leads))
	  (princ (format "%s (%s): %d%s\n"
			 method-name
			 method-string
			 leads
			 (if (= nwbells leads) ""
			   " <--- unexpected number of leads")))))))))

(defun ringing:test-lead-end-orders-in-file (file)
  "Test the regularity tester"
  (interactive "fTest lead end orders in file: ")
  (with-output-to-temp-buffer "*Regularity*"
    (let* ((methods (microSiRiL:parse-file-superficially file)))
      (dolist (methodlist methods)
	(let* ((method-data (ringing:find-named-method (car methodlist)))
	       (regularity (ringing:method-regular-p method-data)))
	  (princ (format "%s %s %s"
			 (change-ringing-method-title method-data)
			 (change-ringing-method-original-description-string method-data)
			 (if (numberp regularity)
			     (format "+%d\n" regularity)
			   "irregular\n"))))))))

(defun ringing:test-recreate-place-notation-for-method (method &optional anyway)
  "Test whether we get back to the original place notation when we re-create place notation."
  (interactive (list (ringing:read-method-name "Test place notation recreation for method: ")))
  (let* ((method-data (ringing:find-named-method method))
	 (original-definition-string (change-ringing-method-original-description-string method-data))
	 (original-places-string (substring original-definition-string
					    (1+ (string-match " " original-definition-string))))
	 (recreated-string (place-notation:recreate-string-from-changes (change-ringing-method-plain method-data) t t))
	 (same (string= original-places-string recreated-string))
	 (name (ringing:method-name method-data)))
    (when (or anyway (not same)) (message "%s: %s: Old: %s; New: %s" (if same "Pass" "Fail") name original-places-string recreated-string))
    (list same name original-places-string recreated-string)))

(defvar ringing:test-recreate-place-notation-skip-methods "\\(Zzz\\)\\|\\(darrowby\\)"
  "Methods not to test.")

(defun ringing:test-recreate-place-notation-for-file (file)
  "Test method string re-creation for all methods in FILE.
If FILE is a directory, do so for all files therein."
  (interactive "fCheck place notation re-creation for file or directory: ")
  (if (file-directory-p file)
      (let* ((all-results (apply 'concatenate 'list
				 (mapcar 'ringing:test-recreate-place-notation-for-file
					 (directory-files file t "[adhopst]p?[0-9]+$")))))
	(with-output-to-temp-buffer "*Methods for which recreated place notation was wrong*"
	  (dolist (method all-results)
	    (unless (first method)
	      (princ (format "%s: %s   !=   %s\n" (second method) (third method) (fourth method)))))))
    (let ((results nil))
      (dolist (method-pair (microSiRiL:parse-file-superficially file))
	(let* ((method-name (car method-pair))
	       (comparison-data (ringing:test-recreate-place-notation-for-method method-name t)))
	  (unless (string-match ringing:test-recreate-place-notation-skip-methods method-name)
	    (push comparison-data results))))
      (let* ((maxname (apply 'max (mapcar (function (lambda (d) (length (second d)))) results)))
	     (maxold (apply 'max (mapcar (function (lambda (d) (length (third d)))) results)))
	     (maxnew (apply 'max (mapcar (function (lambda (d) (length (fourth d)))) results)))
	     (resformat (format "%%s: %%-%ds %%%ds %%2s %%%ds\n" maxname maxold maxnew))
	     (red (ringing:get-face nil nil ringing:red))
	     (green (ringing:get-face nil nil ringing:green))
	     (output-buffer (ringing:get-buffer-create (format "*test of recreating place notation for %s %s*"
							       (capitalize (ringing:kind-by-name (ringing:figure-out-kind (file-name-nondirectory file))))
							       (ringing:stage-name (ringing:figure-out-number-of-bells (file-name-nondirectory file)))))))
	(switch-to-buffer output-buffer)
	(erase-buffer)
	(dolist (comparison results)
	  (let ((text (format resformat
			      (if (first comparison) "pass" "fail")
			      (second comparison)
			      (third comparison)
			      (if (first comparison) "==" "!=")
			      (fourth comparison))))
	    (when t
	      (put-text-property 0 (length text) 'face 
				 (if (first comparison) green red)
				 text))
	    (insert text)))
	(goto-char (point-min)))
      results)))

(defun ringing:test-lengths-in-file (file)
  "Test reading of methods in FILE by checking length of leads."
  (interactive "fCheck method lengths in file: ")
  (with-output-to-temp-buffer (format "*Method lengths for %s*" file)
    (dolist (method-pair (microSiRiL:parse-file-superficially file))
      (let* ((method-name (car method-pair))
	     (method-data (ringing:find-named-method method-name)))
	(princ (format "%s : plain %d; work %d\n"
		       method-name
		       (length (change-ringing-method-plain method-data))
		       (length (change-ringing-method-work method-data))))))))

;;; end of test.el
