;;;; touch-book.el -- part of John's elisp change-ringing system
;;; record touches rung successfully and unsuccessfully
;;; Time-stamp: <99/04/09 13:59:37 johnstu>

(eval-when (compile load eval)
  (provide 'touch-book)
  (require 'method-groups)
  (require 'method-library)
  (require 'cl)
  (require 'teacher))

;; This feature keeps track of what methods you have attempted to ring
;; interactively, and how successfully. It maintains a file containing
;; a lisp alist of (method-name . acheivement-data) where
;; the acheivement data is a list of vectors with one more entry than the
;; number of bells; entry zero is used for plain courses.
;; See documentation for touch-book:get-record for what is in the vectors.

(defvar touch-book:touches nil
  "Record of what the user has rung so far.")

(defvar touch-book:touch-book-file "~/.touch-book.el"
  "Where the record of what the user has rung is kept.")

(defun touch-book:load ()
  "Load the touch book from file, if needed; return the touch list."
  (when (null touch-book:touches)
    (when (file-exists-p touch-book:touch-book-file)
      (load-file touch-book:touch-book-file)))
  ;; convert from old format if necessary
  (when (and (car touch-book:touches)
	     (not (consp (cddar touch-book:touches))))
    (dolist (method touch-book:touches)
      (rplacd method (list (cadr method)
			   (cddr method)
			   (make-vector (length (cadr method)) 0)))))
  touch-book:touches)

(defun touch-book:get-record (method)
  "Get the user's record for METHOD.
The record is a list of (attempts successes faults nrows) where each of the
first three is an array by place bell, with element 0 being for whole plain
courses, and the fourth is how many rows there are per lead of that method.
There may optionally be another array, containing strings giving detail of
every blow; this is as produced by user-mistakes:lead-to-string."
  (touch-book:load)
  (let* ((method-name (ringing:method-name method))
	 (pair (assoc method-name touch-book:touches)))
    (if pair
	(cdr pair)
      (let* ((method-data (ringing:find-named-method method))
	     (n (1+ (change-ringing-method-nbells
		     method-data)))
	     (newlist (list (make-vector n 0)
			    (make-vector n 0)
			    (make-vector n 0)
			    (length (change-ringing-method-plain method-data))
			    (list (cons 'started (current-time-string))))))
	(push (cons method-name newlist) touch-book:touches)
	newlist))))

(defun touch-book:save ()
  "Save the touch book file."
  (save-window-excursion
    (find-file touch-book:touch-book-file)
    (erase-buffer)
    (insert (format ";; Touch book for %s\n;; saved at %s\n"
		    (user-login-name)
		    (current-time-string)))
    (insert "(setq touch-book:touches '(\n")
    (mapcar (function
	     (lambda (method)
	       (insert (format "  %S\n" method))))
	    touch-book:touches)
    (insert "  )\n)\n")
    (basic-save-buffer)
    (bury-buffer)))

(defvar touch-book:methods-rung-this-session nil
  "Methods attempted this session.
May be accessed via (touch-book:methods-rung-this-session-as-group)
or (method-groups:group-methods \"*Methods rung this session*\"...)")

(defvar touch-book:blow-by-blow t
  "*Whether to record full detail in the touch book.")

(defun touch-book:record-clash-level (method &optional lead clashes)
  "Record the user's preformance on METHOD for LEAD as being CLASHES."
  (unless (numberp lead) (setq lead 0))
  (let* ((method-name (ringing:method-name method))
	 (record (touch-book:get-record method-name)))
    (pushnew method-name touch-book:methods-rung-this-session :test 'string=)
    (incf (aref (first record) lead))
    (if (and (numberp clashes) (not (zerop clashes)))
	(incf (aref (third record) lead) clashes)
      (incf (aref (second record) lead)))
    (when touch-book:blow-by-blow

      ;; will call user-mistakes:lead-to-string here


      )
    (touch-book:save)))

(defun touch-book:list-my-methods (&optional detail)
  "List the methods the user has rung; with optional DETAIL for each lead."
  (interactive "P")
  (with-output-to-temp-buffer "*Touch book*"
    (let* ((max-name-width (apply
			    'max
			    (mapcar 'length (mapcar 'car touch-book:touches))))
	   (name-format (format "%%%ds" max-name-width))
	   (padstr (make-string max-name-width ? ))
	   )
      (mapcar (function
	       (lambda (method)
		 (let* ((method-name (first method))
			(attempts (second method))
			(lead-attempts (reduce '+ attempts :start 1))
			(successes (third method))
			(lead-successes (reduce '+ successes :start 1))
			(faults (fourth method))
			(lead-faults (reduce '+ faults :start 1))
			(lead-length (fifth method))
			)
		   (princ (format " %s: %3d plain courses (%d perfect)"
				  (format name-format method-name)
				  (aref attempts 0)
				  (aref successes 0)))
		   (when (> lead-attempts 0)
		     (princ (format " ; % 3d perfect leads out of %d" lead-successes lead-attempts))
		     (when detail
		       (let ((n (length attempts))
			     (i 2))
			 (while (< i n)
			   (let ((this-attempts (aref attempts i))
				 (this-successes (aref successes i))
				 (this-faults (aref faults i)))
			     (when (> this-attempts 0)
			       (princ (format "\n%s%c: %3d/% -3d%s (%d faults -- %f per blow)"
					      padstr
					      (ringing:bell-number-to-char i)
					      this-successes
					      this-attempts
					      (if (zerop this-attempts)
						  ""
						(format " %3d%%" (/ (* 100 this-successes) this-attempts)))
					      this-faults
					      (if (zerop this-faults)
						  0.0
						(/ (float this-faults)
						   (float lead-length)))))))
			   (incf i)))))
		   (princ "\n"))))
	      touch-book:touches))))

(defun touch-book:list-all-leads (&optional percentage)
  "Return a list of all the leads mentioned in the touch book.
Each element is a list of method, lead number, attempts, successes, quality.
With optional argument PERCENTAGE, go by percentage of perfect leads;
otherwise, by average faults per blow."
  (let ((methods (touch-book:load))
	(leads nil))
    (dolist (method methods)
      (let* ((lead-length
	      (if (fifth method)
		  (fifth method)
		(length (change-ringing-method-plain (ringing:find-named-method (first method))))))
	     (attempts (second method))
	     (successes (third method))
	     (faults (fourth method))
	     (n (length attempts))
	     (i 2))
	(while (< i n)
	  (let ((tries (aref attempts i))
		(OKs (aref successes i))
		(clashes (aref faults i))
		)
	    (push (list (car method)
			i
			tries
			OKs
			(if (zerop tries)
			    0
			  (if percentage
			      (/ (* 100 OKs) tries)
			    (- 1.0
			       (/ (float clashes)
				  (float (* tries lead-length)))))))
		  leads))
	  (incf i))))
    leads))

(defconst touch-book:worst-leads-buffer-name "*Leads in improving order*"
  "The name of the buffer containing the list of leads in improving order.
We want to keep this around so we can see whether to update the list.")

(defvar touch-book:leads-keymap
  (make-keymap "Leads")
  "Keympa for leads-list-mode")

(suppress-keymap touch-book:leads-keymap t)

(define-key touch-book:leads-keymap "g" 'touch-book:refill-leads-buffer)
(define-key touch-book:leads-keymap "%" 'touch-book:toggle-percentage-display)
(define-key touch-book:leads-keymap "\r" 'touch-book:leads-buffer-teach-this)
(define-key touch-book:leads-keymap "?" 'touch-book:leads-buffer-show-this)
(define-key touch-book:leads-keymap " " 'touch-book:leads-buffer-ring-this)
(define-key touch-book:leads-keymap "v" 'touch-book:revise-my-methods)
(define-key touch-book:leads-keymap "w" 'touch-book:teach-worst-lead)
(define-key touch-book:leads-keymap "\M-n" 'touch-book:next-lead-of-method)
(define-key touch-book:leads-keymap "\M-p" 'touch-book:previous-lead-of-method)
(define-key touch-book:leads-keymap "n" 'touch-book:next-place-bell-of-method)
(define-key touch-book:leads-keymap "p" 'touch-book:previous-place-bell-of-method)
(define-key touch-book:leads-keymap "<" 'touch-book:worst-lead-of-method)
(define-key touch-book:leads-keymap ">" 'touch-book:best-lead-of-method)
(define-key touch-book:leads-keymap "m" 'touch-book:mark-method)
(define-key touch-book:leads-keymap "\M-." 'touch-book:find-mSiRiL-for-method) 

(put 'leads-list-mode 'mode-class 'special)

(defun leads-list-mode ()
  "Major mode for the touch book lead list buffer.\\<touch-book:leads-keymap>

This buffer lists how well you are doing at each lead of each method that
you've been ringing with the simulator. You can move between the leads of
the same method either in order of how well you're doing (\\[touch-book:next-lead-of-method], \\[touch-book:previous-lead-of-method], with
\\[touch-book:worst-lead-of-method] and \\[touch-book:best-lead-of-method] taking you to worst and best leads) or in order of place bells (\\[touch-book:next-place-bell-of-method], \\[touch-book:previous-place-bell-of-method]).

On any line of this buffer, you can do:
  \\[touch-book:leads-buffer-show-this] to display that lead
  \\[touch-book:leads-buffer-ring-this] to ring that lead
  \\[touch-book:leads-buffer-teach-this] to learn that lead intensively

Having moved up or down leads of a method, or displayed or rung it from this
mode, that method becomes your current method, and its leads will be
highlighted.

You normally get into this mode from mSiRiL-mode by the command \\<mSiRiL-keymap> \\[touch-book:worst-leads-first],
which displays the lead table with the method you were on in the
mSiRiL file as the current method.

Commands available are:
\\{touch-book:leads-keymap}
"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'leads-list-mode
	mode-name "leads")
  (pushnew (list 'touch-book:by-percentage " by perfect leads") minor-mode-alist :test 'equal)
  (use-local-map touch-book:leads-keymap))

(defvar touch-book:lead-reference-format "%s(%c)"
  "The format of a reference to a lead of a method.")

(defvar touch-book:leads-list-line-format-percentage "%3d%%: %s"
  "Format of each line in the leads list book")

(defvar touch-book:leads-list-line-format-faults-per-lead "%f: %s"
  "Format of each line in the leads list book")

(defvar touch-book:by-percentage nil
  "Whether the leads list buffer is done by percentage.")

(defvar touch-book:number-lines t
  "*Whether to number the lines in the touch book")

(defun touch-book:refill-leads-buffer (&optional percentage)
  "Refill the sorted leads buffer.
With optional argument PERCENTAGE, go by percentage of perfect leads;
otherwise, by average faults per blow."
  (interactive "P")
  (message "Preparing leads buffer...")
  (erase-buffer)
  (setq touch-book:by-percentage percentage)
  (let* ((leads (touch-book:list-worst-leads-first percentage))
	 (number-format (if touch-book:number-lines
			    (format "%% %dd: " (1+ (ceiling (log (length leads) 10))))
			  nil))
	 (i 0))
    ;; (message "format is %s" number-format)
    (dolist (this leads)
      (incf i)
      (when number-format (insert (format number-format i)))
      (insert (format (if percentage
			  touch-book:leads-list-line-format-percentage
			touch-book:leads-list-line-format-faults-per-lead)
		      (fifth this)
		      (format touch-book:lead-reference-format
			      (first this)
			      (ringing:bell-number-to-char (second this))))
	      "\n")))
  (touch-book:color-leads-buffer)
  (message "Preparing leads buffer... done"))

(defun touch-book:color-leads-buffer ()
  "Color the leads buffer."
  (when (fboundp 'copy-face)
    (let ((red (ringing:get-face nil nil ringing:red))
	  (green (ringing:get-face nil nil ringing:green))
	  (blue (ringing:get-face nil nil ringing:blue))
	  (redmarked (ringing:get-face t nil ringing:red))
	  (greenmarked (ringing:get-face t nil ringing:green))
	  (bluemarked (ringing:get-face t nil ringing:blue)))
      (save-excursion
	(goto-char (point-min))
	(while (re-search-forward "^\\(.+\\): \\(.+\\) \\([^ ]+ \\)\\([^ ]+\\)(\\(.\\))$" (point-max) t)
	  (let* ((line-match-data (match-data))
		 (marked (and t touch-book:marked-method
			     (string-match touch-book:marked-method (buffer-substring (match-beginning 2) (match-end 4)))))
		 )
	    (store-match-data line-match-data)
	    (put-text-property (match-beginning 2) (match-beginning 3) 'face (if marked bluemarked blue))
	    (put-text-property (match-beginning 3) (match-beginning 4) 'face (if marked greenmarked green))
	    (put-text-property (match-beginning 4) (match-end 4) 'face (if marked bluemarked blue))
	    (put-text-property (match-beginning 5) (match-end 5) 'face (if marked redmarked red)))
	  t)))))

(defvar touch-book:marked-method nil
  "The name of the currently marked method.")

(defun touch-book:mark-method (where)
  "Mark the method at WHERE (point, interactively)."
  (interactive "d")
  (let ((method
	 (cond
	  ((or (numberp where)
	       (markerp where)) (car (touch-book:lead-on-line where)))
	  ((stringp where) where)
	  ((change-ringing-method-p where) (ringing:method-name where)))))
    (unless (equal method touch-book:marked-method)
      (setq touch-book:marked-method method)
      (remove-text-properties 0 (length touch-book:marked-method) '(face nil) touch-book:marked-method)
      (touch-book:color-leads-buffer))))

(defun touch-book:toggle-percentage-display ()
  "Toggle whether the leads-list-mode buffer is sorted by percentage perfect leads."
  (interactive)
  (let ((lead (touch-book:lead-on-line (point))))
    (touch-book:refill-leads-buffer (not touch-book:by-percentage))
    (goto-char (point-min))
    (search-forward (format touch-book:lead-reference-format (car lead)
			    (ringing:bell-number-to-char (cdr lead)))
		    (point-max) t)))

(defun touch-book:lead-on-line (where)
  "Return a cons of the method name and lead number mentioned at WHERE,
in a leads-list-mode buffer."
  (save-excursion
    (goto-char where)
    (beginning-of-line 1)
    (if (looking-at "^ *[0-9]+:? *[.0-9]+%?: \\([A-Za-z ]+\\)(\\([0-9eta-d]\\))$")
	(let ((name (buffer-matched 1))
	      (number-string (buffer-matched 2)))
	  (remove-text-properties 0 (length name) '(face nil) name)
	  (remove-text-properties 0 (length number-string) '(face nil) number-string)
	  (cons name 
		(ringing:bell-char-to-number (aref number-string 0))))
      (error "No lead on this line"))))

(defun touch-book:next-lead-of-method (where)
  "Go to next lead of the same method in leads-list-mode."
  (interactive "d")
  (let* ((which (car (touch-book:lead-on-line where)))
	 (next (save-excursion
		 (end-of-line 1)
		 (search-forward which (point-max) t))))
    (if next
	(progn
	  (goto-char next)
	  (touch-book:mark-method next))
      (error "No more leads of %s" which))))

(defun touch-book:previous-lead-of-method (where)
  "Go to previous lead of the same method in leads-list-mode."
  (interactive "d")
  (let* ((which (car (touch-book:lead-on-line where)))
	 (next (save-excursion
		 (beginning-of-line 1)
		 (search-backward which (point-min) t))))
    (if next
	(progn
	  (goto-char next)
	  (touch-book:mark-method next))
      (error "No more leads of %s" which))))

(defun touch-book:worst-lead-of-method (where)
  "Go to worst lead of the same method in leads-list-mode."
  (interactive "d")
  (let* ((which (car (touch-book:lead-on-line where)))
	 (next (save-excursion
		 (goto-char (point-min))
		 (search-forward which (point-max) t))))
    (if next
	(progn
	  (goto-char next)
	  (touch-book:mark-method next))
      (error "Could not find a lead of %s" which))))

(defun touch-book:best-lead-of-method (where)
  "Go to best lead of the same method in leads-list-mode."
  (interactive "d")
  (let* ((which (car (touch-book:lead-on-line where)))
	 (next (save-excursion
		 (goto-char (point-max))
		 (search-backward which (point-min) t))))
    (if next
	(progn
	  (goto-char next)
	  (touch-book:mark-method next))
      (error "Could not find a lead of %s" which))))

(defun touch-book:next-place-bell-of-method (where)
  "Go to next place bell of the same method in leads-list-mode."
  (interactive "d")
  (let* ((which (touch-book:lead-on-line where))
	 (which-method (car which))
	 (which-bell (1+ (cdr which)))
	 (search-pattern (format touch-book:lead-reference-format which-method
				 (ringing:bell-number-to-char which-bell)))
	 (whither (save-excursion
		    (goto-char (point-min))
		    (search-forward search-pattern (point-max) t))))
    (if (> which-bell
	   (ringing:figure-out-number-of-bells which-method))
	(error "%s does not have a %dth place bell" which-method which-bell)
      (if whither
	  (progn
	    (goto-char whither)
	    (touch-book:mark-method whither))
	(error "%s not found" search-pattern)))))

(defun touch-book:previous-place-bell-of-method (where)
  "Go to previous place bell of the same method in leads-list-mode."
  (interactive "d")
  (let* ((which (touch-book:lead-on-line where))
	 (which-method (car which))
	 (which-bell (1- (cdr which)))
	 (search-pattern (format touch-book:lead-reference-format which-method
				 (ringing:bell-number-to-char which-bell)))
	 (whither (save-excursion
		    (goto-char (point-min))
		    (search-forward search-pattern (point-max) t))))
    (if whither
	(progn
	  (goto-char whither)
	  (touch-book:mark-method whither))
      (if (> which-bell 1)
	  (error "%s not found" search-pattern)
	(error "That would be the treble")))))

(defun touch-book:leads-buffer-teach-this (where)
  "Try (teach) the lead that the cursor is on, from a leads-list-mode buffer.
Then, update the leads list buffer to show how well the user is now doing
on this lead, and move point to the new position of the same lead.

Teaching is done using ringing:teach-lead-until-perfect, which see."
  (interactive "d")
  (let* ((which-lead (touch-book:lead-on-line where))
	 (method-name (car which-lead))
	 (lead-number (cdr which-lead))
	 (the-name (format "%s(%c)" method-name (ringing:bell-number-to-char lead-number))))
    (save-window-excursion
      (ringing:teach-lead-until-perfect
       method-name
       lead-number))
    (touch-book:refill-leads-buffer)
    (goto-char (point-min))
    (when (search-forward the-name (point-max) t)
      (touch-book:mark-method (point)))))

(defun touch-book:switch-to-buffer-other-window-beside (buffer)
  "Like switch-to-buffer-other-window, but splits using
split-window-horizontally if necessary."
  (interactive "BSwitch to buffer in other window: ")
  (if (one-window-p)
      (split-window-horizontally))
  (let ((pop-up-windows t))
    (pop-to-buffer buffer t)))

(defun touch-book:leads-buffer-show-this (where)
  "Show the lead that the cursor is on, from a leads-list-mode buffer."
  (interactive "d")
  (let* ((which-lead (touch-book:lead-on-line where))
	 (method-name (car which-lead))
	 (lead-number (cdr which-lead)))
    (teach-ringing:do-one-lead method-name lead-number nil nil 'touch-book:switch-to-buffer-other-window-beside)
    (other-window -1)
    (touch-book:mark-method where)))

(defun touch-book:leads-buffer-ring-this (where)
  "Get the user to ring the lead that the cursor is on, from a leads-list-mode buffer."
  (interactive "d")
  (let* ((which-lead (touch-book:lead-on-line where))
	 (method-name (car which-lead))
	 (lead-number (cdr which-lead))
	 (the-name (format "%s(%c)" method-name (ringing:bell-number-to-char lead-number))))
    (save-window-excursion
      (ringing:try-user-on-lead method-name lead-number))
    (touch-book:refill-leads-buffer)
    (goto-char (point-min))
    (when (search-forward the-name (point-max) t)
      (touch-book:mark-method (point)))))

(defun touch-book:list-worst-leads-first (&optional percentage)
  "Return a list of leads, as from touch-book:list-all-leads, worst first.
With optional argument PERCENTAGE, go by percentage of perfect leads;
otherwise, by average faults per blow."
  (sort (touch-book:list-all-leads percentage)
	(function
	 (lambda (a b)
	   (< (fifth a) (fifth b))))))

(defun touch-book:worst-lead-matching (name-pattern &optional place-bell)
  "Return the worst lead whose name matches NAME-PATTERN for PLACE-BELL.
If PLACE-BELL is nil, just match the name."
  (let ((leads (touch-book:list-worst-leads-first nil)))
    (catch 'got-it
      (dolist (lead leads)
	(when (and (or (null place-bell)
		       (= place-bell (second lead)))
		   (string-match name-pattern (first lead)))
	  (throw 'got-it lead)))
      nil)))

(defun touch-book:worst-leads-first (&optional percentage)
  "Display a list of leads, as from touch-book:list-all-leads, worst first,
that is, sorted according to the average number of mistakes per row per lead,
or, with optional (prefix) argument, the percentage rate of perfect attempts
for that lead."
  (interactive "P")
  (let ((from-method (if (eq major-mode 'mSiRiL-mode)
			 (mSiRiL-mode:method-at (point))
		       nil)))
    (switch-to-buffer-other-window
     (ringing:get-buffer-create touch-book:worst-leads-buffer-name))
    (leads-list-mode)
    (touch-book:refill-leads-buffer percentage)
    (goto-char (point-min))
    (when (and from-method
	       (search-forward (ringing:method-name from-method) (point-max) t))
      (touch-book:mark-method from-method))
    (other-window -1)))

(defvar touch-book:matches-used nil
  "The match strings given for filtering kinds of touches.")

(defun touch-book:teach-worst-lead (&optional matching)
  "Teach the ringer their current worst lead.
With optional argument (prompted for if prefix given, interactively) pick the
worst whose name matches the argument."
  (interactive
   (list
    (if (null current-prefix-arg)
	nil
      (read-from-minibuffer "Practice worst lead of methods matching: "
			    nil		; initial-contents
			    nil		; keymap
			    nil		; read as lisp
			    'touch-book:matches-used ; history
			    ))))
  (require 'teacher)
  (let ((the-worst
	 (if matching
	     (touch-book:worst-lead-matching matching)
	   (car (touch-book:list-worst-leads-first)))))
    (ringing:teach-lead-until-perfect
     (first the-worst) (second the-worst)))
  ;; update the listing buffer if it's on display
  (when (and (get-buffer touch-book:worst-leads-buffer-name)
	     (get-buffer-window (get-buffer touch-book:worst-leads-buffer-name)))
    (touch-book:worst-leads-first)))

(defun ringing:choose-action (prompt choices)
  "Display PROMPT and read one of CHOICES from the user.
CHOICES is a list of symbols; the user must type the first letter of one of
their print-names."
  (catch 'chosen
    (while t
      (message "%s %s? " prompt choices)
      (let ((char (read-char-exclusive)))
	(dolist (choice choices)
	  (when (= (aref (symbol-name choice) 0)
		   char)
	    (throw 'chosen choice)))))
    nil))

(defun touch-book:revise-my-methods (&optional matching)
  "Take the user through all the leads they've ever tried (with optional
string argument, only those matching that argument), starting with the
one they've got right a lowest proportion of tries so far. Before each lead,
you are offered the choice of try, skip or quit. So you start with what you
most need to practice, and stop when you get to things you're sure you know."
  (interactive
   (list
    (if (null current-prefix-arg)
	nil
      (read-from-minibuffer "Practice worst lead of methods matching: "
			    (let* ((file (buffer-file-name))
				   (kind (ringing:figure-out-kind file))
				   (kindname (if kind (ringing:kind-by-name kind) nil))
				   (stage (ringing:figure-out-number-of-bells file))
				   (stagename (if stage (ringing:stage-name stage) nil)))
			      (if (and kind stage)
				  (concat kindname " " stagename)
				nil))
			    ; initial-contents
			    nil		; keymap
			    nil		; read as lisp
			    'touch-book:matches-used ; history
			    ))))
  (save-window-excursion
    (let ((leads (touch-book:list-worst-leads-first))
	  (teach-ringing:tolerable-faults 1)
	  (teach-ringing:required-perfects-seen 1)
	  (teach-ringing:required-perfects-unseen 0))
      (catch 'revised
	(dolist (lead leads)
	  (let* ((method-name (first lead))
		 (place-bell (second lead)))
	    (when (or (null matching) (string-match matching method-name))
	      (teach-ringing:do-one-lead method-name place-bell nil nil)
	      (case
		  (ringing:choose-action (format "Revise lead %d of %s" place-bell method-name)
					 '(try skip quit))
		(try (ringing:teach-lead-until-perfect method-name place-bell))
		(quit (throw 'revised nil))
		(skip nil)))))))
    (touch-book:worst-leads-first)))

(defun touch-book:find-mSiRiL-for-method (where)
  "Find the method mentioned at point."
  (interactive "d")
  (touch-book:mark-method where)
  (method-library:locate-method (car (touch-book:lead-on-line where))))

;;;; method groups generated from the touch book

(defun touch-book:touch-book-as-group (&optional n)
  "Return the touch book as a group (see method-groups.el).
With optional argument N numeric or string, take that as a number of bells,
and return only the methods at that stage."
  (let ((methods (mapcar 'ringing:name-data-pair
			 (mapcar 'car (touch-book:load)))))
    (when (stringp n)
      (setq n (ringing:figure-out-number-of-bells n)))
    (when (integerp n)
      (setq methods (delete-if-not
		     (function
		      (lambda (method)
			(ringing:method-at-stage-p method n)))
		     methods)))
    (make-method-groups:method-group
     :title (if (null n)
		"*Touch book*"
	      (format "*My %s methods*" (ringing:stage-name n)))
     :methods methods)))

(defun touch-book:methods-rung-this-session-as-group ()
  "Return a group (see method-groups.el) of the methods attempted this session."
  (make-method-groups:method-group
   :title "*Methods rung this session*"
   :methods (mapcar 'ringing:name-data-pair touch-book:methods-rung-this-session)))

(defun touch-book:stages-present-in-touch-book ()
  "Return a list of numbers indicating which stages are in the user's touch book."
  (let ((methods (touch-book:touch-book-as-group))
	(result nil))
    (dolist (method (method-groups:method-group-methods methods))
      (pushnew (ringing:figure-out-number-of-bells (car method))
	       result))
    result))

(defun touch-book:touch-book-stage-groups ()
  "Return an alist of names of groups for stages of methods present in touch book."
  (mapcar
   (function
    (lambda (n)
      (cons (format "*My %s methods*"
		    (ringing:stage-name n))
	    nil)))
   (touch-book:stages-present-in-touch-book)))
	  
;;; end of touch-book.el
