;;;; touches.el -- part of John's elisp change-ringing system
;;; ring touches
;;; Time-stamp: <99/11/06 21:34:44 johnstu>

(eval-when (compile load eval)
  (provide 'touches)
  (require 'ringing-methods)
  (require 'beer)			; but you should probably load the lot anywaye
  (require 'method-display))

;;;; ringing touches

(defun ringing:make-match-call (call-type call method &optional spec-rest)
  "Make a regular expression to spot CALL-TYPE at CALL for METHOD.
CALL is one of 'home, 'wrong, 'middle, 'fifths, 'fourths, 'before, 'in, 'out."
  (let* ((nbells (change-ringing-method-nbells method))
	 (pattern (make-string (1+ nbells) ?.))
	 (tenor-position
	  (case call
	    ('home nbells)
	    ('wrong (1- nbells))
	    ('middle (- nbells 2))
	    ('fifths 5)
	    ('fourths 4)
	    ('before (if (eq call-type 'bob)
			 3
		       2))
	    ('in 2)
	    ('out 3))))
    (aset pattern 1 ?1)
    (aset pattern tenor-position (ringing:bell-number-to-char nbells))

    ;; now ring backwards using the pattern as the row, to step the
    ;; tenor to the relevant place for that call for that method

    (list call-type
	  (concat			; look for something other than a bell
	   "[^0-9ETetA-Da-d]"		; so we know it's the start of a row
	   (substring			; but first chop what was there
	    (ringing:retreat-changes
	     pattern
	     (case call-type
	       ('bob (change-ringing-method-bob method))
	       ('single (change-ringing-method-single method))
	       (t (change-ringing-method-plain-lead-end method))))
	    1))
	  (if spec-rest spec-rest call))))

(defun ringing:convert-to-touch (touch-string method)
  "Convert TOUCH-STRING for METHOD to a list of lead actions headed by 'touch.
This lets us rplacd our way through it as we go. Lower case p, b, s are
plain, bob and single; upper case H W M V F B I O are Home, Wrong, Middle,
Fifths, Fourths, Before, In, Out -- unfortunately resorting to case-significance
to distinguish Before from bob. Upper-case S makes the next H, W etc call a single
instead of a bob.
If TOUCH-STRING is already a touch, a fresh one (sharing structure apart from the holder)
is returned, so you can still rplacd your way along without affecting the original."
  (if (and (listp touch-string)
	   (eq (car touch-string) 'touch))
      (cons 'touch (cdr touch-string)) ; copy the holder so caller can rplacd along it without affecting original
    (let ((call-list (coerce touch-string 'list))
	  (next-call 'bob)
	  (result nil))
      (dolist (this-call call-list)
	(case this-call
	  (?p (push 'plain result))
	  (?b (push 'bob result))
	  (?s (push 'single result))
	  (?S (setq next-call 'single))
	  (?H				; n
	   (push (ringing:make-match-call next-call 'home method) result))
	  (?W				; n-1
	   (push (ringing:make-match-call next-call 'wrong method) result))
	  (?M				; 6
	   (push (ringing:make-match-call next-call 'middle method) result))
	  (?V				; 5
	   (push (ringing:make-match-call next-call 'fifths method) result))
	  (?F				; 4
	   (push (ringing:make-match-call next-call 'fourths method) result))
	  (?B				; 3
	   (push (ringing:make-match-call next-call 'before method) result))
	  (?I				; 2
	   (push (ringing:make-match-call next-call 'in method) result))
	  (?O
	   (push (ringing:make-match-call next-call 'out method) result))
	  (t 'unknown))
	(unless (= this-call ?S) (setq next-call 'bob)))
      (cons 'touch (nreverse result)))))

(defconst ringing:calling-positions
  '(home wrong middle fifths fourths before in out)
  "The possible calling positions, as lisp symbols.")

(defun ringing:complete-method-name (name defaults)
  "Complete the method name NAME from DEFAULTS.
If the stage is missing from NAME, take it from DEFAULTS.
If the kind of method is missing from NAME, take it from DEFAULTS.
Not actually implemented yet."
  (if nil  (let ((stage (ringing:figure-out-number-of-bells name))
		 (kind (ringing:figure-out-kind name)))
	     (cond
	      ((and stage kind)
	       name)
	      ((and (null stage)
		    (null kind))
	       (format "%s %s %s" name (capitalize (ringing:figure-out-kind defaults)) (ringing:stage-name (ringing:figure-out-number-of-bells defaults))))
	      ((null stage)
	       (format "%s %s" name (ringing:stage-name (ringing:figure-out-number-of-bells defaults))))
	      ;;((null kind)
	      ;; (format "%s %s %s " ...))
	      (t name)))
    name)
  )
  
(defun ringing:compile-touch (calls &optional method)
  "Make a proper touch list from CALLS."
  (when (eq (car calls) 'touch) (setq calls (cdr calls)))
  (let* ((result nil)
	 (current-method (if method
			     (ringing:find-named-method method)
			   nil))
	 (previous-method current-method))
    (message "Compiling touch %s" calls)
    (dolist (call calls)
      (message "Compiling call %s" call)
      (push
       (cond
	((and (listp call)
	      (= (length call) 2)
	      (memq (first call) '(go method)))
	 (message "Compiling unconditional method change %s" call)
	 (setq previous-method current-method
	       current-method (ringing:find-named-method (ringing:complete-method-name (second call) previous-method)))
	 (list 'method "" current-method))
	((stringp call)
	 (message "Compiling unconditional method change (just name given) %s" call)
	 (setq previous-method current-method
	       current-method (ringing:find-named-method (ringing:complete-method-name call previous-method)))
	 (list 'method "" current-method))
	((and (listp call)
	      (= (length call) 2)
	      (memq (first call) '(bob single))
	      (memq (second call) ringing:calling-positions))
	 (message "Compiling fully specified call %s" call)
	 (ringing:make-match-call (first call) (second call) current-method))
	((and (listp call)
	      (= (length call) 3)
	      (memq (first call) '(method))
	      (memq (second call) ringing:calling-positions))
	 (message "Compiling fully specified method change %s" call)
	 (setq previous-method current-method
	       current-method (ringing:find-named-method (ringing:complete-method-name (third call) previous-method)))
	 (if nil
	     (list 'method
		   (ringing:make-match-call 'method (second call) current-method)
		   current-method)
	   (ringing:make-match-call 'method (second call) current-method current-method)
	   )
	 )
	((and (listp call)
	      (= (length call) 1)
	      (memq (first call) ringing:calling-positions))
	 (message "Compiling call (implicitly a bob) %s" call)
	 (ringing:make-match-call 'bob (first call) current-method))
	(t
	 (message "Compiling unknown call type (identity) %s" call)
	 call))
       result))
    (cons 'touch (nreverse result))))

(defvar ringing:defined-touches nil
  "List of defined named touches.")

(defun ringing:define-touch (name calls)
  "Define a touch called NAME using CALLS.
See deftouch for details of CALLS format."
  (when (stringp name) (setq name (intern name)))
  (let ((pair (assoc name ringing:defined-touches))
	(touch (ringing:compile-touch calls)))
    (if pair
	(rplacd pair touch)
      (push (cons name touch) ringing:defined-touches))))

(require 'backquote)

(defmacro deftouch (touchname &rest touchcalls)
  "Define TOUCHNAME to call TOUCHCALLS by passing them to ringing:define-touch.
You can access touches thus defined through M-x ringing:show-named-touch and
M-x ringing:ring-named-touch, or from a program by ringing:find-named-touch.
The first argument is the touch name, and the rest are the calls.
Each call may be a list (looking like a Lisp function call) or a symbol or a string.
The `functions' available are
  (method <method-name>)
     Changes the method to the specified one at the next lead start
     `go' is a synonym for `method' in this form
     <method-name> is a string naming a method, e.g \"Cambridge Surprise Minor\"
  <method-name>
     Same as (method <method-name>)
  (bob <position>)
     Calls a bob when <position> occurs.
     <position> may be any of the following Lisp symbols:
       home wrong middle fifths fourths before in out
  (single <position>)
     Calls a single when <position> occurs
  (method <position> <method-name>)
     Changes to <method-name> when <position> occurs
  (<position>)
     Calls a bob when <position> occurs
"
  `(ringing:define-touch ',touchname '(,@touchcalls)))

(defmacro touch (&rest touchcalls)
  "Return a compiled touch containing TOUCHCALLS.
This is so you can write (touch (go \"Cambridge Surprise Minor\")) like a quoted list constant.
See deftouch for details of arguments."
  `(ringing:compile-touch '(,@touchcalls)))

(defun ringing:find-named-touch (name)
  "Return the touch called NAME."
  (when (stringp name) (setq name (intern name)))
  (let ((pair (assoc name ringing:defined-touches)))
    (if pair
	(copy-sequence (cdr pair))
      nil)))

(defun ringing:peek-next-lead-end-action (method row touch-spec &optional remaining-changes)
  "Get the next lead end action according to METHOD, ROW and TOUCH-SPEC.
If REMAINING-CHANGES is given, use ROW advanced thereby before trying a match.
TOUCH-SPEC should be a cons, of which the cdr is the list of successive
actions."
  (let ((action (cadr touch-spec)))
    ;; (message "ringing:peek-next-lead-end-action: action is %S" action)
    (cond
     ((symbolp action)
      action)
     ((and (listp action)
	   (string-match (second action)
			 (ringing:advance-changes row remaining-changes)))
      (if (eq (first action) 'method)
	  (list (first action) (third action))
	(first action)))
     (t 'plain))))

(defun ringing:get-lead-end-action (method row touch-spec &optional remaining-changes)
  "Get the next lead end action according to METHOD, ROW and TOUCH-SPEC.
If REMAINING-CHANGES is given, use ROW advanced thereby before trying a match.
TOUCH-SPEC should be a cons, of which the cdr is the list of successive
actions. It is modified with rplacd to work through the touch."
  (let ((action (cadr touch-spec)))
    ;; (message "ringing:get-lead-end-action: action is %S" action)
    (cond
     ((symbolp action)
      (rplacd touch-spec (cddr touch-spec))
      action)
     ((and (listp action)
	   (string-match (second action)
			 (ringing:advance-changes row remaining-changes)))
      (rplacd touch-spec (cddr touch-spec))
      (if (eq (first action) 'method)
	  (list (first action) (third action))
	(first action)))
     (t 'plain))))

(defun ringing:show-touch-helper-function (row
					   remaining-changes
					   method-holder
					   rounds
					   place-marker
					   touch-spec
					   count-holder
					   truth-holder)
  "Helper function for ringing or showing touches."
  (incf ringing:plain-course-row-count)
  (incf (car count-holder))
  (if (or (string= row rounds)
	  (not (car truth-holder)))
      nil
    (if remaining-changes
	t
      (let ((method (car method-holder)))
	(cons row
	      (case (car place-marker)

		((lead-end)
		 ;; We are already in a lead end, i.e. have just done one and are about to start the
		 ;; next lead, so the only thing we're allowed to do here is change the method.
		 (when ringing:lead-end-callback
		   (apply ringing:lead-end-callback
			  row
			  remaining-changes
			  method
			  ringing:lead-end-callback-args))

		 (rplaca place-marker 'work)
		 (let* ((this-call-descr (ringing:peek-next-lead-end-action method row touch-spec remaining-changes))
			(this-call
			 (if (listp this-call-descr)
			     (first this-call-descr)
			   this-call-descr))
			(next-method (if (and (eq this-call 'method)
					      (listp this-call-descr))
					 (second this-call-descr)
				       nil)))
		   (message "Got work call %s (part of %s), next-method=%s" this-call this-call-descr next-method)
		   (when (not (null next-method))
		     (when (eq this-call 'method)
		       ;; if the call is not a method change, we leave it until the lead end;
		       ;; if it is a method change, act on it, and consume it from the calls list
		       (ringing:get-lead-end-action method row touch-spec remaining-changes))
		     (setq method next-method)
		     (rplaca method-holder next-method)))
		 (change-ringing-method-work method))


		((work)
		 ;; We are already in the work, i.e. have just done the work and are about to start
		 ;; the lead end, so we can do a bob or single here, but not change the method.
		 (rplaca place-marker 'lead-end)
		 (let* ((this-call-descr (ringing:peek-next-lead-end-action method row touch-spec remaining-changes))
			(this-call
			 (if (listp this-call-descr)
			     (first this-call-descr)
			   this-call-descr)))
		   (message "Got lead end call %s (part of %s)" this-call this-call-descr)

		   (if (eq this-call 'method)
		       (progn
			 ;; We can't change the method at this point, so don't consume this entry in
			 ;; the calls list -- just return a plain lead end, and leave the method change
			 ;; for the end of this lead.
			 (change-ringing-method-plain-lead-end method)
			 )
		     (progn
		       ;; consume the call
		       (ringing:get-lead-end-action method row touch-spec remaining-changes)
		       (case this-call
			 ((bob)
			  (change-ringing-method-bob method))
			 ((single)
			  (change-ringing-method-single method))
			 (nil nil)
			 ((plain) (change-ringing-method-plain-lead-end method))
			 (t (change-ringing-method-plain-lead-end method)))
		       ))
		   ))))))))

(defvar ringing:call-touch-at-changes-remaining 2
  "How many rows before running out of place notation a call should come.")

(defun ringing:touch-display-function (row
				       remaining-changes
				       method-holder
				       rounds
				       place-marker
				       calls
				       truth-holder)
  "Calls method-display:main-display-function but adds simple calling.
Arguments (ROW REMAINING-CHANGES METHOD-HOLDER ROUNDS PLACE-MARKER CALLS TRUTH-HOLDER)
CALLS takes the form of a list whose car is 'touch and whose tail is consumed
by the corresponding calls to the helper function."
  (let* ((method (car method-holder))
	 (ringing:prompt-user-extra-string
	  (if (and (eq (car place-marker) 'work)
		   (<= (length remaining-changes) ringing:call-touch-at-changes-remaining))
	      (let* ((this-call-descr (ringing:peek-next-lead-end-action
				       method row calls
				       (cdr remaining-changes)))
		     (this-call
		      (if (listp this-call-descr)
			  (first this-call-descr)
			this-call-descr))
		     (further-detail (if (listp this-call-descr) (third this-call-descr) nil)))
		(message "this-call-descr=%s this-call=%s" this-call-descr this-call)
		(case this-call
		  ((method)  (format "%s! " (ringing:method-name (second this-call-descr))))
		  ((single) (if further-detail (format "Single! (%s)" further-detail) "Single! "))
		  ((bob) (if further-detail (format "Bob! (%s)" further-detail) "Bob! "))
		  (t nil)))
	    nil))
	 (false (and
		 (not (string-match rounds row))
		 (save-excursion (search-backward row (point-min) t)))))
    (prog1 (funcall method-display:main-display-function row
		    remaining-changes
		    method-holder
		    rounds
		    place-marker)
      ;; (unless ringing:test-user (insert (format " %s %d " (symbol-name (car place-marker)) (length remaining-changes))))
      (when ringing:prompt-user-extra-string
	(if ringing:test-user
	    (momentary-string-display ringing:prompt-user-extra-string (point) 0 ringing:prompt-user-extra-string)
	  (when (eq method-display:main-display-function 'method-display:single-column-display-function)
	    (insert ringing:prompt-user-extra-string))))
      (when (and (eq method-display:main-display-function 'method-display:single-column-display-function)
		 (not ringing:test-user)
		 false)
	(insert " False!")))))

(defvar ringing:touch-buffer-header-format "Touch starting with %s (%s)\n"
  "Header format string for buffers containing touches. String arg is for method name.")

(defvar ringing:touch-buffer-name-format "*Touch starting with %s*"
  "Format string for making buffer name for touches. String arg is for method name.")

(defun ringing:show-touch (method touch-calls &optional show-buffer-function helper-function)
  "Show a touch of METHOD with TOUCH-CALLS giving the actions at each lead end.
An optional show-buffer-function is what to do with the buffer first (not available interactively).
for example, switch-to-buffer-other-window.
An optional HELPER-FUNCTION overrides the normal one (not available interactively)."
  (interactive (list
  		(ringing:read-method-name "Show touch of: ")
 		(read-from-minibuffer "Touch: ")))
  (let* ((method-data (ringing:find-named-method method))
	 (nbells (change-ringing-method-nbells method-data))
	 (method-name (ringing:method-name method))
         (symmetrical-starting-bell (if (null ringing:observation)
				  ;; (ringing:bell-number-to-char nbells)
				  (ringing:bell-number-to-char (ringing:lowest-working-bell-making-places
						 (car (change-ringing-method-plain-lead-end method))))
				ringing:observation))
	 (ringing:observation-number
	  (if t
	      nbells
	  symmetrical-starting-bell))
	 (ringing:observation (ringing:bell-number-to-char ringing:observation-number))
	 (ringing:end-of-row-char ? )	; before making the rounds
	 ;; (the-rounds (ringing:make-rounds nbells))
	 (method-display:block-top-marker (make-marker))
	 (ringing:call-touch-at-changes-remaining 1)
	 (calls (ringing:convert-to-touch touch-calls method-data))
	 (place-marker (list 'work))
	 (count-holder (list 0))
	 (truth-holder (list t)))

    (require 'ermm)
    (ermm:init nbells)
    (funcall (if (null show-buffer-function)
		 'switch-to-buffer-other-window
	       show-buffer-function)
	     (ringing:get-buffer-create (format ringing:touch-buffer-name-format method-name)))
    (erase-buffer)
    (insert (format ringing:touch-buffer-header-format
		    (capitalize method-name)
		    (concat (place-notation:recreate-string-from-changes (change-ringing-method-work method))
			    ", "
			    (place-notation:recreate-string-from-changes (change-ringing-method-plain-lead-end method)))
		    (change-ringing-method-original-description-string method)))
    (set-marker method-display:block-top-marker (point))
    (setq truncate-lines t)
    (insert ?\n)
    (when (eq method-display:main-display-function
	      'method-display:multi-column-display-function)
      (method-display:label-column ringing:observation-number (method-display:column-width nbells))
      (insert ?\n))
    (message "Starting touch")
    (ringing:ring-from-rounds method
			      (if helper-function
					  helper-function
					'ringing:show-touch-helper-function)
			      (list place-marker calls count-holder truth-holder)
			      'ringing:touch-display-function
			      (list place-marker calls truth-holder))
    (goto-char (point-max))
    (insert (format "\n%d changes\n" (car count-holder)))
    (message "Completed touch")
    (goto-char (point-min))
    (set-marker method-display:block-top-marker nil)))

(defvar method-name-tidy nil
  "The current method name for ringing a touch.
This global value should never be used; this declaration is to keep the compiler quiet.")

(defun ringing:ring-touch (method-name manual-bell touch-calls
				       &optional helper-function display-function
				       helper-extra-args display-extra-args)
  "Ring a touch of METHOD-NAME, with the user ringing MANUAL-BELL, with TOUCH-CALLS.
An optional HELPER-FUNCTION overrides the normal one (not available interactively).
Likewise DISPLAY-FUNCTION."
  (interactive
   (let* ((method-name (ringing:read-method-name "Ring touch of: "))
	  (manual (ringing:read-number-up-to
		   "Manual bell"
		   (change-ringing-method-nbells
		    (ringing:find-named-method method-name))))
	  (calls (read-from-minibuffer "Calls: ")))
     (list method-name manual calls)))
  (setq ringing:clashes 0
	ringing:plain-course-row-count 0)
  (setq ringing:prev-place (ringing:bell-char-to-number manual-bell))
  (let* ((ringing-buffer (ringing:get-buffer-create
			  (format ringing:touch-buffer-name-format (ringing:method-name method-name))))
	 (method (ringing:find-named-method method-name))
	 (method-name-tidy (capitalize (ringing:method-name method-name)))
	 (ringing:observation-number manual-bell)
	 (ringing:observation (ringing:bell-number-to-char manual-bell))
	 (ringing:allow-timeout-this-row nil) ; will be set to t after one row
	 (nbells (change-ringing-method-nbells method))
	 (ringing:test-user t)
	 (ringing:end-of-row-char ? )	; before making the rounds
	 (method-display:block-top-marker (make-marker))
	 ;; (the-rounds (ringing:make-rounds nbells))
	 (calls (ringing:convert-to-touch touch-calls method))
	 (place-marker (list 'work))
	 (count-holder (list 0))
	 (truth-holder (list t)))
    (switch-to-buffer ringing-buffer)
    (erase-buffer)
    (delete-other-windows)
    (insert (format ringing:touch-buffer-header-format
		    method-name-tidy
		    (change-ringing-method-original-description-string method)))
    (setq mode-line-format '("Random touch of " method-name-tidy))
    (set-marker method-display:block-top-marker (point))
    (require 'ermm)
    (ermm:init (change-ringing-method-nbells method))
    (insert "\n")
    (when (eq method-display:main-display-function
	      'method-display:multi-column-display-function)
      (method-display:label-column ringing:observation-number (method-display:column-width nbells))
      (insert ?\n)
      )
    (let* ((ringing:lead-end-callback 'ringing:lead-end-record-clashes)
	   (ringing:lead-end-callback-args (list (cons ringing:clashes ringing:observation-number)))
	   (start-time (current-time))
	   (ringing-result
	    (ringing:ring-from-rounds method
				      (if helper-function
					  helper-function
					'ringing:show-touch-helper-function)
				      (append (list place-marker calls count-holder truth-holder)
					      helper-extra-args)
				      (if display-function
					  display-function
					'ringing:touch-display-function)
				      (append (list place-marker calls truth-holder)
					      display-extra-args)))
	   (end-time (current-time))
	   (time-taken (seconds-between start-time end-time)))
      (set-marker method-display:block-top-marker nil)
      (if (eq ringing-result 'user-stand)
	  (progn
	    (kill-buffer ringing-buffer)
	    )
	(progn
	  (set-buffer ringing-buffer)
	  (goto-char (point-max))
	  (insert (format "\n%d changes\n" (car count-holder)))
	  (touch-book:record-clash-level method-name
					 (cdar ringing:lead-end-callback-args)
					 (- ringing:clashes (caar ringing:lead-end-callback-args)))
	  (with-output-to-temp-buffer "*Ringing report*"
	    (princ (format "Touch of %d changes finished with %d clashes, in %d seconds\nPeal time would be %s (%d changes per minute)"
			   ringing:plain-course-row-count
			   ringing:clashes
			   time-taken
			   (time-string (* (float time-taken)
					   (/ 5000.0
					      (float ringing:plain-course-row-count))))
			   (floor (/ (float ringing:plain-course-row-count) (/ (float time-taken) 60.0)))
			   ))))))))

(defun ringing:read-touch-from-region (start end)
  "Return a touch read from the region between START and END.
This is for the lines and columns format as used in the RW diary, for example

23456 B M W H
52436     -
42635   -
42356 -
53624   - -
62534     - -
23645 -
23456 -     -
for which this function returns a list of the characters WMBMWWHBBH"
  (save-excursion
    (goto-char start)
    (if (re-search-forward "^[0-9etET]*\\([^,0-9etET
]*H[^,
]*\\)[0-9etET]*$" end t)
	(let ((header-line (buffer-matched 1))
	      (result nil))
	  (while (re-search-forward "^[0-9etET]*\\([ -23xXSs]+\\)[0-9etET]*$" end t)
	    (let* ((this-line (buffer-matched 1))
		   (n (length this-line))
		   (i 0))
	      (while (< i n)
		(let ((c (aref this-line i)))
		  (cond
		   ((eq c ?-)
		    (push (aref header-line i) result))
		   ((eq c ?S)
		    (push ?S result)
		    (push (aref header-line i) result))
		   ((eq c ?2)
		    (push (aref header-line i) result)
		    (push (aref header-line i) result))
		   ((eq c ?3)
		    (push (aref header-line i) result)
		    (push (aref header-line i) result)
		    (push (aref header-line i) result))
		   )
		  )
		(incf i)
		))
	    (goto-char (1+ (point))))
	  (reverse result)
	  )
      (error "Could not find line containing Home"))))

(defun ringing:show-region-as-touch (from to method)
  "Take text between FROM and TO as callings for a touch of METHOD."
  (interactive (list (region-beginning) (region-end)
		     (method-name (ringing:read-method-name "Ring touch of: "))))
  (let ((touch (ringing:read-touch-from-region from to)))
    (ringing:show-touch (ringing:find-named-method method) touch)))

(defun ringing:methods-named-in-buffer ()
  "Return methods listed after Method: or Methods: in this buffer."
  (save-excursion
    (goto-char (point-min))
    (if (re-search-forward "Methods?: \\(.+\\)$" (point-max) t)
	(let ((methods-string (buffer-matched 1))
	      (methods-list nil))
	  (while (string-match "\\([^,]+\\), *" methods-string)
	    (let ((method-name (substring methods-string 0 (match-end 1))))
	      (push (cons method-name
			  (ringing:find-named-method method-name))
		    methods-list)
	      (setq methods-string (substring methods-string (match-end 0)))))
	  (push (cons methods-string
		      (ringing:find-named-method methods-string))
		methods-list)
	  methods-list)
      nil)))

(defun ringing:show-buffer-as-touch (method)
  "Take text of buffer as callings for touch of METHOD."
  (interactive
   (let ((methods (ringing:methods-named-in-buffer)))
     (if methods
	 (if (= (length methods) 1)
	     (list (caar methods))
	   (list (completing-read "Ring touch of: "
				  methods
				  nil t)))
       (list
	(ringing:read-method-name "Ring touch of: ")))))
  (ringing:show-region-as-touch (point-min) (point-max) method))

(defun ringing:starting-method-for-touch (touch)
  "Return the first method specified in TOUCH."
  (catch 'got-method
    (dolist (call (cdr touch))
      (cond
       ((and (listp call)
	     (eq (first call) 'method))
	(throw 'got-method (ringing:find-named-method (third call))))
       (t nil)))
    nil))

(defun ringing:show-named-touch (touchname)
  "Show the touch named TOUCHNAME."
  (interactive (list (completing-read "Show named touch: "
				      (mapcar (function (lambda (sympair) (cons (symbol-name (car sympair)) sympair)))
					      ringing:defined-touches) nil t)))
  (let* ((touch (ringing:find-named-touch touchname))
	 (starting-method (ringing:starting-method-for-touch touch)))
    (rplacd touch (cddr touch))
    (ringing:show-touch starting-method touch)))

(defun ringing:ring-named-touch (touch-name manual-bell)
  "Ring the touch named TOUCH-NAME, starting from MANUAL-BELL."
  (interactive
   (let* ((touch-name (completing-read "Ring named touch: "
				       (mapcar (function (lambda (sympair) (cons (symbol-name (car sympair)) sympair)))
					       ringing:defined-touches) nil t))
	  (touch (ringing:find-named-touch touch-name))
	  (manual (ringing:read-number-up-to
		   "Manual bell"
		   (change-ringing-method-nbells
		    (ringing:find-named-method (ringing:starting-method-for-touch touch)))))	  )
     (list touch-name manual)))
  (let* ((touch (ringing:find-named-touch touch-name))
	 (starting-method (ringing:starting-method-for-touch touch)))
    (rplacd touch (cddr touch))
    (ringing:ring-touch starting-method manual-bell touch)))

;; end of touches.el
