;;;; user-mistakes.el -- part of John's elisp change-ringing system
;;; record the user's ringing mistakes in detail
;;; Time-stamp: <99/02/05 12:12:44 johnstu>

(eval-when (compile load eval)
  (require 'backquote)
  (provide 'user-mistakes))

(defstruct user-mistakes:method
  method				; back to actual method structure
  leads					; array of user-mistakes:lead structures
  )

(defstruct user-mistakes:lead
  method				; back to user-mistakes:method
  place					; which place bell
  row-faults				; each kind of fault, per row
					; (including the non-faults!
					; this is as count vector [ down places up ]
  which-row				; where we are
  attempts                              ; how many times we've tried this lead
  perfects				; how many times we've rung this lead perfectly
  total-faults				; how many faults we've made on this lead altogether
  latest-faults				; how many faults we've made last time we rang this
  )

(defun user-mistakes:lead-to-string (row-faults-array)
  "Convert the ROW-FAULTS-ARRAY to a string for storage in the touch-book.
Successive bytes are succesive elements of the successive row-faults vectors."
  (let* ((n-rows (length row-faults-array))
	 (string (make-string (* n-rows 3) ?.))
	 (j -1))
    (dotimes (i n-rows)
      (let ((row (aref row-faults-array i)))
	(aset string (incf j) (min (aref row 0) 250))
	(aset string (incf j) (min (aref row 1) 250))
	(aset string (incf j) (min (aref row 2) 250))))
    string))

(defun user-mistakes:lead-from-string (string)
  "Convert STRING to a row-faults array, for loading from the touch-book."
  (let ((n-elts (/ (length string) 3)))
    (if (= (* n-elts 3) (length string))
	(let ((row-faults-array (make-vector n-elts nil))
	      (j -1))
	  (dotimes (i n-elts)
	    (let ((this-row-faults (make-vector 3 nil)))
	      (aset row-faults-array i this-row-faults)
	      (aset this-row-faults 0 (aref string (incf j)))
	      (aset this-row-faults 1 (aref string (incf j)))
	      (aset this-row-faults 2 (aref string (incf j)))))
	  row-faults-array)
      (error "Lead faults string not a multiple of 3 bytes"))))

(defvar user-mistakes:methods nil
  "Alist of method names to their user-mistakes:method structures.")

(defvar user-mistakes:fetched-leads nil
  "Leads whose teaching information has been fetched.
This appears to have been added for debugging, and I've now commented out
the use of it.")

(defun user-mistakes:get-teaching-data (method-name &optional place-bell)
  "Get the teaching data relating to METHOD-NAME (for PLACE if given)."
  (let* ((name (ringing:method-name method-name))
	 (pair (assoc name user-mistakes:methods)))
    (unless pair
      (let ((method (ringing:find-named-method method-name)))
	(setq pair (cons name
			 (make-user-mistakes:method
			  :method method
			  :leads (make-vector (1+ (change-ringing-method-nbells method))
					      nil))))
	(push pair user-mistakes:methods)))
    (let ((method-teaching-data (cdr pair)))
      (if place-bell
	  (let* ((leads (user-mistakes:method-leads method-teaching-data))
		 (lead (aref leads place-bell)))
	    (unless lead
	      (setq lead (make-user-mistakes:lead
			  :method method-teaching-data
			  :place place-bell
			  :attempts 0
			  :row-faults nil
			  :perfects 0
			  :total-faults 0
			  :latest-faults 0))
	      (aset leads place-bell lead))
	    ;; (push (list name place-bell lead) user-mistakes:fetched-leads)
	    lead)
	method-teaching-data))))

(defvar user-mistakes:teaching-info nil
  "The currently relevant user-mistakes:lead, or nil if not teaching.")

(defun user-mistakes:merge-teaching-data (new-data)
  "Merge NEW-DATA into the main teaching data."
  (dolist (pair new-data)
    (let ((name (car pair))
	  (data (cdr pair)))
      (when (and (user-mistakes:method-p data)
		 (not (null (user-mistakes:method-leads data))))
	(let* ((leads (user-mistakes:method-leads data))
	       (n (length leads))
	       (i 2))
	  (while (< i n)
	    (let ((new-data (aref leads i)))
	      (when new-data
		(let* ((new-row-fault-arrays (user-mistakes:lead-row-faults new-data))
		       (existing-data (user-mistakes:get-teaching-data name i))
		       (existing-row-fault-arrays (user-mistakes:lead-row-faults existing-data)))

		  (if (and (arrayp existing-row-fault-arrays)
			   (arrayp new-row-fault-arrays))
		      (dotimes (irow (length existing-row-fault-arrays))
			(let ((existing-row (aref existing-row-fault-arrays irow))
			      (new-row (aref new-row-fault-arrays irow)))
			  (if (and (arrayp existing-row)
				   (arrayp new-row))
			      (progn
				(when (and (numberp (aref existing-row 0)) (numberp (aref new-row 0)))
				  (incf (aref existing-row 0) (aref new-row 0)))
				(when (and (numberp (aref existing-row 1)) (numberp (aref new-row 1)))
				  (incf (aref existing-row 1) (aref new-row 1)))
				(when (and (numberp (aref existing-row 2)) (numberp (aref new-row 2)))
				  (incf (aref existing-row 2) (aref new-row 2)))
				)
			    (unless (arrayp existing-row)
			      (setf (aref existing-row-fault-arrays irow)
				    (aref new-row-fault-arrays irow))))))
		    (setf (user-mistakes:lead-row-faults existing-data)
			  (user-mistakes:lead-row-faults new-data)))


		  (message "New data %s; existing data %s" new-data existing-data)

		  ))
	      (incf i))))))))

(defmacro user-mistakes:with-previous-mistakes-hidden (&rest body)
  "Evaluate BODY forms with user mistake data apparently empty.
User mistake data generated during evaluation of BODY is merged back in
at the end."
  `(let ((new-teaching-data nil))
     (let ((user-mistakes:methods nil))	; so we have local teaching data
       ,@body   
       (setq new-teaching-data user-mistakes:methods))
     (user-mistakes:merge-teaching-data new-teaching-data)))

(defun user-mistakes:record-mistake (teaching-info remaining-changes place nbells)
  "Record in TEACHING-INFO that the user went to PLACE amongs NBELLS."
  (let ((which-row (user-mistakes:lead-which-row teaching-info))
	(all-row-faults (user-mistakes:lead-row-faults teaching-info)))
    (when (null all-row-faults)
      (setq all-row-faults (make-vector (* nbells 8) ; should be enough for just about any method!
					nil))
      (setf (user-mistakes:lead-row-faults teaching-info) all-row-faults))
    (let* ((row-faults (aref all-row-faults
			     which-row))
	   (just-rung-row (car remaining-changes))
	   (given-next-place (aref just-rung-row ringing:prev-place))
	   (given-direction (- given-next-place ringing:prev-place))
	   (user-direction (- place ringing:prev-place)))
      (when (null row-faults)
	(setq row-faults (make-vector 3 0))
	(aset row-faults (1+ given-direction) t)
	(aset all-row-faults which-row row-faults))
      (incf (aref row-faults (1+ user-direction))))))

(defun user-mistakes:performance (teaching-info)
  "Return a description of how the user is doing at the current point in TEACHING-INFO."
  (let* ((attempts (user-mistakes:lead-attempts teaching-info))
	 (which-row (user-mistakes:lead-which-row teaching-info))
	 (all-row-faults (user-mistakes:lead-row-faults teaching-info))
	 (row-faults (if (arrayp all-row-faults) (aref all-row-faults which-row) nil))
	 (downs (if (arrayp row-faults) (aref row-faults 0) 0))
	 (places (if (arrayp row-faults) (aref row-faults 1) 0))
	 (ups (if (arrayp row-faults) (aref row-faults 2) 0))
	 (n-faults (+ (if (numberp downs) downs 0)
		      (if (numberp places) places 0)
		      (if (numberp ups) ups 0))))
    (incf (user-mistakes:lead-which-row teaching-info))
    (if (zerop n-faults)
	nil
      (concat
       (if (and (numberp downs) (not (zerop downs))) (format " went down %s" (times-string downs)) "")
       (if (and (numberp places) (not (zerop places))) (format " made places %s" (times-string places)) "")
       (if (and (numberp ups) (not (zerop ups))) (format " went up %s" (times-string ups)) "")
       (if (and (> n-faults 1) (= n-faults attempts)) " -- wrong every time!" "")))))

;; (defun user-mistakes:display-function (row remaining-changes method-holder rounds)
;;   ""
;;   (incf (user-mistakes:lead-which-row teaching-info))
;;   (insert "\n")
;;   (ringing:apply-bell-colours row)
;;   (ermm:output-row row)
;;   (ringing:insert-row-lines row (user-mistakes:performance teaching-info))))

(defun user-mistakes:list-mistakes-in-method (method)
  "List the mistakes the user has made in METHOD."


  ;; change this so it does all the leads in one go, rather than calling ring-from-rounds each time

  (interactive (list (ringing:read-method-name "Show mistakes for: ")))
  (if (user-mistakes:get-teaching-data method)
      (let* ((method-name (ringing:method-name method))
	     (method-data (ringing:find-named-method method))
	     (ringing:show-lines-with-lines ringing:lines-between-rows-when-teaching)
	     (buffer (ringing:get-buffer-create (format "*Mistakes in %s*" method-name)))
	     (nbells (change-ringing-method-nbells method-data))


	     ;; ;; this lot from teach-ringing:do-one-lead
	     ;; (ringing:observation-number manual-bell)
	     ;; (ringing:observation (ringing:bell-number-to-char manual-bell))
	     ;; (teach-ringing:current-method (ringing:find-named-method method-name))
	     ;; (ringing:end-of-row-char ? )	; before making the rounds
	     ;; (ringing:test-user with-user)
	     ;; (ringing:allow-timeout-this-row nil) ; set to t after one row
	     ;; (ringing:green-line-too ringing:green-line-too-while-teaching)
	     ;; (ringing:bell-input-extra-help-function (if with-user 'teach-ringing:extra-help ringing:bell-input-extra-help-function))
	     ;; (place-marker (list 'work))





	     (i 2)
	     (j 2))
	(switch-to-buffer buffer)
	(erase-buffer)
	(insert (format "Mistakes so far in %s\n" method-name))

	(let ((user-mistakes:teaching-info (user-mistakes:get-teaching-data method-data i)))
	  (setf (user-mistakes:lead-which-row user-mistakes:teaching-info) 0)


	  ;; not actually suitable as it calls ring-from-rounds for each lead
	  (teach-ringing:do-one-lead method-name j nil t)





          ;; ;; also from do-one-lead
	  ;; (ringing:ring-from-rounds teach-ringing:current-method
	  ;;					    'ringing:ring-one-lead-helper-function
	  ;;					    (list place-marker)
	  ;;					    'teach-ringing:display-function
	  ;;					    (list user-mistakes:teaching-info place-marker show-performance))


	  (insert "\n\n"))
	(incf i)
	(setq j teach-ringing:observation-last-seen-at)



	(goto-char (point-min)))
    (message "No error data for %s" (ringing:method-name method))))

;;; end of user-mistakes.el
